<?php

/**
 * @link              https://wsform.com/knowledgebase/activecampaign/
 * @since             1.0.0
 * @package           WS_Form_ActiveCampaign_V3
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - ActiveCampaign
 * Plugin URI:        https://wsform.com/knowledgebase/activecampaign/
 * Description:       ActiveCampaign add-on for WS Form PRO
 * Version:           1.0.8
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-activecampaign-v3
 */

	Class WS_Form_Add_On_ActiveCampaign_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.140';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_ActiveCampaign_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-activecampaign-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_activecampaign_v3 = new WS_Form_Add_On_ActiveCampaign_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_activecampaign_v3, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_ActiveCampaign_V3 extends WS_Form_Action {

			public $id = 'activecampaignv3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $api_url;
			private $api_key;
			public $list_id = false;
			public $status;
			public $return_url;
			public $email_type;
			public $opt_in_field;
			public $field_mapping;
			public $field_mapping_custom;
			public $tag_mapping;
			public $tag_mapping_custom;
			public $tag_replace;
			public $automation_mapping;

			// Constants
			const API_PATH = '/api/3/';
			const LIST_ID_PREFIX = 'list_id_';	// Required to avoid clashing between list ID's and field ID's
			const AUTOMATION_ID_PREFIX = 'automation_id_';	// Required to avoid clashing between automation ID's and field ID's
			const CUSTOM_FIELD_EMAIL = 'email';
			const CUSTOM_FIELD_FIRST_NAME = 'firstName';
			const CUSTOM_FIELD_LAST_NAME = 'lastName';
			const CUSTOM_FIELD_PHONE = 'phone';
			const PAGINATION_LIMIT = 100;
			const WS_FORM_LICENSE_ITEM_ID = 5666;
			const WS_FORM_LICENSE_NAME = 'ActiveCampaign add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.8';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('ActiveCampaign', 'ws-form-activecampaign-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to ActiveCampaign', 'ws-form-activecampaign-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('rest_api_init', array($this, 'rest_api_init'), 10, 2);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-activecampaign-v3')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-activecampaign-v3') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-activecampaign-v3') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

					break;
				}
			}

			// Nag
			public function nag() {

				// Check user API email address is valid
				if(!filter_var($this->api_url, FILTER_VALIDATE_EMAIL) AND ($this->api_key == '')) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API URL and key <a href="%s">here</a>.', 'ws-form-activecampaign-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Reset list ID (For multiple actions)
				$this->list_id = false;

				// Load configuration
				self::load_config($config);

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-activecampaign-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }
				$list_id = self::get_list_id();

				// Process field mapping
				$email_address = false;
				$contact = array();
				$field_values = array();

				foreach($this->field_mapping as $field_map) {

					// Get field ID
					$field_id = $field_map['ws_form_field'];

					// Get list field
					$api_field = $field_map['action_' . $this->id . '_list_fields'];
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// Array conversion
					if(is_array($submit_value)) { 

						if(count($submit_value) > 1) {

							$submit_value = '||' . implode('||', $submit_value) . '||';

						} else {

							$submit_value = implode(',', $submit_value);
						}
					}

					switch($api_field) {

						case self::CUSTOM_FIELD_EMAIL :

							// Save email address
							$email_address = $submit_value;
							$contact[$api_field] = $submit_value;
							break;

						case self::CUSTOM_FIELD_FIRST_NAME :
						case self::CUSTOM_FIELD_LAST_NAME :
						case self::CUSTOM_FIELD_PHONE :

							// Save other built in fields
							$contact[$api_field] = $submit_value;
							break;

						default :

							// Normal field
							$field_values[] = array(

								'field' => $api_field,
								'value' => $submit_value
							);
					}
				}

				foreach($this->field_mapping_custom as $field_map) {

					// Get value
					$custom_value = $field_map['action_' . $this->id . '_field_mapping_custom_value'];
					$custom_value = WS_Form_Common::parse_variables_process($custom_value, $form, $submit, 'text/plain');
					if($custom_value == '') { continue; }

					// Get list field
					$api_field = $field_map['action_' . $this->id . '_list_fields'];

					// Array conversion
					if(is_array($custom_value)) { 

						if(count($custom_value) > 1) {

							$custom_value = '||' . implode('||', $custom_value) . '||';

						} else {

							$custom_value = implode(',', $custom_value);
						}
					}

					switch($api_field) {

						case self::CUSTOM_FIELD_EMAIL :

							// Save email address
							$email_address = $custom_value;
							$contact[$api_field] = $custom_value;
							break;

						case self::CUSTOM_FIELD_FIRST_NAME :
						case self::CUSTOM_FIELD_LAST_NAME :
						case self::CUSTOM_FIELD_PHONE :

							// Save other built in fields
							$contact[$api_field] = $custom_value;
							break;

						default :

							// Normal field
							$field_values[] = array(

								'field' => $api_field,
								'value' => $custom_value
							);
					}
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-activecampaign-v3')); }

				// Check user email address is valid
				if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-activecampaign-v3'), $email_address)); }

				// Process tags
				$tags = array();

				foreach($this->tag_mapping as $tag_map) {

					// Get field ID
					$field_id = $tag_map['ws_form_field'];
					if($field_id == '') { continue; }

					// Read submit meta
					$tag_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($tag_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($tag_array)) { $tag_array = array($tag_array); }

					// Process each interest ready for API format
					foreach($tag_array as $tag_id) {

						if($tag_id == '') { continue; }

						$tags[] = array('tag' => $tag_id);
					}
				}

				foreach($this->tag_mapping_custom as $tag_map) {

					// Get value
					$tag_name = $tag_map['action_' . $this->id . '_tag_mapping_custom_tag'];
					$tag_name = WS_Form_Common::parse_variables_process($tag_name, $form, $submit, 'text/plain');

					// Get tag ID
					$tag_id_array = self::get_tag_id_array($tag_name);

					// Add tag IDs to array
					foreach($tag_id_array as $tag_id) {

						if($tag_id !== false) {

							$tags[] = array('tag' => $tag_id);
						}
					}
				}

				// Process automations
				$automations = array();
				foreach($this->automation_mapping as $automation_map) {

					// Get field ID
					$automation_id = $automation_map['action_' . $this->id . '_automation_id'];
					if($automation_id == '') { continue; }
					$automation_id = self::get_automation_id($automation_id);

					$automations[] = array('automation' => $automation_id);
				}

				// Add contact
				$data = array(

					'contact' => $contact
				);

				// Add field values
				if(count($field_values) > 0) {

					$data['contact']['fieldValues'] = $field_values;
				}

				// POST
				$api_response = parent::api_call($this->api_url, 'contact/sync', 'POST', json_encode($data), self::get_api_headers());

				// Check status code
				if(
					($api_response['http_code'] != 200) &&
					($api_response['http_code'] != 201)

				) {

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Existing contact?
				$contact_existing = ($api_response['http_code'] == 200);

				// Process contact data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }
				if(!isset($api_response_decoded->contact)) { return self::error(__('No contact returned', 'ws-form-activecampaign-v3')); }
				if(!isset($api_response_decoded->contact->id)) { return self::error(__('No contact ID returned', 'ws-form-activecampaign-v3')); }
				$contact_id = $api_response_decoded->contact->id;

				// Assign contact to list
				$data = array(

					'contactList' => array(

						'list' 		=> $list_id,
						'contact' 	=> $contact_id,
						'status' 	=> 1,
					)
				);

				// POST
				$api_response = parent::api_call($this->api_url, 'contactLists', 'POST', json_encode($data), self::get_api_headers());

				// Check status code
				if(
					($api_response['http_code'] != 200) &&
					($api_response['http_code'] != 201)
				) {

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Delete existing tags
				if($contact_existing && $this->tag_replace) {

					$contact_tags = array();

					// Get tags
					$api_response = parent::api_call($this->api_url, sprintf('contacts/%u/contactTags', $contact_id), 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

					if(!isset($api_response_decoded->contactTags)) { return self::error(__('No tags returned', 'ws-form-activecampaign-v3')); }
					$api_response_tags = $api_response_decoded->contactTags;

					foreach($api_response_tags as $tag) {

						$contact_tags[parent::get_object_value($tag, 'id')] = parent::get_object_value($tag, 'tag');
					}

					// Check tags to add to see if they match existing
					foreach($tags as $tag_index => $tag) {

						$tag_id = $tag['tag'];

						$tag_existing_index = array_search($tag_id, $contact_tags);

						if($tag_existing_index !== false) {

							// Do not set this tag
							unset($tags[$tag_index]);

							// Do not delete this tag
							unset($contact_tags[$tag_existing_index]);
						}
					}

					// Delete tags
					foreach($contact_tags as $contact_tag_id => $tag_id) {

						// Delete contactTag
						$api_response = parent::api_call($this->api_url, sprintf('contactTags/%u', $contact_tag_id, $tag_id), 'DELETE', null, self::get_api_headers());
					}
				}

				// Process tags
				foreach($tags as $tag) {

					// Add contact ID
					$tag['contact'] = $contact_id;

					// Add fieldValue
					$data = array(

						'contactTag' => $tag
					);

					// POST
					$api_response = parent::api_call($this->api_url, 'contactTags', 'POST', json_encode($data), self::get_api_headers());

					// Check status code
					if(
						($api_response['http_code'] != 200) &&
						($api_response['http_code'] != 201)
					) {
						return self::error(self::get_api_error_detail($api_response['response']));
					}
				}

				// Process automations
				foreach($automations as $automation) {

					// Add contact ID
					$automation['contact'] = $contact_id;

					// Add fieldValue
					$data = array(

						'contactAutomation' => $automation
					);

					// POST
					$api_response = parent::api_call($this->api_url, 'contactAutomations', 'POST', json_encode($data), self::get_api_headers());

					// Check status code
					if(
						($api_response['http_code'] != 200) &&
						($api_response['http_code'] != 201)
					) {

						return self::error(self::get_api_error_detail($api_response['response']));
					}
				}

				// Success
				self::success(sprintf(__('Successfully subscribed %s to list ID: %s', 'ws-form-activecampaign-v3'), $email_address, $list_id));
				return true;
			}

			// Get tag ID from tag name
			public function get_tag_id_array($tag_name, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Return
				$tag_id_array = array();

				// Check to see if lists are cached
				$tags = self::get_tags();

				// Turn tags into lookup array
				$tag_lookup = array();
				foreach($tags as $tag) {

					$tag_lookup[strtolower($tag['label'])] = $tag['id'];
				}

				$tag_name = str_replace("\n", ',', $tag_name);
				$tag_array = explode(',', $tag_name);

				foreach($tag_array as $tag_name) {

					$tag_name = trim($tag_name);
					if($tag_name == '') { continue; }

					// Look for tag in cache

					if(isset($tag_lookup[strtolower($tag_name)])) {

						$tag_id_array[] = $tag_lookup[strtolower($tag_name)];
						continue;
					}

					// Tag not found in cache, so create it
					$params = array(

						'tag'		=> array(

							'tag' => $tag_name,
							'tagType' => 'contact'
						)
					);

					// Create tag
					$api_response = parent::api_call($this->api_url,  'tags', 'POST', json_encode($params), self::get_api_headers());

					// Tag successfully added
					if($api_response['http_code'] == 201) {

						// Process tag data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

						if(!isset($api_response_decoded->tag)) { return self::error(__('No tag returned', '	ws-form-activecampaign-v3')); }
						if(!isset($api_response_decoded->tag->id)) { return self::error(__('No tag ID returned', '	ws-form-activecampaign-v3')); }

						// Get tag ID
						$tag_id = $api_response_decoded->tag->id;
						$tag_id_array[] = $tag_id;
					}

					// Duplicate tag found
					if($api_response['http_code'] == 422) {

						// Retrieve tags so we can get the tag ID (Force update)
						$tags = self::get_tags(1, true);

						// Turn tags into lookup array
						$tag_lookup = array();
						foreach($tags as $tag) {

							$tag_lookup[strtolower($tag['label'])] = $tag['id'];
						}

						if(isset($tag_lookup[strtolower($tag_name)])) {

							$tag_id_array[] = $tag_lookup[strtolower($tag_name)];
						}
					}
				}

				return $tag_id_array;
			}

			// Extract list ID
			public function get_list_id() {

				$list_id = substr($this->list_id, strlen(self::LIST_ID_PREFIX));
				return $list_id;
			}

			// Extract automation ID
			public function get_automation_id($automation_id) {

				return substr($automation_id, strlen(self::AUTOMATION_ID_PREFIX));
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$pagination_offset = 0;

					do {

						// Params
						$params = array(

							'limit'		=> self::PAGINATION_LIMIT,
							'offset'	=> $pagination_offset
						);

						// Get lists
						$api_response = parent::api_call($this->api_url,  'lists', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

						if(!isset($api_response_decoded->lists)) { return self::error(__('No lists returned', 'ws-form-activecampaign-v3')); }
						$api_response_lists = $api_response_decoded->lists;

						// Get total lists
						$total_lists = $api_response_decoded->meta->total;

						// Increment read counter
						$pagination_offset += count($api_response_lists);

						foreach($api_response_lists as $list) {

							$lists[] = array(

								'id' => 			self::LIST_ID_PREFIX . parent::get_object_value($list, 'id'),
								'label' => 			parent::get_object_value($list, 'name'),
								'field_count' => 	false,
								'record_count' => 	false
							);
						}

					} while (count($lists) < $total_lists);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if(true || $fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get list
					$list_id = self::get_list_id();
					$api_response = parent::api_call($this->api_url, 'lists/' . $list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

					// Get list
					if(!isset($api_response_decoded->list)) { return self::error(__('No list returned', 'ws-form-activecampaign-v3')); }
					$api_response_list = $api_response_decoded->list;

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Get list fields
					$api_response = parent::api_call($this->api_url, 'fields', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process field data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

					// Process field options
					if(!isset($api_response_decoded->fieldOptions)) { return self::error(__('No field options returned', 'ws-form-activecampaign-v3')); }
					$api_response_field_options = $api_response_decoded->fieldOptions;
					if(!isset($api_response_decoded->fields)) { return self::error(__('No fields returned', 'ws-form-activecampaign-v3')); }
					$api_response_fields = $api_response_decoded->fields;

					// Process field options
					$field_options_lookup = array();
					foreach($api_response_field_options as $field_option) {

                    	$field_id = $field_option->field;

                    	if(!isset($field_options_lookup[$field_id])) { $field_options_lookup[$field_id] = array(); }
						$field_options_lookup[$field_id][] = array(

							'sort_index' => $field_option->orderid,
							'value' => $field_option->value,
							'text' => $field_option->label,
							'default' => ($field_option->isdefault == 1),
						);
					}

					// Add built in ActiveCampaign fields not returned by API
					$list_fields[] = array(

						'id' => 			self::CUSTOM_FIELD_EMAIL,
						'label' => 			__('Email', 'ws-form-activecampaign-v3'), 
						'label_field' => 	__('Email', 'ws-form-activecampaign-v3'), 
						'type' => 			'email',
						'action_type' =>	'email',
						'required' => 		true, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	1,
						'visible' =>		true,
						'meta' => 			false
					);

					$list_fields[] = array(

						'id' => 			'firstName',
						'label' => 			__('First Name', 'ws-form-activecampaign-v3'), 
						'label_field' => 	__('First Name', 'ws-form-activecampaign-v3'), 
						'type' => 			'text',
						'action_type' =>	'text',
						'required' => 		false, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	2,
						'visible' =>		true,
						'meta' => 			false
					);

					$list_fields[] = array(

						'id' => 			'lastName',
						'label' => 			__('Last Name', 'ws-form-activecampaign-v3'), 
						'label_field' => 	__('Last Name', 'ws-form-activecampaign-v3'), 
						'type' => 			'text',
						'action_type' =>	'text',
						'required' => 		false, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	3,
						'visible' =>		true,
						'meta' => 			false
					);

					$list_fields[] = array(

						'id' => 			'phone',
						'label' => 			__('Phone', 'ws-form-activecampaign-v3'), 
						'label_field' => 	__('Phone', 'ws-form-activecampaign-v3'), 
						'type' => 			'tel',
						'action_type' =>	'tel',
						'required' => 		true, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	4,
						'visible' =>		false,
						'meta' => 			false
					);

					// Process fields
					$sort_index = 5;
					foreach($api_response_fields as $field) {

						$field_id = $field->id;

						$field_options = isset($field_options_lookup[$field_id]) ? $field_options_lookup[$field_id] : array();

						$action_type = parent::get_object_value($field, 'type');

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						$list_fields[] = array(

							'id' => 			$field_id,
							'label' => 			parent::get_object_value($field, 'title'),
							'label_field' => 	parent::get_object_value($field, 'title'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		(parent::get_object_value($field, 'isrequired') == 1),
							'default_value' => 	parent::get_object_value($field, 'defval'),
							'pattern' => 		'',
							'placeholder' => 	'',
							'help' => 			parent::get_object_value($field, 'descript'), 
							'sort_index' => 	$sort_index++,
							'visible' =>		(parent::get_object_value($field, 'visible') == 1),
							'meta' => 			self::action_field_to_ws_form_meta_keys($field, $field_options)
						);
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get automations
			public function get_automations($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if automations are cached
				$automations = WS_Form_Common::option_get('action_' . $this->id . '_automations');

				// Retried if fetch is requested or automations are not cached
				if($fetch || ($automations === false)) {

					$automations = array();

					// Load configuration
					self::load_config();

					// Initialize
					$pagination_offset = 0;
					$automation_count = 0;

					do {

						// Params
						$params = array(

							'limit'		=> self::PAGINATION_LIMIT,
							'offset'	=> $pagination_offset
						);

						// Get automations
						$api_response = parent::api_call($this->api_url,  'automations', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process automation data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

						if(!isset($api_response_decoded->automations)) { return self::error(__('No automations returned', 'ws-form-activecampaign-v3')); }
						$api_response_automations = $api_response_decoded->automations;

						// Get total automations
						$automation_total = $api_response_decoded->meta->total;

						// Increment read counter
						$pagination_offset += count($api_response_automations);

						foreach($api_response_automations as $automation) {

							// Only list active automations
							$status = parent::get_object_value($automation, 'status');
							if($status == 1) {

								$automations[] = array(

									'id' 			=> 	self::AUTOMATION_ID_PREFIX . parent::get_object_value($automation, 'id'),
									'label' 		=> 	parent::get_object_value($automation, 'name')
								);
							}

							$automation_count++;
						}

					} while ($automation_count < $automation_total);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_automations', $automations);
				}

				return $automations;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_tag_mapping'		=>	'tag_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				$tag_categories = array(

					array(

						'id' => 		1,
						'label' => 		__('Tags', 'ws-form-activecampaign-v3'), 
						'sort_index' =>	1,
						'type' => 		'checkbox'
					)
				);

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = 1, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check to see if tags are cached
				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tags');

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Initialize
					$pagination_offset = 0;

					$sort_index = 1;

					do {

						// Params
						$params = array(

							'limit'		=> self::PAGINATION_LIMIT,
							'offset'	=> $pagination_offset
						);

						// Get lists
						$api_response = parent::api_call($this->api_url, 'tags', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-activecampaign-v3')); }

						if(!isset($api_response_decoded->tags)) { return self::error(__('No tags returned', 'ws-form-activecampaign-v3')); }
						$api_response_tags = $api_response_decoded->tags;

						// Get total tags
						$total_tags = $api_response_decoded->meta->total;

						// Increment read counter
						$pagination_offset += count($api_response_tags);

						foreach($api_response_tags as $tag) {

							$tags[] = array(

								'id' => 			parent::get_object_value($tag, 'id'),
								'label' => 			parent::get_object_value($tag, 'tag'),
								'record_count' =>	false,
								'sort_index' =>		$sort_index++
							);
						}

					} while (count($tags) < $total_tags);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tags', $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-activecampaign-v3'); }

				$error_detail_array = array();
				if(isset($api_response_decoded->message)) {

					// Get error detail from JSON response
					$error_detail_array[] = $api_response_decoded->message;
				}
				if(isset($api_response_decoded->errors)) {

					foreach($api_response_decoded->errors as $error) {

						if(!isset($error->title)) { continue; }
						$error_detail_array[] = $error->title;
					}
				}

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_field_mapping_custom',
						'action_' . $this->id . '_tag_mapping',
						'action_' . $this->id . '_tag_mapping_custom',
						'action_' . $this->id . '_tag_replace',
						'action_' . $this->id . '_automation_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-activecampaign-v3') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-activecampaign-v3')); }
				if(strpos($this->list_id, self::LIST_ID_PREFIX) === false) { echo "ERROR"; exit; return self::error(__('List ID is not set', 'ws-form-activecampaign-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('ActiveCampaign List', 'ws-form-activecampaign-v3'),
						'type'							=>	'select',
						'help'							=>	__('Select the ActiveCampaign list to associate this form with.', 'ws-form-activecampaign-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-activecampaign-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'						=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-activecampaign-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-activecampaign-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-activecampaign-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-activecampaign-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to ActiveCampaign fields.', 'ws-form-activecampaign-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Cstom field mapping
					'action_' . $this->id . '_field_mapping_custom'	=> array(

						'label'						=>	__('Custom Field Mapping', 'ws-form-activecampaign-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map custom values to ActiveCampaign fields.', 'ws-form-activecampaign-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_field_mapping_custom_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom field mapping - Value
					'action_' . $this->id . '_field_mapping_custom_value'	=> array(

						'label'						=>	__('Value', 'ws-form-activecampaign-v3'),
						'type'						=>	'text'
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-activecampaign-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to ActiveCampaign tags. The value of the field should be a tag ID.', 'ws-form-activecampaign-v3'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field_choice'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom tag mapping
					'action_' . $this->id . '_tag_mapping_custom'	=> array(

						'label'						=>	__('Custom Tag Mapping', 'ws-form-activecampaign-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map contact to custom tag names. Enter one tag name per row.', 'ws-form-activecampaign-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_tag_mapping_custom_tag'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom tag mapping - Tag ID
					'action_' . $this->id . '_tag_mapping_custom_tag'	=> array(

						'label'						=>	__('Tag Name', 'ws-form-activecampaign-v3'),
						'type'						=>	'text'
					),

					// Replace existing tags
					'action_' . $this->id . '_tag_replace'	=> array(

						'label'						=>	__('Replace Existing Tags', 'ws-form-activecampaign-v3'),
						'type'						=>	'checkbox',
						'default'					=>	'',
					),

					// Automation mapping
					'action_' . $this->id . '_automation_mapping'	=> array(

						'label'						=>	__('Automation Mapping', 'ws-form-activecampaign-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Select which automations to add this contact to. Only active automations are listed.', 'ws-form-activecampaign-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_automation_id'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_automation_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Automation ID
					'action_' . $this->id . '_automation_id'	=> array(

						'label'							=>	__('Automation', 'ws-form-activecampaign-v3'),
						'type'							=>	'select',
						'options'						=>	array(),
						'options_blank'					=>	__('Select...', 'ws-form-activecampaign-v3')
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('ActiveCampaign Field', 'ws-form-activecampaign-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-activecampaign-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Populate automations
				$automations = self::get_automations(WS_Form_Common::can_user('edit_form'));
				if(is_array($automations)) {

					foreach($automations as $automation) {

						$config_meta_keys['action_' . $this->id . '_automation_id']['options'][] = array('value' => $automation['id'], 'text' => $automation['label']);
					}
				}

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-activecampaign-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-activecampaign-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your ActiveCampaign add-on for WS Form PRO license key here.', 'ws-form-activecampaign-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-activecampaign-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_url'	=>	array(

							'label'		=>	__('ActiveCampaign API URL', 'ws-form-activecampaign-v3'),
							'type'		=>	'url',
							'default'	=>	'',
							'help'		=>	__('For help finding your ActiveCampaign API URL and Key, <a href="https://help.activecampaign.com/hc/en-us/articles/207317590-Getting-started-with-the-API#how-to-obtain-your-activecampaign-api-url-and-key" target="_blank">click here</a>.', 'ws-form-activecampaign-v3')
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'		=>	__('ActiveCampaign API Key', 'ws-form-activecampaign-v3'),
							'type'		=>	'password',
							'default'	=>	''
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'text' : return 'text';
					case 'textarea' : return 'textarea';
					case 'date' : return 'datetime';
					case 'dropdown' : return 'select';
					case 'listbox' : return 'select';
					case 'radio' : return 'radio';

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field, $field_options) {

				$type = parent::get_object_value($field, 'type');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids
					case 'dropdown' :
					case 'listbox' :
					case 'radio' :

						// Build new columns
						$columns = array(

							array('id' => 0, 'label' => __('Label', 'ws-form-activecampaign-v3')),
							array('id' => 1, 'label' => __('Value', 'ws-form-activecampaign-v3'))
						);

						$rows = array();
						$row_index = 1;
						foreach($field_options as $field_option) {

							// Build new rows
							$rows[] = array(

								'id'		=> $row_index++,
								'default'	=> $field_option['default'],
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array($field_option['text'], $field_option['value'])
							);
						}

						// Group name
						switch($type) {

							case 'dropdown' : 

								// Build meta data
								$meta_data = array(

									'data_grid_select' => WS_Form_Common::build_data_grid_meta(

										'data_grid_select',
										__('Options', 'ws-form-activecampaign-v3'),
										$columns,
										$rows
									),

									'select_field_label' => 0,
									'select_field_value' => 1
								);

								break;

							case 'listbox' : 

								// Build meta data
								$meta_data = array(

									'data_grid_select' => WS_Form_Common::build_data_grid_meta(

										'data_grid_select',
										__('Options', 'ws-form-activecampaign-v3'),
										$columns,
										$rows
									),

									'select_field_label' 	=> 0,
									'select_field_value' 	=> 1,
									'size'					=> '4'
								);

								break;

							case 'radio' :

								// Build meta data
								$meta_data = array(

									'data_grid_radio' => WS_Form_Common::build_data_grid_meta(

										'data_grid_radio',
										__('Radios', 'ws-form-activecampaign-v3'),
										$columns,
										$rows
									),

									'radio_field_label' 	=> 0,
									'radio_field_value' 	=> 1
								);

								break;
						}

						return $meta_data;

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id', false); }
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');

				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->field_mapping_custom = parent::get_config($config, 'action_' . $this->id . '_field_mapping_custom', array());
				if(!is_array($this->field_mapping_custom)) { $this->field_mapping_custom = array(); }

				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
				$this->tag_mapping_custom = parent::get_config($config, 'action_' . $this->id . '_tag_mapping_custom', array());
				if(!is_array($this->tag_mapping_custom)) { $this->tag_mapping_custom = array(); }

				$this->tag_replace = parent::get_config($config, 'action_' . $this->id . '_tag_replace', false);

				$this->automation_mapping = parent::get_config($config, 'action_' . $this->id . '_automation_mapping', array());
				if(!is_array($this->automation_mapping)) { $this->automation_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = false;

				$this->api_url = WS_Form_Common::option_get('action_' . $this->id . '_api_url', '');
				$this->api_key = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');

				// Check API URL
				if($this->api_url == '') { return $this->configured; }
				$this->api_url = rtrim($this->api_url, '/') . self::API_PATH;

				// Check API key
				if($this->api_key == '') { return $this->configured; }

				$this->configured = true;

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('Api-Token' => $this->api_key);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(51.000000, 62.000000)"><path d="M16.9 34.6c1 0 2-.4 3.1-1.2 1.3-.9 2.4-1.6 2.4-1.6l.4-.3-.4-.3c-.2-.1-17.1-11.9-18.8-13-.8-.6-1.8-.7-2.5-.4-.7.4-1.1 1.1-1.1 2v4l.2.1c.1.1 11.8 8.2 14 9.7.9.7 1.8 1 2.7 1z" fill="#356ae6"/><path d="M39.6 27.5C38.6 26.8 2.3 1.4.7.4L.2 0v6.2c0 2 1.1 2.7 2.3 3.7 0 0 27.9 19.4 31.3 21.8-3.4 2.4-29.6 20.6-31.3 21.6-2 1.3-2.2 2.2-2.2 4V64s38.4-27.5 39.2-28.1c1.8-1.3 2.1-2.9 2.2-4v-.7c.1-1.4-.7-2.7-2.1-3.7z" fill="#356ae6"/></g>';

				return $svg_logo;
			}
		}
	});