<?php

	/**
	 * @link              https://wsform.com/knowledgebase/authorize_accept/
	 * @since             1.0.0
	 * @package           WS_Form_Authorize_Accept
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Authorize Accept
	 * Plugin URI:        https://wsform.com/knowledgebase/authorize_accept/
	 * Description:       Authorize Accept add-on for WS Form PRO
	 * Version:           1.0.12
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-authorize-accept
	 */

	require_once(__DIR__ . '/composer/vendor/autoload.php');

	use net\authorize\api\contract\v1 as AnetAPI;
	use net\authorize\api\controller as AnetController;

	Class WS_Form_Add_On_Authorize_Accept {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.5.37';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Authorize_Accept();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-authorize-accept'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_authorize_accept = new WS_Form_Add_On_Authorize_Accept();	

	register_activation_hook(__FILE__, array($wsf_add_on_authorize_accept, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Authorize_Accept extends WS_Form_Action {

			public $id = 'authorizeaccept';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 10;
			public $can_repost = false;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			public $environment;
			public $sandbox_api_login_id;
			public $sandbox_transaction_key;
			public $sandbox_client_key;
			public $production_api_login_id;
			public $production_transaction_key;
			public $production_client_key;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 3458;
			const WS_FORM_LICENSE_NAME = 'Authorize Accept add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.12';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Authorize Accept', 'ws-form-authorize-accept');

				// Set label for actions pull down
				$this->label_action = __('Pay with Authorize Accept', 'ws-form-authorize-accept');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('wsf_actions_post_submit', array($this, 'actions_post_submit'), 10, 1);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_config_frameworks', array($this, 'config_frameworks'), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();

				// If configured, register the Authorize field type
				if($this->configured) {

					add_filter('wsf_config_field_types', array($this, 'config_field_types'), 10, 2);
				}
			}

			// Actions post filter
			public function actions_post_submit($actions) {

				// Get Authorize descriptor and value
				$descriptor = WS_Form_Common::get_query_var('authorize_accept_descriptor', false);
				$value = WS_Form_Common::get_query_var('authorize_accept_value', false);

				// Check form data
				if(empty($descriptor)) { return $actions; }
				if(empty($value)) { return $actions; }

				// Prepend this action so it runs first
				$actions[] = array(

					'id' => $this->id,
					'meta' => array(
						'action_' . $this->id . '_descriptor' => $descriptor,
						'action_' . $this->id . '_value' => $value
					),
					'events' => array(
						'0' => 'submit'
					),
					'label' => $this->label_action,
					'priority' => $this->priority,
					'row_index' => 0
				);

				return $actions;
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Get Authorize descriptor, value and field ID
				$descriptor = WS_Form_Common::get_query_var('authorize_accept_descriptor', false);
				$value = WS_Form_Common::get_query_var('authorize_accept_value', false);
				$field_id = WS_Form_Common::get_query_var('authorize_accept_field_id', false);

				// Check form data
				if(empty($descriptor)) { return false; }
				if(empty($value)) { return false; }
				if(empty($field_id)) { return false; }

				// Get field ID
				$ws_form_field = new WS_Form_Field();
				$ws_form_field->id = $field_id;
				$field = $ws_form_field->db_read(true, true);	// Will throw an error if field_id is invalid

				// Convert to object
				$field = (object)$field;

				// Read config from field
				$transaction_amount = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_amount', '');
				$transaction_zero = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero', '');
				$transaction_zero_message = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero_message', ''), $form, $submit);

				// Process transaction amount
				switch($transaction_amount) {

					case '' :
					case '#ecommerce_cart_total' :

						// Get calculated transaction amount from query var
						$ecommerce_cart = WS_Form_Common::get_query_var('wsf_ecommerce_cart', false);
						$amount = floatval($ecommerce_cart['total']);
						break;

					default :

						// Use amount specified in field config
						$amount = floatval($transaction_amount);
				}

				// Check for a zero amount
				if($amount == 0) {

					switch($transaction_zero) {

						case 'message' :

							self::error($transaction_zero_message);
							break;

						case '' :

							return false;
					}
				}

				// Get transaction method
				$transaction_type = 'auth_capture_transaction';

				switch($transaction_type) {

					case 'auth_capture_transaction' :

						return self::auth_capture_transaction($form, $submit, $field, $descriptor, $value, $amount);
				}

				return true;
			}

			public function auth_capture_transaction($form, $submit, $field, $descriptor, $value, $amount) {

				// Transaction description
				$transaction_description = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_description', ''), $form, $submit);

				// Customer type
				$customer_type = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_type', 'individual');

				// Customer email
				$customer_email_field_id = intval(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_email', ''));
				$customer_email = ($customer_email_field_id > 0) ? parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $customer_email_field_id, false) : false;
				if($customer_email !== false) {

					$customer_email = WS_Form_Common::parse_variables_process($customer_email, $form, $submit);
				}

				$api_login_id = ($this->environment == 'sandbox') ? $this->sandbox_api_login_id : $this->production_api_login_id;
				$transaction_key = ($this->environment == 'sandbox') ? $this->sandbox_transaction_key : $this->production_transaction_key;

				// Merchant authentication
				$merchant_authentication = new AnetAPI\MerchantAuthenticationType();
				$merchant_authentication->setName($api_login_id);
				$merchant_authentication->setTransactionKey($transaction_key);

				// Ref ID
				$ref_id = 'ref' . time();

				// Create the payment object for a payment nonce
				$opaque_data = new AnetAPI\OpaqueDataType();
				$opaque_data->setDataDescriptor($descriptor);
				$opaque_data->setDataValue($value);

				// Add the payment data to a paymentType object
				$payment_type = new AnetAPI\PaymentType();
				$payment_type->setOpaqueData($opaque_data);

				// Create order information
				$order = new AnetAPI\OrderType();
				if(!empty($transaction_description)) { $order->setDescription($transaction_description); }

				// Billing address mapping
				$billing_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_billing_address_mapping', '');
				$billing_address_obj = self::get_customer_address($submit, $billing_address_mapping);

				// Shipping address mapping
				$shipping_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_address_mapping', '');
				$shipping_address_obj = self::get_customer_address($submit, $shipping_address_mapping);

				// Customer mapping
				$customer_data = new AnetAPI\CustomerDataType();
				$customer_data->setType($customer_type);
				if(!empty($customer_email)) { $customer_data->setEmail($customer_email); }

				// Add values for transaction settings
				$duplicate_window_settings = new AnetAPI\SettingType();
				$duplicate_window_settings->setSettingName("duplicateWindow");
				$duplicate_window_settings->setSettingValue("60");

				// Create a TransactionRequestType object and add the previous objects to it
				$transaction_request = new AnetAPI\TransactionRequestType();
				$transaction_request->setTransactionType('authCaptureTransaction'); 
				$transaction_request->setAmount($amount);
				$transaction_request->setOrder($order);
				$transaction_request->setPayment($payment_type);

				// Billing address
				if($billing_address_obj !== false) {
					$transaction_request->setBillTo($billing_address_obj);
				}

				// Shipping address
				if($shipping_address_obj !== false) {
					$transaction_request->setShipTo($shipping_address_obj);
				}

				// Tracking data - Remote IP
				$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
				if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

					$transaction_request->setCustomerIP($tracking_remote_ip);
				}

				$transaction_request->setCustomer($customer_data);
				$transaction_request->addToTransactionSettings($duplicate_window_settings);

				// Assemble the complete transaction request
				$request = new AnetAPI\CreateTransactionRequest();
				$request->setMerchantAuthentication($merchant_authentication);
				$request->setRefId($ref_id);
				$request->setTransactionRequest($transaction_request);

				// Create the controller and get the response
				$controller = new AnetController\CreateTransactionController($request);
				$environment_constant = ($this->environment == 'sandbox') ? \net\authorize\api\constants\ANetEnvironment::SANDBOX : \net\authorize\api\constants\ANetEnvironment::PRODUCTION;
				$response = $controller->executeWithApiResponse($environment_constant);

				// Error handling
				$error_code = false;
				$error_message = 'No response returned';

				if($response != null) {

					// Check to see if the API request was successfully received and acted upon
					if ($response->getMessages()->getResultCode() == "Ok") {

						// Since the API request was successful, look for a transaction response
						// and parse it to display the results of authorizing the card
						$transaction_response = $response->getTransactionResponse();

						if ($transaction_response != null && $transaction_response->getMessages() != null) {

							// Read transaction data
							$transaction_id = $transaction_response->getTransId();
							$transaction_response_code = $transaction_response->getResponseCode();
							$transaction_message_code = $transaction_response->getMessages()[0]->getCode();
							$transaction_auth_code = $transaction_response->getAuthCode();
							$transaction_description = $transaction_response->getMessages()[0]->getDescription();

							// Set transaction ID
							$submit->meta['ecommerce_transaction_id'] = $transaction_id;

							// Set transaction
							$transaction = array(

								'environment' => $this->environment,
								'amount' => $amount,
								'customer_type' => $customer_type,
								'email' => $customer_email,
								'ref_id' => $ref_id,
								'transaction' => array(

									'id' => $transaction_id,
									'response_code' => $transaction_response_code,
									'message_code' => $transaction_message_code,
									'auth_code' => $transaction_auth_code,
									'description' => $transaction_description,
								)
							);
							$submit->meta['ecommerce_transaction'] = json_encode($transaction);

							// Set status
							$submit->meta['ecommerce_status'] = 'completed';

							self::success(__('Payment successfully processed with transaction ID: ', 'ws-form-authorize-accept') . $transaction_id);
							return $submit;

						} else {

							if ($transaction_response->getErrors() != null) {

								$error_code = $transaction_response->getErrors()[0]->getErrorCode();
								$error_message = $transaction_response->getErrors()[0]->getErrorText();
							}
						}
		
					} else {
		
						$transaction_response = $response->getTransactionResponse();

						if ($transaction_response != null && $transaction_response->getErrors() != null) {

							$error_code = $transaction_response->getErrors()[0]->getErrorCode();
							$error_message = $transaction_response->getErrors()[0]->getErrorText();

						} else {

							$error_code = $response->getMessages()->getMessage()[0]->getCode();
							$error_message = $response->getMessages()->getMessage()[0]->getText();
						}
					}
				}

				// Set error message
				$error_message = (($error_code !== false) ? $error_code . ': ' : '') . $error_message;
				self::error($error_message);

				// Stop process actions
				return 'halt';
			}

			// Get customer address object
			public function get_customer_address($submit, $address_mapping) {

				$address = new AnetAPI\CustomerAddressType();

				if(!is_array($address_mapping)) { return false; }

				$field_map_count = 0;

				foreach($address_mapping as $field_map) {

					// Get submit value
					$field_id = $field_map->ws_form_field;
					if(empty($field_id)) { continue; }
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					// Process authorize field
					$authorize_field = $field_map->{'action_' . $this->id . '_address_part'};
					if(empty($authorize_field)) { continue; }

					switch($authorize_field) {

						case 'first_name' :

							$address->setFirstName($submit_value);
							break;

						case 'last_name' :

							$address->setLastName($submit_value);
							break;

						case 'company' :

							$address->setCompany($submit_value);
							break;

						case 'address_line_1' :

							$address->setAddress($submit_value);
							break;

						case 'city' :

							$address->setCity($submit_value);
							break;

						case 'state' :

							$address->setState($submit_value);
							break;

						case 'zip' :

							$address->setZip($submit_value);
							break;

						case 'country' :

							$address->setCountry($submit_value);
					}

					$field_map_count++;
				}

				if($field_map_count == 0) { return false; }

				return $address;				
			}

			// Config - Frameworks
			public function config_frameworks($frameworks) {

				// WS Form
				$frameworks['types']['ws-form']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('wsf-button');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('wsf-button-full');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';

				// Bootstrap
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';

				// Foundation
				$frameworks['types']['foundation5']['fields']['public']['field_types']['authnet_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('expand');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['authnet_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['authnet_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['authnet_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['authnet_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['authnet_checkout']['class_field_button_type_fallback'] = 'primary';

				return $frameworks;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API keys <a href="%s">here</a>.', 'ws-form-'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-authorize-accept')));

				return $links;
			}

			// Plug-in field types for this action
			public function config_field_types($field_types) {

				// Credentials
				$api_login_id = ($this->environment == 'sandbox') ? $this->sandbox_api_login_id : $this->production_api_login_id;
				$client_key = ($this->environment == 'sandbox') ? $this->sandbox_client_key : $this->production_client_key;

				$field_types['ecommerce']['types']['authnet_checkout'] = array (

					'label'						=>	__('Authorize.Net', 'ws-form'),
					'pro_required'				=>	!WS_Form_Common::is_edition('pro'),
// @if NODE_ENV='license_pro'
					'label_default'				=>	__('Checkout', 'ws-form'),
					'label_position_force'		=>	'top',
					'mask_field'				=>	'<button id="#id" data-ecommerce-payment#attributes>#label</button><div id="#id_trigger" class="AcceptUI" data-apiLoginID="' . htmlentities($api_login_id) . '" data-clientKey="' . htmlentities($client_key) . '" data-responseHandler="authorize_response_handler_#form_instance_id" style="display:none;"#attributes></div><input type="hidden" id="#id_descriptor" name="authorize_accept_descriptor" /><input type="hidden" id="#id_value" name="authorize_accept_value" /><input type="hidden" id="#id_field_id" name="authorize_accept_field_id" value="#field_id" /><script>' . file_get_contents(__DIR__ . '/js/authorize-accept.js') . '</script>',
					'mask_field_attributes'		=>	array('class', 'action_' . $this->id . '_appearance_dbao', 'action_' . $this->id . '_appearance_dAUIFBT', 'action_' . $this->id . '_appearance_dAUIFHT'),
					'mask_field_label'			=>	'#label',
					'submit_save'				=>	false,
					'submit_edit'				=>	false,
					'progress'					=>	false,
					'multiple'					=>	false,
					'icon'						=>	'<svg height="16" width="16" viewBox="0 0 16 16"><path d="M5.4 16H.1v-.6c0-.1.1-.2.2-.2.2-.1.5-.1.7-.1.4-.1.8-.4 1-.8.5-1.3 1-2.5 1.5-3.7.9-2.1 1.7-4.2 2.5-6.3.5-1.2.9-2.3 1.4-3.5.1-.3.2-.5.6-.5.3 0 .5-.1.8-.2 0 .2.1.5.2.7.8 2.2 1.6 4.3 2.4 6.5.7 2 1.5 4.1 2.2 6.1l.6 1.2c.3.4.7.6 1.1.6h.1c.5.1.4 0 .4.4v.2c-.1.2-.2.2-.4.2H9.1v-.6c0-.1.1-.2.2-.2.3 0 .5-.1.8-.1.5-.1.6-.3.4-.8-.4-1.1-.8-2.1-1.2-3.2-.1-.2-.2-.3-.4-.3H5.1c-.1 0-.1 0-.2.1-.1.4-.3.9-.4 1.3l-.6 2.1c0 .5.1.7.5.8.4.1.7.1 1.1.2v.6s0 .1-.1.1zm0-6.5h3.4c-.5-1.6-1.1-3.2-1.7-4.9C6.5 6.3 6 7.9 5.4 9.5z"/></svg>',
					'conditional'				=>	array(

						'exclude_condition'	=>	true,
						'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper')
					),
					'meta_key_parse_variables'	=>	array(

						'action_' . $this->id . '_transaction_amount',
						'action_' . $this->id . '_transaction_zero',
						'action_' . $this->id . '_transaction_zero_message',
						'action_' . $this->id . '_message_method',
						'action_' . $this->id . '_message_clear',
						'action_' . $this->id . '_message_scroll_top',
						'action_' . $this->id . '_message_form_hide',
						'action_' . $this->id . '_message_duration'
					),

					'fieldsets'				=> array(

						// Tab: Basic
						'basic'	=> array(

							'label'			=>	__('Basic', 'ws-form'),

							'fieldsets'		=>	array(

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_amount',
										'action_' . $this->id . '_transaction_description'
									)
								),

								array(

									'label'		=>	__('Customer', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_customer_email',
										'action_' . $this->id . '_customer_type',
									)
								),

								array(

									'label'		=>	__('Appearance', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_appearance_dbao',
										'action_' . $this->id . '_appearance_dAUIFHT',
										'action_' . $this->id . '_appearance_dAUIFBT'
									)
								)
							)
						),

						// Tab: Advanced
						'advanced'	=>	array(

							'label'			=>	__('Advanced', 'ws-form'),

							'fieldsets'	=>	array(

								array(

									'label'		=>	__('Billing Address', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_billing_address_mapping'
									)
								),

								array(

									'label'		=>	__('Shipping Address', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_shipping_address_mapping'
									)
								),

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_zero',
										'action_' . $this->id . '_transaction_zero_message'
									)
								),

								array(

									'label'		=>	__('Messages', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_message_method',
										'action_' . $this->id . '_message_clear',
										'action_' . $this->id . '_message_scroll_top',
										'action_' . $this->id . '_message_form_hide',
										'action_' . $this->id . '_message_duration'
									)
								),

								array(
									'label'		=>	__('Styles', 'ws-form'),
									'meta_keys'	=>	array('class_single_vertical_align_bottom', 'class_field_button_type_primary', 'class_field_full_button_remove')
								),

								array(
									'label'		=>	__('Classes', 'ws-form'),
									'meta_keys'	=>	array('class_field_wrapper', 'class_field')
								),

								array(
									'label'		=>	__('Breakpoints', 'ws-form'),
									'meta_keys'	=> array('breakpoint_sizes'),
									'class'		=>	array('wsf-fieldset-panel')
								)
							)
						)
					)
				);

				// Parse JS
				$mask_field = $field_types['ecommerce']['types']['authnet_checkout']['mask_field'];

				$mask_fields = array(

					'action_' . $this->id . '_environment' => $this->environment,
					'action_' . $this->id . '_sandbox_client_key' => $this->sandbox_client_key,
					'action_' . $this->id . '_production_client_key' => $this->production_client_key
				);

				$field_types['ecommerce']['types']['authnet_checkout']['mask_field'] = WS_Form_Common::mask_parse($mask_field, $mask_fields);

				return $field_types;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-authorize-accept') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-authorize-accept') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { self::error(sprintf(__('Action not configured (%s)', 'ws-form-authorize-accept'), $this->label)); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Transaction - Amount
					'action_' . $this->id . '_transaction_amount'	=> array(

						'label'						=>	__('Amount To Charge', 'ws-form-authorize-accept'),
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('Enter the amount to charge the user. Leave blank to use cart total.', 'ws-form-authorize-accept'),
						'select_list'				=>	true,
						'placeholder'				=>	'#ecommerce_cart_total'
					),

					// Transaction - Description
					'action_' . $this->id . '_transaction_description'	=> array(

						'label'						=>	__('Description', 'ws-form-authorize-accept'),
						'type'						=>	'textarea',
						'default'					=>	'Payment to #blog_name',
						'help'						=>	__('Enter a description for this transaction.', 'ws-form-authorize-accept'),
						'select_list'				=>	true
					),

					// Appearance - Button label
					'action_' . $this->id . '_appearance_dAUIFBT'	=> array(

						'label'						=>	__('Button Label', 'ws-form-authorize-accept'),
						'mask'						=>	'data-acceptUIFormBtnTxt="#value"',
						'mask_disregard_on_empty'	=>	true,
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('This is the text that is displayed on the button. The default is "Pay".', 'ws-form-authorize-accept'),
						'select_list'				=>	true,
						'placeholder'				=>	'Pay'
					),

					// Appearance - Header text
					'action_' . $this->id . '_appearance_dAUIFHT'	=> array(

						'label'						=>	__('Header Text', 'ws-form-authorize-accept'),
						'mask'						=>	'data-acceptUIFormHeaderTxt="#value"',
						'mask_disregard_on_empty'	=>	true,
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('This is the text that is displayed on the header of the payment form. The default is to not display any text.', 'ws-form-authorize-accept'),
						'select_list'				=>	true
					),

					// Customer - Email
					'action_' . $this->id . '_customer_email'	=> array(

						'label'						=>	__('Email', 'ws-form-authorize-accept'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form-authorize-accept'),
						'fields_filter_type'		=>	array('email'),
						'help'						=>	__('Select the email field on your form to use for the customers email address.', 'ws-form-authorize-accept')
					),

					// Customer - Type
					'action_' . $this->id . '_customer_type'	=> array(

						'label'						=>	__('Type', 'ws-form-authorize-accept'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'individual', 'text' => __('Individual', 'ws-form-authorize-accept')),
							array('value' => 'business', 'text' => __('Business', 'ws-form-authorize-accept')),
						),
						'default'					=>	'individual',
						'help'						=>	__('Should transactions be processed as individuals or businesses?', 'ws-form-authorize-accept'),
					),

					// Transaction - Billing address option
					'action_' . $this->id . '_appearance_dbao'	=> array(

						'label'						=>	__('Billing Address Option', 'ws-form-authorize-accept'),
						'mask'						=>	'data-billingAddressOptions=\'#value\'',
						'mask_disregard_on_empty'	=>	true,
						'type'						=>	'select',
						'help'						=>	__('Whether the name and postal code fields should be shown on the form and whether they are required to be filled in.', 'ws-form-authorize-accept'),
						'default'					=>	'{"show":true, "required":true}',
						'options'					=>	array(

							array('value' => '{"show":true, "required":true}', 'text' => __('Shown & required', 'ws-form-authorize-accept')),
							array('value' => '{"show":true, "required":false}', 'text' => __('Shown & not required', 'ws-form-authorize-accept')),
							array('value' => '{"show":false, "required":false}', 'text' => __('Not shown', 'ws-form-authorize-accept'))
						)
					),

					// Transaction - Zero Amount
					'action_' . $this->id . '_transaction_zero'	=> array(

						'label'						=>	__('Zero Amount', 'ws-form-authorize-accept'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-authorize-accept')),
							array('value' => 'submit', 'text' => __('Submit form', 'ws-form-authorize-accept')),
							array('value' => 'message', 'text' => __('Show error message', 'ws-form-authorize-accept')),
						),
						'default'					=>	'message',
						'help'						=>	__('What should happen if the cart total is zero and the Authorize button is clicked?', 'ws-form-authorize-accept'),
					),

					// Transaction - Zero Amount - Message
					'action_' . $this->id . '_transaction_zero_message'	=> array(

						'label'						=>	__('Zero Amount Message', 'ws-form-authorize-accept'),
						'type'						=>	'text',
						'help'						=>	__('Enter a message to show if the Authorize button is clicked and the cart amount is zero.', 'ws-form-authorize-accept'),
						'default'					=>	__('The total cart amount cannot be zero. Please review the form and try again.', 'ws-form-authorize-accept'),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_zero',
								'meta_value'		=>	'message'
							)
						)
					),

					// Billing - Address mapping
					'action_' . $this->id . '_billing_address_mapping'	=> array(

						'label'						=>	__('Billing Address Mapping', 'ws-form-authorize-accept'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Authorize billing address fields.', 'ws-form-authorize-accept'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_address_part'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_address_part'
						),
						'default'					=>	array()
					),

					// Shipping - Address mapping
					'action_' . $this->id . '_shipping_address_mapping'	=> array(

						'label'						=>	__('Shipping Address Mapping', 'ws-form-authorize-accept'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Authorize shipping address fields.', 'ws-form-authorize-accept'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_address_part'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_address_part'
						),
						'default'					=>	array()
					),

					// Address parts
					'action_' . $this->id . '_address_part'	=> array(

						'label'							=>	__('Authorize Field', 'ws-form-authorize-accept'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => 'first_name', 'text' => __('First Name')),
							array('value' => 'last_name', 'text' => __('Last Name')),
							array('value' => 'company', 'text' => __('Company')),
							array('value' => 'address_line_1', 'text' => __('Address')),
							array('value' => 'city', 'text' => __('City')),
							array('value' => 'state', 'text' => __('State')),
							array('value' => 'zip', 'text' => __('Zip')),
							array('value' => 'country', 'text' => __('Country'))
						),
						'options_blank'					=>	__('Select...', 'ws-form-authorize-accept'),
					),

					// Message - Method
					'action_' . $this->id . '_message_method'	=> array(

						'label'						=>	__('Error Messages', 'ws-form-post'),
						'type'						=>	'select',
						'help'						=>	__('Where should error messages be added?', 'ws-form-post'),
						'options'					=>	array(

							array('value' => 'before', 'text' => __('Before Form', 'ws-form-post')),
							array('value' => 'after', 'text' => __('After Form', 'ws-form-post'))
						),
						'default'					=>	'before',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Duration
					'action_' . $this->id . '_message_duration'	=> array(

						'label'						=>	__('Error Message Duration (ms)', 'ws-form-post'),
						'type'						=>	'number',
						'help'						=>	__('Duration in milliseconds error message shown.', 'ws-form-post'),
						'default'					=>	'4000',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Clear other messages
					'action_' . $this->id . '_message_clear'	=> array(

						'label'						=>	__('Clear Other Messages On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Clear any other messages when error message shown?', 'ws-form'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Scroll to top
					'action_' . $this->id . '_message_scroll_top'	=> array(

						'label'						=>	__('Scroll To Top On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Scroll to top of page when error message shown?', 'ws-form'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Form Hide
					'action_' . $this->id . '_message_form_hide'	=> array(

						'label'						=>	__('Hide Form On Error', 'ws-form-post'),
						'type'						=>	'checkbox',
						'help'						=>	__('Hide form when error message shown?', 'ws-form-post'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-authorize-accept'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-authorize-accept'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Authorize Accept add-on for WS Form PRO license key here.', 'ws-form-authorize-accept'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-authorize-accept'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_environment'	=>	array(

							'label'		=>	__('Environment', 'ws-form-authorize-accept'),
							'type'		=>	'select',
							'help'		=>	__('Select which Authorize environment to use.', 'ws-form'),
							'default'	=>	'sandbox',
							'options'	=>	array(

								'sandbox'		=>	array('text' => __('Sandbox', 'ws-form')),
								'production'	=>	array('text' => __('Production', 'ws-form')),
							)				
						),

						'action_' . $this->id . '_sandbox_api_login_id'	=>	array(

							'label'			=>	__('Sandbox API Login ID', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; API Credentials and Keys.', 'ws-form-authorize-accept')
						),

						'action_' . $this->id . '_sandbox_transaction_key'	=>	array(

							'label'			=>	__('Sandbox API Transaction Key', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; API Credentials and Keys.', 'ws-form-authorize-accept')
						),

						'action_' . $this->id . '_sandbox_client_key'	=>	array(

							'label'			=>	__('Sandbox Client Key', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; Manage Public Client Key.', 'ws-form-authorize-accept')
						),

						'action_' . $this->id . '_production_api_login_id'	=>	array(

							'label'			=>	__('Production API Login ID', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; API Credentials and Keys.', 'ws-form-authorize-accept')
						),

						'action_' . $this->id . '_production_transaction_key'	=>	array(

							'label'			=>	__('Production API Transaction Key', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; API Credentials and Keys.', 'ws-form-authorize-accept')
						),

						'action_' . $this->id . '_production_client_key'	=>	array(

							'label'			=>	__('Production Client Key', 'ws-form-authorize-accept'),
							'type'			=>	'text',
							'help'			=>	__('Found in your Authorize account under Account &gt; Settings &gt; Security Settings &gt; General Security Settings &gt; Manage Public Client Key.', 'ws-form-authorize-accept')
						),
					)
				);

				return $options;
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->environment  			= WS_Form_Common::option_get('action_' . $this->id . '_environment', '');

				$this->sandbox_api_login_id 	= WS_Form_Common::option_get('action_' . $this->id . '_sandbox_api_login_id', '');
				$this->sandbox_transaction_key 	= WS_Form_Common::option_get('action_' . $this->id . '_sandbox_transaction_key', '');
				$this->sandbox_client_key  		= WS_Form_Common::option_get('action_' . $this->id . '_sandbox_client_key', '');

				$this->production_api_login_id 	= WS_Form_Common::option_get('action_' . $this->id . '_production_api_login_id', '');
				$this->production_transaction_key 	= WS_Form_Common::option_get('action_' . $this->id . '_production_transaction_key', '');
				$this->production_client_key 	= WS_Form_Common::option_get('action_' . $this->id . '_production_client_key', '');

				// Check configuration
				if(
					(empty($this->sandbox_api_login_id) || empty($this->sandbox_transaction_key) || empty($this->sandbox_client_key)) &&
					(empty($this->production_api_login_id) || empty($this->production_transaction_key) || empty($this->production_client_key))
				) {

					$this->configured = false;
				}

				return $this->configured;
			}
		}
	});