<?php

	/**
	 * @link              https://wsform.com/knowledgebase/aweber/
	 * @since             1.0.0
	 * @package           WS_Form_AWeber
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - AWeber
	 * Plugin URI:        https://wsform.com/knowledgebase/aweber/
	 * Description:       AWeber add-on for WS Form PRO
	 * Version:           1.0.3
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-aweber-v1
	 */

	Class WS_Form_Add_On_AWeber_V1 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.0.3';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_AWeber_V1();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if(!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-aweber-v1'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_aweber = new WS_Form_Add_On_AWeber_V1();	

	register_activation_hook(__FILE__, array($wsf_add_on_aweber, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_AWeber_V1 extends WS_Form_Action {

			public $id = 'aweberv1';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Subscribers';

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint;
			private $api_access_token;
			private $api_access_secret;
			private $account_id = '';

			public $list_id = false;
			public $ad_tracking;
			public $misc_notes;
			public $opt_in_field;
			public $field_mapping;
			public $tag_mapping;

			// Constants
			const CONSUMER_KEY = 'AkYojPmB4ZELWyQtHd1NUXWh';
			const CONSUMER_SECRET = 'UHZoH5DK0YOt6i03w3lsu5PhQ02fQu3QskgzhKAD';
			const WS_FORM_LICENSE_ITEM_ID = 3354;
			const WS_FORM_LICENSE_NAME = 'AWeber add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.3';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';
			const LISTS_FETCH_COUNT = 100;
			const EMAIL_MERGE_FIELD = 'email';
			const NAME_MERGE_FIELD = 'name';

			public function __construct() {

				// Set label
				$this->label = __('AWeber', 'ws-form-aweber-v1');

				// Set label for actions pull down
				$this->label_action = __('Add to an AWeber List', 'ws-form-aweber-v1');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				$access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($access_token === false) { return; }

				$access_secret = WS_Form_Common::get_query_var($this->id . '_access_secret', false);
				if($access_secret === false) { return; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $access_token);
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_secret', $access_secret);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-aweber-v1') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-aweber-v1') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-aweber-v1') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/aweber-v1';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-aweber-v1'));
						}

						break;

					case 'account_list_action_' . $this->id :

						$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="account_list_refresh" value="' . __('Refresh', 'ws-form-aweber-v1') . '" />';
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;

							case 'account_list_refresh' :

								self::account_list_refresh();
								break;						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_secret', '');
				WS_Form_Common::option_set('action_' . $this->id . '_account_list', false);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your AWeber account <a href="%s">here</a>.', 'ws-form-aweber-v1'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-aweber-v1'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$custom_fields = array();
				$data = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// AWeber cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($api_field_string) {

						case self::EMAIL_MERGE_FIELD :
						case self::NAME_MERGE_FIELD :

							$data[$api_field_string] = $submit_value;
							break;

						default :

							$custom_fields[$api_field_string] = $submit_value;
							break;
					}
				}

				// Check email
				if(
					!isset($data[self::EMAIL_MERGE_FIELD]) ||
					(!filter_var($data[self::EMAIL_MERGE_FIELD], FILTER_VALIDATE_EMAIL))
				) {

					return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-aweber-v1'), $email_address));
				}

				// Process tags
				$tags = array();
				$tags_push = false;
				foreach($this->tag_mapping as $field_map) {

					// Get field ID
					$field_id = $field_map['ws_form_field'];
					if($field_id == '') { continue; }

					$tags_push = true;

					// Read submit meta
					$tags_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($tags_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($tags_array)) { $tags_array = array($tags_array); }

					// Process each tag ready for API format
					foreach($tags_array as $tag) {

						if($tag == '') { continue; }

						$tags[] = $tag;
					}
				}

				// Build post data
				if(count($custom_fields) > 0) { $data['custom_fields'] = json_encode($custom_fields); }
				if($tags_push) { $data['tags'] = json_encode($tags); }

				// Add ad_tracking
				if(strlen($this->ad_tracking) > 20) { $this->ad_tracking = substr($this->ad_tracking, 0, 20); }
				$data['ad_tracking'] = $this->ad_tracking;

				// Tracking data - Remote IP
				$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
				if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

					$data['ip_address'] = $tracking_remote_ip;
				}

				// Add misc_notes
				if(strlen($this->misc_notes) > 60) { $this->misc_notes = substr($this->misc_notes, 0, 60); }
				$data['misc_notes'] = $this->misc_notes;

				// Process according to status
				$do_update = false;
				if($this->status == 'subscribed') {

					// Params
					$query_params = array(

						'ws.op'		=> 'create'
					);

					// Get oAuth request
					$oauth_request = self::get_oauth_request('POST', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', $data, $query_params);

					// Get lists
					$post_data = array_merge($oauth_request, $data);

					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers?ws.op=create', 'POST', $post_data, array(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

					// Process response
					switch($api_response['http_code']) {

						case 201 :

							self::success(__('Successfully added to list: ', 'ws-form-aweber-v1') . $this->list_id);
							return true;

						default :

							$do_update = true;
					}
				}

				if(
					($this->status == 'unsubscribed') ||
					$do_update
				) {

					// Params
					$query_params = array(

						'ws.op'		=> 'find',
						'email'		=> $data[self::EMAIL_MERGE_FIELD],
					);

					// Get oAuth request
					$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', array(), $query_params);

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', 'GET', $oauth_request);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }

					if(!isset($api_response_decoded->entries)) { return false; }
					if(!isset($api_response_decoded->entries[0])) { return false; }

					$api_response_decoded = $api_response_decoded->entries[0];

					$subscriber_id = $api_response_decoded->id;

					// Build update tag element
					$old_tags = $api_response_decoded->tags;
					$tags_removed = array_diff($old_tags,$tags);
					$tags_added = array_diff($tags,$old_tags);
					$tags_new = array();
					if(count($tags_added) > 0) { $tags_new['add'] = array_values($tags_added); }
					if(count($tags_removed) > 0) { $tags_new['remove'] = array_values($tags_removed); }
					if(count($tags_new) > 0) {

						$data['tags'] = json_encode($tags_new);

					} else {

						unset($data['tags']);
					}

					// Set status
					$data['status'] = $this->status;

					// Remove IP
					if(isset($data['ip_address'])) { unset($data['ip_address']); }

					// Get oAuth request
					$oauth_request = self::get_oauth_request('PATCH', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers/' . $subscriber_id, $data);

					// Get lists
					$post_data = array_merge($oauth_request, $data);

					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers/' . $subscriber_id, 'PATCH', $post_data, array(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

					// Process response
					switch($api_response['http_code']) {

						case 209 :

							self::success(__('Successfully updated.', 'ws-form-aweber-v1'));
							return true;

						default :

							// Get invalid merge fields
							return self::error(self::get_api_error_detail($api_response['response']));
							return false;
					}
				}

				if($this->status == 'delete') {

					// Params
					$query_params = array(

						'ws.op'		=> 'find',
						'email'		=> $data[self::EMAIL_MERGE_FIELD],
					);

					// Get oAuth request
					$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', array(), $query_params);

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', 'GET', $oauth_request);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }

					if(!isset($api_response_decoded->entries)) { return false; }
					if(!isset($api_response_decoded->entries[0])) { return false; }

					$api_response_decoded = $api_response_decoded->entries[0];

					$subscriber_id = $api_response_decoded->id;

					// Get oAuth request
					$oauth_request = self::get_oauth_request('DELETE', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers/' . $subscriber_id);

					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers/' . $subscriber_id, 'DELETE', $oauth_request, array(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

					// Process response
					switch($api_response['http_code']) {

						case 200 :

							self::success(__('Successfully deleted from list: ', 'ws-form-aweber-v1') . $this->list_id);
							return true;

						default :

							// Get invalid merge fields
							return self::error(self::get_api_error_detail($api_response['response']));
							return false;
					}
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-aweber-v1')); }

				// Params
				$query_params = array(

					'ws.op'		=> 'find',
					'email'		=> $user_email,
				);

				// Get oAuth request
				$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', array(), $query_params);

				// Get lists
				$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/subscribers', 'GET', $oauth_request);

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }

				if(!isset($api_response_decoded->entries)) { return false; }
				if(!isset($api_response_decoded->entries[0])) { return false; }

				$api_response_decoded = $api_response_decoded->entries[0];

				// Process fields
				$fields_return = array(

					self::NAME_MERGE_FIELD =>	parent::get_object_value($api_response_decoded, self::NAME_MERGE_FIELD),
					self::EMAIL_MERGE_FIELD =>	parent::get_object_value($api_response_decoded, self::EMAIL_MERGE_FIELD),
				);

				$custom_fields = parent::get_object_value($api_response_decoded, 'custom_fields');
				if($custom_fields !== false) {

					foreach($custom_fields as $id => $value) {

						$fields_return[$id] = (string) $value;
					}
				}

				// Process tags
				$tags_return = array();

				$tags = parent::get_object_value($api_response_decoded, 'tags');
				if($tags !== false) {

					foreach($tags as $tag) {

						$tags_return[$tag] = true;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => $tags_return);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists_' . $this->account_id);

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$query_params = array(

							'ws.start'		=> $total_lists_retrieved,
							'ws.size'		=> self::LISTS_FETCH_COUNT
						);

						// Get oAuth request
						$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists', array(), $query_params);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists', 'GET', $oauth_request);

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }
						if(!isset($api_response_decoded->entries)) { return self::error(__('Lists not found in API response data', 'ws-form-aweber-v1')); }

						// Get total field count
						if(isset($api_response_decoded->total_size)) {

							$total_lists = $api_response_decoded->total_size;
						}

						// Increment read counter
						$total_lists_retrieved += count($api_response_decoded->entries);

						foreach($api_response_decoded->entries as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	false, 
								'record_count' => 	parent::get_object_value($list, 'total_subscribed_subscribers')
							);
						}

					} while ($total_lists_retrieved < $total_lists);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists_' . $this->account_id, $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get oAuth request
					$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id, 'GET', $oauth_request);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Process fields
					$sort_index = 0;

					// Fixed fields
					$fields = array(

						(object) array('id' => self::EMAIL_MERGE_FIELD, 'name' => 'Email', 'type' => 'email', 'required' => true),
						(object) array('id' => self::NAME_MERGE_FIELD, 'name' => 'Name', 'type' => 'text', 'required' => false),
					);

					// Process fields
					$sort_index_offset = 0;
					foreach($fields as $field) {

						$type = parent::get_object_value($field, 'type');

						$list_field = array(

							'id' => 			parent::get_object_value($field, 'id'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$type,
							'required' => 		parent::get_object_value($field, 'required'), 
							'default_value' => 	'',
							'pattern' => 		false,
							'placeholder' => 	false,
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		true,
							'meta' => 			false
						);

						$list_fields[] = $list_field;

						$sort_index++;
					}

					// API custom fields
					$total_fields = 0;
					$total_fields_retrieved = 0;
					$api_list_fields = array();

					do {

						// Params
						$query_params = array(

							'ws.start'		=> $total_fields_retrieved,
							'ws.size'		=> self::LISTS_FETCH_COUNT
						);

						// Get oAuth request
						$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/custom_fields', array(), $query_params);

						// Get list fields
						$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/custom_fields', 'GET', $oauth_request);

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }
						if(!isset($api_response_decoded->entries)) { return self::error(__('Lists not found in API response data', 'ws-form-aweber-v1')); }

						// Get total field count
						if(isset($api_response_decoded->total_size)) {

							$total_fields = $api_response_decoded->total_size;
						}

						// Get the fields
						$api_list_fields_read = $api_response_decoded->entries;

						// Increment read counter
						$total_fields_retrieved += count($api_list_fields_read);

						$api_list_fields = array_merge($api_list_fields, $api_list_fields_read);

					} while ($total_fields_retrieved < $total_fields);

					// Custom fields
					foreach($api_list_fields as $field) {

						$list_field = array(

							'id' => 			parent::get_object_value($field, 'name'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			'text',
							'action_type' =>	'text',
							'required' => 		false,
							'default_value' => 	'',
							'pattern' => 		false,
							'placeholder' => 	false,
							'input_mask' => 	false,
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		true
						);

						$list_fields[] = $list_field;

						$sort_index++;
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'recaptcha' => array(

						'type'			=>	'recaptcha',
						'label'			=>	__('reCAPTCHA', 'ws-form')
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_tag_mapping'		=>	'tag_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				$tag_categories = array(

					array(

						'id' => 		1,
						'label' => 		__('Tags', 'ws-form-aweber-v1'), 
						'sort_index' =>	1,
						'type' => 		'checkbox'
					)
				);

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = false, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id);

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Get oAuth request
					$oauth_request = self::get_oauth_request('GET', $this->api_endpoint . 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/tags');

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'accounts/' . $this->account_id . '/lists/' . $this->list_id . '/tags', 'GET', $oauth_request);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process tag data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }

					$sort_index = 1;

					// Sort alphabetically
					sort($api_response_decoded);

					// Process tags
					foreach($api_response_decoded as $tag) {

						$tags[] = array(

							'id' => 			$tag,
							'label' => 			$tag,
							'record_count' =>	false,
							'sort_index' =>		$sort_index
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id, $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-aweber-v1'); }

				if(!isset($api_response_decoded->error)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-aweber-v1'); }

				if(!isset($api_response_decoded->error->message)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-aweber-v1'); }

				return $api_response_decoded->error->message;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_status',
						'action_' . $this->id . '_ad_tracking',
						'action_' . $this->id . '_misc_notes',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_tag_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-aweber-v1') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-aweber-v1')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('AWeber List', 'ws-form-aweber-v1'),
						'type'						=>	'select',
						'help'						=>	__('Select the AWeber list to associate this form with.', 'ws-form-aweber-v1'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-aweber-v1'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Status
					'action_' . $this->id . '_status'	=> array(

						'label'						=>	__('Status', 'ws-form-aweber-v1'),
						'type'						=>	'select',
						'help'						=>	__("Set subscriber's status.", 'ws-form-aweber-v1'),
						'options'					=>	array(

							array('value' => 'subscribed', 'text' => __('Subscribe', 'ws-form-aweber-v1')),
							array('value' => 'unsubscribed', 'text' => __('Unsubscribe', 'ws-form-aweber-v1')),
							array('value' => 'delete', 'text' => __('Delete', 'ws-form-aweber-v1'))
						),
						'default'					=>	'subscribed'
					),

					// Ad tracking
					'action_' . $this->id . '_ad_tracking'	=> array(

						'label'							=>	__('Ad Tracking', 'ws-form-aweber-v1'),
						'type'							=>	'text',
						'default'						=>	'',
						'help'							=>	__('The customer ad tracking field (Maximum characters: 20)', 'ws-form-aweber-v1'),
						'select_list'					=>	true
					),

					// Miscellaneous notes
					'action_' . $this->id . '_misc_notes'	=> array(

						'label'							=>	__('Miscellaneous Notes', 'ws-form-aweber-v1'),
						'type'							=>	'textarea',
						'default'						=>	'',
						'help'							=>	__('Notes added to each subscriber (Maximum characters: 60)', 'ws-form-aweber-v1'),
						'select_list'					=>	true
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-aweber-v1'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-aweber-v1'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-aweber-v1')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-aweber-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to AWeber fields.', 'ws-form-aweber-v1'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-aweber-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to AWeber tags.', 'ws-form-aweber-v1'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field_choice'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('AWeber Field', 'ws-form-aweber-v1'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-aweber-v1'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-aweber-v1'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-aweber-v1'),
							'type'		=>	'text',
							'help'		=>	__('Enter your AWeber add-on for WS Form PRO license key here.', 'ws-form-aweber-v1'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-aweber-v1'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('AWeber Status', 'ws-form-aweber-v1'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your AWeber account, or <a href=\"https://www.aweber.com/order.htm\" target=\"blank\">create a new account</a>.", 'ws-form-aweber-v1'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				// Get account list
				$account_list = WS_Form_Common::option_get('action_' . $this->id . '_account_list');

				if($this->configured && ($account_list === false)) {

					self::account_list_refresh();
				}

				$options['action_' . $this->id]['fields']['action_' . $this->id . '_account_id'] = WS_Form_Common::option_get('action_' . $this->id . '_account_list');

				return $options;
			}

			// Get payload
			public function get_oauth_request($method, $url, $post_params = array(), $query_params = array()) {

				$oauth_request_base = array(

					'oauth_consumer_key' 		=>	self::CONSUMER_KEY,
					'oauth_nonce'				=>	self::get_nonce(),
					'oauth_signature_method'	=>	'HMAC-SHA1',
					'oauth_timestamp'			=>	time(),
					'oauth_token'				=>	$this->api_access_token,
					'oauth_version'				=>	'1.0'
				);

				$oauth_request_full = $oauth_request_base;

				$oauth_request_full = array_merge($oauth_request_full, $post_params);

				$oauth_request_full = array_merge($oauth_request_full, $query_params);

				$oauth_signature = self::oauth_signature($method, $url, $oauth_request_full);

				$oauth_request_base['oauth_signature'] = $oauth_signature;
				$oauth_request_full['oauth_signature'] = $oauth_signature;

				return ($method == 'GET') ? $oauth_request_full : $oauth_request_base;
			}

			// Get nonce
			public function get_nonce() {

				return base64_encode(bin2hex(openssl_random_pseudo_bytes(16)));
			}

			// oAuth signature
			public function oauth_signature($method, $url, $oauth_request) {

				// Sort keys
				ksort($oauth_request);

				// RFC 3986 encode
				$oauth_signature_string = http_build_query($oauth_request, null, '&', PHP_QUERY_RFC3986);

				// Build base string
				$oauth_signature_string = strtoupper($method) . '&' . rawurlencode($url) . '&' . rawurlencode($oauth_signature_string);

				// Build signing key
				$oauth_signing_key = rawurlencode(self::CONSUMER_SECRET) . '&' . rawurlencode($this->api_access_secret);

				// Build signature
				$oauth_signature = base64_encode(hash_hmac('sha1', $oauth_signature_string, $oauth_signing_key, true));

				return $oauth_signature;
			}

			// Update account list
			public function account_list_refresh() {

				// Get oauth_payload
				$oauth_payload = self::get_oauth_request('GET', $this->api_endpoint . 'accounts');

				// Request accounts
				$api_response = parent::api_call($this->api_endpoint, 'accounts', 'GET', $oauth_payload);

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-aweber-v1')); }
				if(!isset($api_response_decoded->entries)) { return self::error(__('Account entries not found', 'ws-form-aweber-v1')); }

				$account_id_array = array();

				if(count($api_response_decoded->entries) == 0) {

					// No accounts returned
					$this->account_id = '';
					WS_Form_Common::option_set('action_' . $this->id . '_account_id', $this->account_id);

				} else {

					// Check to see if the current account ID can be found
					$account_id_found = false;
					foreach($api_response_decoded->entries as $account) {

						if($account->id == $this->account_id) { $account_id_found = true; }
					}

					// If account ID not found, clear it
					if(!$account_id_found) { $this->account_id = ''; }

					// Set the default 
					foreach($api_response_decoded->entries as $account) {

						$account_id_array[$account->id] = array('text' => $account->id);

						// Client ID not set, so set it
						if($this->account_id === '') {

							$this->account_id = $account->id;
							WS_Form_Common::option_set('action_' . $this->id . '_account_id', $this->account_id);
						}
					}

					// Clients found
					$account_list = array(

						'label'			=>	__('AWeber Account ID', 'ws-form-aweber-v1'),
						'type'			=>	'select',
						'options'		=>	$account_id_array,
						'default'		=>	$this->account_id,
						'help'			=>	__('Select the AWeber account to use.', 'ws-form-aweber-v1'),
						'button'		=>	'account_list_action_' . $this->id
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_account_list', $account_list);
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->ad_tracking = parent::get_config($config, 'action_' . $this->id . '_ad_tracking');
				$this->status = parent::get_config($config, 'action_' . $this->id . '_status');
				$this->misc_notes = parent::get_config($config, 'action_' . $this->id . '_misc_notes');
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token 	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_secret 	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_secret', '');
				$this->account_id  			= WS_Form_Common::option_get('action_' . $this->id . '_account_id', '');

				// Set endpoint URL
				$this->api_endpoint = 'https://api.aweber.com/1.0/';

				// Build endpoint using API key
				if(empty($this->api_access_token) || empty($this->api_access_secret)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}
		
			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(47.000000, 65.000000)">';
				$svg_logo .= '<path d="M20.512 0h17.441a1.154 1.154 0 0 0 .275.144 11.673 11.673 0 0 1 9.57 9.282c.051.21.134.414.202.62v28.047a7.997 7.997 0 0 0-.273.822A11.747 11.747 0 0 1 36.46 48H21.134c-.765-2.102-1.604-4.19-2.289-6.327a60.07 60.07 0 0 1-2.83-19.973 62.373 62.373 0 0 1 1.53-12.284A65.61 65.61 0 0 1 20.511 0zm-1.327 24.332c0 .256-.006.511.002.767.008.278.033.556.045.834a32.082 32.082 0 0 0 3.929 14.227 50.256 50.256 0 0 0 2.855 4.446 1.96 1.96 0 0 0 2.783.431 2.007 2.007 0 0 0 .514-2.726c-.135-.216-.29-.42-.438-.628a29.158 29.158 0 0 1-5.46-20.509 28.79 28.79 0 0 1 5.042-13.622c.302-.44.644-.855.92-1.31a1.976 1.976 0 0 0-.382-2.48 1.925 1.925 0 0 0-2.435-.23 3.489 3.489 0 0 0-.77.795 33.18 33.18 0 0 0-6.605 20.005zm10.216.449c.044.576.08 1.6.212 2.61a24.38 24.38 0 0 0 .536 2.729 23.096 23.096 0 0 0 4.336 8.72 1.36 1.36 0 0 0 1.92.275 1.336 1.336 0 0 0 .276-1.86c-.151-.233-.332-.448-.498-.671a20.624 20.624 0 0 1-.21-24.22c.243-.34.521-.657.745-1.009a1.342 1.342 0 0 0-1.025-2.068 1.459 1.459 0 0 0-1.35.701 23.654 23.654 0 0 0-4.942 14.793z" fill="#0179c2"/><path d="M12.14 0a66.7 66.7 0 0 0-3.06 11.696 67.794 67.794 0 0 0-.958 11.574 68.775 68.775 0 0 0 4.355 24.044c.065.174.13.348.192.523.007.02-.004.045-.013.125a10.258 10.258 0 0 1-3.812-.3A11.64 11.64 0 0 1 .257 38.77 5.92 5.92 0 0 0 0 38.093V9.907A4.711 4.711 0 0 0 .243 9.3 11.667 11.667 0 0 1 9.703.152 1.86 1.86 0 0 0 10.047 0z" fill="#0179c1"/>';
				$svg_logo .= '</g>';

				return $svg_logo;
			}
		}
	});