<?php

	/**
	 * @link              https://wsform.com/knowledgebase/campaign-monitor/
	 * @since             1.0.0
	 * @package           WS_Form_Campaign_Monitor_V3
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Campaign Monitor
	 * Plugin URI:        https://wsform.com/knowledgebase/campaign-monitor/
	 * Description:       Campaign Monitor add-on for WS Form PRO
	 * Version:           1.0.3
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-campaign-monitor-v3
	 */

	Class WS_Form_Add_On_Campaign_Monitor_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.0.7';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Campaign_Monitor_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-campaign-monitor-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_campaign_monitor_v3 = new WS_Form_Add_On_Campaign_Monitor_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_campaign_monitor_v3, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Campaign_Monitor_V3 extends WS_Form_Action {

			public $id = 'campaignmonitorv3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Subscribers';

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_refresh_token;
			private $client_id = '';

			public $list_id = false;
			public $resubscribe = false;
			public $restart_autoresponders = false;
			public $consent_to_track = false;
			public $opt_in_field = false;
			public $name_mask = false;
			public $field_mapping = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 2697;
			const WS_FORM_LICENSE_NAME = 'Campaign Monitor add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.3';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const SLACK_FIELD_SHORT_THRESHOLD = 20;
			const EMAIL_MERGE_FIELD = 'EmailAddress';

			public function __construct() {

				// Set label
				$this->label = __('Campaign Monitor', 'ws-form-campaign-monitor-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to Campaign Monitor', 'ws-form-campaign-monitor-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save access token expiry
				$this->api_access_token_expiry = time() + intval(WS_Form_Common::get_query_var($this->id . '_expires_in', false)) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = '/campaign-monitor-v3?refresh_token=' . rawurlencode($this->api_refresh_token);
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

					// Save access token
					$this->api_access_token = parent::get_object_value($api_response_decoded, $this->id . '_access_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save refresh token
					$this->api_refresh_token = parent::get_object_value($api_response_decoded, $this->id . '_refresh_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

					// Save access token expiry
					$this->api_access_token_expiry = time() + intval(parent::get_object_value($api_response_decoded, $this->id . '_expires_in')) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-campaign-monitor-v3')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-campaign-monitor-v3') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-campaign-monitor-v3') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-campaign-monitor-v3') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/campaign-monitor-v3';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-campaign-monitor-v3'));
						}

						break;

					case 'client_list_action_' . $this->id :

						$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="client_list_refresh" value="' . __('Refresh', 'ws-form-campaign-monitor-v3') . '" />';

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;

							case 'client_list_refresh' :

								self::client_list_refresh();
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_client_list', false);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Campaign Monitor account <a href="%s">here</a>.', 'ws-form-campaign-monitor-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check access token
				self::access_token_check();

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-campaign-monitor-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$custom_fields = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					$custom_field_key = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$custom_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Campaign Monitor cannot accept arrays
					if(is_array($custom_field_value)) { $custom_field_value = implode(',', $custom_field_value); }

					switch($custom_field_key) {

						case self::EMAIL_MERGE_FIELD :

							$email_address = $custom_field_value;
							break;
					}

					// Normal merge field
					$custom_fields[] = array('Key' => $custom_field_key, 'Value' => $custom_field_value);
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-campaign-monitor-v3')); }

				// Get name
				$name = WS_Form_Common::parse_variables_process($this->name_mask, $form, $submit, 'text/plain');

				// Build post data
				$data = array(

					'EmailAddress'								=>	$email_address,
					'Name'										=>	$name,
					'Resubscribe'								=>	($this->resubscribe == 'on'),
					'RestartSubscriptionBasedAutoresponders'	=>	($this->restart_autoresponders == 'on'),
					'ConsentToTrack'							=>	($this->consent_to_track == 'on') ? 'Yes' : 'No',
				);
				if(count($custom_fields) > 0) { $data['CustomFields'] = $custom_fields; }

				// Data for API call
				$data_json = json_encode($data);

				switch($this->method) {

					case 'update' :

						$api_path = 'subscribers/' . $this->list_id . '.json?email=' . rawurlencode($email_address);
						$api_method = 'PUT';
						break;						

					case 'unsubscribe' :

						$api_path = 'subscribers/' . $this->list_id . '/unsubscribe.json';
						$api_method = 'POST';
						break;

					case 'delete' :

						$api_path = 'subscribers/' . $this->list_id . '.json?email=' . rawurlencode($email_address);
						$api_method = 'DELETE';
						break;						

					default :

						$api_path = 'subscribers/' . $this->list_id . '.json';
						$api_method = 'POST';
				}

				// Call API
				$api_response = parent::api_call($this->api_endpoint, $api_path, $api_method, $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 200 :
					case 201 :

						self::success(__('Successfully processed list: ', 'ws-form-campaign-monitor-v3') . $this->list_id);
						return true;

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-campaign-monitor-v3')); }

				// Check access token
				self::access_token_check();

				// Get member
				$api_response = parent::api_call($this->api_endpoint, 'subscribers/' . $this->list_id . '.json?email=' . rawurlencode($user_email) . '&includetrackingpreference=true', 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

				// Process fields
				$fields_return = array(

					self::EMAIL_MERGE_FIELD	=>	parent::get_object_value($api_response_decoded, self::EMAIL_MERGE_FIELD),
					'Name'					=>	parent::get_object_value($api_response_decoded, 'Name'),
				);

				$custom_fields = parent::get_object_value($api_response_decoded, 'CustomFields');
				if($custom_fields !== false) {

					foreach($custom_fields as $custom_field) {

						$fields_return[$custom_field->Key] = (string) $custom_field->Value;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => array());

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists_' . $this->client_id);

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'clients/' . $this->client_id . '/lists.json', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

					foreach($api_response_decoded as $list) {

						$lists[] = array(

							'id' => 			parent::get_object_value($list, 'ListID'), 
							'label' => 			parent::get_object_value($list, 'Name'), 
							'field_count' => 	false, 
							'record_count' => 	false
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists_' . $this->client_id, $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->client_id . '_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '.json', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(sprintf(__('HTTP Status Code: %s', 'ws-form-campaign-monitor-v3'), $api_response['http_code'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_decoded, 'Title')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->client_id . '_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->client_id . '_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Initialize
					$api_list_fields = array();
					$total_fields;
					$total_fields_retrieved = 0;

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/customfields.json', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

					// Process fields
					$sort_index_offset = 2;

					foreach($api_response_decoded as $field) {

						$action_type = parent::get_object_value($field, 'DataType');
						$action_options = parent::get_object_value($field, 'FieldOptions');

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						// Strip [ and ] from key
						$id = parent::get_object_value($field, 'Key');
						$id = str_replace('[', '', $id);
						$id = str_replace(']', '', $id);

						$list_fields[] = array(

							'id' => 			$id,
							'label' => 			parent::get_object_value($field, 'FieldName'), 
							'label_field' => 	parent::get_object_value($field, 'FieldName'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		false,	// API does not return this parameter 
							'default_value' => 	'',
							'pattern' => 		'',
							'placeholder' => 	'',
							'input_mask' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index_offset,
							'visible' =>		true,
							'meta' => 			self::action_field_to_ws_form_meta_keys($field)
						);

						$sort_index_offset++;
					}

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_MERGE_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Email', 
						'label_field' => 	'Email', 
						'label_sub' =>		false,
						'type' => 			'email', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Add name
					$list_fields[] = array(

						'id' => 			'Name', 
						'id_sub' =>			false,
						'label' => 			'Name', 
						'label_field' => 	'Name', 
						'label_sub' =>		false,
						'type' => 			'text', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	1,
						'visible' =>		true
					);

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->client_id . '_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
							'action_' . $this->id . '_name_mask'		=>	'#field(#Name)'
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->Message);

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_method',
						'action_' . $this->id . '_resubscribe',
						'action_' . $this->id . '_restart_autoresponders',
						'action_' . $this->id . '_consent_to_track',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_name_mask',
						'action_' . $this->id . '_field_mapping',
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-campaign-monitor-v3'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-campaign-monitor-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('Campaign Monitor List', 'ws-form-campaign-monitor-v3'),
						'type'							=>	'select',
						'help'							=>	__('Which Campaign Monitor list should this subscriber be added to?', 'ws-form-campaign-monitor-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-campaign-monitor-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Method
					'action_' . $this->id . '_method'	=> array(

						'label'						=>	__('Method', 'ws-form-campaign-monitor-v3'),
						'type'						=>	'select',
						'help'						=>	__('Campaign Monitor method.', 'ws-form-campaign-monitor-v3'),
						'options'					=>	array(

							array('value' => 'add', 'text' => __('Add', 'ws-form-campaign-monitor-v3')),
							array('value' => 'update', 'text' => __('Update', 'ws-form-campaign-monitor-v3')),
							array('value' => 'unsubscribe', 'text' => __('Unsubscribe', 'ws-form-campaign-monitor-v3')),
							array('value' => 'delete', 'text' => __('Delete', 'ws-form-campaign-monitor-v3')),
						),
						'default'					=>	'add'
					),

					// Resubscribe
					'action_' . $this->id . '_resubscribe'	=> array(

						'label'						=>	__('Resubscribe', 'ws-form-campaign-monitor-v3'),
						'type'						=>	'select',
						'help'						=>	__('If set to on and the subscriber is in an inactive state or has previously been unsubscribed or added to the suppression list, they will be re-added to the list.', 'ws-form-campaign-monitor-v3'),
						'options'					=>	array(

							array('value' => 'on', 'text' => __('On', 'ws-form-campaign-monitor-v3')),
							array('value' => '', 'text' => __('Off', 'ws-form-campaign-monitor-v3')),
						),
						'default'					=>	'on'
					),

					// Restart Subscription Based Autoresponders
					'action_' . $this->id . '_restart_autoresponders'	=> array(

						'label'						=>	__('Restart Autoresponders', 'ws-form-campaign-monitor-v3'),
						'type'						=>	'select',
						'help'						=>	__('Whether autoresponder sequences should be restarted if resubscribed.', 'ws-form-campaign-monitor-v3'),
						'options'					=>	array(

							array('value' => 'on', 'text' => __('On', 'ws-form-campaign-monitor-v3')),
							array('value' => '', 'text' => __('Off', 'ws-form-campaign-monitor-v3')),
						),
						'default'					=>	'on'
					),

					// Restart Subscription Based Autoresponders
					'action_' . $this->id . '_consent_to_track'	=> array(

						'label'						=>	__('Consent To Track', 'ws-form-campaign-monitor-v3'),
						'type'						=>	'select',
						'help'						=>	__('Whether the subscriber has given permission to have their email opens and clicks tracked.', 'ws-form-campaign-monitor-v3'),
						'options'					=>	array(

							array('value' => 'on', 'text' => __('On', 'ws-form-campaign-monitor-v3')),
							array('value' => '', 'text' => __('Off', 'ws-form-campaign-monitor-v3')),
						),
						'default'					=>	'on'
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-campaign-monitor-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-campaign-monitor-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-campaign-monitor-v3')
					),

					// List ID
					'action_' . $this->id . '_name_mask'	=> array(

						'label'							=>	__('Name Mask', 'ws-form-campaign-monitor-v3'),
						'type'							=>	'text',
						'help'							=>	__('Enter a mask for the name field.', 'ws-form-campaign-monitor-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-campaign-monitor-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Campaign Monitor merge fields.', 'ws-form-campaign-monitor-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Campaign Monitor Field', 'ws-form-campaign-monitor-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-campaign-monitor-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-campaign-monitor-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-campaign-monitor-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Campaign Monitor add-on for WS Form PRO license key here.', 'ws-form-campaign-monitor-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-campaign-monitor-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Campaign Monitor Status', 'ws-form-campaign-monitor-v3'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Campaign Monitor account, or <a href=\"https://www.campaignmonitor.com/signup\" target=\"blank\">create a new account</a>.", 'ws-form-campaign-monitor-v3'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				// Get client list
				$client_list = WS_Form_Common::option_get('action_' . $this->id . '_client_list');

				if($this->configured && ($client_list === false)) {

					self::client_list_refresh();
				}

				$options['action_' . $this->id]['fields']['action_' . $this->id . '_client_id'] = WS_Form_Common::option_get('action_' . $this->id . '_client_list');

				return $options;
			}

			// Update client list
			public function client_list_refresh() {

				// Check access token
				self::access_token_check();

				// Request clients
				$api_response = parent::api_call($this->api_endpoint, 'clients.json', 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-campaign-monitor-v3')); }

				$client_id_array = array();

				if(count($api_response_decoded) == 0) {

					// No clients returned
					$this->client_id = '';
					WS_Form_Common::option_set('action_' . $this->id . '_client_id', $this->client_id);

				} else {

					// Check to see if the current client ID can be found
					$client_id_found = false;
					foreach($api_response_decoded as $client) {

						if($client->ClientID == $this->client_id) { $client_id_found = true; }
					}

					// If client ID not found, clear it
					if(!$client_id_found) { $this->client_id = ''; }

					// Set the default 
					foreach($api_response_decoded as $client) {

						$client_id_array[$client->ClientID] = array('text' => $client->Name);

						// Client ID not set, so set it
						if($this->client_id === '') {

							$this->client_id = $client->ClientID;
							WS_Form_Common::option_set('action_' . $this->id . '_client_id', $this->client_id);
						}
					}

					// Clients found
					$client_list = array(

						'label'			=>	__('Campaign Monitor Client', 'ws-form-campaign-monitor-v3'),
						'type'			=>	'select',
						'options'		=>	$client_id_array,
						'default'		=>	$this->client_id,
						'help'			=>	__('Select the Campaign Monitor client to use.', 'ws-form-campaign-monitor-v3'),
						'button'		=>	'client_list_action_' . $this->id
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_client_list', $client_list);
				}
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'Text' : return 'text';
					case 'Number' : return 'number';
					case 'MultiSelectOne' : return 'select';
					case 'MultiSelectMany' : return 'checkbox';
					case 'Date' : return 'datetime';

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$type = parent::get_object_value($field, 'DataType');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids for radio and select
					case 'MultiSelectOne' :
					case 'MultiSelectMany' :

						switch($type) {

							case 'MultiSelectOne' : $meta_key = 'data_grid_select'; break;
							case 'MultiSelectMany' : $meta_key = 'data_grid_checkbox'; break;
						}

						// Get options
						$options = parent::get_object_value($field, 'FieldOptions');

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys[$meta_key])) { return false; }
						if(!isset($meta_keys[$meta_key]['default'])) { return false; }

						$meta = $meta_keys[$meta_key]['default'];

						// Build new rows
						$rows = array();
						$id = 1;
						$default_value = parent::get_object_value($field, 'default_value');
						if($default_value === false) { $default_value = ''; }
						foreach($options as $option) {

							$rows[] = array(

								'id'		=> $id,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array($option)
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						return(array($meta_key => $meta));

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->method = parent::get_config($config, 'action_' . $this->id . '_method');
				$this->resubscribe = parent::get_config($config, 'action_' . $this->id . '_resubscribe');
				$this->restart_autoresponders = parent::get_config($config, 'action_' . $this->id . '_restart_autoresponders');
				$this->consent_to_track = parent::get_config($config, 'action_' . $this->id . '_consent_to_track');
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->name_mask = parent::get_config($config, 'action_' . $this->id . '_name_mask');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry  = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');
				$this->client_id  				= WS_Form_Common::option_get('action_' . $this->id . '_client_id', '');

				// Set endpoint URL
				$this->api_endpoint = 'https://api.createsend.com/api/v3.2/';

				// Build endpoint using API key
				if(empty($this->api_access_token)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(43.000000, 69.000000)"><path d="M56.7.9c-.7-1-2.1-1.3-3.2-.6l-56 39.2c.4.6 1.1 1 1.9 1h55.4c1.3 0 2.3-1 2.3-2.3v-36c0-.4-.1-.9-.4-1.3M.7.4c-1-.7-2.5-.5-3.2.5-.3.4-.4.8-.4 1.3v36.1l26.7-21.8L.7.4z" fill="#7856FF"/></g>';

				return $svg_logo;
			}
		}
	});