<?php

/**
 * @link              https://wsform.com/knowledgebase/constant-contact/
 * @since             1.0.0
 * @package           WS_Form_Constant_Contact
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - Constant Contact
 * Plugin URI:        https://wsform.com/knowledgebase/constant-contact/
 * Description:       Constant Contact add-on for WS Form PRO
 * Version:           1.0.3
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-constant-contact-v2
 */

	Class WS_Form_Add_On_Constant_Contact_V2 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.0.2';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Constant_Contact_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-constant-contact-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_constant_contact = new WS_Form_Add_On_Constant_Contact_V2();	

	register_activation_hook(__FILE__, array($wsf_add_on_constant_contact, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Constant_Contact_V2 extends WS_Form_Action {

			public $id = 'constantcontactv2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint;
			private $api_access_token;
			public $list_id = false;
			public $action_by;
			public $opt_in_field;
			public $field_mapping;

			// Constants
			const API_KEY = '6b2csgzm5ntfafas5xm4bwfd';
			const EMAIL_MERGE_FIELD = 'email_addresses|email_address';
			const WS_FORM_LICENSE_ITEM_ID = 610;
			const WS_FORM_LICENSE_NAME = 'Constant Contact add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.3';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Constant Contact', 'ws-form-constant-contact-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to a Constant Contact List', 'ws-form-constant-contact-v2');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				$access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($access_token === false) { return; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $access_token);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-constant-contact-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-constant-contact-v2') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-constant-contact-v2') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/constant-contact-v2';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-constant-contact-v2'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
								break;
						}

						break;
				}
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Constant Contact account <a href="%s">here</a>.', 'ws-form-constant-contact-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-constant-contact-v2'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$api_fields = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];
					$api_field_array = explode('|', $api_field_string);

					$api_field = $api_field_array[0];
					$api_field_sub = (isset($api_field_array[1])) ? $api_field_array[1] : false;

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// Constant Contact cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($api_field_string) {

						case self::EMAIL_MERGE_FIELD :

							$email_address = $submit_value;
							break;
					}

					if($api_field_sub === false) {

						// Normal field
						$api_fields[$api_field] = $submit_value;

					} else {

						// Merge field with sub component (e.g. Address)
						if(!isset($api_fields[$api_field])) { $api_fields[$api_field] = array(); $api_fields[$api_field][0] = array(); }
						$api_fields[$api_field][0][$api_field_sub] = $submit_value;
					}
				}

				// Check user email address is valid
				if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-constant-contact-v2'), $email_address)); }

				// Parameters
				$params = array(

					'api_key'	=>	self::API_KEY,
					'email'		=>	$email_address
				);

				// Get contact
				$api_response = parent::api_call($this->api_endpoint, 'contacts/', 'GET', $params, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process response
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-constant-contact-v2')); }

				if(!isset($api_response_decoded->results)) { return self::error(__('No result set returned in API response', 'ws-form-constant-contact-v2')); }

				$results = $api_response_decoded->results;

				// Add list ID
				$api_fields['lists'] = array(array('id' => $this->list_id));

				// Build JSON for API request
				$data_json = json_encode($api_fields);

				// Build query string
				$query_string_array = array(

					'api_key'	=> self::API_KEY,
					'action_by' => $this->action_by
				);
				$query_string = http_build_query($query_string_array);

				if(count($results) == 0) {

					// Contact does not exists
					$api_response = parent::api_call($this->api_endpoint, 'contacts?' . $query_string, 'POST', $data_json, self::get_api_headers());

					// Process response
					switch($api_response['http_code']) {

						case 201 :

							self::success(__('Successfully added to list: ', 'ws-form-constant-contact-v2') . $this->list_id);
							return true;

						default :

							// Invalid request
							return self::error(self::get_api_error_detail($api_response['response']));
					}
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-constant-contact-v2')); }

				// Parameters
				$params = array(

					'api_key'	=>	self::API_KEY,
					'email'		=>	$user_email
				);

				// Get contact
				$api_response = parent::api_call($this->api_endpoint, 'contacts/', 'GET', $params, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-constant-contact-v2')); }

				if(!isset($api_response_decoded->results)) { return false; }
				if(!isset($api_response_decoded->results[0])) { return false; }

				$api_response_decoded = $api_response_decoded->results[0];

				// Process fields
				$fields_return = array(

					'status' =>	parent::get_object_value($api_response_decoded, 'status'),
				);

				$contact_fields = get_object_vars($api_response_decoded);

				foreach($contact_fields as $id => $value) {

					if(is_null($value)) { continue; }

					switch($id) {

						// Process subs
						case 'addresses' :

							if(is_array($value) && isset($value[0])) {

								$address = $value[0];

								$subs = get_object_vars($address);
								foreach($subs as $id_sub => $value_sub) {

									$fields_return[$id . '|' . $id_sub] = (string) $value_sub;
								}
							}

							break;

						case 'email_addresses' :

							if(is_array($value) && isset($value[0])) {

								$email = $value[0];

								$subs = get_object_vars($email);
								foreach($subs as $id_sub => $value_sub) {

									$fields_return[$id . '|' . $id_sub] = (string) $value_sub;
								}
							}

							break;

						// Process as normal string
						default :

							$fields_return[$id] = (string) $value;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Parameters
					$params = array(

						'api_key' =>	self::API_KEY
					);

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'lists', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-constant-contact-v2')); }
					if(!is_array($api_response_decoded)) { return self::error(__('Lists not found in API response data', 'ws-form-constant-contact-v2')); }

					foreach($api_response_decoded as $list) {

						$lists[] = array(

							'id' => 			parent::get_object_value($list, 'id'), 
							'label' => 			parent::get_object_value($list, 'name'), 
							'field_count' => 	false,
							'record_count' => 	parent::get_object_value($list, 'contact_count')
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Parameters
					$params = array(

						'api_key' =>	self::API_KEY
					);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-constant-contact-v2')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// We cannot get the fields from Constant Contact API V2, so we'll simulate it
					$fields = array(

						(object) array('id' => 'email_addresses', 'name' => 'Email', 'action_type' => 'email_addresses', 'type' => 'email', 'required' => true, 'display_order' => 1),
//						(object) array('id' => 'prefix_name', 'name' => 'Prefix', 'action_type' => 'text', 'type' => 'text', 'required' => false, 'display_order' => 2),
						(object) array('id' => 'first_name', 'name' => 'First Name', 'action_type' => 'text', 'type' => 'text', 'required' => false, 'display_order' => 3),
						(object) array('id' => 'last_name', 'name' => 'Last Name', 'action_type' => 'text', 'type' => 'text', 'required' => false, 'display_order' => 4),
						(object) array('id' => 'job_title', 'name' => 'Job Title', 'action_type' => 'text', 'type' => 'text', 'required' => false, 'display_order' => 5),
						(object) array('id' => 'company_name', 'name' => 'Company Name', 'action_type' => 'text', 'type' => 'text', 'required' => false, 'display_order' => 6),
						(object) array('id' => 'home_phone', 'name' => 'Phone (Home)', 'action_type' => 'tel', 'type' => 'tel', 'required' => false, 'display_order' => 7),
						(object) array('id' => 'work_phone', 'name' => 'Phone (Work)', 'action_type' => 'tel', 'type' => 'tel', 'required' => false, 'display_order' => 8),
						(object) array('id' => 'cell_phone', 'name' => 'Phone (Cell)', 'action_type' => 'tel', 'type' => 'tel', 'required' => false, 'display_order' => 9),
						(object) array('id' => 'fax', 'name' => 'Fax', 'type' => 'tel', 'action_type' => 'tel', 'required' => false, 'display_order' => 10),
						(object) array('id' => 'addresses', 'name' => 'Address', 'action_type' => 'addresses', 'type' => 'text', 'required' => false, 'display_order' => 11),
					);

					// Process fields
					$sort_index_offset = 0;
					foreach($fields as $field) {

						$type = parent::get_object_value($field, 'type');
						$action_type = parent::get_object_value($field, 'action_type');

						$list_field = array(

							'id' => 			parent::get_object_value($field, 'id'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		parent::get_object_value($field, 'required'), 
							'default_value' => 	parent::get_object_value($field, 'default_value'),
							'pattern' => 		self::action_field_to_ws_form_pattern($action_type),
							'placeholder' => 	self::action_field_to_ws_form_placeholder($action_type),
							'help' => 			parent::get_object_value($field, 'help_text'), 
							'sort_index' => 	intval(parent::get_object_value($field, 'display_order')) + $sort_index_offset,
							'visible' =>		true,
							'meta' => 			false
						);

						$sort_index_offset += self::list_field_explode($list_fields, $list_field);
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get API error title
			public function get_api_error_title($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-constant-contact-v2'); }

				// Get error title from JSON response
				$error_title = $api_response_decoded->detail;

				return $error_title;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-constant-contact-v2'); }

				if(!is_array($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-constant-contact-v2'); }

				$error_detail_array = array();
				foreach($api_response_decoded as $error) {

					$error_detail_array[] = $error->error_message;
				}

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_action_by',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-constant-contact-v2') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-constant-contact-v2')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('Constant Contact List', 'ws-form-constant-contact-v2'),
						'type'						=>	'select',
						'help'						=>	__('Select the Constant Contact list to associate this form with.', 'ws-form-constant-contact-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-constant-contact-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Action by
					'action_' . $this->id . '_action_by'	=> array(

						'label'							=>	__('Who Is Submitting?', 'ws-form-constant-contact-v2'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => 'ACTION_BY_VISITOR', 'text' => __('Visitors (Welcome message sent)', 'ws-form')),
							array('value' => 'ACTION_BY_OWNER', 'text' => __('Account Owner', 'ws-form')),
						),
						'default'						=>	'ACTION_BY_VISITOR',
						'help'							=>	__('Determines autoresponder/confirm opt-in messages.', 'ws-form-constant-contact-v2')
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-constant-contact-v2'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-constant-contact-v2'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-constant-contact-v2')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-constant-contact-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Constant Contact fields.', 'ws-form-constant-contact-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Constant Contact Field', 'ws-form-constant-contact-v2'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-constant-contact-v2'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-constant-contact-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-constant-contact-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Constant Contact add-on for WS Form PRO license key here.', 'ws-form-constant-contact-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-constant-contact-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Constant Contact Status', 'ws-form-constant-contact-v2'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Constant Contact account, or <a href=\"https://www.constantcontact.com/signup.jsp?pn=westguard\" target=\"blank\">create a new account</a>.", 'ws-form-constant-contact-v2'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				return $options;
			}

			// Explode list field (e.g. for address fields) - Returns number of addition fields added
			public function list_field_explode(&$list_fields, $list_field) {

				if(!isset($list_field['action_type'])) { $list_fields[] = $list_field; return 0; }

				switch($list_field['action_type']) {

					case 'addresses' :

						$list_fields[] = self::list_field_get_sub($list_field, 'line1', 'Address Line 1', 'Address Line 1', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'line2', 'Address Line 2', 'Address Line 2', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'line3', 'Address Line 3', 'Address Line 3', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'city', 'City', 'City', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'state', 'State', 'State', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'postal_code', 'Zip', 'Zip', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'country_code', 'Country', 'Country', 'select', false, self::get_data_grid_meta_country());
						$list_fields[] = self::list_field_get_sub($list_field, 'address_type', 'Address Type', 'Address Type', 'select', false, self::get_data_grid_meta_address_type());

						return 7;

					case 'email_addresses' :

						$list_fields[] = self::list_field_get_sub($list_field, 'email_address', 'Email', 'Email', 'email');
						return 1;

					default :

						$list_fields[] = $list_field;
						return 0;
				}
			}

			// Build sub data
			public function list_field_get_sub(&$list_field, $id_sub, $label_sub, $label_field, $action_type, $help = '', $meta = '') {

				$list_field_return = array(

					'id' => 			$list_field['id'] . '|' . $id_sub,
					'label' => 			$label_sub,
					'label_field' => 	$label_field,
					'type' => 			$action_type,
					'action_type' =>	$action_type,
					'required' => 		$list_field['required'], 
					'default_value' => 	'', 
					'pattern' => 		self::action_field_to_ws_form_pattern($action_type),
					'placeholder' => 	self::action_field_to_ws_form_placeholder($action_type),
					'help' => 			$help, 
					'sort_index' => 	$list_field['sort_index'],
					'visible' =>		$list_field['visible'],
					'meta'	=>			$meta
				);

				$list_field['sort_index'] = $list_field['sort_index'] + 1;

				return $list_field_return;
			}

			// Convert action field type to WS Form pattern
			public function action_field_to_ws_form_pattern($type, $options = false) {

				switch($type) {

					case 'zip' : return '(\d{5}([\-]\d{4})?)';

					default :

						return false;
				}
			}

			// Convert action field type to WS Form pattern
			public function action_field_to_ws_form_placeholder($type, $options = false) {

				switch($type) {

					case 'zip' : return '#####';

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->action_by = parent::get_config($config, 'action_' . $this->id . '_action_by');
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token  = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');

				// Build endpoint using API key
				if(!empty($this->api_access_token )) {

					// Set endpoint URL
					$this->api_endpoint = 'https://api.constantcontact.com/v2/';

				} else {

					$this->api_endpoint = false;
					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}
		
			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g id="logo" transform="translate(47.000000, 76.000000)">';
				$svg_logo .= '<defs><path d="M10.9.3C8 .8 5.2 1.5 2.5 2.4a62 62 0 0 0-2.3 6A60.5 60.5 0 0 1 9 6.7L11.1.2H11z" id="a"/><linearGradient x1="5.8%" y1="54.9%" x2="105.4%" y2="54.9%" id="b"><stop stop-color="#FDE1A4" offset="0%"/><stop stop-color="#FDE1A4" offset="10%"/><stop stop-color="#FCD787" offset="30%"/><stop stop-color="#FBCE6B" offset="60%"/><stop stop-color="#FAC85A" offset="80%"/><stop stop-color="#FAC654" offset="100%"/><stop stop-color="#FAC654" offset="100%"/></linearGradient><path d="M3.4 3A65 65 0 0 0 0 8.4c1.6-1 3.4-2.1 5.2-3C6 3.5 7.2 1.8 8.2 0L3.3 3" id="d"/><linearGradient x1="-6.5%" y1="43.6%" x2="93.2%" y2="43.6%" id="e"><stop stop-color="#FDE1A4" offset="0%"/><stop stop-color="#FDE1A4" offset="10%"/><stop stop-color="#FCD88A" offset="20%"/><stop stop-color="#FBD072" offset="30%"/><stop stop-color="#FACA61" offset="50%"/><stop stop-color="#FAC757" offset="70%"/><stop stop-color="#FAC654" offset="100%"/><stop stop-color="#FAC654" offset="100%"/></linearGradient><path d="M2.3 3.4C1.5 5.2.7 7 0 8.6c2.5-1 5-2 7.5-2.7L9.8.2c-2.6 1-5.1 2-7.5 3.2" id="g"/><linearGradient x1="-7.1%" y1="36.5%" x2="92.6%" y2="36.5%" id="h"><stop stop-color="#8AA4C4" offset="0%"/><stop stop-color="#8AA4C4" offset="20%"/><stop stop-color="#86A1C2" offset="50%"/><stop stop-color="#7B98BC" offset="80%"/><stop stop-color="#6D8CB5" offset="100%"/></linearGradient><path d="M3.8 2.6a64 64 0 0 0-3.4 6c2.5-.8 5-1.4 7.6-2h1C9.7 4.2 10.7 2 12 0 9.2.7 6.4 1.5 3.8 2.6" id="j"/><linearGradient x1="5.7%" y1="61.5%" x2="105.6%" y2="61.5%" id="k"><stop stop-color="#164989" offset="0%"/><stop stop-color="#1E446D" offset="100%"/></linearGradient><path d="M3 3L.5 7.6l4.2-3C5 3.2 5.6 1.8 6.3.4L3.1 3" id="m"/><linearGradient x1="-9%" y1="37.9%" x2="90.5%" y2="37.9%" id="n"><stop stop-color="#164989" offset="0%"/><stop stop-color="#1E446D" offset="100%"/></linearGradient><path d="M2.2 3.4L.2 8c1.7-1.2 3.6-2.2 5.5-3 .6-1.8 1.3-3.4 2-5C5.8 1 4 2.2 2.2 3.4" id="p"/><linearGradient x1="9%" y1="60.7%" x2="108.5%" y2="60.7%" id="q"><stop stop-color="#F7B219" offset="0%"/><stop stop-color="#F2AB1C" offset="30%"/><stop stop-color="#E49823" offset="80%"/><stop stop-color="#DD8F27" offset="100%"/></linearGradient><path d="M2.2 1A62 62 0 0 0 0 7.8c3.3-.2 6.5-.3 9.8 0 .7-2.5 1.5-5 2.4-7.4h-.4C8.5.3 5.3.5 2.2 1" id="s"/><linearGradient x1="-1.9%" y1="45.9%" x2="97.7%" y2="45.9%" id="t"><stop stop-color="#8AA4C4" offset="0%"/><stop stop-color="#8AA4C4" offset="20%"/><stop stop-color="#86A1C2" offset="50%"/><stop stop-color="#7B98BC" offset="80%"/><stop stop-color="#6D8CB5" offset="100%"/></linearGradient><path d="M3.7 1.3A66.1 66.1 0 0 0 .4 8c3-.5 6.1-.7 9.2-.7h.8c1-2.5 2.2-5 3.6-7.3C10.5.2 7 .6 3.8 1.3" id="v"/><linearGradient x1=".5%" y1="52.5%" x2="100.2%" y2="52.5%" id="w"><stop stop-color="#F7B219" offset="0%"/><stop stop-color="#F2AB1C" offset="30%"/><stop stop-color="#E49823" offset="80%"/><stop stop-color="#DD8F27" offset="100%"/></linearGradient><path d="M8.6.6A60 60 0 0 0 1.9 2L.4 7.9c3.1-.5 6.2-.8 9.3-.8A54 54 0 0 1 11 .3L8.6.5" id="y"/><linearGradient x1="-1.3%" y1="48%" x2="98.4%" y2="48%" id="z"><stop stop-color="#F7B219" offset="0%"/><stop stop-color="#F2AB1C" offset="30%"/><stop stop-color="#E49823" offset="80%"/><stop stop-color="#DD8F27" offset="100%"/></linearGradient><path d="M1.5 4L0 7.8a42 42 0 0 1 4.9-3.3c.3-1.3.7-2.7 1.2-4A51.4 51.4 0 0 0 1.5 4" id="B"/><linearGradient x1="14.5%" y1="67.9%" x2="114.4%" y2="67.9%" id="C"><stop stop-color="#FDE1A4" offset="0%"/><stop stop-color="#FDE1A4" offset="10%"/><stop stop-color="#FCD88A" offset="20%"/><stop stop-color="#FBD072" offset="30%"/><stop stop-color="#FACA61" offset="50%"/><stop stop-color="#FAC757" offset="70%"/><stop stop-color="#FAC654" offset="100%"/><stop stop-color="#FAC654" offset="100%"/></linearGradient><path d="M1.7 3L.2 8C2.7 7.3 5.3 6.6 8 6c.3-2 .8-4 1.4-5.8A56 56 0 0 0 1.7 3" id="E"/><linearGradient x1="-8.8%" y1="37.8%" x2="91%" y2="37.8%" id="F"><stop stop-color="#164989" offset="0%"/><stop stop-color="#1E446D" offset="100%"/></linearGradient></defs><g fill="none" fill-rule="evenodd"><g transform="translate(23.7 9.1)"><mask id="c" fill="#fff"><use xlink:href="#a"/></mask><path fill="url(#b)" fill-rule="nonzero" mask="url(#c)" d="M.2.2h11v8.1H.1z"/></g><g transform="translate(11.3 9.1)"><mask id="f" fill="#fff"><use xlink:href="#d"/></mask><path fill="url(#e)" fill-rule="nonzero" mask="url(#f)" d="M0 0h8.2v8.4H.1z"/></g><g transform="translate(15 11.8)"><mask id="i" fill="#fff"><use xlink:href="#g"/></mask><path fill="url(#h)" fill-rule="nonzero" mask="url(#i)" d="M0 .2h9.7v8.4H.1z"/></g><g transform="translate(26.3 1.6)"><mask id="l" fill="#fff"><use xlink:href="#j"/></mask><path fill="url(#k)" fill-rule="nonzero" mask="url(#l)" d="M.4 0H12v8.6H.4z"/></g><g transform="translate(2.2 20.4)"><mask id="o" fill="#fff"><use xlink:href="#m"/></mask><path fill="url(#n)" fill-rule="nonzero" mask="url(#o)" d="M.4.4h6v7.2h-6z"/></g><g transform="translate(8 16.1)"><mask id="r" fill="#fff"><use xlink:href="#p"/></mask><path fill="url(#q)" fill-rule="nonzero" mask="url(#r)" d="M.2 0h7.5v8H.2z"/></g><g transform="translate(33.9 8)"><mask id="u" fill="#fff"><use xlink:href="#s"/></mask><path fill="url(#t)" fill-rule="nonzero" mask="url(#u)" d="M0 .3h12.2v7.4H-.1z"/></g><g transform="translate(36)"><mask id="x" fill="#fff"><use xlink:href="#v"/></mask><path fill="url(#w)" fill-rule="nonzero" mask="url(#x)" d="M.4 0H14v8H.4z"/></g><g transform="translate(21.5 16.7)"><mask id="A" fill="#fff"><use xlink:href="#y"/></mask><path fill="url(#z)" fill-rule="nonzero" mask="url(#A)" d="M.5.3H11V8H.5z"/></g><g transform="translate(0 26.3)"><mask id="D" fill="#fff"><use xlink:href="#B"/></mask><path fill="url(#C)" fill-rule="nonzero" mask="url(#D)" d="M0 .4h6v7.4H0z"/></g><g transform="translate(12.9 18.8)"><mask id="G" fill="#fff"><use xlink:href="#E"/></mask><path fill="url(#F)" fill-rule="nonzero" mask="url(#G)" d="M.2.2h9.2V8H.2z"/></g></g>';
				$svg_logo .= '</g>';

				return $svg_logo;
			}

			// Address type data grid meta
			public function get_data_grid_meta_address_type() {

				$group_name = __('Country', 'ws-form');

				$columns = array(

					array('id' => 0, 'label' => __('Label', 'ws-form')),
					array('id' => 1, 'label' => __('Value', 'ws-form'))
				);

				$rows = array(

					array(

						'id'		=> 1,
						'default'	=> 'on',
						'required'	=> '',
						'disabled'	=> '',
						'hidden'	=> '',
						'data'		=> array('Personal', 'PERSONAL')
					),

					array(

						'id'		=> 2,
						'default'	=> '',
						'required'	=> '',
						'disabled'	=> '',
						'hidden'	=> '',
						'data'		=> array('Business', 'BUSINESS')
					)
				);

				return array('data_grid_select' => WS_Form_Common::build_data_grid_meta('data_grid_select', $group_name, $columns, $rows), 'select_field_label' => 0, 'select_field_value' => 1);
			}

			// Country data grid meta
			public function get_data_grid_meta_country() {

				$countries = json_decode('[{"n":"Afghanistan","c":"af"},{"n":"\u00c5land Islands","c":"ax"},{"n":"Albania","c":"al"},{"n":"Algeria","c":"dz"},{"n":"American Samoa","c":"as"},{"n":"Andorra","c":"ad"},{"n":"Angola","c":"ao"},{"n":"Anguilla","c":"ai"},{"n":"Antarctica","c":"aq"},{"n":"Antigua and Barbuda","c":"ag"},{"n":"Argentina","c":"ar"},{"n":"Armenia","c":"am"},{"n":"Aruba","c":"aw"},{"n":"Australia","c":"au"},{"n":"Austria","c":"at"},{"n":"Azerbaijan","c":"az"},{"n":"Bahamas","c":"bs"},{"n":"Bahrain","c":"bh"},{"n":"Bangladesh","c":"bd"},{"n":"Barbados","c":"bb"},{"n":"Belarus","c":"by"},{"n":"Belgium","c":"be"},{"n":"Belize","c":"bz"},{"n":"Benin","c":"bj"},{"n":"Bermuda","c":"bm"},{"n":"Bhutan","c":"bt"},{"n":"Bolivia (Plurinational State of)","c":"bo"},{"n":"Bonaire, Sint Eustatius and Saba","c":"bq"},{"n":"Bosnia and Herzegovina","c":"ba"},{"n":"Botswana","c":"bw"},{"n":"Bouvet Island","c":"bv"},{"n":"Brazil","c":"br"},{"n":"British Indian Ocean Territory","c":"io"},{"n":"Brunei Darussalam","c":"bn"},{"n":"Bulgaria","c":"bg"},{"n":"Burkina Faso","c":"bf"},{"n":"Burundi","c":"bi"},{"n":"Cabo Verde","c":"cv"},{"n":"Cambodia","c":"kh"},{"n":"Cameroon","c":"cm"},{"n":"Canada","c":"ca"},{"n":"Cayman Islands","c":"ky"},{"n":"Central African Republic","c":"cf"},{"n":"Chad","c":"td"},{"n":"Chile","c":"cl"},{"n":"China","c":"cn"},{"n":"Christmas Island","c":"cx"},{"n":"Cocos (Keeling) Islands","c":"cc"},{"n":"Colombia","c":"co"},{"n":"Comoros","c":"km"},{"n":"Congo","c":"cg"},{"n":"Congo (Democratic Republic of the)","c":"cd"},{"n":"Cook Islands","c":"ck"},{"n":"Costa Rica","c":"cr"},{"n":"C\u00f4te d\'Ivoire","c":"ci"},{"n":"Croatia","c":"hr"},{"n":"Cuba","c":"cu"},{"n":"Cura\u00e7ao","c":"cw"},{"n":"Cyprus","c":"cy"},{"n":"Czechia","c":"cz"},{"n":"Denmark","c":"dk"},{"n":"Djibouti","c":"dj"},{"n":"Dominica","c":"dm"},{"n":"Dominican Republic","c":"do"},{"n":"Ecuador","c":"ec"},{"n":"Egypt","c":"eg"},{"n":"El Salvador","c":"sv"},{"n":"Equatorial Guinea","c":"gq"},{"n":"Eritrea","c":"er"},{"n":"Estonia","c":"ee"},{"n":"Ethiopia","c":"et"},{"n":"Falkland Islands (Malvinas)","c":"fk"},{"n":"Faroe Islands","c":"fo"},{"n":"Fiji","c":"fj"},{"n":"Finland","c":"fi"},{"n":"France","c":"fr"},{"n":"French Guiana","c":"gf"},{"n":"French Polynesia","c":"pf"},{"n":"French Southern Territories","c":"tf"},{"n":"Gabon","c":"ga"},{"n":"Gambia","c":"gm"},{"n":"Georgia","c":"ge"},{"n":"Germany","c":"de"},{"n":"Ghana","c":"gh"},{"n":"Gibraltar","c":"gi"},{"n":"Greece","c":"gr"},{"n":"Greenland","c":"gl"},{"n":"Grenada","c":"gd"},{"n":"Guadeloupe","c":"gp"},{"n":"Guam","c":"gu"},{"n":"Guatemala","c":"gt"},{"n":"Guernsey","c":"gg"},{"n":"Guinea","c":"gn"},{"n":"Guinea-Bissau","c":"gw"},{"n":"Guyana","c":"gy"},{"n":"Haiti","c":"ht"},{"n":"Heard Island and McDonald Islands","c":"hm"},{"n":"Holy See","c":"va"},{"n":"Honduras","c":"hn"},{"n":"Hong Kong","c":"hk"},{"n":"Hungary","c":"hu"},{"n":"Iceland","c":"is"},{"n":"India","c":"in"},{"n":"Indonesia","c":"id"},{"n":"Iran (Islamic Republic of)","c":"ir"},{"n":"Iraq","c":"iq"},{"n":"Ireland","c":"ie"},{"n":"Isle of Man","c":"im"},{"n":"Israel","c":"il"},{"n":"Italy","c":"it"},{"n":"Jamaica","c":"jm"},{"n":"Japan","c":"jp"},{"n":"Jersey","c":"je"},{"n":"Jordan","c":"jo"},{"n":"Kazakhstan","c":"kz"},{"n":"Kenya","c":"ke"},{"n":"Kiribati","c":"ki"},{"n":"Korea (Democratic People\'s Republic of)","c":"kp"},{"n":"Korea (Republic of)","c":"kr"},{"n":"Kuwait","c":"kw"},{"n":"Kyrgyzstan","c":"kg"},{"n":"Lao People\'s Democratic Republic","c":"la"},{"n":"Latvia","c":"lv"},{"n":"Lebanon","c":"lb"},{"n":"Lesotho","c":"ls"},{"n":"Liberia","c":"lr"},{"n":"Libya","c":"ly"},{"n":"Liechtenstein","c":"li"},{"n":"Lithuania","c":"lt"},{"n":"Luxembourg","c":"lu"},{"n":"Macao","c":"mo"},{"n":"Macedonia (the former Yugoslav Republic of)","c":"mk"},{"n":"Madagascar","c":"mg"},{"n":"Malawi","c":"mw"},{"n":"Malaysia","c":"my"},{"n":"Maldives","c":"mv"},{"n":"Mali","c":"ml"},{"n":"Malta","c":"mt"},{"n":"Marshall Islands","c":"mh"},{"n":"Martinique","c":"mq"},{"n":"Mauritania","c":"mr"},{"n":"Mauritius","c":"mu"},{"n":"Mayotte","c":"yt"},{"n":"Mexico","c":"mx"},{"n":"Micronesia (Federated States of)","c":"fm"},{"n":"Moldova (Republic of)","c":"md"},{"n":"Monaco","c":"mc"},{"n":"Mongolia","c":"mn"},{"n":"Montenegro","c":"me"},{"n":"Montserrat","c":"ms"},{"n":"Morocco","c":"ma"},{"n":"Mozambique","c":"mz"},{"n":"Myanmar","c":"mm"},{"n":"Namibia","c":"na"},{"n":"Nauru","c":"nr"},{"n":"Nepal","c":"np"},{"n":"Netherlands","c":"nl"},{"n":"New Caledonia","c":"nc"},{"n":"New Zealand","c":"nz"},{"n":"Nicaragua","c":"ni"},{"n":"Niger","c":"ne"},{"n":"Nigeria","c":"ng"},{"n":"Niue","c":"nu"},{"n":"Norfolk Island","c":"nf"},{"n":"Northern Mariana Islands","c":"mp"},{"n":"Norway","c":"no"},{"n":"Oman","c":"om"},{"n":"Pakistan","c":"pk"},{"n":"Palau","c":"pw"},{"n":"Palestine, State of","c":"ps"},{"n":"Panama","c":"pa"},{"n":"Papua New Guinea","c":"pg"},{"n":"Paraguay","c":"py"},{"n":"Peru","c":"pe"},{"n":"Philippines","c":"ph"},{"n":"Pitcairn","c":"pn"},{"n":"Poland","c":"pl"},{"n":"Portugal","c":"pt"},{"n":"Puerto Rico","c":"pr"},{"n":"Qatar","c":"qa"},{"n":"R\u00e9union","c":"re"},{"n":"Romania","c":"ro"},{"n":"Russian Federation","c":"ru"},{"n":"Rwanda","c":"rw"},{"n":"Saint Barth\u00e9lemy","c":"bl"},{"n":"Saint Helena, Ascension and Tristan da Cunha","c":"sh"},{"n":"Saint Kitts and Nevis","c":"kn"},{"n":"Saint Lucia","c":"lc"},{"n":"Saint Martin (French part)","c":"mf"},{"n":"Saint Pierre and Miquelon","c":"pm"},{"n":"Saint Vincent and the Grenadines","c":"vc"},{"n":"Samoa","c":"ws"},{"n":"San Marino","c":"sm"},{"n":"Sao Tome and Principe","c":"st"},{"n":"Saudi Arabia","c":"sa"},{"n":"Senegal","c":"sn"},{"n":"Serbia","c":"rs"},{"n":"Seychelles","c":"sc"},{"n":"Sierra Leone","c":"sl"},{"n":"Singapore","c":"sg"},{"n":"Sint Maarten (Dutch part)","c":"sx"},{"n":"Slovakia","c":"sk"},{"n":"Slovenia","c":"si"},{"n":"Solomon Islands","c":"sb"},{"n":"Somalia","c":"so"},{"n":"South Africa","c":"za"},{"n":"South Georgia and the South Sandwich Islands","c":"gs"},{"n":"South Sudan","c":"ss"},{"n":"Spain","c":"es"},{"n":"Sri Lanka","c":"lk"},{"n":"Sudan","c":"sd"},{"n":"Suriname","c":"sr"},{"n":"Svalbard and Jan Mayen","c":"sj"},{"n":"Swaziland","c":"sz"},{"n":"Sweden","c":"se"},{"n":"Switzerland","c":"ch"},{"n":"Syrian Arab Republic","c":"sy"},{"n":"Taiwan, Province of China","c":"tw"},{"n":"Tajikistan","c":"tj"},{"n":"Tanzania, United Republic of","c":"tz"},{"n":"Thailand","c":"th"},{"n":"Timor-Leste","c":"tl"},{"n":"Togo","c":"tg"},{"n":"Tokelau","c":"tk"},{"n":"Tonga","c":"to"},{"n":"Trinidad and Tobago","c":"tt"},{"n":"Tunisia","c":"tn"},{"n":"Turkey","c":"tr"},{"n":"Turkmenistan","c":"tm"},{"n":"Turks and Caicos Islands","c":"tc"},{"n":"Tuvalu","c":"tv"},{"n":"Uganda","c":"ug"},{"n":"Ukraine","c":"ua"},{"n":"United Arab Emirates","c":"ae"},{"n":"United Kingdom of Great Britain and Northern Ireland","c":"gb"},{"n":"United States of America","c":"us"},{"n":"United States Minor Outlying Islands","c":"um"},{"n":"Uruguay","c":"uy"},{"n":"Uzbekistan","c":"uz"},{"n":"Vanuatu","c":"vu"},{"n":"Venezuela (Bolivarian Republic of)","c":"ve"},{"n":"Viet Nam","c":"vn"},{"n":"Virgin Islands (British)","c":"vg"},{"n":"Virgin Islands (U.S.)","c":"vi"},{"n":"Wallis and Futuna","c":"wf"},{"n":"Western Sahara","c":"eh"},{"n":"Yemen","c":"ye"},{"n":"Zambia","c":"zm"},{"n":"Zimbabwe","c":"zw"}]');

				$group_name = __('Country', 'ws-form');

				$columns = array(

					array('id' => 0, 'label' => __('Label', 'ws-form')),
					array('id' => 1, 'label' => __('Value', 'ws-form'))
				);

				$rows = array();
				$row_index = 1;
				foreach($countries as $country) {

					$rows[] = array(

						'id'		=> $row_index++,
						'default'	=> (($country->c == 'us') ? 'on' : ''),
						'required'	=> '',
						'disabled'	=> '',
						'hidden'	=> '',
						'data'		=> array($country->n, $country->c)
					);
				}

				return array('data_grid_select' => WS_Form_Common::build_data_grid_meta('data_grid_select', $group_name, $columns, $rows), 'select_field_label' => 0, 'select_field_value' => 1);
			}
		}
	});