<?php

/**
 * @link              https://wsform.com/knowledgebase/convertkit/
 * @since             1.0.0
 * @package           WS_Form_ConvertKit_V3
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - ConvertKit
 * Plugin URI:        https://wsform.com/knowledgebase/convertkit/
 * Description:       ConvertKit add-on for WS Form PRO
 * Version:           1.0.0
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-convertkit-v3
 */

	Class WS_Form_Add_On_ConvertKit_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.8.36';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_ConvertKit_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-convertkit-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_convertkit_v3 = new WS_Form_Add_On_ConvertKit_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_convertkit_v3, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_ConvertKit_V3 extends WS_Form_Action {

			public $id = 'convertkitv3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_key;

			public $list_id = false;
			public $sequence_id = false;
			public $opt_in_field = false;
			public $field_mapping = false;
			public $custom_mapping = false;
			public $tag_mapping = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 39652;
			const WS_FORM_LICENSE_NAME = 'ConvertKit add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.0';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const EMAIL_FIELD = 'email';
			const FIRST_NAME_FIELD = 'first_name';

			public function __construct() {

				// Set label
				$this->label = __('ConvertKit', 'ws-form-convertkit-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to ConvertKit', 'ws-form-convertkit-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-convertkit-v3')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-convertkit-v3') . '" />';

						} else {

							$value = '<input class="button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-convertkit-v3') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your ConvertKit account <a href="%s">here</a>.', 'ws-form-convertkit-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-convertkit-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$custom_field_values = array();
				$data = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($api_field_string) {

						case self::EMAIL_FIELD :
						case self::FIRST_NAME_FIELD :

							$data[$api_field_string] = $submit_value;
							break;

						default :

							$custom_field_values[$api_field_string] = $submit_value;
					}
				}

				foreach($this->custom_mapping as $custom_map) {

					// Get value
					$custom_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$custom_value = WS_Form_Common::parse_variables_process($custom_value, $form, $submit, 'text/plain');
					if($custom_value == '') { continue; }

					// Get attribute
					$custom_field = $custom_map['action_' . $this->id . '_list_fields'];
					if($custom_field == '') { continue; }

					// Array conversion
					if(is_array($custom_value)) { $custom_value = implode(',', $custom_value); }

					// Drip cannot accept arrays
					switch($custom_field) {

						case self::EMAIL_MERGE_FIELD :
						case self::FIRST_NAME_FIELD :

							$data[$api_field_string] = $submit_value;
							break;

						default :

							$custom_field_values[$api_field_string] = $submit_value;
					}
				}

				// Check email
				if(
					!isset($data[self::EMAIL_FIELD]) ||
					(!filter_var($data[self::EMAIL_FIELD], FILTER_VALIDATE_EMAIL))
				) {

					return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-convertkit-v3'), $email_address));
				}

				// Process tags
				$tags = array();
				foreach($this->tag_mapping as $tag_map) {

					// Get field ID
					$field_id = $tag_map['ws_form_field'];
					if($field_id == '') { continue; }

					// Read submit meta
					$tags_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($tags_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($tags_array)) { $tags_array = array($tags_array); }

					// Process each tag ready for API format
					foreach($tags_array as $tag) {

						if($tag == '') { continue; }

						$tags[] = $tag;
					}
				}

				// Build fields data
				if(count($custom_field_values) > 0) { $data['fields'] = $custom_field_values; }

				// Build tags data
				if(count($tags) > 0) { $data['tags'] = $tags; }

				// API key
				$data['api_key'] = $this->api_key;

				// JSON encode
				$data_json = json_encode($data);

				// Post
				$api_response = parent::api_call($this->api_endpoint, 'forms/' . $this->list_id . '/subscribe', 'POST', $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						self::success(__('Successfully added to form: ', 'ws-form-convertkit-v3') . $this->list_id);
						break;

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Process sequence
				if(empty($this->sequence_id)) { return true; }

				// Data
				$data_sequence = array(

					'api_key' => $this->api_key,
					'email' => $data[self::EMAIL_FIELD]
				);

				// JSON encode
				$data_json = json_encode($data_sequence);

				// Post
				$api_response = parent::api_call($this->api_endpoint, 'sequences/' . $this->sequence_id . '/subscribe', 'POST', $data_json, self::get_api_headers());
				// Process response
				switch($api_response['http_code']) {

					case 200 :

						self::success(__('Successfully added to sequence: ', 'ws-form-convertkit-v3') . $this->sequence_id);
						break;

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}

				return true;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'api_key'		=> $this->api_key
					);

					// Get forms
					$api_response = parent::api_call($this->api_endpoint, 'forms', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

					foreach($api_response_decoded->forms as $form) {

						$lists[] = array(

							'id' => 			parent::get_object_value($form, 'id'), 
							'label' => 			parent::get_object_value($form, 'name'), 
							'field_count' => 	false, 
							'record_count' => 	false
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'api_key'		=> $this->api_key
					);

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'forms/' . $this->list_id, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_decoded, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);				
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Email', 
						'label_field' => 	'Email', 
						'label_sub' =>		false,
						'type' => 			'email', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Add first name
					$list_fields[] = array(

						'id' => 			self::FIRST_NAME_FIELD, 
						'id_sub' =>			false,
						'label' => 			'First Name', 
						'label_field' => 	'First Name', 
						'label_sub' =>		false,
						'type' => 			'text', 
						'required' => 		false, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	1,
						'visible' =>		true
					);

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Params
					$params = array(

						'api_key'		=> $this->api_key
					);

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'custom_fields', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

					// Process fields
					$sort_index = 2;
					foreach($api_response_decoded->custom_fields as $field) {

						$list_fields[] = array(

							'id' => 			parent::get_object_value($field, 'key'),
							'label' => 			parent::get_object_value($field, 'label'), 
							'label_field' => 	parent::get_object_value($field, 'label'), 
							'type' => 			'text',
							'action_type' =>	'text',
							'required' => 		false,
							'default_value' => 	'',
							'pattern' => 		'',
							'placeholder' => 	'',
							'input_mask' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		true,
							'meta' => 			false
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_custom_mapping'	=>	'custom_mapping',
							'action_' . $this->id . '_tag_mapping'		=>	'tag_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field'
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get sequences
			public function get_sequences($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$sequences = WS_Form_Common::option_get('action_' . $this->id . '_sequences');

				// Retried if fetch is requested or sequences are not cached
				if($fetch || ($sequences === false)) {

					$sequences = array();

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'api_key'		=> $this->api_key
					);
					// Get sequences
					$api_response = parent::api_call($this->api_endpoint, 'sequences', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process sequence data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

					foreach($api_response_decoded->courses as $course) {

						$sequences[] = array(

							'id' => 			parent::get_object_value($course, 'id'), 
							'label' => 			parent::get_object_value($course, 'name')
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_sequences', $sequences);
				}

				return $sequences;
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				$tag_categories = array(

					array(

						'id' => 		1,
						'label' => 		__('Tags', 'ws-form-convertkit-v3'), 
						'sort_index' =>	1,
						'type' => 		'checkbox'
					)
				);

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = false, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check to see if tags are cached
				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id);

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'api_key'		=> $this->api_key
					);

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'tags', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

					$sort_index = 1;

					foreach($api_response_decoded->tags as $tag) {

						$tags[] = array(

							'id' => 			parent::get_object_value($tag, 'id'),
							'label' => 			parent::get_object_value($tag, 'name'),
							'record_count' =>	false,
							'sort_index' =>		$sort_index
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id, $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-convertkit-v3')); }

				$error_detail_array[] = $api_response_decoded->error;

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_sequence_id',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_tag_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-convertkit-v3'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-convertkit-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('ConvertKit Form', 'ws-form-convertkit-v3'),
						'type'							=>	'select',
						'help'							=>	__('Which ConvertKit form should this subscriber be added to?', 'ws-form-convertkit-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-convertkit-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Sequence
					'action_' . $this->id . '_sequence_id'	=> array(

						'label'							=>	__('ConvertKit Sequence', 'ws-form-convertkit-v3'),
						'type'							=>	'select',
						'help'							=>	__('Which ConvertKit sequence should this subscriber be added to?', 'ws-form-convertkit-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-convertkit-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'sequences',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'sequences/fetch'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-convertkit-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-convertkit-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-convertkit-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-convertkit-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to ConvertKit merge fields.', 'ws-form-convertkit-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom field mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-sendinblue-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map custom values to Sendinblue attributes.', 'ws-form-sendinblue-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom field mapping - Value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'	=>	__('Value', 'ws-form-sendinblue-v3'),
						'type'	=>	'text'
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-convertkit-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to AWeber tags.', 'ws-form-convertkit-v3'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('ConvertKit Field', 'ws-form-convertkit-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-convertkit-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-convertkit-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-convertkit-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your ConvertKit add-on for WS Form PRO license key here.', 'ws-form-convertkit-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-convertkit-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'		=>	__('ConvertKit API Key', 'ws-form-convertkit-v3'),
							'type'		=>	'text',
							'default'	=>	'',
							'help'		=>	__('<a href="https://wsform.com/knowledgebase/convertkit/" target="_blank">Learn more</a>', 'ws-form-convertkit-v3')
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('ConvertKit Status', 'ws-form-convertkit-v3'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Enter your <a href=\"https://app.convertkit.com/account/edit\" target=\"blank\">ConvertKit API key</a> above, or <a href=\"https://convertkit.com/\" target=\"blank\">create a new account</a>.", 'ws-form-convertkit-v3')
						)
					)
				);

				return $options;
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				if($this->sequence_id === false) { $this->sequence_id = parent::get_config($config, 'action_' . $this->id . '_sequence_id'); }
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_key = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');

				// Set endpoint URL
				$this->api_endpoint = 'https://api.convertkit.com/v3/';

				// Build endpoint using API key
				if(empty($this->api_key)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/sequences/', array('methods' => 'GET', 'callback' => array($this, 'api_get_sequences'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/sequences/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_sequences'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Sequences
			public function api_get_sequences() {

				// Get sequences
				$sequences = self::get_sequences();

				// Process response
				self::api_response($sequences);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Sequences with fetch
			public function api_fetch_sequences() {

				// Get sequences
				$sequences = self::get_sequences(true);

				// Process response
				self::api_response($sequences);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('X-Auth-Token' => 'api-key ' . $this->api_key);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(38.000000, 62.000000)"><path d="M30.9 47.2c11.1 0 19.7-8.4 19.7-18.9 0-9.8-8.1-15.7-13.4-15.7-7.4 0-13.4 5.3-14.3 13.1-.2 1.4-1.3 2.6-2.7 2.6h-8.1c-1 0-1.7-.8-1.7-1.7.3-6.8 2.6-13.1 6.9-18C22 3.2 29 0 37.2 0 51 0 64 12.6 64 28.3c0 17.4-14.3 31.4-32.8 31.4C18 59.7 4.7 51 0 38.9v-.1c.1-.4.2-.7.3-1 .2-.5.3-.9.3-1.5l-.1-1.7c0-.8.4-1.6 1.2-2l.7-.3c.8-.4 1.3-1.1 1.5-1.9.2-1.2 1.3-2.1 2.6-2.1 2 0 3.7 1.3 4.3 3.2 3.1 9 8 15.7 20.1 15.7z" fill="#fb6970"/></g>';

				return $svg_logo;
			}
		}
	});