<?php

/**
 * @link              https://wsform.com/knowledgebase/dotdigital/
 * @since             1.0.0
 * @package           WS_Form_Dotdigital
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - dotdigital
 * Plugin URI:        https://wsform.com/knowledgebase/dotdigital/
 * Description:       dotdigital add-on for WS Form PRO
 * Version:           1.0.0
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-dotdigital-v2
 */

	Class WS_Form_Add_On_Dotdigital_V2 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.0.3';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Dotdigital_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-dotdigital-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_digital = new WS_Form_Add_On_Dotdigital_V2();	

	register_activation_hook(__FILE__, array($wsf_add_on_digital, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Dotdigital_V2 extends WS_Form_Action {

			public $id = 'dotmailerv2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_email;
			private $api_password;
			public $list_id = false;
			public $status;
			public $return_url;
			public $email_type;
			public $opt_in_field;
			public $field_mapping;

			// Constants
			const API_PATH = '/v2/';
			const DEFAULT_STATUS = 'Subscribed';
			const DEFAULT_EMAIL_TYPE = 'Html';
			const EMAIL_MERGE_FIELD = 'Email';
			const LISTS_FETCH_COUNT = 1000;
			const OPT_IN_TYPE = 'Single';
			const WS_FORM_LICENSE_ITEM_ID = 15740;
			const WS_FORM_LICENSE_NAME = 'dotdigital add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.0';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('dotdigital', 'ws-form-dotdigital-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to dotdigital', 'ws-form-dotdigital-v2');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('rest_api_init', array($this, 'rest_api_init'), 10, 2);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-dotdigital-v2')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-dotdigital-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-dotdigital-v2') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							default :

								self::account_info();
						}

					break;
				}
			}

			// Get account info
			public function account_info() {

				$this->api_email = WS_Form_Common::option_get('action_' . $this->id . '_api_email', '');
				$this->api_password = WS_Form_Common::option_get('action_' . $this->id . '_api_password', '');

				// Check user API email address is valid
				if(!filter_var($this->api_email, FILTER_VALIDATE_EMAIL)) { self::reset_api_endpoint(); return false; }

				// Check API password
				if($this->api_password == '') { self::reset_api_endpoint(); return false; }

				// Get account info
				$api_response = parent::api_call('https://api.dotmailer.com/v2/', 'account-info', 'GET', null, array(), 'basic', $this->api_email, $this->api_password);

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { self::reset_api_endpoint(); return self::error(__('Invalid API response data', 'ws-form-dotdigital-v2')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					self::reset_api_endpoint();

					switch($api_response_decoded->message) {

						case 'Authorization has been denied for this request.' :

							$message = sprintf(__('The %s API email and password is invalid. Please enter valid credentials <a href="%s">here</a>.', 'ws-form-dotdigital-v2'), $this->label, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id));
							break;

						default :

							$message = sprintf(__('%s API Error: %s', 'ws-form-dotdigital-v2'), $this->label, $api_response_decoded->message);
					}

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					return self::error(self::get_api_error_detail($api_response['response']));
				}


				// Check that we got properties
				if(!isset($api_response_decoded->properties) || !is_array($api_response_decoded->properties)) { self::reset_api_endpoint(); return false; }

				// Process returned properties
				$properties = array();
				foreach($api_response_decoded->properties as $property) {

					$properties[$property->name] = $property->value;
				}

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_api_endpoint', isset($properties['ApiEndpoint']) ? $properties['ApiEndpoint'] . self::API_PATH : '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_calls_in_last_hour', isset($properties['ApiCallsInLastHour']) ? $properties['ApiCallsInLastHour'] : 0);
				WS_Form_Common::option_set('action_' . $this->id . '_api_calls_remaining', isset($properties['ApiCallsRemaining']) ? $properties['ApiCallsRemaining'] : 0);
			}

			// Reset API endpoint
			public function reset_api_endpoint() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_endpoint', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_calls_in_last_hour', 0);
				WS_Form_Common::option_set('action_' . $this->id . '_api_calls_remaining', 0);
			}

			// Nag
			public function nag() {

				// Check user API email address is valid
				if(!filter_var($this->api_email, FILTER_VALIDATE_EMAIL) AND ($this->api_password == '')) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API email and password <a href="%s">here</a>.', 'ws-form-dotdigital-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-dotdigital-v2'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Process field mapping
				$email_address = false;
				$api_fields = array();

				$first_name = '';
				$last_name = '';

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];
					$api_field_array = explode('|', $api_field_string);

					$api_field = $api_field_array[0];
					$api_field_sub = (isset($api_field_array[1])) ? $api_field_array[1] : false;

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// Constant Contact cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($api_field_string) {

						case self::EMAIL_MERGE_FIELD :

							// Save email address
							$email_address = $submit_value;
							break;

						default :

							// Normal field
							$api_fields[] = array('key' => $api_field, 'value' => $submit_value);
							if($api_field == 'FIRSTNAME') { $first_name = $submit_value; }
							if($api_field == 'LASTNAME') { $last_name = $submit_value; }
							break;
					}
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-dotdigital-v2')); }

				// Check user email address is valid
				if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-dotdigital-v2'), $email_address)); }

				// Full name
				$api_fields[] = array('key' => 'FULLNAME', 'value' => ($first_name . (($last_name != '') ? ' ' . $last_name : '')));

				// Data
				switch($this->status) {

					case 'subscribe_address_book' :

						// Subscribe
						$data = array(

							self::EMAIL_MERGE_FIELD	=>	$email_address,
							'OptInType'				=>	self::OPT_IN_TYPE,
							'EmailType'				=>	$this->email_type,
							'DataFields'			=>	$api_fields
						);

						$message_success = sprintf(__('Successfully subscribed %s to address book: %s', 'ws-form-dotdigital-v2'), $email_address, $this->list_id);

						$method = 'address-books/' . $this->list_id . '/contacts';

						break;

					case 'unsubscribe_address_book' :

						// Unsubscribe
						$data = array(

							self::EMAIL_MERGE_FIELD	=>	$email_address
						);

						$message_success = sprintf(__('Successfully unsubscribed %s from address book: %s', 'ws-form-dotdigital-v2'), $email_address, $this->list_id);

						$method = 'address-books/' . $this->list_id . '/contacts/unsubscribe';

						break;

					case 'create' :

						// Create
						$data = array(

							self::EMAIL_MERGE_FIELD	=>	$email_address
						);

						$message_success = sprintf(__('Successfully created %s', 'ws-form-dotdigital-v2'), $email_address);

						$method = 'contacts';

						break;

					case 'suppress' :

						// Suppress
						$data = array(

							self::EMAIL_MERGE_FIELD	=>	$email_address
						);

						$message_success = sprintf(__('Successfully suppressed %s', 'ws-form-dotdigital-v2'), $email_address);

						$method = 'contacts/unsubscribe';

						break;

					case 'resubscribe_address_book' :

						// Resubscribe
						$data = array(

							'UnsubscribedContact' => array(

								self::EMAIL_MERGE_FIELD	=>	$email_address,
								'DataFields'			=>	$api_fields
							)
						);

						$message_success = sprintf(__('Successfully resubscribed %s to address book: %s', 'ws-form-dotdigital-v2'), $email_address, $this->list_id);

						$method = 'address-books/' . $this->list_id . '/contacts/resubscribe';

						break;

					case 'unsuppress' :

						// Resubscribe
						$data = array(

							'UnsubscribedContact' 			=> array(

								self::EMAIL_MERGE_FIELD		=>	$email_address,
								'DataFields'				=>	$api_fields
							),
							'PreferredLocale'				=> get_locale(),
							'ReturnUrlToUseIfChallenged'	=> WS_Form_Common::parse_variables_process($this->return_url, $form, $submit)
						);

						$message_success = sprintf(__('Successfully unsuppressed %s', 'ws-form-dotdigital-v2'), $email_address, $this->list_id);

						$method = 'contacts/resubscribe';

						break;
				}

				// Data for API call
				$data_json = json_encode($data);

				// POST (Adds new records, updates existing)
				$api_response = parent::api_call($this->api_endpoint, $method, 'POST', $data_json, array(), 'basic', $this->api_email, $this->api_password);

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						// Success
						self::success($message_success);
						return true;

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get field types
				$list_fields = self::get_list_fields();
				$list_field_type = array();
				foreach($list_fields as $list_field) {

					$list_field_type[$list_field['id']] = $list_field['type'];
				}

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-dotdigital-v2')); }

				// Get member
				$api_response = parent::api_call($this->api_endpoint, 'contacts/' . urlencode($user_email), 'GET', null, array(), 'basic', $this->api_email, $this->api_password);

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-dotdigital-v2')); }

				// Process fields
				$fields_return = array(

					'email_format' =>			parent::get_object_value($api_response_decoded, 'emailType'),
					'status' =>					parent::get_object_value($api_response_decoded, 'status'),
					self::EMAIL_MERGE_FIELD	=>	parent::get_object_value($api_response_decoded, 'email'),
				);

				$api_fields = parent::get_object_value($api_response_decoded, 'dataFields');

				if($api_fields !== false) {

					foreach($api_fields as $field) {

						if(is_null($field->value)) { continue; }

						$type = isset($list_field_type[$field->key]) ? $list_field_type[$field->key] : 'text';

						switch($type) {

							case 'datetime' :

								// dotdigital returns hours:mins:seconds so we strip that
								$fields_return[$field->key] = date('Y-m-d', strtotime($field->value));
								break;

							default :

								$fields_return[$field->key] = (string) $field->value;
						}
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$params = array(

							'Count'	=> self::LISTS_FETCH_COUNT,
							'Skip'	=> $total_lists_retrieved
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'address-books', 'GET', $params, array(), 'basic', $this->api_email, $this->api_password);

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-dotdigital-v2')); }

						// Get total lists
						$total_lists = count($api_response_decoded);

						// Increment read counter
						$total_lists_retrieved += $total_lists;

						foreach($api_response_decoded as $list) {

							// Skip 'Test' and 'All contacts' address book
							$name = parent::get_object_value($list, 'name');
							if(($name == 'Test') || ($name == 'All contacts')) { continue; }

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'),
								'label' => 			$name,
								'field_count' => 	false,
								'record_count' => 	parent::get_object_value($list, 'contacts')
							);
						}

					} while ($total_lists == self::LISTS_FETCH_COUNT);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'address-books/' . $this->list_id, 'GET', null, array(), 'basic', $this->api_email, $this->api_password);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-dotdigital-v2')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'data-fields', 'GET', null, array(), 'basic', $this->api_email, $this->api_password);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Add built in dotdigital fields not returned by API
					$list_fields[] = array(

						'id' => 			self::EMAIL_MERGE_FIELD,
						'label' => 			__('Email', 'ws-form-dotdigital-v2'), 
						'label_field' => 	__('Email', 'ws-form-dotdigital-v2'), 
						'type' => 			'email',
						'action_type' =>	'email',
						'required' => 		true, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	1,
						'visible' =>		true,
						'meta' => 			false
					);

					// Process list data
					$api_list_fields = json_decode($api_response['response']);

					// Checks
					if(is_null($api_list_fields)) { return self::error(__('Invalid API response data', 'ws-form-dotdigital-v2')); }

					// Process fields
					$sort_index_offset = 0;
					foreach($api_list_fields as $field) {

						$action_type = parent::get_object_value($field, 'type');

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						$list_fields[] = array(

							'id' => 			parent::get_object_value($field, 'name'),
							'label' => 			parent::get_object_value($field, 'name'),
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		false,
							'default_value' => 	self::action_to_ws_form_default_value(parent::get_object_value($field, 'defaultValue'), $type),
							'pattern' => 		'',
							'placeholder' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index_offset,
							'visible' =>		self::action_to_ws_form_visible(parent::get_object_value($field, 'visibility')),
							'meta' => 			self::action_field_to_ws_form_meta_keys($field)
						);

						$sort_index_offset ++;
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-dotdigital-v2'); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->message);

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_status',
						'action_' . $this->id . '_return_url',
						'action_' . $this->id . '_email_type',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-dotdigital-v2') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-dotdigital-v2')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('dotdigital Address Book', 'ws-form-dotdigital-v2'),
						'type'						=>	'select',
						'help'						=>	__('Select the dotdigital address book to associate this form with.', 'ws-form-dotdigital-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-dotdigital-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Status (subscribe, unsubscribe)
					'action_' . $this->id . '_status'	=> array(

						'label'						=>	__('Status', 'ws-form-dotdigital-v2'),
						'type'						=>	'select',
						'help'						=>	__("Set subscriber's status.", 'ws-form-dotdigital-v2'),
						'options'					=>	array(

							array('value' => 'subscribe_address_book', 'text' => __('Subscribe To Address Book', 'ws-form-dotdigital-v2')),
							array('value' => 'unsubscribe_address_book', 'text' => __('Unsubscribe From Address Book', 'ws-form-dotdigital-v2')),
							array('value' => 'resubscribe_address_book', 'text' => __('Resubscribe To Address Book', 'ws-form-dotdigital-v2')),
							array('value' => 'create', 'text' => __('Create', 'ws-form-dotdigital-v2')),
							array('value' => 'suppress', 'text' => __('Suppress', 'ws-form-dotdigital-v2')),
							array('value' => 'unsuppress', 'text' => __('Unsuppress', 'ws-form-dotdigital-v2')),
						),
						'default'					=>	'subscribe_address_book'
					),

					// Field mapping
					'action_' . $this->id . '_return_url'	=> array(

						'label'						=>	__('Return URL', 'ws-form-dotdigital-v2'),
						'type'						=>	'text',
						'help'						=>	__('The URL you would like to redirect challenged contacts to after they have completed their resubscription.', 'ws-form-dotdigital-v2'),
						'default'					=>	'#blog_url',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_status',
								'meta_value'	=>	'unsuppress'
							)
						)
					),

					// Email type
					'action_' . $this->id . '_email_type'	=> array(

						'label'						=>	__('Email Type', 'ws-form-dotdigital-v2'),
						'type'						=>	'select',
						'help'						=>	__('Type of email this subscriber will receive.', 'ws-form-dotdigital-v2'),
						'options'					=>	array(

							array('value' => 'Html', 'text' => __('HTML', 'ws-form-dotdigital-v2')),
							array('value' => 'PlainText', 'text' => __('Plain text', 'ws-form-dotdigital-v2'))
						),
						'default'					=>	'html'
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-dotdigital-v2'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-dotdigital-v2'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-dotdigital-v2')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-dotdigital-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to dotdigital merge fields.', 'ws-form-dotdigital-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('dotdigital Field', 'ws-form-dotdigital-v2'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-dotdigital-v2'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-dotdigital-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-dotdigital-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your dotdigital add-on for WS Form PRO license key here.', 'ws-form-dotdigital-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-dotdigital-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_email'	=>	array(

							'label'		=>	__('dotdigital API Email', 'ws-form-dotdigital-v2'),
							'type'		=>	'email',
							'default'	=>	'',
							'help'		=>	__('For more information about the dotdigital API, <a href="https://developer.dotdigital.com/docs/" target="_blank">click here</a>.', 'ws-form-dotdigital-v2')
						),

						'action_' . $this->id . '_api_password'	=>	array(

							'label'		=>	__('dotdigital API Password', 'ws-form-dotdigital-v2'),
							'type'		=>	'password',
							'default'	=>	''
						),

						'action_' . $this->id . '_api_calls_remaining'	=>	array(

							'label'		=>	__('Calls Remaining', 'ws-form-dotdigital-v2'),
							'type'		=>	'static',
							'default'	=>	0
						),

						'action_' . $this->id . '_api_calls_in_last_hour'	=>	array(

							'label'		=>	__('Calls In Last Hour', 'ws-form-dotdigital-v2'),
							'type'		=>	'static',
							'default'	=>	0
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'String' : return 'text';
					case 'Numeric' : return 'number';
					case 'Date' : return 'datetime';
					case 'Boolean' : return 'select';

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$type = parent::get_object_value($field, 'type');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids for radio and select
					case 'Boolean' :

						$group_name = __('Yes/No', 'ws-form-dotdigital-v2');

						// Build new columns
						$columns = array(

							array('id' => 0, 'label' => __('Label', 'ws-form-dotdigital-v2')),
							array('id' => 1, 'label' => __('Value', 'ws-form-dotdigital-v2'))
						);

						// Build new rows
						$rows = array(

							array(

								'id'		=> 1,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array('Yes', '1'),
							),

							array(

								'id'		=> 2,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array('No', '0')
							)
						);

						return array('data_grid_select' => WS_Form_Common::build_data_grid_meta('data_grid_select', $group_name, $columns, $rows), 'select_field_label' => 0, 'select_field_value' => 1);

					default :

						return false;
				}
			}

			// Convert action required to WS Form required
			public function action_to_ws_form_required($required) {

				return ($required == 1);
			}

			// Convert action default value to WS Form default value
			public function action_to_ws_form_default_value($default_value, $type = 'text') {

				switch($type) {

					case 'datetime' :

						return ($default_value !== '') ? date('Y-m-d', strtotime($default_value)) : '';

					default :

						return $default_value;
				}
			}

			// Convert action visibility to WS Form visibility
			public function action_to_ws_form_visible($visible) {

				return ($visible == 'Public');
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->status = parent::get_config($config, 'action_' . $this->id . '_status', self::DEFAULT_STATUS);
				$this->return_url = parent::get_config($config, 'action_' . $this->id . '_return_url', '#blog_url');
				$this->email_type = parent::get_config($config, 'action_' . $this->id . '_email_type', self::DEFAULT_EMAIL_TYPE);
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = false;

				$this->api_email = WS_Form_Common::option_get('action_' . $this->id . '_api_email', '');
				$this->api_password = WS_Form_Common::option_get('action_' . $this->id . '_api_password', '');
				$this->api_endpoint = WS_Form_Common::option_get('action_' . $this->id . '_api_endpoint', '');

				// Check user API email address is valid
				if(!filter_var($this->api_email, FILTER_VALIDATE_EMAIL)) { return $this->configured; }

				// Check API password
				if($this->api_password == '') { return $this->configured; }

				// Set endpoint URL
				if($this->api_endpoint == '') { return $this->configured; }

				// Check API endpoint
				$this->configured = true;

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}
		
			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(47.000000, 62.000000)"><path d="M0 24.959C-.083 10.78 12.437.086 24.378.001 35.904-.082 49.917 8.956 50 24.959c.077 14.925-11.194 24.787-25.207 24.958C11.194 50.083.08 38.641 0 24.959zm8.738-.118c-.017 8.984 7.153 16.428 15.998 16.503 9.212.078 16.72-7.073 16.668-16.178-.052-9.164-8.531-16.408-16.942-16.293-8.044-.249-15.707 7.113-15.724 15.968z" fill="#81c243"/><path d="M31.41 24.922c-.002 3.538-2.915 6.416-6.461 6.442-3.465.025-6.418-3.213-6.384-6.658.036-3.737 2.937-6.215 6.412-6.245 3.535-.032 6.435 2.914 6.433 6.461z" fill="#e90a8c"/></g>';

				return $svg_logo;
			}
		}
	});