<?php

	/**
	 * @link              https://wsform.com/knowledgebase/drip/
	 * @since             1.0.0
	 * @package           WS_Form_Drip
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Drip
	 * Plugin URI:        https://wsform.com/knowledgebase/drip/
	 * Description:       Drip add-on for WS Form PRO
	 * Version:           1.0.7
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-drip-v2
	 */

	Class WS_Form_Add_On_Drip_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.2.16';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Drip_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-drip-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_drip = new WS_Form_Add_On_Drip_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_drip, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Drip_V2 extends WS_Form_Action {

			public $id = 'dripv2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'People';

			// Licensing
			private $licensing;

			// Config
			private $api_access_token;
			private $api_endpoint;
			private $account_id;
			public $list_id = false;
			public $method;
			public $double_opt_in;
			public $prospect;
			public $base_lead_score;
			public $opt_in_field;
			public $field_mapping;
			public $custom_mapping;
			public $tag_mapping;
			public $tag_merge = false;

			// Constants
			const DEFAULT_STATUS = 'subscribed';
			const DEFAULT_EMAIL_TYPE = 'html';
			const EMAIL_MERGE_FIELD = 'email';
			const LIST_FIELDS_FETCH_COUNT = 50;
			const LISTS_FETCH_COUNT = 10;
			const TAG_CATEGORIES_FETCH_COUNT = 10;
			const TAGS_FETCH_COUNT = 10;
			const WS_FORM_LICENSE_ITEM_ID = 3533;
			const WS_FORM_LICENSE_NAME = 'Drip add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.7';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Drip', 'ws-form-drip-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to Drip', 'ws-form-drip-v2');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Get portal ID
				self::account_info();

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-drip-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-drip-v2') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-drip-v2') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/drip-v2';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-drip-v2'));
						}

						break;

					case 'account_list_action_' . $this->id :

						$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="account_list_refresh" value="' . __('Refresh', 'ws-form-drip-v2') . '" />';

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;

							case 'account_list_refresh' :

								self::account_list_refresh();
								break;
						}

					break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_account_list', false);
			}

			// Get account info
			public function account_info() {

				self::load_config_plugin();

				// Get account info
				$api_response = parent::api_call($this->api_endpoint, 'integrations/v1/me', 'GET', null, self::get_api_headers());

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					$message = sprintf(__('%s API Error: %s', 'ws-form-hubspot'), $this->label, $api_response_decoded->message);

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					self::error(self::get_api_error_detail($api_response['response']));
				}

				// Check that we got properties
				if(!isset($api_response_decoded->portalId)) { return false; }

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_portal_id', $api_response_decoded->portalId);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Drip account  <a href="%s">here</a>.', 'ws-form-drip-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value) && ($this->method == 'create')) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-drip-v2'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$custom_fields = new stdClass();
				$custom_fields_push = false;
				$custom_fields_update_rules = array();

				foreach($this->field_mapping as $field_map) {

					// Get merge field
					$custom_field = $field_map['action_' . $this->id . '_list_fields'];
					if($custom_field == '') { continue; }

					// Get field ID
					$field_id = $field_map['ws_form_field'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, '');

					// Drip cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($custom_field) {

						case self::EMAIL_MERGE_FIELD :

							$email_address = $submit_value;
							break;

						default :

							// Normal merge field
							$custom_fields->{$custom_field} = $submit_value;

							// Get update rule
							$api_update_rule = isset($field_map['action_' . $this->id . '_update_rule']) ? $field_map['action_' . $this->id . '_update_rule'] : '';
							if($api_update_rule != '') { $custom_fields_update_rules[$custom_field] = $api_update_rule; }
					}

					$custom_fields_push = true;
				}

				foreach($this->custom_mapping as $custom_map) {

					// Get value
					$custom_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$custom_value = WS_Form_Common::parse_variables_process($custom_value, $form, $submit, 'text/plain');
					if($custom_value == '') { continue; }

					// Get attribute
					$custom_field = $custom_map['action_' . $this->id . '_list_fields'];
					if($custom_field == '') { continue; }

					// Array conversion
					if(is_array($custom_value)) { $custom_value = implode(',', $custom_value); }

					// Drip cannot accept arrays
					switch($custom_field) {

						case self::EMAIL_MERGE_FIELD :

							$email_address = $submit_value;
							break;

						default :

							// Normal merge field
							$custom_fields->{$custom_field} = $custom_value;

							// Get update rule
							$api_update_rule = isset($custom_map['action_' . $this->id . '_update_rule']) ? $custom_map['action_' . $this->id . '_update_rule'] : '';
							if($api_update_rule != '') { $custom_fields_update_rules[$custom_field] = $api_update_rule; }
					}

					$custom_fields_push = true;
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-drip-v2')); }

				// Process according to method
				switch($this->method) {

					// Create or Update
					case 'create' :

						// Get subscriber
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/subscribers/' . urlencode($email_address), 'GET', null, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] == 200) {

							// Process list data
							$api_response_decoded = json_decode($api_response['response']);
							if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
							if(!isset($api_response_decoded->subscribers)) { return self::error(__('No subscribers found', 'ws-form-drip-v2')); }
							if(count($api_response_decoded->subscribers) == 0) { return self::error(__('No subscribers found', 'ws-form-drip-v2')); }

							// Get first subscriber
							$subscriber = $api_response_decoded->subscribers[0];

							$old_tags = parent::get_object_value($subscriber, 'tags');

						} else {

							$subscriber = false;

							$old_tags = array();
						}

						// Build post data
						$data = array(

							'email'				=>	$email_address,
							'prospect'			=>	($this->prospect == 'on'),
							'base_lead_score'	=>	intval($this->base_lead_score),
						);
						if($custom_fields_push) { $data['custom_fields'] = $custom_fields; }

						// Process update rules
						if(
							$custom_fields_push &&
							($subscriber !== false) &&
							(count($custom_fields_update_rules) > 0)
						) {

							foreach($custom_fields_update_rules as $custom_field => $update_rule) {

								// Get subscribers existing custom fields
								if(!isset($subscriber->custom_fields)) { break; }
								$subscriber_custom_fields = $subscriber->custom_fields;

								// Get subscribers existing field
								if(!isset($subscriber_custom_fields->{$custom_field})) { break; }
								$subscriber_custom_field = $subscriber_custom_fields->{$custom_field};

								switch($update_rule) {

									// Do not update if an existing value exists
									case 'ignore' :

										if($subscriber_custom_field != '') {

											unset($data['custom_fields']->{$custom_field});
										}

										break;

									// Merge as comma separated values
									case 'merge' :

										$custom_field_value = $data['custom_fields']->{$custom_field};

										if($subscriber_custom_field !== $custom_field_value) {

											// Convert old and new values to arrays
											$subscriber_custom_field_array = explode(',', $subscriber_custom_field);
											$custom_field_value_array = explode(',', $custom_field_value);

											// Merge arrays
											$subscriber_custom_field_array = array_merge($subscriber_custom_field_array, $custom_field_value_array);

											// Get unique values
											$subscriber_custom_field_array = array_unique($subscriber_custom_field_array);

											// Sort alphabetically
											sort($subscriber_custom_field_array, SORT_NATURAL | SORT_FLAG_CASE);

											// Convert back to string
											$data['custom_fields']->{$custom_field} = implode(',', $subscriber_custom_field_array);
										}

										break;
								}
							}
						}

						// Process tags
						$tags = array();
						foreach($this->tag_mapping as $field_map) {

							// Get field ID
							$field_id = $field_map['ws_form_field'];
							if($field_id == '') { continue; }

							// Read submit meta
							$tags_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
							if($tags_array === false) { continue; }

							// Turn into array if it isn't already
							if(!is_array($tags_array)) { $tags_array = array($tags_array); }

							// Process each tag ready for API format
							foreach($tags_array as $tag) {

								if($tag == '') { continue; }

								$tags[] = $tag;
							}
						}

						// Build tag data
						$tags_removed = array_diff($old_tags,$tags);
						$tags_added = array_diff($tags,$old_tags);
						if(count($tags_added) > 0) { $data['tags'] = array_values($tags_added); }
						if(
							!$this->tag_merge && 
							(count($tags_removed) > 0)
						) {
							$data['remove_tags'] = array_values($tags_removed);
						}

						// Tracking data - Remote IP
						$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
						if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

							$data['ip_address'] = $tracking_remote_ip;
						}

						// GDPR
						$data['eu_consent'] = ($opted_in) ? 'granted' : 'denied';

						// Double opt-in
						if($this->list_id != '0') {

							$data['double_optin'] = ($this->double_opt_in == 'on');
						}

						// Wrap data
						$data = array('subscribers' => array($data));

						// Data for API call
						$data_json = json_encode($data);

						// POST (Adds new records, updates existing)
						if($this->list_id != '0') {

							$api_path = $this->account_id . '/campaigns/' . $this->list_id . '/subscribers';
							$success_message = __('Successfully added to campaign: ', 'ws-form-drip-v2') . $this->list_id;

						} else {

							$api_path = $this->account_id . '/subscribers';
							$success_message = __('Successfully added to account: ', 'ws-form-drip-v2') . $this->account_id;
						}
						$api_response = parent::api_call($this->api_endpoint, $api_path, 'POST', $data_json, self::get_api_headers());

						// Process response
						switch($api_response['http_code']) {

							case 200 :
							case 201 :

								self::success($success_message);
								return true;

							default :

								// Get invalid merge fields
								self::error(self::get_api_error_detail($api_response['response']));
								return false;
						}

					// Remove
					case 'remove' :

						// POST
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/subscribers/' . urlencode($email_address) . '/remove' . (($this->list_id != '0') ? '?campaign_id=' . $this->list_id : ''), 'POST', null, self::get_api_headers());

						// Process response
						switch($api_response['http_code']) {

							case 200 :

								if($this->list_id != '0') {

									self::success(__('Successfully removed from campaign: ', 'ws-form-drip-v2') . $this->list_id);
								} else {

									self::success(__('Successfully removed from all campaigns', 'ws-form-drip-v2'));
								}
								return true;

							case 404 :

								self::success(__('Email address not found', 'ws-form-drip-v2'));
								return false;

							default :

								// Get invalid merge fields
								self::error(self::get_api_error_detail($api_response['response']));
								return false;
						}

					// Unsubscribe
					case 'unsubscribe' :

						// POST
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/subscribers/' . urlencode($email_address) . '/unsubscribe_all' , 'POST', null, self::get_api_headers());

						// Process response
						switch($api_response['http_code']) {

							case 200 :

								self::success(__('Successfully unsubscribed from all campaigns', 'ws-form-drip-v2'));
								return true;

							case 404 :

								self::success(__('Email address not found', 'ws-form-drip-v2'));
								return true;

							default :

								// Get invalid merge fields
								self::error(self::get_api_error_detail($api_response['response']));
								return false;
						}

					// Delete
					case 'delete' :

						// POST
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/subscribers/' . urlencode($email_address), 'DELETE', null, self::get_api_headers());

						// Process response
						switch($api_response['http_code']) {

							case 204 :

								self::success(__('Successfully deleted', 'ws-form-drip-v2'));
								return true;

							case 404 :

								self::success(__('Email address not found', 'ws-form-drip-v2'));
								return true;

							default :

								// Get invalid merge fields
								self::error(self::get_api_error_detail($api_response['response']));
								return false;
						}
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-drip-v2')); }

				// Get subscriber
				$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/subscribers/' . urlencode($user_email), 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
				if(!isset($api_response_decoded->subscribers)) { return self::error(__('No subscribers found', 'ws-form-drip-v2')); }
				if(count($api_response_decoded->subscribers) == 0) { return self::error(__('No subscribers found', 'ws-form-drip-v2')); }

				// Get first subscriber
				$subscriber = $api_response_decoded->subscribers[0];

				// Process fields
				$fields_return = array(

					self::EMAIL_MERGE_FIELD	=>	parent::get_object_value($subscriber, 'email'),
				);

				$custom_fields = parent::get_object_value($subscriber, 'custom_fields');
				if($custom_fields !== false) {

					$custom_fields = get_object_vars($custom_fields);
					foreach($custom_fields as $id => $value) {

						if(is_null($value)) { continue; }

						$fields_return[$id] = (string) $value;
					}
				}

				// Process tags
				$tags_return = array();

				$tags = parent::get_object_value($subscriber, 'tags');
				if($tags !== false) {

					foreach($tags as $tag) {

						$tags_return[$tag] = true;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => $tags_return);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists_' . $this->account_id);

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array(

						array(

							'id' => 			'0', 
							'label' => 			'People', 
							'field_count' => 	false, 
							'record_count' => 	false
						)
					);

					// Load configuration
					self::load_config();

					// Initialize
					$total_pages = 0;
					$page = 0;

					do {

						$page++;

						// Params
						$params = array(

							'page'	=> $page
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/campaigns', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
						if(!isset($api_response_decoded->meta)) { return self::error(__('List meta not found in API response data', 'ws-form-drip-v2')); }
						if(!isset($api_response_decoded->campaigns)) { return self::error(__('Lists not found in API response data', 'ws-form-drip-v2')); }

						// Get total field count
						if(isset($api_response_decoded->meta->total_pages)) {

							$total_pages = $api_response_decoded->meta->total_pages;
						}

						foreach($api_response_decoded->campaigns as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	false, 
								'record_count' => 	parent::get_object_value($list, 
								'active_subscriber_count')
							);
						}

					} while ($page < $total_pages);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists_' . $this->account_id, $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->account_id . '_' . $this->list_id);

				if($fetch || ($list === false)) {

					if($this->list_id == '0') {

						$list_name = 'People';

					} else {

						// Load configuration
						self::load_config();

						// Get list
						$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/campaigns/' . $this->list_id, 'GET', null, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_list = json_decode($api_response['response']);
						if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
						if(!isset($api_response_list->campaigns)) { return self::error(__('No campaigns returned', 'ws-form-drip-v2')); }
						if(count($api_response_list->campaigns) == 0) { return self::error(__('No campaigns returned', 'ws-form-drip-v2')); }

						$campaign = $api_response_list->campaigns[0];

						$list_name = parent::get_object_value($campaign, 'name');
					}

					// Build list
					$list = array(

						'label' => $list_name
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->account_id . '_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->account_id . '_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/custom_field_identifiers', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
					if(!isset($api_response_decoded->custom_field_identifiers)) { return self::error(__('List fields not found in API response data', 'ws-form-drip-v2')); }

					// Get the fields
					$api_list_fields = $api_response_decoded->custom_field_identifiers;

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_MERGE_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Email', 
						'label_field' => 	'Email', 
						'label_sub' =>		false,
						'type' => 			'email', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Process fields
					$sort_index = 1;
					foreach($api_list_fields as $name) {

						$list_fields[] = array(

							'id' => 			$name,
							'label' => 			$name, 
							'label_field' => 	$name, 
							'type' => 			'text',
							'action_type' =>	'text',
							'required' => 		false, 
							'default_value' => 	'',
							'pattern' => 		'',
							'placeholder' => 	'',
							'input_mask' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		true,
							'meta' => 			false
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->account_id . '_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_tag_mapping'		=>	'tag_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				$tag_categories = array(

					array(

						'id' => 		1,
						'label' => 		__('Tags', 'ws-form-drip-v2'), 
						'sort_index' =>	1,
						'type' => 		'checkbox'
					)
				);

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = false, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check tag category ID is set
				if($tag_category_id === false) { return self::error(__('Tag category ID is not set', 'ws-form-drip-v2')); }

				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $this->account_id . '_' . $tag_category_id);

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Get tags
					$api_response = parent::api_call($this->api_endpoint, $this->account_id . '/tags', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process tag data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
					if(!isset($api_response_decoded->tags)) { return self::error(__('Tags not found in API response data', 'ws-form-drip-v2')); }

					// Get the tags
					$tags_api = $api_response_decoded->tags;

					$sort_index = 0;

					foreach($tags_api as $tag) {

						$tags[] = array(

							'id' => 			$tag,
							'label' => 			$tag,
							'record_count' =>	false,
							'sort_index' =>		$sort_index
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $this->account_id . '_' . $tag_category_id, $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-drip-v2'); }
				if(!isset($api_response_decoded->errors)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-drip-v2'); }

				// Get error title from JSON response
				$error_detail_array = array();
				foreach($api_response_decoded->errors as $error) {

					$error_detail_array[] = $error->message;
				}

				return implode(', ', $error_detail_array);
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_method',
						'action_' . $this->id . '_double_opt_in',
						'action_' . $this->id . '_prospect',
						'action_' . $this->id . '_base_lead_score',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_tag_mapping',
						'action_' . $this->id . '_tag_merge'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-drip-v2') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-drip-v2')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('Drip Campaign', 'ws-form-drip-v2'),
						'type'						=>	'select',
						'help'						=>	__('Select the Drip campaign to associate this form with.', 'ws-form-drip-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-drip-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Method
					'action_' . $this->id . '_method'	=> array(

						'label'						=>	__('Method', 'ws-form-drip-v2'),
						'type'						=>	'select',
						'help'						=>	__("Select method to apply to this person.", 'ws-form-drip-v2'),
						'options'					=>	array(

							array('value' => 'create', 'text' => __('Create or update', 'ws-form-drip-v2')),
							array('value' => 'remove', 'text' => __('Remove from campaign(s)', 'ws-form-drip-v2')),
							array('value' => 'unsubscribe', 'text' => __('Unsubscribe', 'ws-form-drip-v2')),
							array('value' => 'delete', 'text' => __('Delete', 'ws-form-drip-v2'))
						),
						'default'					=>	'create'
					),

					// Double opt-in
					'action_' . $this->id . '_double_opt_in'	=> array(

						'label'						=>	__('Double Opt-In', 'ws-form-drip-v2'),
						'type'						=>	'checkbox',
						'help'						=>	__('If checked, a double opt-in confirmation email is sent.', 'ws-form-drip-v2'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'!=',
								'meta_key'			=>	'action_' . $this->id . '_list_id',
								'meta_value'		=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_method',
								'meta_value'		=>	'create',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Prospect
					'action_' . $this->id . '_prospect'	=> array(

						'label'						=>	__('Prospect', 'ws-form-drip-v2'),
						'type'						=>	'checkbox',
						'help'						=>	__('Whether a lead score should be attached to the person (when lead scoring is enabled).', 'ws-form-drip-v2'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'create'
							)
						)
					),

					// Base Lead Score
					'action_' . $this->id . '_base_lead_score'	=> array(

						'label'						=>	__('Base Lead Score', 'ws-form-drip-v2'),
						'type'						=>	'number',
						'help'						=>	__('The starting value for lead score calculation for new people.', 'ws-form-drip-v2'),
						'default'					=>	30,
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'create'
							)
						)
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-drip-v2'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-drip-v2'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-drip-v2'),
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'create'
							)
						)
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-drip-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Drip merge fields.', 'ws-form-drip-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'create'
							)
						)
					),

					// Custom field mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-drip-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map custom values to Drip attributes.', 'ws-form-drip-v2'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Update rules
					'action_' . $this->id . '_update_rule'	=> array(

						'label'							=>	__('On Update', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Update', 'ws-form-pipedrive-v1')),
							array('value' => 'merge', 'text' => __('Merge', 'ws-form-pipedrive-v1')),
							array('value' => 'ignore', 'text' => __('Ignore', 'ws-form-pipedrive-v1'))
						)
					),

					// Custom field mapping - Value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'	=>	__('Value', 'ws-form-drip-v2'),
						'type'	=>	'text'
					),

					// Tags
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-drip-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Drip tags.', 'ws-form-drip-v2'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field_choice'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'create'
							)
						)
					),

					// Keep existing tags
					'action_' . $this->id . '_tag_merge'	=> array(

						'label'						=>	__('Merge existing tags', 'ws-form-drip-v2'),
						'type'						=>	'checkbox',
						'help'						=>	__('Merge new tags with old tags.', 'ws-form-drip-v2')
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Drip Field', 'ws-form-drip-v2'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-drip-v2'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-drip-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-drip-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Drip add-on for WS Form PRO license key here.', 'ws-form-drip-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-drip-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Drip Status', 'ws-form-drip-v2'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Drip account, or <a href=\"https://www.getdrip.com/signup/starter\" target=\"blank\">create a new account</a>.", 'ws-form-drip-v2'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				// Get account list
				$account_list = WS_Form_Common::option_get('action_' . $this->id . '_account_list');

				if($this->configured && ($account_list === false)) {

					self::account_list_refresh();
				}

				$options['action_' . $this->id]['fields']['action_' . $this->id . '_account_id'] = WS_Form_Common::option_get('action_' . $this->id . '_account_list');

				return $options;
			}

			// Update account list
			public function account_list_refresh() {

				// Request accounts
				$api_response = parent::api_call($this->api_endpoint, 'accounts', 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-drip-v2')); }
				if(!isset($api_response_decoded->accounts)) { return self::error(__('No accounts found', 'ws-form-drip-v2')); }

				$account_id_array = array();

				if(count($api_response_decoded->accounts) == 0) {

					// No accounts returned
					$this->account_id = '';
					WS_Form_Common::option_set('action_' . $this->id . '_account_id', $this->account_id);

				} else {

					// Check to see if the current account ID can be found
					$account_id_found = false;
					foreach($api_response_decoded->accounts as $account) {

						if($account->id == $this->account_id) { $account_id_found = true; }
					}

					// If account ID not found, clear it
					if(!$account_id_found) { $this->account_id = ''; }

					// Set the default 
					foreach($api_response_decoded->accounts as $account) {

						$account_id_array[$account->id] = array('text' => $account->name);

						// Client ID not set, so set it
						if($this->account_id === '') {

							$this->account_id = $account->id;
							WS_Form_Common::option_set('action_' . $this->id . '_account_id', $this->account_id);
						}
					}

					// Accounts
					$account_list = array(

						'label'			=>	__('Drip Account', 'ws-form-drip-v2'),
						'type'			=>	'select',
						'options'		=>	$account_id_array,
						'default'		=>	$this->account_id,
						'help'			=>	__('Select the Drip Account to use.', 'ws-form-drip-v2'),
						'button'		=>	'account_list_action_' . $this->id
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_account_list', $account_list);
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->method = parent::get_config($config, 'action_' . $this->id . '_method', 'create');
				$this->double_opt_in = parent::get_config($config, 'action_' . $this->id . '_double_opt_in', false);
				$this->prospect = parent::get_config($config, 'action_' . $this->id . '_prospect', true);
				$this->base_lead_score = parent::get_config($config, 'action_' . $this->id . '_base_lead_score', 30);
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
				$this->tag_merge = parent::get_config($config, 'action_' . $this->id . '_tag_merge', false);
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->account_id  				= WS_Form_Common::option_get('action_' . $this->id . '_account_id', '');

				$this->api_endpoint = 'https://api.getdrip.com/v2/';

				if(empty($this->api_access_token)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Convert email to action record ID
			public function ws_form_email_to_action_record_id($email_address) {

				return md5($email_address);
			}
		
			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<style>.drip_logo{fill:#f224f2}</style><g transform="translate(53.000000, 66.000000)"><path class="drip_logo" d="M34.54 27.86h-5.19c.01.18.02.36.02.53 0 7.67-6.28 12.42-12.09 12.42s-12.1-4.74-12.1-12.42c0-.17.01-.35.02-.53H.01c-.01.18-.01.36-.01.53C0 38.96 8.64 46 17.28 46s17.28-7.04 17.28-17.61c-.01-.17-.01-.35-.02-.53zM17.28 6.58l3.7 4.99h6.53C22.73 5 17.28 0 17.28 0S11.82 5 7.04 11.57h6.53l3.71-4.99z"/><circle class="drip_logo" cx="17.28" cy="19.83" r="3.46"/><circle class="drip_logo" cx="27.64" cy="19.83" r="3.46"/><circle class="drip_logo" cx="6.91" cy="19.83" r="3.46"/></g>';

				return $svg_logo;
			}
		}
	});