<?php

	/**
	 * @link              https://wsform.com/knowledgebase/dropbox/
	 * @since             1.0.0
	 * @package           WS_Form_Dropbox_V2
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Dropbox
	 * Plugin URI:        https://wsform.com/knowledgebase/dropbox/
	 * Description:       Dropbox add-on for WS Form PRO
	 * Version:           1.0.2
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-dropbox-v2
	 */

	Class WS_Form_Add_On_Dropbox_V2 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.145';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_File_Handler_Dropbox_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-dropbox-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_file_handler_dropbox_v2 = new WS_Form_Add_On_Dropbox_V2();	

	register_activation_hook(__FILE__, array($wsf_file_handler_dropbox_v2, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_File_Handler_Dropbox_V2 extends WS_Form_File_Handler {

			public $id = 'dropboxv2';
			public $pro_required = false;
			public $label;
			public $public = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 28974;
			const WS_FORM_LICENSE_NAME = 'Dropbox add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.2';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const FILE_UPLOAD_SESSION_SIZE = 150000000;	// If file > 150,000,000 bytes, we'll use upload sessions
			const PREVIEW_URL = 'https://www.dropbox.com/preview';
			const DEFAULT_UPLOAD_PATH = '/Apps/WS Form/#form_id/#submit_hash/#field_id/';

			public function __construct() {

				// Set label
				$this->label = __('Dropbox', 'ws-form');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();

				// Create intial file handler
				add_filter('wsf_file_handler_' . $this->id, array($this, 'handler'), 10, 5);
			}

			// Handler
			public function handler($file_objects, $submit, $field, $section_repeatable_index) {

				// Check form ID
				WS_Form_Common::check_form_id($submit->form_id);

				// Field ID
				$field_id = intval($field->id);
				if($field_id == 0) { parent::db_throw_error(__('Invalid field ID', 'ws-form')); }

				foreach($file_objects as $file_object_index => $file_object) {

					// Get file name
					if(!isset($file_object['name'])) { parent::db_throw_error(__('File name not found in file object', 'ws-form')); }
					$file_name = $file_object['name'];

					// Get file type
					if(!isset($file_object['type'])) { parent::db_throw_error(__('File type not found in file object', 'ws-form')); }
					$file_type = $file_object['type'];

					// Get file path
					if(!isset($file_object['path'])) { parent::db_throw_error(__('File source path not found in file object', 'ws-form')); }
					$file_path = $file_object['path'];

					// Get file size
					if(!isset($file_object['size'])) { parent::db_throw_error(__('File size not found in file object', 'ws-form')); }
					$file_size = $file_object['size'];

					// Get form
					$form = isset($submit->form_object) ? $submit->form_object : false;

					// Get path
					$upload_path = WS_Form_Common::get_object_meta_value($field, 'file_handler_' . $this->id . '_upload_path', self::DEFAULT_UPLOAD_PATH);

					// Parse with additional parameters
					$variables = array(

						'form_id' => $submit->form_id,
						'submit_hash' => $submit->hash,
						'field_id' => $field->id
					);
					$upload_path = WS_Form_Common::mask_parse($upload_path, $variables);

					// Regular parse
					if(strpos($upload_path, '#') !== false) {

						$upload_path = WS_Form_Common::parse_variables_process($upload_path, $form, $submit, 'text/plain');
					}

					// Check Dropbox upload path
					if(substr($upload_path, -1) !== '/') { $upload_path .= '/'; }

					if($file_size < self::FILE_UPLOAD_SESSION_SIZE) {

						// Get file data
						$file_data = file_get_contents($file_path);

						// Build headers
						$headers = self::get_api_headers();

						// Build Dropbox API arg
						$dropbox_api_arg = array(

							'path' => sprintf('%s%s', $upload_path, $file_name),
							'mode' => 'add',
							'autorename' => true,
							'mute' => false,
							'strict_conflict' => false
						);
						$headers['Dropbox-API-Arg'] = json_encode($dropbox_api_arg);

						// Push to Dropbox
						$api_response = parent::api_call('https://content.dropboxapi.com/2/', 'files/upload', 'POST', $file_data, $headers, false, false, false, 'application/json', 'application/octet-stream');

						// Check status code
						if($api_response['http_code'] != 200) {

							parent::db_throw_error(self::get_api_error_detail($api_response['response']));
						}

						// Process upload data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { parent::db_throw_error(__('Invalid API response data', 'ws-form-dropbox-v2')); }

						// Set path
						if(!isset($api_response_decoded->path_lower)) { parent::db_throw_error(__('No file path returned', 'ws-form-dropbox-v2')); }
						$path_lower = $api_response_decoded->path_lower;

					} else {

						// Upload according to file size
						$bytes_uploaded = 0;

						// Session ID
						$session_id = false;

						// Upload session
						do {

							// Get file data
							$file_data = file_get_contents($file_path, false, null, $bytes_uploaded, self::FILE_UPLOAD_SESSION_SIZE);

							// Build headers
							$headers = self::get_api_headers();

							if($session_id === false) {

								// Create Dropbox upload session

								// Build Dropbox API arg
								$dropbox_api_arg = array(

									'close' => false
								);
								$headers['Dropbox-API-Arg'] = json_encode($dropbox_api_arg);

								// Push to Dropbox
								$api_response = parent::api_call('https://content.dropboxapi.com/2/', 'files/upload_session/start', 'POST', $file_data, $headers, false, false, false, 'application/json', 'application/octet-stream');

								// Check status code
								if($api_response['http_code'] != 200) {

									parent::db_throw_error(self::get_api_error_detail($api_response['response']));
								}

								// Process upload data
								$api_response_decoded = json_decode($api_response['response']);
								if(is_null($api_response_decoded)) { parent::db_throw_error(__('Invalid API response data', 'ws-form-dropbox-v2')); }

								// Set session ID
								if(!isset($api_response_decoded->session_id)) { parent::db_throw_error(__('No session ID returned', 'ws-form-dropbox-v2')); }
								$session_id = $api_response_decoded->session_id;

							} else {

								// Append Dropbox upload session

								// Is this the final session?
								$upload_session_finish = (($bytes_uploaded + self::FILE_UPLOAD_SESSION_SIZE) >= $file_size);

								// Build Dropbox API arg
								$dropbox_api_arg = array(

									'cursor' => array(

										'session_id' => $session_id,
										'offset' => $bytes_uploaded
									)
								);

								if($upload_session_finish) {

									$dropbox_api_arg['commit'] = array(

										'path' => sprintf('%s%s', $upload_path, $file_name),
										'mode' => 'add',
										'autorename' => true,
										'mute' => false,
										'strict_conflict' => false
									);

								} else {

									$dropbox_api_arg['close'] = false;
								}

								$headers['Dropbox-API-Arg'] = json_encode($dropbox_api_arg);

								// Determine API path
								$api_path = $upload_session_finish ? 'files/upload_session/finish' : 'files/upload_session/append_v2';

								// Push to Dropbox
								$api_response = parent::api_call('https://content.dropboxapi.com/2/', $api_path, 'POST', $file_data, $headers, false, false, false, 'application/json', 'application/octet-stream');

								// Check status code
								if($api_response['http_code'] != 200) {

									parent::db_throw_error(self::get_api_error_detail($api_response['response']));
								}

								if($upload_session_finish) {

									// Process upload data
									$api_response_decoded = json_decode($api_response['response']);
									if(is_null($api_response_decoded)) { parent::db_throw_error(__('Invalid API response data', 'ws-form-dropbox-v2')); }

									// Set path
									if(!isset($api_response_decoded->path_lower)) { parent::db_throw_error(__('No file path returned', 'ws-form-dropbox-v2')); }
									$path_lower = $api_response_decoded->path_lower;

									break;
								}
							}

							$bytes_uploaded += self::FILE_UPLOAD_SESSION_SIZE;

	 					} while ($bytes_uploaded < $file_size);
	 				}

					// Set path
					$file_objects[$file_object_index]['path'] = $path_lower;
					if(isset($api_response_decoded->id)) {

						$file_objects[$file_object_index]['dropbox_id'] = $api_response_decoded->id;
					}
					if(isset($api_response_decoded->rev)) {

						$file_objects[$file_object_index]['dropbox_rev'] = $api_response_decoded->rev;
					}
					if(isset($api_response_decoded->content_hash)) {

						$file_objects[$file_object_index]['dropbox_content_hash'] = $api_response_decoded->content_hash;
					}
				}

				self::touch($file_objects);

				return $file_objects;
			}

			// Get URL
			public function get_url($file_object, $field_id = 0, $file_object_index = 0, $submit_hash = '') {

				// Ensure this file object belongs to this file handler
				if(!isset($file_object['handler']) || ($file_object['handler'] != $this->id)) { return false; }

				// Check file path exists
				if(!isset($file_object['path']) || ($file_object['path'] == '')) { return false; }

				return self::PREVIEW_URL . $file_object['path'];
			}

			// Copy to file
			public function copy_to_temp_file($file_object, $temp_path = false) {

				// Ensure this file object belongs to this file handler
				if(!isset($file_object['handler']) || ($file_object['handler'] != $this->id)) { return false; }

				// Check path
				if(!isset($file_object['path']) || ($file_object['path'] == '')) { return false; }
				$path = $file_object['path'];

				// Check name
				if(!isset($file_object['name']) || ($file_object['name'] == '')) { return false; }
				$name = $file_object['name'];
				if(in_array($name, array( '.', '/', '\\'))) { return false; }

				// Check dropbox_content_hash
				if(!isset($file_object['dropbox_content_hash']) || ($file_object['dropbox_content_hash'] == '')) { return false; }
				$dropbox_content_hash = $file_object['dropbox_content_hash'];
				if($dropbox_content_hash == '') { return false; }
				if(in_array($dropbox_content_hash, array( '.', '/', '\\'))) { return false; }

				// Build headers
				$headers = self::get_api_headers();

				// Build Dropbox API arg
				$dropbox_api_arg = array(

					'path' => $path
				);
				$headers['Dropbox-API-Arg'] = json_encode($dropbox_api_arg);

				// Push to Dropbox
				$api_response = parent::api_call('https://content.dropboxapi.com/2/', 'files/download', 'POST', false, $headers, false, false, false, 'application/json', 'application/octet-stream');

				// Check status code
				if($api_response['http_code'] != 200) {

					parent::db_throw_error(self::get_api_error_detail($api_response['response']));
				}

				// Get file data
				$file_data = $api_response['response'];

				// Get file path to copy to
				require_once(ABSPATH . 'wp-admin/includes/file.php');

				if($temp_path === false) {

					$file_path_copy_to = wp_tempnam();

				} else {

					if(!file_exists($temp_path)) {

						mkdir($temp_path);
					}

					if(!isset($file_object['name']) || ($file_object['name'] == '')) { return false; }

					$file_path_copy_to = $temp_path . '/' . $file_object['name'];
				}

				// Create temporary file
				return file_put_contents($file_path_copy_to, $file_data) ? $file_path_copy_to : false;
			}

			// Get temp file
			public function get_temp_file($file_object, $temp_path) {

				$copy_to_temp_file_return = self::copy_to_temp_file($file_object, $temp_path);
				if($copy_to_temp_file_return === false) { return false; }

				return array(

					'path' 				=> $copy_to_temp_file_return,
					'unlink_after_use' 	=> true
				);
			}			

			// Delete file
			public function delete($file_object) {

				// Ensure this file object belongs to this file handler
				if(!isset($file_object['handler']) || ($file_object['handler'] != $this->id)) { return false; }

				// Check attachment id
				if(!isset($file_object['path']) || ($file_object['path'] == '')) { return false; }
				$path = $file_object['path'];

				// Build headers
				$headers = self::get_api_headers();

				// Build Dropbox params
				$params = array(

					'path' => $path
				);

				// Push to Dropbox
				$api_response = parent::api_call('https://api.dropboxapi.com/2/', 'files/delete_v2', 'POST', json_encode($params), $headers, false, false, false, 'application/json', 'application/json');

				return true;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save access token expiry
				$this->api_access_token_expiry = time() + intval(WS_Form_Common::get_query_var($this->id . '_expires_in', false)) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = '/dropbox-v2?refresh_token=' . rawurlencode($this->api_refresh_token);
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if(
					($api_response['http_code'] == 200) &&
					($api_response['error'] === false)
				) {

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return parent::db_throw_error(__('Invalid API response data', 'ws-form-dropbox-v2')); }

					// Save access token
					$this->api_access_token = isset($api_response_decoded->{$this->id . '_access_token'}) ? $api_response_decoded->{$this->id . '_access_token'} : '';
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save access token expiry
					$this->api_access_token_expiry = time() + intval(isset($api_response_decoded->{$this->id . '_expires_in'}) ? $api_response_decoded->{$this->id . '_expires_in'} : '') - self::ACCESS_TOKEN_EXPIRY_WINDOW;
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-dropbox-v2')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-dropbox-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-dropbox-v2') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-dropbox-v2') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/dropbox-v2';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-dropbox-v2'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Dropbox account <a href="%s">here</a>.', 'ws-form-dropbox-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { 

					if(is_string($response)) {

						return $response;

					} else {

						return __('Invalid API response data', 'ws-form-dropbox-v2');
					}
				}

				return $api_response_decoded->error_summary;
			}

			// Get settings
			public function get_file_handler_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'file_handler_' . $this->id . '_upload_path'
					)
				);

				return $settings;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Upload path
					'file_handler_' . $this->id . '_upload_path'	=> array(

						'label'			=>	__('Upload Path', 'ws-form-dropbox-v2'),
						'type'			=>	'text',
						'help'			=>	__('Enter the Dropbox path to upload the file(s) to.', 'ws-form-dropbox-v2'),
						'default'		=>	self::DEFAULT_UPLOAD_PATH,
						'placeholder'	=>	'/',
						'condition'		=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'file_handler',
								'meta_value'	=>	$this->id
							)
						),
						'select_list'	=>	true
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-dropbox-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-dropbox-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Dropbox add-on for WS Form PRO license key here.', 'ws-form-dropbox-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-dropbox-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Dropbox Status', 'ws-form-dropbox-v2'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Dropbox account, or <a href=\"https://www.dropbox.com/\" target=\"blank\">create a new account</a>.", 'ws-form-dropbox-v2'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				return $options;
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry  = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');

				// Set endpoint URL
				$this->api_endpoint = 'https://content.dropboxapi.com/2/';

				// Build endpoint using API key
				if(empty($this->api_access_token)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Get API headers
			public function get_api_headers() {

				// Access token check
				self::access_token_check();

				// Build header
				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}
		}
	});