<?php

/**
 * @link              https://wsform.com/knowledgebase/getresponse/
 * @since             1.0.0
 * @package           WS_Form_GetResponse_V3
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - GetResponse
 * Plugin URI:        https://wsform.com/knowledgebase/getresponse/
 * Description:       GetResponse add-on for WS Form PRO
 * Version:           1.1.0
 * Author:            20script
 * Author URI:        https://www.20script.ir/
 * Text Domain:       ws-form-getresponse-v3
 */

	Class WS_Form_Add_On_GetResponse_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.124';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_GetResponse_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-getresponse-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_getresponse_v3 = new WS_Form_Add_On_GetResponse_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_getresponse_v3, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_GetResponse_V3 extends WS_Form_Action {

			public $id = 'getresponsev3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_key;

			public $list_id = false;
			public $opt_in_field = false;
			public $field_mapping = false;
			public $tag_mapping = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 3372;
			const WS_FORM_LICENSE_NAME = 'GetResponse add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.1.0';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const EMAIL_MERGE_FIELD = 'email';
			const NAME_MERGE_FIELD = 'name';
			const LISTS_FETCH_COUNT = 100;

			public function __construct() {

				// Set label
				$this->label = __('GetResponse', 'ws-form-getresponse-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to GetResponse', 'ws-form-getresponse-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-getresponse-v3')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-getresponse-v3') . '" />';

						} else {

							$value = '<input class="button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-getresponse-v3') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your GetResponse account <a href="%s">here</a>.', 'ws-form-getresponse-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-getresponse-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$custom_field_values = array();
				$data = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					switch($api_field_string) {

						case self::EMAIL_MERGE_FIELD :
						case self::NAME_MERGE_FIELD :

							$data[$api_field_string] = $submit_value;
							break;

						default :

							if(!is_array($submit_value)) { $submit_value = array($submit_value); }

							$custom_field_values[] = array(

								'customFieldId' 	=> $api_field_string,
								'value' 			=> $submit_value
							);

							break;
					}
				}

				// Check email
				if(
					!isset($data[self::EMAIL_MERGE_FIELD]) ||
					(!filter_var($data[self::EMAIL_MERGE_FIELD], FILTER_VALIDATE_EMAIL))
				) {

					return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-getresponse-v3'), $email_address));
				}

				// Process tags
				$tags = array();
				$tags_push = false;
				foreach($this->tag_mapping as $field_map) {

					// Get field ID
					$field_id = $field_map['ws_form_field'];
					if($field_id == '') { continue; }

					$tags_push = true;

					// Read submit meta
					$tags_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($tags_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($tags_array)) { $tags_array = array($tags_array); }

					// Process each tag ready for API format
					foreach($tags_array as $tag) {

						if($tag == '') { continue; }

						$tags[] = array('tagId' => $tag);
					}
				}

				// Build post data
				if(count($custom_field_values) > 0) { $data['customFieldValues'] = $custom_field_values; }
				if($tags_push) { $data['tags'] = $tags; }

				// Tracking data - Remote IP
				$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
				if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

					$data['ipAddress'] = $tracking_remote_ip;
				}

				// List ID
				$data['campaign']['campaignId'] = $this->list_id;

				// JSON encode
				$data_json = json_encode($data);

				// Post
				$api_response = parent::api_call($this->api_endpoint, 'contacts', 'POST', $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 202 :

						self::success(__('Successfully added to list: ', 'ws-form-getresponse-v3') . $this->list_id);
						return true;

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-getresponse-v3')); }

				// Parameters
				$params = array(

					'query[email]' 	=> $user_email,
					'fields'		=> 'contactId',
					'page'			=> 1,
					'perPage'		=> 1
				);

				// Find contact
				$api_response = parent::api_call($this->api_endpoint, 'contacts', 'GET', $params, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

				if(count($api_response_decoded) != 1) { return false; }

				$contact_id = $api_response_decoded[0]->contactId;

				// Get contact
				$api_response = parent::api_call($this->api_endpoint, 'contacts/' . $contact_id, 'GET', array(), self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

				// Process fields
				$fields_return = array(

					self::EMAIL_MERGE_FIELD	=>	parent::get_object_value($api_response_decoded, self::EMAIL_MERGE_FIELD),
					self::NAME_MERGE_FIELD	=>	parent::get_object_value($api_response_decoded, self::NAME_MERGE_FIELD),
				);

				$custom_field_values = parent::get_object_value($api_response_decoded, 'customFieldValues');
				if($custom_field_values !== false) {

					foreach($custom_field_values as $custom_field) {

						$fields_return[$custom_field->customFieldId] = (count($custom_field->values) == 1) ? $custom_field->values[0] : $custom_field->values;
					}
				}

				$tags_return = array();

				$tags = parent::get_object_value($api_response_decoded, 'tags');
				if($tags !== false) {

					foreach($tags as $id => $value) {

						$tags_return[$value->tagId] = true;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => $tags_return);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$page = 0;

					do {

						$page++;

						// Params
						$params = array(

							'perPage'		=> self::LISTS_FETCH_COUNT,
							'page'			=> $page,
							'sort[name]'	=> 'asc',
							'fields'		=> 'campaignId,name',
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'campaigns', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Get total pages
						$total_pages = $api_response['headers']['totalpages'];

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

						foreach($api_response_decoded as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'campaignId'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	false, 
								'record_count' => 	false
							);
						}

					} while ($page < $total_pages);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'fields'		=> 'name',
					);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'campaigns/' . $this->list_id, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(sprintf(__('HTTP Status Code: %s', 'ws-form-getresponse-v3'), $api_response['http_code'])); return false; }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); return false; }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_decoded, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();
					$page = 0;

					do {

						$page++;

						// Params
						$params = array(

							'perPage'		=> self::LISTS_FETCH_COUNT,
							'page'			=> $page,
							'sort[name]'	=> 'asc'
						);

						// Get list fields
						$api_response = parent::api_call($this->api_endpoint, 'custom-fields', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Get total pages
						$total_pages = $api_response['headers']['totalpages'];

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);

						// Checks
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

						$api_list_fields = array_merge($api_list_fields, $api_response_decoded);

					} while ($page < $total_pages);

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_MERGE_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Email', 
						'label_field' => 	'Email', 
						'label_sub' =>		false,
						'type' => 			'email', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Add name
					$list_fields[] = array(

						'id' => 			self::NAME_MERGE_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Name', 
						'label_field' => 	'Name', 
						'label_sub' =>		false,
						'type' => 			'text', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	1,
						'visible' =>		true
					);

					// Process fields
					$sort_index = 2;
					foreach($api_list_fields as $field) {

						$field_type = parent::get_object_value($field, 'fieldType');

						$action_options = parent::get_object_value($field, 'values');

						// WS Form type
						$type = self::action_field_type_to_ws_form_field_type($field);

						// Meta data
						$meta = self::action_field_to_ws_form_meta_keys($field);

						// Determine default value
						$default_value = (($meta === false) && (count($action_options) == 1)) ? $action_options[0] : '';

						$list_fields[] = array(

							'id' => 			parent::get_object_value($field, 'customFieldId'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$field_type,
							'required' => 		false,
							'default_value' => 	$default_value,
							'pattern' => 		'',
							'placeholder' => 	'',
							'input_mask' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		!(parent::get_object_value($field, 'hidden')),
							'meta' => 			$meta
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
							'action_' . $this->id . '_name_mask'		=>	'#field(#Name)'
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				$tag_categories = array(

					array(

						'id' => 		1,
						'label' => 		__('Tags', 'ws-form-getresponse-v3'), 
						'sort_index' =>	1,
						'type' => 		'checkbox'
					)
				);

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = false, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check to see if tags are cached
				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id);

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Initialize
					$page = 0;

					do {

						$page++;

						// Params
						$params = array(

							'perPage'		=> self::LISTS_FETCH_COUNT,
							'page'			=> $page,
							'fields'		=> 'tagId,name',
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'tags', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Get total pages
						$total_pages = $api_response['headers']['totalpages'];

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

						$sort_index = 1;

						foreach($api_response_decoded as $tag) {

							$tags[] = array(

								'id' => 			parent::get_object_value($tag, 'tagId'),
								'label' => 			parent::get_object_value($tag, 'name'),
								'record_count' =>	false,
								'sort_index' =>		$sort_index
							);

							$sort_index++;
						}

					} while ($page < $total_pages);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id, $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-getresponse-v3')); }

				// Get error detail from JSON response
				$error_detail_array = array();
				if(
					(isset($api_response_decoded->context)) &&
					(is_array($api_response_decoded->context))
				) {

					foreach($api_response_decoded->context as $context) {

						if(isset($context->fieldName)) { $error_detail_array[] = is_array($context->fieldName) ? implode(', ', $context->fieldName) : $context->fieldName; }
						if(isset($context->errorDescription)) { $error_detail_array[] = $context->errorDescription; }
						if(is_string($context)) { $error_detail_array[] = $context; }
					}
				}
				$error_detail_array[] = $api_response_decoded->codeDescription;

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_tag_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-getresponse-v3'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-getresponse-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('GetResponse List', 'ws-form-getresponse-v3'),
						'type'							=>	'select',
						'help'							=>	__('Which GetResponse list should this subscriber be added to?', 'ws-form-getresponse-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-getresponse-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-getresponse-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-getresponse-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-getresponse-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-getresponse-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to GetResponse merge fields.', 'ws-form-getresponse-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-getresponse-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to AWeber tags.', 'ws-form-getresponse-v3'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('GetResponse Field', 'ws-form-getresponse-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-getresponse-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-getresponse-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-getresponse-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your GetResponse add-on for WS Form PRO license key here.', 'ws-form-getresponse-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-getresponse-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'		=>	__('GetResponse API Key', 'ws-form-getresponse-v3'),
							'type'		=>	'text',
							'default'	=>	'',
							'help'		=>	__('<a href="https://wsform.com/knowledgebase/getresponse/" target="_blank">Learn more</a>', 'ws-form-getresponse-v3')
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('GetResponse Status', 'ws-form-getresponse-v3'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Enter your <a href=\"https://app.getresponse.com/api\" target=\"blank\">GetResponse API key</a> above, or <a href=\"https://www.getresponse.com/start-free-trial\" target=\"blank\">create a new account</a>.", 'ws-form-getresponse-v3')
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($field) {

				$field_type = parent::get_object_value($field, 'fieldType');
				$type = parent::get_object_value($field, 'type');

				// Field type
				switch($field_type) {

					case 'radio' : return 'radio';
					case 'checkbox' : return 'checkbox';
					case 'multi_select' : return 'select';
					case 'single_select' : return 'select';
				}

				// Type
				switch($type) {

					case 'country' :
					case 'currency' :
					case 'gender' :

						return 'select';

					case 'date' :
					case 'datetime' :

						switch($field_type) {

							case 'text' : return 'datetime';
						}
						break;

					case 'ip' : return 'text';

					case 'number' :

						switch($field_type) {

							case 'text' : return 'number';
							case 'textarea' : return 'textarea';
						}
						break;

					case 'phone' : return 'tel';

					case 'text' :

						switch($field_type) {

							case 'text' : return 'text';
							case 'textarea' : return 'textarea';
						}
						break;

					case 'url' : return 'url';
				}

				return 'text';
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$field_type = parent::get_object_value($field, 'fieldType');
				$type = parent::get_object_value($field, 'type');

				$multiple = false;

				// Get WS Form meta configurations for action field types
				switch($field_type) {

					// Build meta for datetime
					case 'text' :

						switch($type) {

							case 'datetime' :

								return array('input_type_datetime' => 'datetime-local');

							case 'phone' :

								return array('input_mask' => '+19999999999');
						}
						break;

					// Build data grids for radio and select
					case 'single_select' :
					case 'multi_select' :
					case 'checkbox' :
					case 'radio' :

						switch($field_type) {

							case 'single_select' : $meta_key = 'data_grid_select'; break;
							case 'multi_select' : $meta_key = 'data_grid_select'; $multiple = true; break;
							case 'checkbox' : $meta_key = 'data_grid_checkbox'; break;
							case 'radio' : $meta_key = 'data_grid_radio'; break;
						}

						// Get multiple
						$multiple = parent::get_object_value($field, 'values');

						// Get options
						$options = parent::get_object_value($field, 'values');

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys[$meta_key])) { return false; }
						if(!isset($meta_keys[$meta_key]['default'])) { return false; }

						$meta = $meta_keys[$meta_key]['default'];

						// Build new rows
						$rows = array();
						$id = 1;
						$default_value = parent::get_object_value($field, 'default_value');
						if($default_value === false) { $default_value = ''; }
						foreach($options as $option) {

							$rows[] = array(

								'id'		=> $id,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array($option)
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						$return_array = array($meta_key => $meta);

						// Multiple
						if($type = 'multi_select') { $return_array['multiple'] = 'on'; }

						return($return_array);

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_key = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');

				// Set endpoint URL
				$this->api_endpoint = 'https://api.getresponse.com/v3/';

				// Build endpoint using API key
				if(empty($this->api_key)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists')));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list')));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields')));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists')));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list')));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields')));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('X-Auth-Token' => 'api-key ' . $this->api_key);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(47.000000, 75.000000)"><path d="M48 28.5c0 1.1-.9 2.1-2.1 2.1H2.1C1 30.6 0 29.7 0 28.5V2.1C0 .9.9 0 2.1 0H46c1.1 0 2 .9 2 2.1v26.4z" fill="#00adee"/><path d="M47.5.8C41.9 9.6 33 17.7 23.9 17.5c-3.4-.1-7.2-1-10.4-3.7-2.4-2-4.8-5.1-5.8-9.4h-1c-1.5 0-2.4 1.3-2.2 2.5 0 .3.1.6.1.9.8 4.3 3.4 9.1 7.6 12.2 3.3 2.4 7.4 4 11.8 4 8 0 16.6-5.1 24-18.2V2.1c0-.5-.2-1-.5-1.3z" fill="#fff"/></g>';

				return $svg_logo;
			}
		}
	});