<?php

	/**
	* @link              https://wsform.com/knowledgebase/groundhogg/
	* @since             1.0.0
	* @package           WS_Form_Groundhogg
	*
	* @wordpress-plugin
	* Plugin Name:       WS Form PRO - Groundhogg
	* Plugin URI:        https://wsform.com/knowledgebase/groundhogg/
	* Description:       Groundhogg add-on for WS Form PRO
	* Version:           1.0.4
	* Author:            WS Form
	* Author URI:        https://www.wsform.com/
	* Text Domain:       ws-form-groundhogg
	*/

	Class WS_Form_Add_On_Groundhogg {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.8.11';
		const GROUNDHOGG_VERSION_MIN	= '2.2.22';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Groundhogg();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0) &&
				defined('GROUNDHOGG_VERSION') &&
				(version_compare(GROUNDHOGG_VERSION, self::GROUNDHOGG_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if ( ! current_user_can( 'update_plugins' ) ) {

				return;
			}
			if ( plugin_basename(__FILE__) != $file ) {

				return;
			}

			// Build notice
			printf( '<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>',

				/* translators: %1$s: WS Form PRO product link, %2$s: WS Form PRO minimum version, %3$s: Minimum WooCommerce version */
				sprintf(__('This add-on requires <a href="%1$s" target="_blank">WS Form PRO</a> (version %2$s or later) and Groundhogg (version %3$s or later) to be installed and activated.', 'ws-form-woocommerce')

					, esc_url('https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins')
					, esc_html(self::WS_FORM_PRO_VERSION_MIN)
					, esc_html(self::GROUNDHOGG_VERSION_MIN)
				)
			);
		}
	}

	$wsf_add_on_groundhogg = new WS_Form_Add_On_Groundhogg();	

	register_activation_hook(__FILE__, array($wsf_add_on_groundhogg, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Groundhogg extends WS_Form_Action {

			public $id = 'groundhogg';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			public $opt_in_field;
			public $opt_in_status;
			public $field_mapping;
			public $custom_mapping;
			public $attachment_mapping;
			public $meta;
			public $owner_id;
			public $tag_mapping;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 22144;
			const WS_FORM_LICENSE_NAME = 'Groundhogg add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.4';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Groundhogg', 'ws-form-groundhogg');

				// Set label for actions pull down
				$this->label_action = __('Add to Groundhogg', 'ws-form-groundhogg');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('rest_api_init', array($this, 'rest_api_init'));

				// Groundhogg actions
				add_action('groundhogg/steps/init', array($this, 'groundhogg_steps_init'));

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_template_form_config_files', array( $this, 'template_config_files' ), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Groundhogg - Steps - Init
			public function groundhogg_steps_init($manager) {

				$manager->add_step(new WS_Form_Groundhogg_Benchmark());
			}

			// Template - Config files
			public function template_config_files($config_files) {

				$config_files[] = plugin_dir_path(__FILE__) . 'includes/template/config.json';

				return $config_files;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-groundhogg')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-groundhogg') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-groundhogg') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

					break;
				}
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('The WS Form Groundhogg add-on requires the Groundhogg plugin to be installed and activated.', 'ws-form-groundhogg'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				$fields = array();
				$map = array();

				// Build source page
				$tracking_protocol = ((isset($_SERVER) && isset($_SERVER['HTTPS'])) ? $_SERVER['HTTPS'] : false) ? 'https://' : 'http://';
				$tracking_host = parent::get_submit_value($submit, 'tracking_host', '');
				$tracking_pathname = parent::get_submit_value($submit, 'tracking_pathname', '');
				$tracking_query_string = parent::get_submit_value($submit, 'tracking_query_string', '');
				$tracking_source_page = $tracking_host . $tracking_pathname . $tracking_query_string;
				if(empty($tracking_source_page)) { $tracking_source_page = false; } else { $tracking_source_page = $tracking_protocol . $tracking_source_page; }

				// Tracking mapping
				$tracking_mappings = array(

					'tracking_remote_ip' => 'ip_address',
					'tracking_utm_source' => 'utm_content',
					'tracking_utm_medium' => 'utm_medium',
					'tracking_utm_campaign' => 'utm_campaign',
					'tracking_utm_term' => 'utm_term',
					'tracking_utm_content' => 'utm_content',
					'tracking_ip_lookup_city' => 'city',
					'tracking_ip_lookup_region' => 'region',
					'tracking_ip_lookup_country' => 'country',
					'tracking_ip_lookup_time_zone' => 'time_zone',
					'tracking_source_page' => 'source_page'
				);

				// Tracking
				foreach($tracking_mappings as $submit_meta_key => $api_field) {

					// Check to see if submit value exists
					switch($submit_meta_key) {

						case 'tracking_source_page' :

							$submit_value = $tracking_source_page;
							break;

						default :

							$submit_value = parent::get_submit_value($submit, $submit_meta_key, false);
					}

					// If value not found, skip it
					if($submit_value === false) { continue; }

					// If it has already been mapped and is empty, skip it
					if(isset($fields[$api_field]) && !empty($fields[$api_field])) { continue; }

					// Add value to fields
					$fields[$api_field] = $submit_value;

					// Add field to map
					$map[$api_field] = $api_field;
				}

				// Owner
				if($this->owner_id > 0) {

					// Add value to fields
					$fields['owner_id'] = $this->owner_id;

					// Add field to map
					$map['owner_id'] = 'owner_id';
				}

				// Opt-in status
				if($this->opt_in_status > 0) {

					$fields['optin_status'] = $this->opt_in_status;
					$map['optin_status'] = 'optin_status';
				}

				// Field mapping
				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					$api_field = $field_map['action_' . $this->id . '_list_fields'];
					if($api_field === '') { continue; }

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($submit_value === false) { continue; }

					// Constant Contact cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					// Add value to fields
					$fields[$api_field] = $submit_value;

					// Add field to map
					$map[$api_field] = $api_field;
				}

				// Custom mapping
				foreach($this->custom_mapping as $custom_map) {

					$field_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$api_field = $custom_map['action_' . $this->id . '_list_fields'];
					if($api_field === '') { continue; }

					// Parse field value
					$field_value = WS_Form_Common::parse_variables_process($field_value, $form, $submit, 'text/plain');

					// Add value to fields
					$fields[$api_field] = $field_value;

					// Add field to map
					$map[$api_field] = $api_field;
				}

				// Notes
				foreach($this->notes as $note_index => $note) {

					$note = $note['action_' . $this->id . '_note'];

					if($note !== '') {

						// Get note ID
						$note_id = sprintf('note_%u', $note_index);

						// Add value to fields
						$fields[$note_id] = WS_Form_Common::parse_variables_process($note, $form, $submit, 'text/plain');

						// Add field to map
						$map[$note_id] = 'notes';
					}
				}

				// Tag mapping
				foreach($this->tag_mapping as $tag_index => $tag) {

					if($tag !== '') {

						// Get tag ID
						$tag_id = sprintf('tag_%u', $tag_index);

						// Add value to fields
						$fields[$tag_id] = WS_Form_Common::parse_variables_process($tag, $form, $submit, 'text/plain');

						// Add field to map
						$map[$tag_id] = 'tags';
					}
				}

				// Custom meta
				foreach($this->meta as $meta) {

					$meta_key = $meta['action_' . $this->id . '_meta_key'];
					if($meta_key === '') { continue; }

					$meta_value = $meta['action_' . $this->id . '_meta_value'];
					if($meta_value === '') { continue; }

					// Add value to fields
					$fields[$meta_key] = WS_Form_Common::parse_variables_process($meta_value, $form, $submit, 'text/plain');

					// Add field to map
					$map[$meta_key] = 'meta';
				}

				// Check email is mapped and valid
				if(!isset($fields['email'])) { return self::error(__('Email address not mapped', 'ws-form-sendinblue-v3')); }
				if(!filter_var($fields['email'], FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid email address', 'ws-form-sendinblue-v3')); }

				try {

					// Generate contact with map
					$contact = \Groundhogg\generate_contact_with_map($fields, $map);

					// After form submit handler
					\Groundhogg\after_form_submit_handler($contact);

				} catch (Exception $e) {

					return self::error($e->getMessage());
				}

				// Attachment mapping
				$files = array();
				foreach($this->attachment_mapping as $attachment_map) {

					$field_id = $attachment_map['ws_form_field'];

					// Get submit value
					$get_submit_value_repeatable_return = parent::get_submit_value_repeatable($submit, WS_FORM_FIELD_PREFIX . $field_id, array(), true);
					if(
						!is_array($get_submit_value_repeatable_return) ||
						!is_array($get_submit_value_repeatable_return['value']) ||
						!isset($get_submit_value_repeatable_return['value'][0])
					) { continue; }

					// Add each value
					foreach($get_submit_value_repeatable_return['value'] as $repeater_index => $meta_value) {

						$meta_value = $get_submit_value_repeatable_return['value'][$repeater_index];
						if(!is_array($meta_value)) { continue; }

						foreach($meta_value as $file) {

							// Check submit file data
							if(
								!isset($file['name']) ||
								!isset($file['hash']) ||
								!isset($file['type']) ||
								!isset($file['size']) ||
								!isset($file['path'])

							) { continue; }

							// Source file
							$source_filename = wp_upload_dir()['basedir'] . '/' . $file['path'];

							// Get uploads folder
							$contact->get_uploads_folder();

							// Check path exists
							$destination_path = $contact->upload_paths['path'];
							if(!file_exists($destination_path)) {

								add_filter('upload_dir', [$contact, 'map_upload']);
								wp_upload_dir();
								remove_filter('upload_dir', [$contact, 'map_upload']);
							}

							// Copy file to Groundhogg contact path
							$filename = wp_unique_filename($destination_path, $file['name']);
							$destination_filename = "$destination_path/$filename";
							copy($source_filename, $destination_filename);
						}
					}
				}

				// Success
				self::success(__('Successfully created contact', 'ws-form-groundhogg'));

				// Do action
				do_action('wsf_action_' . $this->id . '_post', $submit, $contact);

				return true;
			}

			// Get from API
			public function get($form = false, $current_user = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check user is logged in
				if(!$current_user) { return false; }

				// Get Groundhogg contact record
				$contact = \Groundhogg\get_contactdata();
				if($contact === false) { return false; }

				$fields_return = array();

				// Get Groundhogg fields
				$fields_groundhogg = self::get_groundhogg_fields(true);

				foreach($fields_groundhogg as $key => $value) {

					$meta_value = $contact->{$key};

					if(!$meta_value) {

						if(defined('GROUNDHOGG_BETTER_META_VERSION')) {

							$custom_field = \GroundhoggBetterMeta\Tab_Api\Fields::instance()->get($key);

							if($custom_field && isset($custom_field['meta'])) {

								$meta_value = $contact->{$custom_field['meta']};

							} else {

								$meta_value = false;
							}
						}
					}

					if($meta_value === false) { $meta_value = ''; }

					$fields_return[$key] = $meta_value;
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}


			// Get lists
			public function get_lists_private($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$lists = array(

					array(

						'id' => 			'contact', 
						'label' => 			__('Contacts', 'ws-form-groundhogg'), 
						'field_count' => 	false,
						'record_count' => 	false
					)
				);

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Load configuration
				self::load_config();

				// Build list
				$list = array(

					'label' => __('Contacts', 'ws-form-groundhogg')
				);

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false, $acf = true) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config();

				$list_fields = array();

				// Get Groundhogg mappable fields
				$fields_groundhogg = self::get_groundhogg_fields(true);

				// Build our fields
				$sort_index = 1;
				foreach($fields_groundhogg as $value => $text) {

					$list_fields[] = array(

						'id' => 			$value,
						'label' => 			$text, 
						'label_field' => 	$text, 
						'type' => 			'text',
						'action_type' =>	'text',
						'required' => 		false, 
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	$sort_index++,
						'section_index' =>	0,
						'visible' =>		true,
						'meta' => 			false,
						'no_map' =>			false
					);
				}

				return $list_fields;
			}

			// Get Groundhogg fields
			public function get_groundhogg_fields($populate = false) {

				// Get Groundhogg mappable fields
				$fields_groundhogg = \Groundhogg\get_mappable_fields();

				// Bypass these Groundhogg fields (We handle those automatically)
				$keys_groundhogg_bypass = array('copy_file', 'ip_address', 'source_page', 'file', 'meta', 'utm_campaign', 'utm_content', 'utm_medium', 'utm_term', 'utm_source');

				if($populate) {

					$keys_groundhogg_bypass[] = 'notes';
					$keys_groundhogg_bypass[] = 'tags';
				}

				foreach($keys_groundhogg_bypass as $key_groundhogg_bypass) {

					unset($fields_groundhogg[$key_groundhogg_bypass]);
				}

				// Convert custom fields to familiar format
				if(defined('GROUNDHOGG_BETTER_META_VERSION')) {

					foreach($fields_groundhogg as $field_groundhogg_key => $field_groundhogg) {

						if(is_array($field_groundhogg)) {

							unset($fields_groundhogg[$field_groundhogg_key]);

							foreach($field_groundhogg as $field_id => $field_label) {

								$field = \GroundhoggBetterMeta\Tab_Api\Fields::instance()->get($field_id);

								if(is_array($field) && isset($field['meta'])) {

									$fields_groundhogg[$field['id']] = $field['name'] . ' (' . $field['meta'] . ')';
								}
							}
						}
					}
				}

				asort($fields_groundhogg);

				return $fields_groundhogg;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_opt_in_status',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_owner_id',
						'action_' . $this->id . '_tag_mapping',
						'action_' . $this->id . '_notes',
						'action_' . $this->id . '_attachment_mapping',
						'action_' . $this->id . '_meta'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-groundhogg') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-groundhogg')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Get Groundhogg mappable fields
				$fields_groundhogg = self::get_groundhogg_fields();

				// Build our fields
				$fields = array();
				foreach($fields_groundhogg as $value => $text) {

					$fields[] = array('value' => $value, 'text' => $text);
				}

				// Sort fields alphabetically
				usort($fields, function ($field_1, $field_2) {

					if ($field_1['text'] == $field_2['text']) return 0;
					return $field_1['text'] < $field_2['text'] ? -1 : 1;
				});

				// Build config_meta_keys
				$config_meta_keys = array(

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field (Optional)', 'ws-form-groundhogg'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-groundhogg'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-groundhogg')
					),

					// Opt-In status
					'action_' . $this->id . '_opt_in_status'	=> array(

						'label'							=>	__('Opt-In Status', 'ws-form-groundhogg'),
						'type'							=>	'select',
						'options'						=>	array(),
						'default'						=>	'1'
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-groundhogg'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Groundhogg fields.', 'ws-form-groundhogg'),
						'meta_keys'					=>	array(

							'ws_form_field_no_file',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						)
					),

					// Fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Groundhogg Field', 'ws-form-groundhogg'),
						'type'							=>	'select',
						'options'						=>	$fields,
						'options_blank'					=>	__('Select...', 'ws-form-groundhogg')
					),

					// Custom mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-groundhogg'),
						'type'						=>	'repeater',
						'help'						=>	__('Map values or variables to Groundhogg fields.', 'ws-form-groundhogg'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						)
					),

					// Custom mapping value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'							=>	__('Value', 'ws-form-groundhogg'),
						'type'							=>	'text'
					),

					// Notes
					'action_' . $this->id . '_notes'	=> array(

						'label'						=>	__('Notes', 'ws-form-groundhogg'),
						'type'						=>	'repeater',
						'help'						=>	__('Add notes to the Groundhogg contact.', 'ws-form-groundhogg'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_note'
						)
					),

					// Note
					'action_' . $this->id . '_note'	=> array(

						'label'							=>	__('Note', 'ws-form-groundhogg'),
						'type'							=>	'textarea'
					),

					// Attachment mapping
					'action_' . $this->id . '_attachment_mapping'	=> array(

						'label'						=>	__('Attachment Mapping', 'ws-form-post'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form file and signature fields to the Groundhogg contact.', 'ws-form-post'),
						'meta_keys'					=>	array(

							'ws_form_field_file'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field_file'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Owner
					'action_' . $this->id . '_owner_id'	=> array(

						'label'							=>	__('Owner', 'ws-form-groundhogg'),
						'type'							=>	'select',
						'options'						=>	array(),
						'options_blank'					=>	__('Select...', 'ws-form-groundhogg')
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'							=> 	__('Tag(s)', 'ws-form-groundhogg'),
						'type'							=>	'select',
						'select2'						=>	true,
						'select2_tags'					=>	true,
						'select_ajax_method_search'		=>	'action_' . $this->id . '_tag_search',
						'select_ajax_method_cache'		=>	'action_' . $this->id . '_tag_cache',
						'select_ajax_placeholder'		=>	__('Search or add new tags...', 'ws-form'),
						'multiple'						=>	true,
						'help'						=>	__('Choose which tags to assign to this contact.', 'ws-form-groundhogg')
					),

					// Custom meta mapping
					'action_' . $this->id . '_meta'	=> array(

						'label'						=>	__('Custom Meta', 'ws-form-post'),
						'type'						=>	'repeater',
						'help'						=>	__('Add custom meta to this contact.', 'ws-form-post'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_meta_key',
							'action_' . $this->id . '_meta_value'
						)
					),

					// Meta key
					'action_' . $this->id . '_meta_key'	=> array(

						'label'						=>	__('Meta Key', 'ws-form-post'),
						'type'						=>	'text'
					),

					// Meta value
					'action_' . $this->id . '_meta_value'	=> array(

						'label'						=>	__('Meta Value', 'ws-form-post'),
						'type'						=>	'text'
					)
				);

				// Add opt-in status
				$opt_in_statuses = \Groundhogg\Preferences::get_preference_names();
				$config_meta_keys['action_' . $this->id . '_opt_in_status']['options'][] = array('value' => '0', 'text' => 'Not Set');
				foreach($opt_in_statuses as $value => $text) {

					$config_meta_keys['action_' . $this->id . '_opt_in_status']['options'][] = array('value' => $value, 'text' => $text);
				}

				// Add users
				$args = array(

					'role__in' => \Groundhogg\Main_Roles::get_owner_roles(),
					'fields' => array('ID', 'display_name', 'user_email')
				);
				$users = get_users($args);
				foreach($users as $user) {

					$config_meta_keys['action_' . $this->id . '_owner_id']['options'][] = array('value' => $user->ID, 'text' => sprintf('%s (%s)', $user->display_name, $user->user_email));
				}

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-groundhogg'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-groundhogg'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Groundhogg add-on for WS Form PRO license key here.', 'ws-form-groundhogg'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-groundhogg'),
							'type'		=>	'static'
						)
					)
				);

				return $options;
			}

			// Tag search
			public function tag_search($parameters) {

				global $wpdb;

				$term = WS_Form_Common::get_query_var_nonce('term', '', $parameters);
				$term = sanitize_text_field( stripslashes( $term ) );

				// Get tags
				$tags = \Groundhogg\Plugin::$instance->dbs->get_db('tags')->search($term);

				// Build results
				$results = array();
				foreach($tags as $tag) {

					$results[] = array(
						'id'   => $tag->tag_id,
						'text' => $tag->tag_name
					);
				}

				// Sort results
				usort($results, function($tag_1, $tag_2) {

					if ($tag1['text'] == $tag2['text']) return 0;
					return $tag1['text'] < $tag2['text'] ? -1 : 1;
				});

				return array('results' => $results);
			}

			// Tag cache
			public function tag_cache($parameters) {

				$return_array = array();

				$term_ids = WS_Form_Common::get_query_var_nonce('ids', '', $parameters);

				// Get tags
				$tags = \Groundhogg\Plugin::$instance->dbs->get_db('tags')->query();

				// Build tag array
				$tag_array = array();
				foreach($tags as $tag) {

					$tag_array[$tag->tag_id] = $tag->tag_name;
				}

				// Build results
				$results = array();
				foreach($term_ids as $term_id) {

					if(isset($tag_array[$term_id])) {

						// Tag found by ID
						$results[$term_id] = $tag_array[$term_id];

					} else {

						// Tag not found so return tag as is
						$results[$term_id] = $term_id;
					}
				}

				return $results;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9_-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// Tag search
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/select2/action_' . $this->id . '_tag_search/', array('methods' => 'GET', 'callback' => array($this, 'tag_search'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/select2/action_' . $this->id . '_tag_cache/', array('methods' => 'POST', 'callback' => array($this, 'tag_cache'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists_private();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(false, false);

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists_private(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true, false);

				// Process response
				self::api_response($list_fields);
			}

			// Load config for this action
			public function load_config($config = array()) {

				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->opt_in_status = absint(parent::get_config($config, 'action_' . $this->id . '_opt_in_status', '1'));
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->notes = parent::get_config($config, 'action_' . $this->id . '_notes', array());
				if(!is_array($this->notes)) { $this->notes = array(); }
				$this->attachment_mapping = parent::get_config($config, 'action_' . $this->id . '_attachment_mapping', array());
				if(!is_array($this->attachment_mapping)) { $this->attachment_mapping = array(); }
				$this->meta = parent::get_config($config, 'action_' . $this->id . '_meta', array());
				if(!is_array($this->meta)) { $this->meta = array(); }
				$this->owner_id = absint(parent::get_config($config, 'action_' . $this->id . '_owner_id'));
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = defined('GROUNDHOGG_VERSION');

				return $this->configured;
			}
		}

		if(class_exists('Groundhogg\Steps\Benchmarks\Benchmark')) {

			class WS_Form_Groundhogg_Benchmark extends Groundhogg\Steps\Benchmarks\Benchmark {

				public function get_name() {

					return WS_FORM_NAME_PRESENTABLE;
				}

				public function get_type() {

					return WS_FORM_IDENTIFIER . '_submit';
				}

				public function get_description() {

					return __( 'Runs whenever WS Form creates or updates a contact in Groundhogg.', 'ws-form-groundhogg' );
				}

				public function get_icon() {

					return plugin_dir_url(__FILE__) . 'images/wsform.png';
				}

				protected function get_complete_hooks() {

					return [

						'wsf_action_groundhogg_post' => 2
					];
				}

				public function admin_scripts() {
				}

				public function setup( $submit, $contact ) {

					$this->add_data( 'form_id', $submit->form_id );
					$this->add_data( 'post_mode', $submit->post_mode );
					$this->add_data( 'email', $contact->email );
				}

				protected function get_the_contact() {

					return Groundhogg\get_contactdata( $this->get_data( 'email' ) );
				}

				protected function can_complete_step() {

					// Get data
					$form_id = $this->get_data( 'form_id' );
					$post_mode = $this->get_data( 'post_mode' );

					// Get settings
					$setting_form_id  = $this->get_setting( 'form_id' );
					$setting_post_mode  = $this->get_setting( 'post_mode' );

					// Check form ID
					if(is_array($setting_form_id) && (count($setting_form_id) > 0) && !in_array($form_id, $setting_form_id)) { return false; }

					// Check post mode
					if(is_array($setting_post_mode) && (count($setting_post_mode) > 0) && !in_array($post_mode, $setting_post_mode)) { return false; }

					return true;
				}

				public function settings( $step ) {

					// Get all forms
					$forms = wsf_form_get_all($published = false, $order_by = 'label');

					// Build form options
					$options_forms = array();

					foreach($forms as $form) {

						$options_forms[$form['id']] = sprintf(__('%s (ID: %u)', 'ws-form-groundhogg'), $form['label'], $form['id']);
					}

					Groundhogg\html()->start_form_table();

					Groundhogg\html()->start_row();

					Groundhogg\html()->th( __( 'Form(s):', 'ws-form-groundhogg' ) );
					Groundhogg\html()->td( [
						Groundhogg\html()->select2(
							[
								'id'       => $this->setting_id_prefix( 'form_id' ),
								'name'     => $this->setting_name_prefix( 'form_id' ) . '[]',
								'options'  => $options_forms,
								'multiple' => true,
								'selected' => $this->get_setting( 'form_id' )
							]
						),
						Groundhogg\html()->description( __( 'Please select one or more forms. Leave blank for any form.', 'ws-form-groundhogg' ) )
					] );

					Groundhogg\html()->end_row();

					Groundhogg\html()->start_row( [
						'class' => 'condition'
					] );

					Groundhogg\html()->th( sprintf( __( 'Mode(s):', 'ws-form-groundhogg' ) ) );
					Groundhogg\html()->td( [
						Groundhogg\html()->select2(
							[
								'id'       => $this->setting_id_prefix( 'post_mode' ),
								'name'     => $this->setting_name_prefix( 'post_mode' ) . '[]',
								'options'  => array(

									'submit'	=>	__('Submitted', 'ws-form-groundhogg'),
									'save'		=>	__('Saved', 'ws-form-groundhogg'),
									'action'	=>	__('Action', 'ws-form-groundhogg')
								),
								'multiple' => true,
								'selected' => $this->get_setting( 'post_mode' )
							]
						),
						Groundhogg\html()->description( __( 'Please select one or more modes. Leave blank or any mode.', 'ws-form-groundhogg' ) )
					] );

					Groundhogg\html()->end_row();

					Groundhogg\html()->end_form_table();
				}

				public function save( $step ) {

					$this->save_setting( 'form_id', wp_parse_id_list( $this->get_posted_data( 'form_id' ) ) );
					$this->save_setting( 'post_mode', wp_parse_list( $this->get_posted_data( 'post_mode' ) ) );
				}
			}
		}
	});