<?php

	/**
	 * @link              https://wsform.com/knowledgebase/hubspot/
	 * @since             1.0.0
	 * @package           WS_Form_HubSpot
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - HubSpot
	 * Plugin URI:        https://wsform.com/knowledgebase/hubspot/
	 * Description:       HubSpot add-on for WS Form PRO
	 * Version:           1.1.2
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-hubspot
	 */

	Class WS_Form_Add_On_Hubspot {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.72';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Hubspot();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-hubspot'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_hubspot = new WS_Form_Add_On_Hubspot();	

	register_activation_hook(__FILE__, array($wsf_add_on_hubspot, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Hubspot extends WS_Form_Action {

			public $id = 'hubspot';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 25;
			public $can_repost = true;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_refresh_token;
			public $portal_id;
			public $inline_message = false;
			public $list_id = false;
			public $opt_in_field;
			public $field_mapping;
			public $custom_mapping;
			public $deal_hubspot_owner_id;

			// Constants
			const LISTS_FETCH_COUNT = 100;
			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const WS_FORM_LICENSE_ITEM_ID = 2287;
			const WS_FORM_LICENSE_NAME = 'HubSpot add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.1.2';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('HubSpot', 'ws-form-hubspot');

				// Set label for actions pull down
				$this->label_action = __('Add to HubSpot', 'ws-form-hubspot');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save access token expiry
				$this->api_access_token_expiry = time() + intval(WS_Form_Common::get_query_var($this->id . '_expires_in', false)) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Get portal ID
				self::account_info();

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = $this->id . '/?refresh_token=' . rawurlencode($this->api_refresh_token);
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); }

					// Save access token
					$this->api_access_token = parent::get_object_value($api_response_decoded, $this->id . '_access_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save refresh token
					$this->api_refresh_token = parent::get_object_value($api_response_decoded, $this->id . '_refresh_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

					// Save access token expiry
					$this->api_access_token_expiry = time() + intval(parent::get_object_value($api_response_decoded, $this->id . '_expires_in')) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-hubspot')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_portal_id' :

						$value = $this->portal_id;
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-hubspot') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-hubspot') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-hubspot') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/hubspot';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-hubspot'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_portal_id', '');
			}

			// Get account info
			public function account_info() {

				self::load_config_plugin();

				// Get account info
				$api_response = parent::api_call($this->api_endpoint, 'integrations/v1/me', 'GET', null, self::get_api_headers());

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); return false; }

				// Check status code
				if($api_response['http_code'] != 200) { 

					$message = sprintf(__('%s API Error: %s', 'ws-form-hubspot'), $this->label, $api_response_decoded->message);

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Check that we got properties
				if(!isset($api_response_decoded->portalId)) { return false; }

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_portal_id', $api_response_decoded->portalId);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your HubSpot account <a href="%s">here</a>.', 'ws-form-hubspot'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-hubspot'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Defaults
				$deal_id = false;
				$organization_id = false;
				$person_id = false;

				// Process field mapping
				$data_deals = array();
				$data_deals['properties'] = array();
				$data_companies = array();
				$data_companies['properties'] = array();
				$data_contacts = array();
				$data_contacts['properties'] = array();
				$email = false;

				// Process field mapping
				foreach($this->field_mapping as $field_map) {

					// Get API field
					$api_field = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$field_id = $field_map['ws_form_field'];
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false, true);
					if($submit_value === false) { continue; }
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					// Get API key
					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];
					if(strpos($api_field_string, ':') === false) {

						$api_field_array = array('contacts', $api_field_string);

					} else {

						$api_field_array = explode(':', $api_field_string);
					}
					if(count($api_field_array) != 2) { continue; }

					// Get API field type
					$api_field_type = $api_field_array[0];
					if(!in_array($api_field_type, array('contacts', 'companies', 'deals'))) { continue; }

					// Get API field
					$api_field = $api_field_array[1];
					if($api_field == '') { continue; }

					// Store email
					if($api_field == 'email') { $email = $submit_value; }

					// Assign to array
					$api_field_type = 'data_' . $api_field_type;
					$$api_field_type['properties'][] = array((($api_field_type == 'data_contacts') ? 'property' : 'name') => $api_field, 'value' => $submit_value);
				}

				// Process custom mapping
				foreach($this->custom_mapping as $custom_map) {

					// Get field value
					$field_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$field_value = WS_Form_Common::parse_variables_process($field_value, $form, $submit, 'text/plain');

					// Get API key
					$api_field_string = $custom_map['action_' . $this->id . '_list_fields'];
					if(strpos($api_field_string, ':') === false) { continue; }
					$api_field_array = explode(':', $api_field_string);
					if(count($api_field_array) != 2) { continue; }

					// Get API field type
					$api_field_type = $api_field_array[0];
					if(!in_array($api_field_type, array('contacts', 'companies', 'deals'))) { continue; }

					// Get API field
					$api_field = $api_field_array[1];

					// Store email
					if($api_field == 'email') { $email = $submit_value; }

					// Assign to array
					$api_field_type = 'data_' . $api_field_type;
					$$api_field_type['properties'][] = array((($api_field_type == 'data_contacts') ? 'property' : 'name') => $api_field, 'value' => $field_value);
				}

				// Build contacts form endpoint data
				if(false && count($data_contacts['fields']) > 0) {

					// HS Context
					$context = array();

					// HS Context - HubSpot Tracking Cookie
					if(isset($_COOKIE) && isset($_COOKIE['hubspotutk'])) {

						$context['hutk'] = $_COOKIE['hubspotutk'];
					}

					// HS Context - IP Address
					$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
					if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

						$context['ipAddress'] = $tracking_remote_ip;
					}

					// Get referring page
					$referrer = WS_Form_Common::get_http_env(array('HTTP_REFERER'));

					if($referrer != '') {

						$context['pageUri'] = $referrer;

						// Get post ID of referrer
						$post_id = url_to_postid($referrer);

						if($post_id > 0) {

							if(!empty($title = get_the_title($post_id))) {

								$context['pageName'] = $title;
							}						
						}
					}

					if(count($context) > 0) {

						$data_contacts['context'] = $context;
					}

					// Submitted at
					$data_contacts['submittedAt'] = (time() * 1000);

					// Skip validation
					$data_contacts['skipValidation'] = true;

					// Legal consent options
					if($opted_in) {

						$legalConsentOptions = array(

							'consent' => array(

								'consentToProcess' 	=> true,
								'text' 				=> $opt_in_field_value[0]
							)
						);

						$data_contacts['legalConsentOptions'] = $legalConsentOptions;
					}
				}

				// Add objects
				$contact_id = self::add_object($data_contacts, 'contacts', $email);
				$company_id = self::add_object($data_companies, 'companies');

				if($contact_id !== false) {

					if(!isset($data_deals['associations'])) { $data_deals['associations'] = array(); }
					$data_deals['associations']['associatedVids'] = array($contact_id);
				}
				if($company_id !== false) {

					if(!isset($data_deals['associations'])) { $data_deals['associations'] = array(); }
					$data_deals['associations']['associatedCompanyIds'] = array($company_id);
				}
				if(count($data_deals['properties']) > 0) {

					if(!empty($this->deal_hubspot_owner_id)) {

						$data_deals['properties'][] = array(

							'name' => 'hubspot_owner_id',
							'value' => $this->deal_hubspot_owner_id
						);
					}
				}
				$deal_id = self::add_object($data_deals, 'deals');

				// Add associations
				if(($contact_id !== false) && ($company_id !== false)) { self::add_association($contact_id, $company_id, 1); }
			}

			// Add association
			public function add_association($from_object_id, $to_object_id, $definition_id) {

				// Check access token
				self::access_token_check();

				$data = array(

					'fromObjectId' => $from_object_id,
					'toObjectId' => $to_object_id,
					'category' => 'HUBSPOT_DEFINED',
					'definitionId' => $definition_id
				);

				$data_json = json_encode($data);

				// Add organization
				$api_response = parent::api_call($this->api_endpoint, 'crm-associations/v1/associations', 'PUT', $data_json, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/json');

				// Process response
				switch($api_response['http_code']) {

					case 204 :

						return true;

					default :

						return false;
				}
			}

			// Add object
			public function add_object($data, $object, $email = false) {

				if(
					(count($data['properties']) == 0)
				) {

					return false;
				}

				// Check access token
				self::access_token_check();

				// Work out endpoint and path
				switch($object) {

					case 'contacts' :

						if(empty($email)) { return false; }
						$path = sprintf('contacts/v1/contact/createOrUpdate/email/%s/', $email);
						$log_description = __('Successfully added contact', 'ws-form-hubspot');
						$id_field = 'vid';
						break;

					case 'companies' :

						$path = 'companies/v2/companies';
						$log_description = __('Successfully added company', 'ws-form-hubspot');
						$id_field = 'companyId';
						break;

					case 'deals' :

						$path = 'deals/v1/deal';
						$log_description = __('Successfully added deal', 'ws-form-hubspot');
						$id_field = 'dealId';
						break;
				}

				$data_json = json_encode($data);

				// Add organization
				$api_response = parent::api_call($this->api_endpoint, $path, 'POST', $data_json, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/json');

				// Process response
				switch($api_response['http_code']) {

					case 200 :
					case 409 :

						self::success($log_description);					
						return self::get_api_response_id($api_response, $id_field);

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));

					case 500 :

						// Server error
						return self::error('500 Server Error');
				}
			}

			// Get API response ID
			public function get_api_response_id($api_response, $id_field) {

				// Get organization ID
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->{$id_field})) { return false; }

				return $api_response_decoded->{$id_field};
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$params = array(

							'Limit'		=> self::LISTS_FETCH_COUNT,
							'Offset'	=> $total_lists_retrieved
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'forms/v2/forms', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); return false; }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-dotmailer')); return false; }

						// Get total lists
						$total_lists = count($api_response_decoded);

						// Increment read counter
						$total_lists_retrieved += $total_lists;

						foreach($api_response_decoded as $list) {

							// Get field count
							$field_count = 0;
							$formFieldGroups = parent::get_object_value($list, 'formFieldGroups');
							foreach($formFieldGroups as $formFieldGroup) {

								$field_count += count(parent::get_object_value($formFieldGroup, 'fields'));
							}

							// Build list
							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'guid'),
								'label' => 			parent::get_object_value($list, 'name'),
								'field_count' => 	$field_count,
								'record_count' => 	false
							);
						}

					} while ($total_lists == self::LISTS_FETCH_COUNT);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'forms/v2/forms/' . rawurlencode($this->list_id), 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Save inlineMessage
					$this->inline_message = parent::get_object_value($api_response_list, 'inlineMessage');

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Check access token
					self::access_token_check();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'forms/v2/forms/' . rawurlencode($this->list_id), 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_list_fields = json_decode($api_response['response']);

					// Checks
					if(is_null($api_list_fields)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); }

					// Process fields
					$sort_index_offset = 0;

					$formFieldGroups = parent::get_object_value($api_list_fields, 'formFieldGroups');

					$sort_index = 0;
					$rich_text_index = 0;

					foreach($formFieldGroups as $formFieldGroup) {

						$fields = parent::get_object_value($formFieldGroup, 'fields');

						// Check for rich text
						$rich_text = parent::get_object_value($formFieldGroup, 'richText');
						if(is_object($rich_text)) {

							if(!empty($rich_text->content)) {

								$list_fields[] = array(

									'id' => 			'contacts:rich_text_' . $rich_text_index,
									'label' => 			__('Rich Text', 'ws-form-hubspot'),
									'label_field' => 	__('Rich Text', 'ws-form-hubspot'), 
									'type' => 			'texteditor',
									'action_type' =>	'richText',
									'required' => 		false,
									'default_value' => 	'',
									'pattern' => 		'',
									'placeholder' => 	'',
									'help' => 			'',  
									'sort_index' => 	$sort_index,
									'visible' =>		true,  
									'meta' => 			array('text_editor' => $rich_text->content)
								);

								$sort_index++;
								$rich_text_index++;
							}
						}

						foreach($fields as $field) {

							$enabled = parent::get_object_value($field, 'enabled');
							if(!$enabled) { continue; }

							$type = self::action_field_type_to_ws_form_field_type($field);

							$list_fields[] = array(

								'id' => 			'contacts:' . parent::get_object_value($field, 'name'),
								'label' => 			'Contacts:' . parent::get_object_value($field, 'label'),
								'label_field' => 	parent::get_object_value($field, 'label'), 
								'type' => 			$type,
								'action_type' =>	parent::get_object_value($field, 'fieldType'),
								'required' => 		parent::get_object_value($field, 'required'),
								'default_value' => 	self::action_to_ws_form_default_value(parent::get_object_value($field, 'defaultValue'), $type),
								'pattern' => 		'',
								'placeholder' => 	parent::get_object_value($field, 'placeholder'),
								'help' => 			parent::get_object_value($field, 'description'),  
								'sort_index' => 	$sort_index,
								'visible' =>		!parent::get_object_value($field, 'hidden'),  
								'meta' => 			self::action_field_to_ws_form_meta_keys($field)
							);

							$sort_index++;
						}
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get contact properties
			public function get_object_properties($endpoint, $prefix, $label, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$object_properties = WS_Form_Common::option_get('action_' . $this->id . '_' . $endpoint . '_properties');

				if($fetch || ($object_properties === false)) {

					$object_properties = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_object_properties = array();

					// Check access token
					self::access_token_check();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'properties/v1/' . $endpoint . '/properties', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_object_properties = json_decode($api_response['response']);

					// Checks
					if(is_null($api_object_properties)) { return self::error(__('Invalid API response data', 'ws-form-hubspot')); }

					// Process fields
					$sort_index = 0;

					foreach($api_object_properties as $object_property) {

						$readOnlyValue = parent::get_object_value($object_property, 'readOnlyValue');
						if($readOnlyValue) { continue; }

						$type = self::action_field_type_to_ws_form_field_type($object_property);

						$object_properties[] = array(

							'id' => 			sprintf('%s:%s', $prefix, parent::get_object_value($object_property, 'name')),
							'label' => 			sprintf('%s: %s', $label, parent::get_object_value($object_property, 'label')),
							'sort_index' => 	$sort_index++
						);
					}

					// Sort by label
					usort($object_properties, function ($a, $b) { return $a['label'] > $b['label']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_' . $endpoint . '_properties', $object_properties);
				}

				return $object_properties;
			}

			// Get form fields
			public function get_fields() {

				switch($this->list_id) {

					default :

						$form_fields = array(

							'opt_in_field' => array(

								'type'	=>	'checkbox',
								'label'	=>	__('Opt-In', 'ws-form'),
								'meta'	=>	array(

									'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

										array(

											'id'		=> 1,
											'default'	=> '',
											'required'	=> '',
											'disabled'	=> '',
											'hidden'	=> '',
											'data'		=> array(__('I agree to allow #blog_name to store and process my personal data.', 'ws-form'))
										)
									))
								)
							),

							'submit' => array(

								'type'			=>	'submit',
								'label'			=>	__('Submit', 'ws-form-hubspot')
							)
						);

						break;
				}

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config();

				// Process inline message
				if(empty($this->inline_message)) { $this->inline_message = __('Thanks for submitting the form.', 'ws-form-hubspot'); }

				switch($this->list_id) {

					default :

						$form_actions = array(

							$this->id => array(

								'meta'	=> array(

									'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
									'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
								)
							),

							'message' => array(

								'meta'	=> array(

									'action_message_message' => $this->inline_message
								)
							),

							'database'
						);

						break;
				}

				return $form_actions;
			}

			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-hubspot'); }

				$error_detail_array = array();

				// Get error detail from JSON response
				if(isset($api_response_decoded->message)) { $error_detail_array[] = $api_response_decoded->message; }

				if(isset($api_response_decoded->validationResults) && is_array($api_response_decoded->validationResults)) {

					foreach($api_response_decoded->validationResults as $validation_result) {

						$message_array = array();

						if(isset($validation_result->error)) { $message_array[] = $validation_result->error; }
						if(isset($validation_result->name)) { $message_array[] = $validation_result->name; }
						if(isset($validation_result->message)) { $message_array[] = $validation_result->message; }

						if(count($message_array) > 0) { $error_detail_array[] = implode(' - ', $message_array); }
					}
				}

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_deal_hubspot_owner_id'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Get owners
			public function get_owners($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$owners = WS_Form_Common::option_get('action_' . $this->id . '_owners');

				// Retried if fetch is requested or owners are not cached
				if($fetch || ($owners === false)) {

					$owners = array();

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get owners
					$api_response = parent::api_call($this->api_endpoint, 'owners/v2/owners', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); return false; }

					// Process owner data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-dotmailer')); return false; }

					foreach($api_response_decoded as $owner) {

						// Build owner
						$owners[] = array(

							'id' => 			parent::get_object_value($owner, 'ownerId'),
							'label' => 			parent::get_object_value($owner, 'firstName') . ' ' . parent::get_object_value($owner, 'lastName')
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_owners', $owners);
				}

				return $owners;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-hubspot'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-hubspot')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-hubspot'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-hubspot'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-hubspot')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-hubspot'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to HubSpot fields.', 'ws-form-hubspot'),
						'meta_keys'					=>	array(

							'ws_form_field_save',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'properties_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true
					),

					// Custom mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-hubspot'),
						'type'						=>	'repeater',
						'help'						=>	__('Map values or variables to HubSpot fields.', 'ws-form-hubspot'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'properties_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						)
					),

					// Deal owner ID
					'action_' . $this->id . '_deal_hubspot_owner_id'	=> array(

						'label'							=>	__('Deal Owner ID', 'ws-form-hubspot'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-hubspot'),
						'options_action_id'				=>	$this->id,
						'options_action_api_populate'	=>	'owners',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'owners_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'help'							=>	__('Owner new deals are assigned to.', 'ws-form-hubspot')
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Hubspot Field', 'ws-form-hubspot'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-hubspot'),
						'options_action_id'				=>	$this->id,
						'options_action_api_populate'	=>	'properties'
					),

					// Custom mapping value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'							=>	__('Value', 'ws-form-hubspot'),
						'type'							=>	'text'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-hubspot'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-hubspot'),
							'type'		=>	'text',
							'help'		=>	__('Enter your HubSpot add-on for WS Form PRO license key here.', 'ws-form-hubspot'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-hubspot'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('HubSpot Status', 'ws-form-hubspot'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your HubSpot account, or <a href=\"http://www.hubspot.com/\" target=\"blank\">create a new account</a>.", 'ws-form-hubspot'),
							'button'		=>	'oauth_action_' . $this->id,
						),

						'action_' . $this->id . '_portal_id'	=>	array(

							'label'			=>	__('HubSpot Portal ID', 'ws-form-hubspot'),
							'type'			=>	'static'
						)
					)
				);

				return $options;
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$type = parent::get_object_value($field, 'fieldType');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids for radio and select
					case 'select' :
					case 'radio' :

						switch($type) {

							case 'radio' : $meta_key = 'data_grid_radio'; break;
							case 'select' : $meta_key = 'data_grid_select'; break;
						}

						// Get options
						$options = parent::get_object_value($field, 'options');

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys[$meta_key])) { return false; }
						if(!isset($meta_keys[$meta_key]['default'])) { return false; }

						$meta = $meta_keys[$meta_key]['default'];

						// Reconfigure columns
						$meta['columns'] = array(

							array('id' => 0, 'label' => 'Label'),
							array('id' => 1, 'label' => 'Value')
						);

						// Build new rows
						$rows = array();
						$id = 1;
						$default_value = parent::get_object_value($field, 'default_value');
						if($default_value === false) { $default_value = ''; }
						foreach($options as $option) {

							$rows[] = array(

								'id'		=> $id,
								'default'	=> (($default_value == $option) ? 'on' : ''),
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> isset($option->hidden) ? $option->hidden : false,
								'data'		=> array($option->label, $option->value)
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						// Build return array
						switch($type) {

							case 'radio' : 

								return(array($meta_key => $meta, 'radio_field_value' => 1, 'radio_field_text' => 0));

							case 'select' :

								return(array($meta_key => $meta, 'select_field_value' => 1, 'select_field_text' => 0));
						}

					default :

						return false;
				}
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($field) {

				// Set type by name
				$name = parent::get_object_value($field, 'name');
				switch($name) {

					case 'email' : 			return 'email';
					case 'phone' : 			return 'tel';
					case 'mobilephone' : 	return 'tel';
					case 'fax' : 			return 'tel';
					case 'website' :		return 'url';
				}

				// Set type by type
				$type = parent::get_object_value($field, 'fieldType');
				switch($type) {

					case 'select' :			return 'select';
					case 'radio' :			return 'radio';

					default :

						return 'text';
				}
			}

			// Convert action default value to WS Form default value
			public function action_to_ws_form_default_value($default_value, $type = 'text') {

				switch($type) {

					default :

						return $default_value;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->deal_hubspot_owner_id = parent::get_config($config, 'action_' . $this->id . '_deal_hubspot_owner_id');
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry  = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');
				$this->portal_id 				= WS_Form_Common::option_get('action_' . $this->id . '_portal_id', '');

				// Build endpoint using API key
				if(!empty($this->api_access_token )) {

					// Set endpoint URL
					$this->api_endpoint = 'https://api.hubapi.com/';

				} else {

					$this->api_endpoint = false;
					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/properties/', array('methods' => 'GET', 'callback' => array($this, 'api_get_properties'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/owners/', array('methods' => 'GET', 'callback' => array($this, 'api_get_owners'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/properties_fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_properties'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/owners_fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_owners'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - Properties
			public function api_get_properties($parameters) {

				// Get lists
				$list_fields = self::get_object_properties('contacts', 'contacts', 'Contacts');
				$list_fields = array_merge($list_fields, self::get_object_properties('companies', 'companies', 'Companies'));
				$list_fields = array_merge($list_fields, self::get_object_properties('deals', 'deals', 'Deals'));

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Owners
			public function api_get_owners($parameters) {

				// Get owners
				$owners = self::get_owners();

				// Process response
				self::api_response($owners);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - Properties with fetch
			public function api_fetch_properties($parameters) {

				// Get lists
				$list_fields = self::get_object_properties('contacts', 'contacts', 'Contacts', true);
				$list_fields = array_merge($list_fields, self::get_object_properties('companies', 'companies', 'Companies', true));
				$list_fields = array_merge($list_fields, self::get_object_properties('deals', 'deals', 'Deals', true));

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Owners with fetch
			public function api_fetch_owners($parameters) {

				// Get owners
				$owners = self::get_owners(true);

				// Process response
				self::api_response($owners);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(40.000000, 54.000000)"><path d="M45.22 20.36v-7.306a5.633 5.633 0 0 0 3.25-5.079v-.17a5.644 5.644 0 0 0-5.629-5.63h-.17a5.644 5.644 0 0 0-5.63 5.63v.17a5.633 5.633 0 0 0 3.25 5.079v7.321a15.964 15.964 0 0 0-7.592 3.34L12.606 8.066a6.41 6.41 0 1 0-3.003 3.906l19.756 15.38a16.01 16.01 0 0 0 .247 18.051l-6.012 6.012a5.164 5.164 0 0 0-1.492-.243 5.219 5.219 0 1 0 5.22 5.22 5.147 5.147 0 0 0-.244-1.492l5.948-5.949a16.036 16.036 0 1 0 12.183-28.59m-2.463 24.07a7 7 0 1 1 .02.004" fill="#ff7a59"/></g>';

				return $svg_logo;
			}
		}
	});