<?php

	/**
	 * @link              https://wsform.com/knowledgebase/icontact/
	 * @since             1.0.0
	 * @package           WS_Form_IContact
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - iContact
	 * Plugin URI:        https://wsform.com/knowledgebase/icontact/
	 * Description:       iContact add-on for WS Form PRO
	 * Version:           1.0.1
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-icontact-v2
	 */

	Class WS_Form_Add_On_IContact_V2 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.0';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_IContact_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-icontact-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_icontact = new WS_Form_Add_On_IContact_V2();	

	register_activation_hook(__FILE__, array($wsf_add_on_icontact, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_IContact_V2 extends WS_Form_Action {

			public $id = 'icontactv2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_account_id = false;
			private $api_client_folder_id = false;
			private $api_app_id;
			private $api_username;
			private $api_password;
			public $list_id = false;
			public $status;
			public $return_url;
			public $email_type;
			public $opt_in_field;
			public $field_mapping;

			// Constants
			const API_PATH = '/v2/';
			const DEFAULT_STATUS = 'Subscribed';
			const DEFAULT_EMAIL_TYPE = 'Html';
			const EMAIL_MERGE_FIELD = 'email';
			const LISTS_FETCH_COUNT = 1000;
			const OPT_IN_TYPE = 'Single';
			const WS_FORM_LICENSE_ITEM_ID = 4671;
			const WS_FORM_LICENSE_NAME = 'iContact add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.1';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('iContact', 'ws-form-icontact-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to iContact', 'ws-form-icontact-v2');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('rest_api_init', array($this, 'rest_api_init'), 10, 2);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-icontact-v2')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-icontact-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-icontact-v2') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							default :

								self::account_info();
						}

					break;
				}
			}

			// Get account info
			public function account_info() {

				$this->api_app_id = WS_Form_Common::option_get('action_' . $this->id . '_api_app_id', '');
				$this->api_username = WS_Form_Common::option_get('action_' . $this->id . '_api_username', '');
				$this->api_password = WS_Form_Common::option_get('action_' . $this->id . '_api_password', '');

				// Check API app ID
				if(
					($this->api_app_id == '') ||
					($this->api_username == '') ||
					($this->api_password == '')
				) {

					 self::reset_api_endpoint();
					 return false;
				}

				// Get account info
				$api_response = parent::api_call('https://app.icontact.com/', 'icp/a/', 'GET', null, self::get_api_headers());

				// Process returned data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { self::reset_api_endpoint(); return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					self::reset_api_endpoint();

					$message = sprintf(__('%s API Error: %s', 'ws-form-icontact-v2'), $this->label, $api_response_decoded->message);

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Get account ID
				if(!isset($api_response_decoded->accounts)) { return false; }
				if(!isset($api_response_decoded->accounts[0])) { return false; }
				if(!isset($api_response_decoded->accounts[0]->accountId)) { return false; }
				$this->api_account_id = $api_response_decoded->accounts[0]->accountId;

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_api_account_id', $this->api_account_id );

				// Get client folder ID
				$api_response = parent::api_call('https://app.icontact.com/', 'icp/a/' .$this->api_account_id  . '/c', 'GET', null, self::get_api_headers());

				// Process returned data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { self::reset_api_endpoint(); return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					self::reset_api_endpoint();

					$message = sprintf(__('%s API Error: %s', 'ws-form-icontact-v2'), $this->label, $api_response_decoded->message);

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Get account ID
				if(!isset($api_response_decoded->clientfolders)) { return false; }
				if(!isset($api_response_decoded->clientfolders[0])) { return false; }
				if(!isset($api_response_decoded->clientfolders[0]->clientFolderId)) { return false; }
				$this->api_client_folder_id = $api_response_decoded->clientfolders[0]->clientFolderId;

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_api_client_folder_id', $this->api_client_folder_id);
			}

			// Reset API endpoint
			public function reset_api_endpoint() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_endpoint', '');
			}

			// Nag
			public function nag() {

				// Check user API credentials are valid
				if(($this->api_app_id == '') || ($this->api_username == '') || ($this->api_password == '')) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API ID, username and password <a href="%s">here</a>.', 'ws-form-icontact-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-icontact-v2'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Process field mapping
				$email_address = false;
				$api_fields = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];
					$api_field_array = explode('|', $api_field_string);

					$api_field = $api_field_array[0];
					$api_field_sub = (isset($api_field_array[1])) ? $api_field_array[1] : false;

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// Constant Contact cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					$api_fields[$api_field] = $submit_value;
				}

				// Check email address is configured
				if(!isset($api_fields[self::EMAIL_MERGE_FIELD])) { return self::error(__('Email address not mapped', 'ws-form-icontact-v2')); }

				// Check user email address is valid
				if(!filter_var($api_fields[self::EMAIL_MERGE_FIELD], FILTER_VALIDATE_EMAIL)) { return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-icontact-v2'), $api_fields[self::EMAIL_MERGE_FIELD])); }

				// Status
				$api_fields['status'] = $this->status_contact;

				// Build data
				$data = array(

					'contacts' => $api_fields
				);

				// Data for API call
				$data_json = json_encode($data);

				// POST (Adds new records, updates existing)
				$api_response = parent::api_call($this->api_endpoint, 'contacts/', 'POST', $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						// Get contact ID
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }
						if(!isset($api_response_decoded->contacts)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }
						if(!isset($api_response_decoded->contacts[0])) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }
						if(!isset($api_response_decoded->contacts[0]->contactId)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

						$contact_id = $api_response_decoded->contacts[0]->contactId;

						self::success(__(sprintf('Successfully added contact ID: %s', $contact_id), 'ws-form'));

						// Success, subscribe to list
						$data = array(

							array(

								'contactId' => $contact_id,
								'listId'	=> $this->list_id,
								'status'	=> $this->status_subscription
							)
						);

						// Data for API call
						$data_json = json_encode($data);

						// POST (Adds new records, updates existing)
						$api_response = parent::api_call($this->api_endpoint, 'subscriptions/', 'POST', $data_json, self::get_api_headers());

						self::success(__(sprintf('Successfully subscribed contact to list ID: %s', $this->list_id), 'ws-form'));
						return true;

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get list fields
				$list_fields = self::get_list_fields();

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-icontact-v2')); }

				// Get member
				$api_response = parent::api_call($this->api_endpoint, 'contacts/?email=' . urlencode($user_email), 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }
				if(!isset($api_response_decoded->contacts)) { return self::error(__('Contact not found', 'ws-form-icontact-v2')); }
				if(!isset($api_response_decoded->contacts[0])) { return self::error(__('Contact not found', 'ws-form-icontact-v2')); }

				$api_response_decoded = $api_response_decoded->contacts[0];

				// Process fields
				$fields_return = array(

					'status_contact' => parent::get_object_value($api_response_decoded, 'status')
				);

				$api_fields = parent::get_object_value($api_response_decoded, 'dataFields');

				foreach($list_fields as $list_field) {

					$fields_return[$list_field['id']] =	parent::get_object_value($api_response_decoded, $list_field['id']);
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$params = array(

							'limit'		=> self::LISTS_FETCH_COUNT,
							'offset'	=> $total_lists_retrieved
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'lists/', 'GET', $params, $this->get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

						$api_response_decoded = $api_response_decoded->lists;

						// Get total lists
						$total_lists = count($api_response_decoded);

						// Increment read counter
						$total_lists_retrieved += $total_lists;

						foreach($api_response_decoded as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'listId'),
								'label' => 			parent::get_object_value($list, 'name'),
								'field_count' => 	false,
								'record_count' => 	false
							);
						}

					} while ($total_lists == self::LISTS_FETCH_COUNT);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

					$api_response_list = $api_response_list->list;

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'customfields/', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Default iContact fields
					$icontact_fields = array(

						array('id' => self::EMAIL_MERGE_FIELD, 'type' => 'email', 'label' => __('Email', 'ws-form-icontact-v2'), 'required' => true),
						array('id' => 'firstName', 'type' => 'text', 'label' => __('First Name', 'ws-form-icontact-v2')),
						array('id' => 'lastName', 'type' => 'text', 'label' => __('Last Name', 'ws-form-icontact-v2')),
						array('id' => 'prefix', 'type' => 'text', 'label' => __('Prefix', 'ws-form-icontact-v2')),
						array('id' => 'suffix', 'type' => 'text', 'label' => __('Suffix', 'ws-form-icontact-v2')),
						array('id' => 'street', 'type' => 'text', 'label' => __('Address 1', 'ws-form-icontact-v2')),
						array('id' => 'street2', 'type' => 'text', 'label' => __('Address 2', 'ws-form-icontact-v2')),
						array('id' => 'city', 'type' => 'text', 'label' => __('City', 'ws-form-icontact-v2')),
						array('id' => 'state', 'type' => 'text', 'label' => __('State', 'ws-form-icontact-v2')),
						array('id' => 'postalCode', 'type' => 'text', 'label' => __('Zip', 'ws-form-icontact-v2')),
						array('id' => 'business', 'type' => 'text', 'label' => __('Business', 'ws-form-icontact-v2')),
						array('id' => 'phone', 'type' => 'tel', 'label' => __('Phone', 'ws-form-icontact-v2')),
						array('id' => 'fax', 'type' => 'tel', 'label' => __('Fax', 'ws-form-icontact-v2'))
					);

					$sort_index_offset = 0;
					foreach($icontact_fields as $icontact_field) {

						// Add built in icontact fields not returned by API
						$list_fields[] = array(

							'id' => 			$icontact_field['id'],
							'label' => 			$icontact_field['label'], 
							'label_field' => 	$icontact_field['label'], 
							'type' => 			$icontact_field['type'],
							'action_type' =>	$icontact_field['type'],
							'required' => 		isset($icontact_field['required']) ? $icontact_field['required'] : false, 
							'default_value' => 	'',
							'pattern' => 		'',
							'placeholder' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index_offset,
							'visible' =>		true,
							'meta' => 			false
						);

						$sort_index_offset++;
					}

					// Process list data
					$api_list_fields = json_decode($api_response['response']);

					// Checks
					if(is_null($api_list_fields)) { return self::error(__('Invalid API response data', 'ws-form-icontact-v2')); }

					$api_list_fields = $api_list_fields->customfields;

					// Process fields
					foreach($api_list_fields as $field) {

						$action_type = parent::get_object_value($field, 'fieldType');

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						$list_fields[] = array(

							'id' => 			parent::get_object_value($field, 'privateName'),
							'label' => 			parent::get_object_value($field, 'publicName'),
							'label_field' => 	parent::get_object_value($field, 'publicName'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		false,
							'default_value' => 	'',
							'pattern' => 		'',
							'placeholder' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index_offset,
							'visible' =>		parent::get_object_value($field, 'displayToUser'),
							'meta' => 			false
						);

						$sort_index_offset++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-icontact-v2'); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->message);

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_status_contact',
						'action_' . $this->id . '_status_subscription',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-icontact-v2') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-icontact-v2')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('iContact List', 'ws-form-icontact-v2'),
						'type'						=>	'select',
						'help'						=>	__('Select the iContact list to associate this form with.', 'ws-form-icontact-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-icontact-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Contact Status
					'action_' . $this->id . '_status_contact'	=> array(

						'label'						=>	__('Contact Status', 'ws-form-icontact-v2'),
						'type'						=>	'select',
						'help'						=>	__("Set contact status.", 'ws-form-icontact-v2'),
						'options'					=>	array(

							array('value' => 'normal', 'text' => __('Normal', 'ws-form-icontact-v2')),
							array('value' => 'bounced', 'text' => __('Bounced', 'ws-form-icontact-v2')),
							array('value' => 'donotcontact', 'text' => __('Do not contact', 'ws-form-icontact-v2')),
							array('value' => 'pending', 'text' => __('Pending', 'ws-form-icontact-v2')),
							array('value' => 'invitable', 'text' => __('Invitable', 'ws-form-icontact-v2')),
							array('value' => 'deleted', 'text' => __('Deleted', 'ws-form-icontact-v2')),
						),
						'default'					=>	'normal'
					),

					// Subscription Status
					'action_' . $this->id . '_status_subscription'	=> array(

						'label'						=>	__('Subscription Status', 'ws-form-icontact-v2'),
						'type'						=>	'select',
						'help'						=>	__("Set subscription status.", 'ws-form-icontact-v2'),
						'options'					=>	array(

							array('value' => 'normal', 'text' => __('Normal', 'ws-form-icontact-v2')),
							array('value' => 'pending', 'text' => __('Pending', 'ws-form-icontact-v2')),
							array('value' => 'unsubscribed', 'text' => __('Unsubscribed', 'ws-form-icontact-v2'))
						),
						'default'					=>	'normal'
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-icontact-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to iContact fields.', 'ws-form-icontact-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-dotmailer-v2'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-dotmailer-v2'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-dotmailer-v2')
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('iContact Field', 'ws-form-icontact-v2'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-icontact-v2'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-icontact-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-icontact-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your icontact add-on for WS Form PRO license key here.', 'ws-form-icontact-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-icontact-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_app_id'	=>	array(

							'label'		=>	__('iContact API App ID', 'ws-form-icontact-v2'),
							'type'		=>	'text',
							'default'	=>	'',
							'help'		=>	__('For more information about the iContact API, <a href="https://www.icontact.com/developerportal/documentation/register-your-app" target="_blank">click here</a>.', 'ws-form-icontact-v2')
						),

						'action_' . $this->id . '_api_username'	=>	array(

							'label'		=>	__('iContact API Username', 'ws-form-icontact-v2'),
							'type'		=>	'text',
							'default'	=>	''
						),

						'action_' . $this->id . '_api_password'	=>	array(

							'label'		=>	__('iContact API Password', 'ws-form-icontact-v2'),
							'type'		=>	'password',
							'default'	=>	''
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'checkbox' : return 'checkbox';
					case 'text' : return 'text';
					case 'number' : return 'number';
					case 'decimalOne' : return 'number';
					case 'decimalTwo' : return 'number';
					case 'decimalThree' : return 'number';
					case 'decimalFour' : return 'number';
					case 'date' : return 'datetime';

					default :

						return 'text';
				}
			}

			// Convert action required to WS Form required
			public function action_to_ws_form_required($required) {

				return ($required == 1);
			}

			// Convert action default value to WS Form default value
			public function action_to_ws_form_default_value($default_value, $type = 'text') {

				switch($type) {

					case 'datetime' :

						return ($default_value !== '') ? date('Y-m-d', strtotime($default_value)) : '';

					default :

						return $default_value;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->status_contact = parent::get_config($config, 'action_' . $this->id . '_status_contact', self::DEFAULT_STATUS);
				$this->status_subscription = parent::get_config($config, 'action_' . $this->id . '_status_subscription', self::DEFAULT_STATUS);
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = false;

				$this->api_app_id = WS_Form_Common::option_get('action_' . $this->id . '_api_app_id', '');
				$this->api_username = WS_Form_Common::option_get('action_' . $this->id . '_api_username', '');
				$this->api_password = WS_Form_Common::option_get('action_' . $this->id . '_api_password', '');
				$this->api_account_id = WS_Form_Common::option_get('action_' . $this->id . '_api_account_id', '');
				$this->api_client_folder_id = WS_Form_Common::option_get('action_' . $this->id . '_api_client_folder_id', '');

				// Check user API email address is valid
				if(
					($this->api_app_id == '') ||
					($this->api_username == '') ||
					($this->api_password == '') ||
					($this->api_account_id == '') ||
					($this->api_client_folder_id == '')
				) {

					return $this->configured;
				}

				// API endpoint
				$this->api_endpoint = 'https://app.icontact.com/icp/a/' . $this->api_account_id . '/c/' . $this->api_client_folder_id . '/';

				// Check API endpoint
				$this->configured = true;

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array(

					'API-Version' 	=> '2.2',
					'API-AppId' 	=> $this->api_app_id,
					'API-Username' 	=> $this->api_username,
					'API-Password' 	=> $this->api_password
				);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g id="logo" transform="translate(18.000000, 84.000000)" fill-rule="nonzero"><path d="M29.1709998,21.6171282 C29.3407319,22.2697239 29.4920594,22.8552303 29.6556566,23.4895288 C28.7170174,24.0526721 27.6843098,24.2092137 26.6536472,24.3433923 C24.9542808,24.5629572 23.2528695,24.5954853 21.5862226,24.1400915 C17.7212377,23.1073233 15.5801588,20.5457329 14.9032752,16.6687862 C14.1548178,12.3791388 15.7703406,8.19114181 19.1608934,6.1174734 C22.4962323,4.08446516 26.0299327,4.14138939 29.7231405,5.32460019 C29.4961493,6.07884625 29.3141474,6.71314482 29.1096508,7.37590551 C25.1976317,6.25368496 21.4819293,6.06258218 18.8071144,9.75452515 C16.931881,12.3445776 16.9830051,16.9717044 18.8786882,19.4560405 C21.6046273,23.036168 25.2896552,22.6397314 29.1709998,21.6171282 Z" fill="#0F5183"></path><path d="M46.7231405,10.834825 L48.8511786,10.834825 C48.9249124,11.4801592 48.9945498,12.1155037 49.0846689,12.9006936 C50.3770578,11.192456 51.9869115,10.4352372 54.0002526,10.5031672 C56.6034639,10.5950723 58.5696974,12.5390667 58.6700572,15.3721438 C58.7745134,18.3690519 58.692587,21.3539723 58.692587,24.4028268 C57.873323,24.5786454 57.0438183,24.4547732 56.1241945,24.4787485 C56.1241945,22.1031994 56.1241945,19.8155595 56.1241945,17.5259218 C56.1241945,16.9265401 56.1241945,16.3271585 56.0566052,15.7277769 C55.8681745,13.8257392 54.8871059,12.7628358 53.224,12.6189843 C51.468727,12.4671409 49.9121255,13.5799928 49.4308079,15.3082098 C49.3298532,15.7282665 49.2885038,16.1598561 49.3079183,16.5908864 L49.3079183,24.4208083 L46.7231405,24.4208083 L46.7231405,10.834825 Z" fill="#0F5183"></path><path d="M92.2457602,21.6946655 L92.6486128,23.4753897 C90.4732087,25.1533395 85.8504749,24.7023905 83.7536271,22.6049533 C81.3606826,20.2117774 81.0303434,15.551272 83.0425922,12.5582291 C84.9964274,9.65747346 89.3593212,8.61924205 92.7231405,10.3181662 C92.5418568,10.9473973 92.3565446,11.593408 92.1692182,12.2415161 C89.0168965,11.1802129 86.7689789,11.6122849 85.39928,13.4811015 C84.0618093,15.3142616 83.9892959,18.405884 85.2320962,20.2851877 C86.5957523,22.3343839 88.6120296,22.7538713 92.2457602,21.6946655 Z" fill="#0F5183"></path><path d="M67.5809534,22.1445965 L67.5809534,24.0935241 C66.5935429,24.5604329 65.5350388,24.5750238 64.506157,24.3832577 C62.8828541,24.0810176 61.9092674,22.7824274 61.8500227,20.7667986 C61.7749795,18.2363194 61.814476,15.7016715 61.8046019,13.1711923 C61.8046019,12.7980821 61.8046019,12.424972 61.8046019,11.9747384 L59.7231405,11.9747384 L59.7231405,9.9069993 L61.780904,9.9069993 L61.780904,7.31190331 L64.2119087,6.49898167 L64.2119087,9.88615515 L67.7231405,9.88615515 L67.7231405,11.8976151 L64.2395562,11.8976151 L64.2395562,15.1805679 C64.2395562,16.7417943 64.2020346,18.3071895 64.2494303,19.8642471 C64.3086749,21.7944149 64.9899882,22.3988951 66.7949746,22.2446484 C67.0161546,22.2258887 67.2353597,22.1883692 67.5809534,22.1445965 Z" fill="#0F5183"></path><path d="M102.574956,22.1642431 L102.574956,24.0874987 C101.574166,24.5264198 100.466645,24.6169402 99.4136814,24.3458776 C97.7876009,23.9770626 96.8965166,22.7247585 96.8451459,20.647309 C96.785872,17.8051413 96.8313154,14.9629737 96.8313154,12.0895505 C96.1338592,11.7832465 95.4304756,12.0770483 94.7231405,11.8811804 L94.7231405,10.0058498 C95.3474922,9.74121986 96.0568031,10.0454402 96.795751,9.79747978 L96.795751,7.30745752 L99.2042469,6.49898167 L99.2042469,9.8641582 L102.72314,9.8641582 L102.72314,11.9478588 L99.281303,11.9478588 C99.281303,14.7921102 99.2081985,17.5259254 99.3089642,20.2534896 C99.3682381,21.8558554 100.047912,22.3371902 101.587058,22.262177 C101.8775,22.2392563 102.167942,22.2017497 102.574956,22.1642431 Z" fill="#0F5183"></path><path d="M11.1280729,24.0335242 C10.065874,24.5568988 8.8458925,24.6424258 7.7231405,24.2722274 C8.24422244,20.2264097 8.77127097,16.1441796 9.31423039,11.9931706 C10.374294,11.4409165 11.4602129,11.3559543 12.7231405,11.7160321 C12.1848217,15.8818758 11.6531325,19.9877065 11.1280729,24.0335242 Z" fill="#0F5183"></path><path d="M0.725776316,8.7274781 C0.875963294,5.18021987 2.71738624,2.59762049 5.88654914,0.763954589 C6.50471004,0.405769018 7.15552027,0.336574078 7.56907862,1.02852348 C7.95216424,1.66959424 7.56907862,2.10918562 6.98792031,2.45312518 C4.68287669,3.79225078 3.16359393,5.62388154 2.86757322,8.2329378 C2.7298016,9.45351341 2.9948886,10.6842746 3.6272146,11.7598446 C3.97329764,12.3703882 4.11042488,12.9646506 3.37690297,13.3452228 C2.58678887,13.7522518 2.08616561,13.292309 1.73572933,12.6532734 C1.03728128,11.4499198 0.689017927,10.0961829 0.725776316,8.7274781 L0.725776316,8.7274781 Z" fill="#105183"></path><path d="M5.75440738,12.2596645 C4.39428085,10.3408431 4.37456887,7.54726491 5.72483942,5.40052921 C6.62764805,3.95707195 7.80642438,3.24076984 8.50422843,3.58372661 C8.74818123,3.80156929 8.79492301,4.19140974 8.61067311,4.47150709 C8.50225723,4.56918465 8.36230218,4.71461568 8.20657755,4.83616998 C6.1111942,6.44676441 5.42127494,8.56094091 6.57048331,11.0441215 C7.14213069,12.2770294 6.21961008,12.8630948 5.75440738,12.2596645 Z" fill="#105283"></path><path d="M9.90014463,10.1604992 C9.62515601,9.01709542 9.59327327,7.93597819 10.3744004,7.13292609 C11.1555275,6.329874 12.1558484,6.43887553 13.1043599,6.65687859 C13.6403884,6.77922725 13.7499854,7.33758203 13.7181026,7.89148777 C13.6643005,8.82356209 13.6324178,9.81124943 12.6858989,10.1983161 C11.791828,10.6119515 10.7849412,10.5982829 9.90014463,10.1604992 Z" fill="#0F5183"></path><path d="M37.8237109,9.49898167 C33.5517808,9.49898167 30.7356249,12.5023539 30.7231824,17.065372 C30.7107399,21.4913942 33.4709044,24.4989817 37.5665656,24.4989817 C41.8799708,24.4989817 44.7376017,21.4492416 44.7230854,16.8609319 C44.7064954,12.4096182 41.9815847,9.49898167 37.8237109,9.49898167 Z M40.7850295,20.9244418 C40.059747,21.9113052 38.92006,22.4940729 37.708093,22.4978031 C36.4961259,22.5015332 35.3529871,21.9257914 34.6218371,20.9434105 C33.0333427,18.8821487 32.9835726,15.2675639 34.5119282,13.1451809 C35.2648753,12.0649883 36.4970703,11.4360799 37.7992185,11.4673584 C39.1013667,11.498637 40.302952,12.1860065 41.0048472,13.3011455 C41.7306606,14.4224045 41.9774372,15.6806593 42.0417235,17.0021431 C41.9670684,18.414255 41.668448,19.7652455 40.7850295,20.9223342 L40.7850295,20.9244418 Z" fill="#0F5183"></path><path d="M79.5727902,21.163522 C79.5487342,19.0661861 79.5908323,16.9688501 79.5587576,14.8715142 C79.5126501,12.0652787 78.2336707,10.1776764 75.9142674,9.69528912 C73.7919579,9.22105126 71.5766513,9.61346365 69.7218417,10.7921958 C69.938346,11.4088126 70.1227757,11.9499252 70.3232427,12.5308873 C71.2744757,11.9569431 72.3330658,11.6049395 73.4264719,11.498998 C75.8140339,11.2745831 76.9927799,12.4176312 76.9847612,15.0351064 C76.6559953,15.0351064 76.3272295,15.0162304 76.0024729,15.0351064 C74.0920224,15.1714332 72.1935998,15.4000429 70.5758311,16.599719 C69.0643099,17.7175991 68.4809509,19.290601 68.8137262,21.176106 C69.1264547,22.9378682 70.2169952,24.0305802 71.8668386,24.3682513 C73.5788268,24.7185064 75.1805582,24.3829327 76.5477432,23.1434071 C76.7482102,22.9546469 76.8724997,22.6001971 77.2573963,22.6316572 C77.4578634,23.1035577 77.2012656,23.6929091 77.5981903,24.1144737 L79.7231405,24.1144737 C79.6750284,23.0993631 79.5848183,22.1324912 79.5727902,21.163522 Z M77.1110554,19.3178664 C76.9290065,21.2124454 75.3672132,22.6303454 73.5487568,22.5519584 C72.4261416,22.480649 71.5921988,21.7906254 71.3676758,20.7482495 C71.0818562,19.5912901 71.5668757,18.3766925 72.5544404,17.7763245 C73.9336534,16.9373901 75.455198,16.868178 77.1210788,16.9059301 C77.1210788,17.7972978 77.1671862,18.5670201 77.1110554,19.3178664 Z" fill="#0F5183"></path></g>';

				return $svg_logo;
			}
		}
	});