<?php

	/**
	 * @link              https://wsform.com/knowledgebase/mailchimp/
	 * @since             1.0.0
	 * @package           WS_Form_Mailchimp
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Mailchimp
	 * Plugin URI:        https://wsform.com/knowledgebase/mailchimp/
	 * Description:       Mailchimp add-on for WS Form PRO
	 * Version:           1.0.10
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-mailchimp-v3
	 */

	Class WS_Form_Add_On_Mailchimp_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.71';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Mailchimp_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-mailchimp-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_mailchimp = new WS_Form_Add_On_Mailchimp_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_mailchimp, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Mailchimp_V3 extends WS_Form_Action {

			public $id = 'mailchimpv3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Subscribers';

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_access_token;
			private $api_key;
			public $list_id = false;
			public $status;
			public $vip;
			public $email_type;
			public $opt_in_field;
			public $field_mapping;
			public $tag_mapping;
			public $tag_category_label_prefix;
			public $tags;

			// Constants
			const DEFAULT_STATUS = 'subscribed';
			const DEFAULT_EMAIL_TYPE = 'html';
			const EMAIL_MERGE_FIELD = 'EMAIL';
			const LIST_FIELDS_FETCH_COUNT = 50;
			const LISTS_FETCH_COUNT = 10;
			const TAG_CATEGORIES_FETCH_COUNT = 10;
			const TAGS_FETCH_COUNT = 10;
			const WS_FORM_LICENSE_ITEM_ID = 580;
			const WS_FORM_LICENSE_NAME = 'Mailchimp add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.10';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Mailchimp', 'ws-form-mailchimp-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to Mailchimp', 'ws-form-mailchimp-v3');

				// Set tag category label prefix
				$this->tag_category_label_prefix = __('Select an interest: ', 'ws-form-mailchimp-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save API endpoint
				$this->api_endpoint = WS_Form_Common::get_query_var($this->id . '_api_endpoint', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_endpoint', $this->api_endpoint);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-mailchimp-v3') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-mailchimp-v3') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-mailchimp-v3') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/mailchimp-v3';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-mailchimp-v3'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

					break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_endpoint', '');
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Mailchimp account  <a href="%s">here</a>.', 'ws-form-mailchimp-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-mailchimp-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Process field mapping
				$email_address = false;
				$merge_fields = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					if($field_id == '') { continue; }

					$merge_field_string = $field_map['action_' . $this->id . '_list_fields'];
					if($merge_field_string == '') { continue; }
					$merge_field_array = explode('|', $merge_field_string);

					$merge_field = $merge_field_array[0];
					$merge_field_sub = (isset($merge_field_array[1])) ? $merge_field_array[1] : false;

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($submit_value === false) { continue; }

					// Mailchimp cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($merge_field) {

						case self::EMAIL_MERGE_FIELD :

							$email_address = $submit_value;
							break;
					}

					if($merge_field_sub === false) {

						// Normal merge field
						$merge_fields[$merge_field] = $submit_value;

					} else {

						// Merge field with sub component (e.g. Address)
						if(!isset($merge_fields[$merge_field])) { $merge_fields[$merge_field] = array(); }
						$merge_fields[$merge_field][$merge_field_sub] = $submit_value;
					}
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-mailchimp-v3')); }

				// Process interests
				$interests = new stdClass();
				$interests_push = false;
				foreach($this->tag_mapping as $tag_map) {

					// Get field ID
					$field_id = $tag_map['ws_form_field'];
					if($field_id == '') { continue; }

					$interests_push = true;

					// Read submit meta
					$interests_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($interests_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($interests_array)) { $interests_array = array($interests_array); }

					// Process each interest ready for API format
					foreach($interests_array as $interest) {

						if($interest == '') { continue; }

						$interests->{$interest} = true;
					}
				}

				// Process tags
				$tags = array();
				foreach($this->tags as $tag_map) {

					// Get tag
					$tag = $tag_map['action_' . $this->id . '_tag'];
					if($tag == '') { continue; }

					$tags[] = array(

						'name' => WS_Form_Common::parse_variables_process($tag, $form, $submit, 'text/plain'),
						'status' => 'active'
					);
				}

				// Build post data
				$data = array(

					'email_address'	=>	$email_address,
					'email_type'	=>	$this->email_type,
					'vip'			=>	($this->vip == 'on'),
					'status'		=>	$this->status
				);
				if(count($merge_fields) > 0) { $data['merge_fields'] = $merge_fields; }
				if($interests_push) { $data['interests'] = $interests; }

				// Tracking data - Remote IP
				$tracking_remote_ip = parent::get_submit_value($submit, 'tracking_remote_ip', '');
				if(filter_var($tracking_remote_ip, FILTER_VALIDATE_IP) !== false) {

					$data['ip_signup'] = $tracking_remote_ip;
					if($opted_in) { $data['ip_opt'] = $tracking_remote_ip; }
				}

				// Tracking data - Geo Location
				$tracking_geo_location = parent::get_submit_value($submit, 'tracking_geo_location', '');
				if($tracking_geo_location != '') {

					$tracking_geo_location_array = explode(',', $tracking_geo_location);
					if(
						(count($tracking_geo_location_array) == 2) &&
						is_numeric($tracking_geo_location_array[0]) &&
						is_numeric($tracking_geo_location_array[1])
					) {

						$data['location'] = array(

							'latitude'	=>	$tracking_geo_location_array[0],
							'longitude'	=>	$tracking_geo_location_array[1]
						);
					}
				}

				// Data for API call
				$data_json = json_encode($data);

				// PUT (Adds new records, updates existing)
				$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/members/' . self::ws_form_email_to_action_record_id($email_address), 'PUT', $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						// Process tags
						if(count($tags) > 0) {

							// JSON encode tag data
							$tags_json = json_encode(array('tags' => $tags));

							// Add tags
							$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/members/' . self::ws_form_email_to_action_record_id($email_address) . '/tags', 'POST', $tags_json, self::get_api_headers());
						}

						self::success(__('Successfully added to list: ', 'ws-form-mailchimp-v3') . $this->list_id);
						return true;

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-mailchimp-v3')); }

				// Build member hash
				$member_id = self::ws_form_email_to_action_record_id($user_email);

				// Get member
				$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/members/' . $member_id, 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }

				// Process fields
				$fields_return = array(

					'email_format' =>			parent::get_object_value($api_response_decoded, 'email_type'),
					'status' =>					parent::get_object_value($api_response_decoded, 'status'),
					'vip' =>					parent::get_object_value($api_response_decoded, 'vip') ? 'on' : '',
					self::EMAIL_MERGE_FIELD	=>	parent::get_object_value($api_response_decoded, 'email_address'),
				);

				$merge_fields = parent::get_object_value($api_response_decoded, 'merge_fields');
				if($merge_fields !== false) {

					$merge_fields = get_object_vars($merge_fields);
					foreach($merge_fields as $id => $value) {

						if(is_null($value)) { continue; }

						switch(gettype($value)) {

							// Process subs
							case 'object' :

								$subs = get_object_vars($value);
								foreach($subs as $id_sub => $value_sub) {

									$fields_return[$id . '|' . $id_sub] = (string) $value_sub;
								}

								break;

							// Process as normal string
							default :

								$fields_return[$id] = (string) $value;
						}
					}
				}

				// Process tags
				$tags_return = array();

				$tags = parent::get_object_value($api_response_decoded, 'interests');
				if($tags !== false) {

					$tags = get_object_vars($tags);
					foreach($tags as $id => $value) {

						$tags_return[$id] = ($value == 1);
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => $tags_return);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$params = array(

							'count'		=> self::LISTS_FETCH_COUNT,
							'offset'	=> $total_lists_retrieved
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'lists', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }
						if(!isset($api_response_decoded->lists)) { return self::error(__('Lists not found in API response data', 'ws-form-mailchimp-v3')); }

						// Get total field count
						if(isset($api_response_decoded->total_items)) {

							$total_lists = $api_response_decoded->total_items;
						}

						// Increment read counter
						$total_lists_retrieved += count($api_response_decoded->lists);

						foreach($api_response_decoded->lists as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	parent::get_object_value($list->stats, 'merge_field_count'), 
								'record_count' => 	parent::get_object_value($list->stats, 'member_count')
							);
						}

					} while ($total_lists_retrieved < $total_lists);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();
					$total_fields = 0;
					$total_fields_retrieved = 0;

					do {

						// Params
						$params = array(

							'count'		=> self::LIST_FIELDS_FETCH_COUNT,
							'offset'	=> $total_fields_retrieved
						);

						// Get list fields
						$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/merge-fields', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);

						// Checks
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }
						if(!isset($api_response_decoded->merge_fields)) { return self::error(__('List fields not found in API response data', 'ws-form-mailchimp-v3')); }

						// Get total field count
						if(isset($api_response_decoded->total_items)) {

							$total_fields = $api_response_decoded->total_items;
						}

						// Get the fields
						$api_list_fields_read = $api_response_decoded->merge_fields;

						// Increment read counter
						$total_fields_retrieved += count($api_list_fields_read);

						$api_list_fields = array_merge($api_list_fields, $api_list_fields_read);

					} while ($total_fields_retrieved < $total_fields);

					// Sort by display_order
					usort($api_list_fields, function ($a, $b) { return $a->display_order - $b->display_order; });

					// Process fields
					$sort_index_offset = 0;
					foreach($api_list_fields as $field) {

						$action_type = parent::get_object_value($field, 'type');
						$action_options = parent::get_object_value($field, 'options');

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						$list_field = array(

							'id' => 			parent::get_object_value($field, 'tag'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		self::action_to_ws_form_required(parent::get_object_value($field, 'required')), 
							'default_value' => 	self::action_to_ws_form_default_value(parent::get_object_value($field, 'default_value'), $type),
							'pattern' => 		self::action_field_to_ws_form_pattern($action_type, $action_options),
							'placeholder' => 	self::action_field_to_ws_form_placeholder($action_type, $action_options),
							'input_mask' => 	self::action_field_to_ws_form_input_mask($action_type, $action_options),
							'help' => 			parent::get_object_value($field, 'help_text'), 
							'sort_index' => 	intval(parent::get_object_value($field, 'display_order')) + $sort_index_offset,
							'visible' =>		self::action_to_ws_form_visible(parent::get_object_value($field, 'public')),
							'meta' => 			self::action_field_to_ws_form_meta_keys($field)
						);

						$sort_index_offset += self::list_field_explode($list_fields, $list_field);
					}

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_MERGE_FIELD, 
						'id_sub' =>			false,
						'label' => 			'Email', 
						'label_field' => 	'Email', 
						'label_sub' =>		false,
						'type' => 			'email', 
						'required' => 		true, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_tag_mapping'		=>	'tag_mapping',
							'action_' . $this->id . '_tags'				=>	'tags',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get tag categories
			public function get_tag_categories($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$tag_categories = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id);

				if($fetch || ($tag_categories === false)) {

					$tag_categories = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_tag_categories = array();
					$total_tag_categories = 0;
					$total_tag_categories_retrieved = 0;

					do {

						// Params
						$params = array(

							'count'		=> self::TAG_CATEGORIES_FETCH_COUNT,
							'offset'	=> $total_tag_categories_retrieved
						);

						// Get tag_categories
						$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/interest-categories', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process tag_category data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }
						if(!isset($api_response_decoded->categories)) { return self::error(__('Tag categories not found in API response data', 'ws-form-mailchimp-v3')); }

						// Get total tag category count
						if(isset($api_response_decoded->total_items)) {

							$total_tag_categories = $api_response_decoded->total_items;
						}

						// Get the tag categories
						$api_tag_categories_read = $api_response_decoded->categories;

						// Increment read counter
						$total_tag_categories_retrieved += count($api_tag_categories_read);

						$api_tag_categories = array_merge($api_tag_categories, $api_tag_categories_read);

					} while ($total_tag_categories_retrieved < $total_tag_categories);

					foreach($api_tag_categories as $tag_category) {

						$type = parent::get_object_value($tag_category, 'type');

						// Skip hidden types
						if($type == 'hidden') { continue; }

						$tag_categories[] = array(

							'id' => 		parent::get_object_value($tag_category, 'id'),
							'label' => 		parent::get_object_value($tag_category, 'title'), 
							'sort_index' =>	parent::get_object_value($tag_category, 'display_order'),
							'type' => 		self::action_tag_category_type_to_ws_form_field_type($type)
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id, $tag_categories);
				}

				return $tag_categories;
			}

			// Get tags
			public function get_tags($tag_category_id = false, $fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check tag category ID is set
				if($tag_category_id === false) { return self::error(__('Tag category ID is not set', 'ws-form-mailchimp-v3')); }

				$tags = WS_Form_Common::option_get('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id);

				if($fetch || ($tags === false)) {

					$tags = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_tags = array();
					$total_tags = 0;
					$total_tags_retrieved = 0;

					do {

						// Params
						$params = array(

							'count'		=> self::TAGS_FETCH_COUNT,
							'offset'	=> $total_tags_retrieved
						);

						// Get tags
						$api_response = parent::api_call($this->api_endpoint, 'lists/' . $this->list_id . '/interest-categories/' . $tag_category_id . '/interests', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process tag data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-mailchimp-v3')); }
						if(!isset($api_response_decoded->interests)) { return self::error(__('Tags not found in API response data', 'ws-form-mailchimp-v3')); }

						// Get total tag category count
						if(isset($api_response_decoded->total_items)) {

							$total_tags = $api_response_decoded->total_items;
						}

						// Get the tags
						$api_tags_read = $api_response_decoded->interests;

						// Increment read counter
						$total_tags_retrieved += count($api_tags_read);

						$api_tags = array_merge($api_tags, $api_tags_read);

					} while ($total_tags_retrieved < $total_tags);

					foreach($api_tags as $tag) {

						$tags[] = array(

							'id' => 			parent::get_object_value($tag, 'id'),
							'label' => 			parent::get_object_value($tag, 'name'),
							'record_count' =>	parent::get_object_value($tag, 'subscriber_count'),
							'sort_index' =>		parent::get_object_value($tag, 'display_order')
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_tag_categories_' . $this->list_id . '_tags_' . $tag_category_id, $tags);
				}

				return $tags;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-mailchimp-v3'); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->detail);

				if(isset($api_response_decoded->errors) && is_array($api_response_decoded->errors)) {

					$error_detail_array = array_merge($error_detail_array, self::get_api_errors($api_response));
				}

				return $error_detail_array;
			}

			// Get API errors
			public function get_api_errors($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-mailchimp-v3'); }

				// Get errors from JSON response
				$errors = $api_response_decoded->errors;

				$error_return = array();
				foreach($errors as $error) {

					$error_return[] = $error->field . ' - ' . $error->message;
				}

				return $error_return;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_status',
						'action_' . $this->id . '_vip',
						'action_' . $this->id . '_email_type',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_tag_mapping',
						'action_' . $this->id . '_tags'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-mailchimp-v3') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-mailchimp-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('Mailchimp List', 'ws-form-mailchimp-v3'),
						'type'						=>	'select',
						'help'						=>	__('Select the Mailchimp list to associate this form with.', 'ws-form-mailchimp-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-mailchimp-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Status (subscribed, unsubscribed, cleaned, pending)
					'action_' . $this->id . '_status'	=> array(

						'label'						=>	__('Status', 'ws-form-mailchimp-v3'),
						'type'						=>	'select',
						'help'						=>	__("Set subscriber's status to this.", 'ws-form-mailchimp-v3'),
						'options'					=>	array(

							array('value' => 'subscribed', 'text' => __('Subscribed', 'ws-form-mailchimp-v3')),
							array('value' => 'unsubscribed', 'text' => __('Unsubscribed', 'ws-form-mailchimp-v3')),
							array('value' => 'cleaned', 'text' => __('Cleaned', 'ws-form-mailchimp-v3')),
							array('value' => 'pending', 'text' => __('Pending', 'ws-form-mailchimp-v3'))
						),
						'default'					=>	'subscribed'
					),

					// VIP
					'action_' . $this->id . '_vip'	=> array(

						'label'						=>	__('VIP', 'ws-form-mailchimp-v3'),
						'type'						=>	'checkbox',
						'help'						=>	__('VIP status for the subscriber.', 'ws-form-mailchimp-v3'),
						'default'					=>	''
					),

					// Email type
					'action_' . $this->id . '_email_type'	=> array(

						'label'						=>	__('Email Type', 'ws-form-mailchimp-v3'),
						'type'						=>	'select',
						'help'						=>	__('Type of email this subscriber will receive.', 'ws-form-mailchimp-v3'),
						'options'					=>	array(

							array('value' => 'html', 'text' => __('HTML', 'ws-form-mailchimp-v3')),
							array('value' => 'text', 'text' => __('Plain text', 'ws-form-mailchimp-v3'))
						),
						'default'					=>	'html'
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-mailchimp-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-mailchimp-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-mailchimp-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-mailchimp-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Mailchimp merge fields.', 'ws-form-mailchimp-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Interest mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Interest Mapping', 'ws-form-mailchimp-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Mailchimp list interests.', 'ws-form-mailchimp-v3'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'meta_keys_unique'			=>	array(
							'ws_form_field_choice'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Tags
					'action_' . $this->id . '_tags'	=> array(

						'label'						=>	__('Tags', 'ws-form-mailchimp-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Member tagging.', 'ws-form-mailchimp-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_tag'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Mailchimp Field', 'ws-form-mailchimp-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-mailchimp-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// List fields
					'action_' . $this->id . '_tag'	=> array(

						'label'							=>	__('Tag', 'ws-form-mailchimp-v3'),
						'type'							=>	'text'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-mailchimp-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-mailchimp-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Mailchimp add-on for WS Form PRO license key here.', 'ws-form-mailchimp-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-mailchimp-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Mailchimp Status', 'ws-form-mailchimp-v3'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Mailchimp account, or <a href=\"https://login.mailchimp.com/signup/\" target=\"blank\">create a new account</a>.", 'ws-form-mailchimp-v3'),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				return $options;
			}

			// Explode list field (e.g. for address fields) - Returns number of addition fields added
			public function list_field_explode(&$list_fields, $list_field) {

				if(!isset($list_field['action_type'])) { $list_fields[] = $list_field; return 0; }

				switch($list_field['action_type']) {

					case 'address' :

						$list_fields[] = self::list_field_get_sub($list_field, 'addr1', 'Address Line 1', 'Address Line 1', 'text', $list_field['help']);
						$list_fields[] = self::list_field_get_sub($list_field, 'addr2', 'Address Line 2', 'Address Line 2', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'city', 'City', 'City', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'state', 'State', 'State', 'text');
						$list_fields[] = self::list_field_get_sub($list_field, 'zip', 'Zip', 'Zip', 'zip');
						$list_fields[] = self::list_field_get_sub($list_field, 'country', 'Country', 'Country', 'text');
						return 6;

					default :

						$list_fields[] = $list_field;
						return 0;
				}
			}

			// Build sub data
			public function list_field_get_sub(&$list_field, $id_sub, $label_sub, $label_field, $action_type, $help = '') {

				$type = self::action_field_type_to_ws_form_field_type($action_type);

				$list_field_return = array(

					'id' => 			$list_field['id'] . '|' . $id_sub,
					'label' => 			$label_sub,
					'label_field' => 	$label_field,
					'type' => 			$type,
					'action_type' =>	$action_type,
					'required' => 		$list_field['required'], 
					'default_value' => 	'', 
					'pattern' => 		self::action_field_to_ws_form_pattern($action_type),
					'placeholder' => 	self::action_field_to_ws_form_placeholder($action_type),
					'input_mask' =>		self::action_field_to_ws_form_input_mask($action_type),
					'help' => 			$help, 
					'sort_index' => 	$list_field['sort_index'],
					'visible' =>		$list_field['visible']
				);

				$list_field['sort_index'] = $list_field['sort_index'] + 1;

				return $list_field_return;
			}

			// Convert action tag category type to WS Form field type
			public function action_tag_category_type_to_ws_form_field_type($type) {

				// Convert action tag category types to WS Form types
				switch($type) {

					case 'checkboxes' : return 'checkbox';
					case 'dropdown' : return 'select';
					case 'radio' : return 'radio';

					default :

						return 'text';
				}
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'birthday' : return 'text';
					case 'url' : return 'url';
					case 'imageurl' : return 'url';
					case 'zip' : return 'text';
					case 'date' : return 'datetime';
					case 'dropdown' : return 'select';
					case 'number' : return 'number';
					case 'radio' : return 'radio';
					case 'phone' : return 'tel';

					default :

						return 'text';
				}
			}

			// Convert action field type to WS Form pattern
			public function action_field_to_ws_form_pattern($type, $options = false) {

				switch($type) {

					case 'birthday' : 

						$date_format = isset($options->date_format) ? $options->date_format : false;

						switch($date_format) {

							case 'dd/mm' : return '(0[1-9]|1[0-9]|2[0-9]|3[01]).(0[1-9]|1[012])';

							default :

								return '(0[1-9]|1[012]).(0[1-9]|1[0-9]|2[0-9]|3[01])';
						}

					case 'zip' : return '(\d{5}([\-]\d{4})?)';

					default :

						return false;
				}
			}

			// Convert action field type to WS Form pattern
			public function action_field_to_ws_form_placeholder($type, $options = false) {

				switch($type) {

					// Birthday
					case 'birthday' : 

						$date_format = isset($options->date_format) ? $options->date_format : false;

						switch($date_format) {

							case 'dd/mm' : return 'dd/mm';

							default :

								return 'mm/dd';
						}

					// Tel
					case 'phone' :

						if(
							($options !== false) &&
							isset($options->phone_format) &&
							($options->phone_format == 'US')
						) {

							return '(###) ###-####';
						}

						return false;

					// Zip
					case 'zip' : return '#####';

					default :

						return false;
				}
			}

			// Convert action field type to WS Form input mask
			public function action_field_to_ws_form_input_mask($type, $options = false) {

				switch($type) {

					case 'birthday' : 

						return '99/99';

					// Tel
					case 'phone' :

						if(
							($options !== false) &&
							isset($options->phone_format) &&
							($options->phone_format == 'US')
						) {

							return '(999) 999-9999';
						}

						return false;

					// Zip
					case 'zip' :

						return '99999';

					default :

						return false;
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$type = parent::get_object_value($field, 'type');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids for radio and select
					case 'radio' :
					case 'dropdown' :

						switch($type) {

							case 'radio' : $meta_key = 'data_grid_radio'; break;
							case 'dropdown' : $meta_key = 'data_grid_select'; break;
						}

						// Get options
						$options = parent::get_object_value($field, 'options');
						$choices = parent::get_object_value($options, 'choices');
						if($choices === false) { return false; }

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys[$meta_key])) { return false; }
						if(!isset($meta_keys[$meta_key]['default'])) { return false; }

						$meta = $meta_keys[$meta_key]['default'];

						// Build new rows
						$rows = array();
						$id = 1;
						$default_value = parent::get_object_value($field, 'default_value');
						if($default_value === false) { $default_value = ''; }
						foreach($choices as $choice) {

							$rows[] = array(

								'id'		=> $id,
								'default'	=> (($default_value == $choice) ? 'on' : ''),
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array($choice)
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						return(array($meta_key => $meta));

					default :

						return false;
				}
			}

			// Convert action required to WS Form required
			public function action_to_ws_form_required($required) {

				return ($required == 1);
			}

			// Convert action default value to WS Form default value
			public function action_to_ws_form_default_value($default_value, $type = 'text') {

				switch($type) {

					case 'datetime' :

						return ($default_value !== '') ? date('Y-m-d', strtotime($default_value)) : '';

					default :

						return $default_value;
				}
			}

			// Convert action visibility to WS Form visibility
			public function action_to_ws_form_visible($visible) {

				return ($visible == 1);
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->status = parent::get_config($config, 'action_' . $this->id . '_status', self::DEFAULT_STATUS);
				$this->vip = parent::get_config($config, 'action_' . $this->id . '_vip');
				$this->email_type = parent::get_config($config, 'action_' . $this->id . '_email_type', self::DEFAULT_EMAIL_TYPE);
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
				$this->tags = parent::get_config($config, 'action_' . $this->id . '_tags', array());
				if(!is_array($this->tags)) { $this->tags = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_endpoint = WS_Form_Common::option_get('action_' . $this->id . '_api_endpoint', '');

				if(empty($this->api_endpoint) || empty($this->api_access_token)) {

					$this->api_endpoint = false;
					$this->configured = false;

				} else {

					$this->api_endpoint .= '/3.0/';
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Convert email to action record ID
			public function ws_form_email_to_action_record_id($email_address) {

				return md5($email_address);
			}
		
			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'OAuth ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(47.000000, 66.000000)"><path d="M34.6 23.07c.36-.04.7-.04 1.01 0 .18-.42.21-1.13.05-1.92-.24-1.16-.57-1.86-1.25-1.76-.68.11-.7.95-.46 2.11.14.67.38 1.22.65 1.57zM28.77 23.99c.49.21.78.35.9.23.08-.08.05-.22-.06-.41-.24-.39-.73-.79-1.26-1.01-1.07-.46-2.35-.31-3.33.4-.33.24-.63.57-.59.77.01.06.06.11.18.13.27.03 1.2-.44 2.28-.51.77-.05 1.4.19 1.88.4zM27.8 24.55c-.63.1-.98.31-1.21.5-.19.17-.31.35-.31.48 0 .06.03.1.05.12.03.03.06.04.1.04.14 0 .46-.13.46-.13.88-.32 1.46-.28 2.04-.21.32.04.47.06.54-.05.02-.03.05-.1-.02-.2-.14-.25-.78-.66-1.65-.55zM32.63 26.59c.43.21.9.13 1.06-.19.15-.31-.07-.74-.5-.95-.43-.21-.9-.13-1.06.19-.15.32.07.74.5.95zM35.4 24.18c-.35-.01-.64.38-.65.86-.01.48.27.87.62.88.35.01.64-.38.65-.86 0-.48-.28-.87-.62-.88zM11.95 32.81c-.09-.11-.23-.08-.37-.04-.1.02-.21.05-.33.05a.7.7 0 0 1-.6-.3c-.16-.24-.15-.61.03-1.02.02-.06.05-.12.08-.19.28-.63.75-1.68.22-2.68-.4-.75-1.04-1.22-1.82-1.32-.75-.09-1.51.18-2 .72-.77.85-.89 2.01-.74 2.42.05.15.14.19.2.2.13.02.33-.08.45-.41a.53.53 0 0 0 .03-.11c.05-.17.16-.5.32-.75.2-.31.51-.53.87-.6.37-.08.75-.01 1.07.2.54.35.75 1.02.52 1.65-.12.33-.31.95-.27 1.47.09 1.04.73 1.46 1.3 1.5.56.02.95-.29 1.05-.52.08-.15.03-.23-.01-.27z"/><path d="M44.2 30.91c-.02-.08-.16-.58-.35-1.19s-.39-1.04-.39-1.04c.76-1.14.78-2.17.68-2.74-.11-.72-.41-1.33-1.01-1.96-.6-.63-1.83-1.28-3.56-1.77-.2-.06-.85-.23-.91-.25 0-.04-.05-2.14-.09-3.04-.03-.65-.08-1.67-.4-2.68-.38-1.36-1.03-2.55-1.85-3.31 2.26-2.34 3.67-4.93 3.67-7.14C39.98 1.52 34.75.23 28.3 2.9c-.01 0-1.36.58-1.37.58l-2.51-2.46C17.07-5.39-5.92 20.16 1.43 26.37l1.61 1.36c-.42 1.08-.58 2.32-.45 3.65.17 1.71 1.05 3.35 2.48 4.61 1.36 1.2 3.14 1.96 4.87 1.96 2.86 6.59 9.4 10.63 17.06 10.86 8.22.24 15.12-3.61 18.01-10.54.19-.49.99-2.68.99-4.61 0-1.95-1.1-2.75-1.8-2.75zM10.57 36.1c-.25.04-.5.06-.76.05-2.48-.07-5.16-2.3-5.43-4.95-.29-2.93 1.2-5.18 3.85-5.72.32-.06.7-.1 1.11-.08 1.49.08 3.67 1.22 4.17 4.46.45 2.86-.25 5.78-2.94 6.24zM7.8 23.73c-1.65.32-3.1 1.26-3.99 2.55-.53-.44-1.52-1.3-1.7-1.64-1.42-2.7 1.55-7.94 3.62-10.9C10.86 6.43 18.89.89 22.6 1.89c.6.17 2.61 2.49 2.61 2.49s-3.72 2.06-7.16 4.93C13.41 12.89 9.9 18.09 7.8 23.73zM33.86 35a.14.14 0 0 0 .08-.15c-.01-.08-.07-.13-.15-.12 0 0-3.89.58-7.56-.77.4-1.3 1.46-.83 3.07-.7 2.9.17 5.5-.25 7.42-.8 1.66-.48 3.85-1.42 5.55-2.76.57 1.26.77 2.64.77 2.64s.44-.08.81.15c.35.22.61.66.43 1.82-.36 2.16-1.28 3.92-2.82 5.53-.94 1.01-2.08 1.89-3.39 2.53-.69.36-1.43.68-2.21.93-5.82 1.9-11.79-.19-13.71-4.68-.15-.34-.28-.69-.39-1.06-.82-2.96-.12-6.51 2.05-8.74.13-.14.27-.31.27-.52 0-.18-.11-.36-.21-.49-.76-1.1-3.39-2.98-2.87-6.62.38-2.61 2.66-4.45 4.79-4.34l.54.03c.92.05 1.73.17 2.49.2 1.27.05 2.42-.13 3.77-1.26.46-.38.82-.71 1.44-.82.07-.01.23-.07.55-.05.33.02.65.11.93.3 1.09.72 1.24 2.47 1.3 3.75.03.73.12 2.5.15 3.01.07 1.16.37 1.32.99 1.53.35.11.67.2 1.14.33 1.44.4 2.29.81 2.83 1.34.32.33.47.68.52 1.01.17 1.24-.96 2.76-3.95 4.15-3.27 1.52-7.23 1.9-9.97 1.6-.21-.02-.96-.11-.96-.11-2.19-.3-3.44 2.54-2.13 4.48.85 1.25 3.16 2.06 5.47 2.06 5.3 0 9.37-2.26 10.88-4.21.05-.06.05-.06.12-.17.07-.11.01-.17-.08-.11-1.24.85-6.73 4.21-12.61 3.2 0 0-.71-.12-1.37-.37-.52-.2-1.6-.7-1.73-1.81 4.76 1.46 7.75.07 7.75.07zm-7.52-.89c0 .01 0 .01 0 0 0 .01 0 0 0 0zm-9.07-20.4c1.82-2.11 4.07-3.94 6.08-4.96.07-.04.14.04.11.11-.16.29-.47.91-.56 1.38-.02.07.06.13.13.09 1.25-.85 3.43-1.77 5.33-1.88.08-.01.12.1.06.15-.29.22-.61.53-.84.84-.04.05 0 .13.06.13 1.34.01 3.23.48 4.46 1.17.08.05.02.21-.07.19-1.86-.43-4.91-.75-8.08.02-2.83.69-4.98 1.75-6.56 2.9-.09.04-.19-.07-.12-.14z"/></g>';

				return $svg_logo;
			}
		}
	});