<?php

	/**
	* @link              https://wsform.com/knowledgebase/mailpoet/
	* @since             1.0.0
	* @package           WS_Form_MailPoet
	*
	* @wordpress-plugin
	* Plugin Name:       WS Form PRO - MailPoet
	* Plugin URI:        https://wsform.com/knowledgebase/mailpoet/
	* Description:       MailPoet add-on for WS Form PRO
	* Version:           1.0.1
	* Author:            Westguard Solutions
	* Author URI:        https://www.westguardsolutions.com/
	* Text Domain:       ws-form-mailpoet
	*/

	Class WS_Form_Add_On_MailPoet {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.0.0';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_MailPoet();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-mailpoet'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_mailpoet = new WS_Form_Add_On_MailPoet();	

	register_activation_hook(__FILE__, array($wsf_add_on_mailpoet, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_MailPoet extends WS_Form_Action {

			public $id = 'mailpoet';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $mailpoet_api = false;
			public $list_id = false;
			public $send_confirmation_email = 'on';
			public $schedule_welcome_email = 'on';
			public $skip_subscriber_notification = '';
			public $method;
			public $opt_in_field;
			public $field_mapping;

			// Constants
			const EMAIL_MERGE_FIELD = 'email';
			const WS_FORM_LICENSE_ITEM_ID = 5354;
			const WS_FORM_LICENSE_NAME = 'MailPoet add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.1';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('MailPoet', 'ws-form-mailpoet');

				// Set label for actions pull down
				$this->label_action = __('Add to MailPoet', 'ws-form-mailpoet');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('rest_api_init', array($this, 'rest_api_init'), 10, 2);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-mailpoet')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-mailpoet') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-mailpoet') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

					break;
				}
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('The WS Form MailPoet add-on required the MailPoet plugin to be installed and activated.', 'ws-form-mailpoet'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				if($this->method == 'subscribe') {

					// Opt-In Check
					$opted_in = false;

					// Get opt in value (False if field not submitted)
					$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

					if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

						// End user did not opt in, exit gracefully
						if(empty($opt_in_field_value)) {

							self::success(__('User did not opt in, no data pushed to action', 'ws-form-mailpoet'));
							return true;

						} else {

							$opted_in = true;
						}
					}
				}

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Process field mapping
				$email_address = false;
				$subscriber = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];

					$api_field = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					if($submit_value === false) { continue; }

					// Constant Contact cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					if($api_field == self::EMAIL_MERGE_FIELD) {

						// Save email address
						$email_address = $submit_value;
					}

					$subscriber[$api_field] = $submit_value;
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-mailpoet')); }

				// Check user email address is valid
				if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return self::error(sprintf(__('Invalid user email address: %s', 'ws-form-mailpoet'), $email_address)); }

				// Process according to method
				switch($this->method) {

					// Unsubscribe
					case 'unsubscribe' :

						// Call API
						try {

							// Get subscriber by email address
							$subscriber = $this->mailpoet_api->getSubscriber($email_address);
							if(!is_array($subscriber) || !isset($subscriber['id'])) { return self::error(__('Email address not found', 'ws-form-mailpoet')); }

							// Get subscriber ID
							$subscriber_id = $subscriber['id'];

							// Unsubscribe from list
							$response = $this->mailpoet_api->unsubscribeFromList($subscriber_id, $this->list_id);

						} catch (\MailPoet\API\MP\v1\APIException $e) {

							return self::error($e->getMessage()); 
						}

						// Success
						self::success(sprintf(__('Successfully unsubscribed %s from list: %s', 'ws-form-mailpoet'), $email_address, $this->list_id));

						break;

					// Subscribe
					default :

						// Call API
						try {

							$this->mailpoet_api->addSubscriber($subscriber, array($this->list_id), array(

								'send_confirmation_email' => ($this->send_confirmation_email == 'on'),
								'schedule_welcome_email' => ($this->send_confirmation_email == 'on'),
								'skip_subscriber_notification' => ($this->skip_subscriber_notification == 'on')
							));

						} catch (\MailPoet\API\MP\v1\APIException $e) {

							return self::error($e->getMessage()); 
						}

						// Success
						self::success(sprintf(__('Successfully subscribed %s to list: %s', 'ws-form-mailpoet'), $email_address, $this->list_id));
				}

				return true;
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get field types
				$list_fields = self::get_list_fields();
				$list_field_type = array();
				foreach($list_fields as $list_field) {

					$list_field_type[$list_field['id']] = $list_field['type'];
				}

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-mailpoet')); }

				try {

					$api_fields = $this->mailpoet_api->getSubscriber($user_email);

				} catch (\MailPoet\API\MP\v1\APIException $e) {

					$api_fields = array();
				}

				if($api_fields !== false) {

					foreach($api_fields as $key => $value) {

						if(is_null($value)) { continue; }

						$fields_return[$key] = $value;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$lists = array();

				// Get MailPoet lists
				$mailpoet_lists = $this->mailpoet_api->getLists();

				foreach($mailpoet_lists as $mailpoet_list) {

					$lists[] = array(

						'id' => 			$mailpoet_list['id'],
						'label' => 			$mailpoet_list['name'],
						'field_count' => 	false,
						'record_count' => 	false
					);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get MailPoet lists
				$mailpoet_lists = $this->mailpoet_api->getLists();

				// Run through lists until we find list_id
				foreach($mailpoet_lists as $mailpoet_list) {

					if($mailpoet_list['id'] == $this->list_id) {

						return array(

							'label' => $mailpoet_list['name']
						);
					}
				}

				return false;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = array();

				$mailpoet_subscriber_fields = $this->mailpoet_api->getSubscriberFields();

				$sort_index = 0;

				foreach($mailpoet_subscriber_fields as $mailpoet_subscriber_field) {

					$list_field_id = $mailpoet_subscriber_field['id'];
					$list_field_type = $mailpoet_subscriber_field['type'];
					$list_field_label = $mailpoet_subscriber_field['name'];
					$list_field_label_field = $mailpoet_subscriber_field['name'];
					$list_field_params = $mailpoet_subscriber_field['params'];
					$list_field_pattern = '';
					$list_field_required = false;
					$list_field_meta = false;
					$meta_key = false;

					// Process IDs
					switch($list_field_id) {

						case 'email' :

							$list_field_type = 'email';
							break;
					}

					// Process types
					switch($list_field_type) {

						case 'select' :

							$meta_key = 'data_grid_select';
							break;

						case 'checkbox' :

							$meta_key = 'data_grid_checkbox';
							break;

						case 'radio' :

							$meta_key = 'data_grid_radio';
							break;

						case 'date' :

							$list_field_type = 'datetime';
							break;
					}

					// Process params
					if(is_array($list_field_params)) {

						foreach($list_field_params as $key => $value) {

							switch($key) {

								// Label
								case 'label' :

									$list_field_label_field = $value;
									break;

								// Validate
								case 'validate' :

									switch($value) {

										case 'number' :

											$list_field_pattern = '^[0-9]+$';
											break;

										case 'alphanum' :

											$list_field_pattern = '^[a-zA-Z0-9]+$';
											break;

										case 'phone' :

											$list_field_pattern = '\([0-9]{3}\) [0-9]{3}-[0-9]{4}';
											break;
									}
									break;

								// Required
								case 'required' :

									$list_field_required = ($value == '1');
									break;

								// Values
								case 'values' :

									if($meta_key === false) { break; }

									// Get base meta
									$meta_keys = WS_Form_Config::get_meta_keys();
									if(!isset($meta_keys[$meta_key])) { break; }
									if(!isset($meta_keys[$meta_key]['default'])) { break; }

									$meta = $meta_keys[$meta_key]['default'];

									// Build new rows
									$rows = array();
									$id = 1;

									foreach($value as $row) {

										$default = isset($row['is_checked']) ? ($row['is_checked'] == '1') : false;

										$rows[] = array(

											'id'		=> $id,
											'default'	=> $default ? 'on' : '',
											'required'	=> '',
											'disabled'	=> '',
											'hidden'	=> '',
											'data'		=> array(isset($row['value']) ? $row['value'] : '')
										);

										$id++;
									}

									// Modify meta
									$meta['groups'][0]['rows'] = $rows;

									// Set list field meta
									$list_field_meta = 	array($meta_key => $meta);

									break;
							}
						}
					}

					$list_fields[] = array(

						'id' => 			$list_field_id,
						'label' => 			$list_field_label, 
						'label_field' => 	$list_field_label_field, 
						'type' => 			$list_field_type,
						'action_type' =>	$mailpoet_subscriber_field['type'],
						'required' => 		$list_field_required, 
						'default_value' => 	'',
						'pattern' => 		$list_field_pattern,
						'placeholder' => 	'',
						'help' => 			'', 
						'sort_index' => 	$sort_index,
						'visible' =>		true,
						'meta' => 			$list_field_meta
					);

					$sort_index++;
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_method',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_send_confirmation_email',
						'action_' . $this->id . '_schedule_welcome_email',
						'action_' . $this->id . '_skip_subscriber_notification'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-mailpoet') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-mailpoet')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('MailPoet List', 'ws-form-mailpoet'),
						'type'						=>	'select',
						'help'						=>	__('Select the MailPoet list to associate this form with.', 'ws-form-mailpoet'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-mailpoet'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-mailpoet'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-mailpoet'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended.', 'ws-form-mailpoet'),
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'subscribe'
							)
						)
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-mailpoet'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to MailPoet merge fields.', 'ws-form-mailpoet'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('MailPoet Field', 'ws-form-mailpoet'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-mailpoet'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// Method
					'action_' . $this->id . '_method'	=> array(

						'label'						=>	__('Method', 'ws-form-mailpoet'),
						'type'						=>	'select',
						'help'						=>	__('Subscribe or unsubscribe?', 'ws-form-mailpoet'),
						'options'					=>	array(

							array('value' => 'subscribe', 'text' => __('Subscribe', 'ws-form-mailpoet')),
							array('value' => 'unsubscribe', 'text' => __('Unsubscribe', 'ws-form-mailpoet'))
						),
						'default'					=>	'subscribe'
					),

					// Send confirmation email
					'action_' . $this->id . '_send_confirmation_email'	=> array(

						'label'						=>	__('Send Confirmation Email', 'ws-form-mailpoet'),
						'type'						=>	'checkbox',
						'help'						=>	__('It is strongly recommended to keep this setting on.', 'ws-form-mailpoet'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'subscribe'
							)
						)
					),

					// Schedule welcome email
					'action_' . $this->id . '_schedule_welcome_email'	=> array(

						'label'						=>	__('Schedule Welcome Email', 'ws-form-mailpoet'),
						'type'						=>	'checkbox',
						'help'						=>	__('Can be used to disable a welcome email.', 'ws-form-mailpoet'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'subscribe'
							)
						)
					),

					// Skip subscriber notification
					'action_' . $this->id . '_skip_subscriber_notification'	=> array(

						'label'						=>	__('Skip Subscriber Notification', 'ws-form-mailpoet'),
						'type'						=>	'checkbox',
						'help'						=>	__('Can be used to disable an admin notification email.', 'ws-form-mailpoet'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_method',
								'meta_value'	=>	'subscribe'
							)
						)
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-mailpoet'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-mailpoet'),
							'type'		=>	'text',
							'help'		=>	__('Enter your MailPoet add-on for WS Form PRO license key here.', 'ws-form-mailpoet'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-mailpoet'),
							'type'		=>	'static'
						)
					)
				);

				return $options;
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->send_confirmation_email = parent::get_config($config, 'action_' . $this->id . '_send_confirmation_email', 'on');
				$this->schedule_welcome_email = parent::get_config($config, 'action_' . $this->id . '_schedule_welcome_email', 'on');
				$this->skip_subscriber_notification = parent::get_config($config, 'action_' . $this->id . '_skip_subscriber_notification', '');
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->method = parent::get_config($config, 'action_' . $this->id . '_method', 'subscribe');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = class_exists(\MailPoet\API\API::class);

				if($this->configured) {

					$this->mailpoet_api = \MailPoet\API\API::MP('v1');
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}
		
			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(39.000000, 63.000000)"><path style="fill:#fe5301" d="M15.9 37.5c1.5 0 2.5-.3 3-1 .6-.6.9-1.4.8-2.3V15l7.2 19c.3.9.8 1.7 1.6 2.3.7.5 1.7.8 3 .8.9 0 1.7-.2 2.5-.6.8-.5 1.4-1.2 1.6-2.1l7.6-21.1V34c-.1.9.3 1.9.9 2.6.6.6 1.6.9 2.9.9 1.5 0 2.5-.3 3-1 .6-.6.9-1.4.8-2.3V3.7c.1-1.1-.5-2.2-1.4-2.8-.8-.6-2-.9-3.5-.9-1.4-.1-2.7.2-4 .8-1.1.6-1.8 1.7-2.1 2.9l-8.3 24.5-8.6-24.7c-.3-1-1-2-1.9-2.6-.9-.6-2.1-.9-3.7-.9-1.3-.1-2.6.2-3.7.9-1 .6-1.5 1.6-1.5 3V34c0 1.2.3 2 .8 2.6.7.6 1.6.9 3 .9zM62.7 49.1l-1-.8c-1.2-.9-2.8-.9-4 .1-1.5 1.2-3.3 1.9-5.3 1.9h-9c-4.2 0-8.1 1.6-11.1 4.6l-.3.5-.5-.4c-2.9-2.9-6.9-4.6-11.1-4.6h-9.1c-1.9 0-3.8-.7-5.2-1.9-1.1-1-2.8-1.1-4-.1l-1 .8c-.6.5-1.1 1.3-1.1 2.2 0 .9.3 1.8 1 2.4 2.8 2.6 6.5 4 10.4 4h9.1c3.7 0 7 2.5 8 6.1v.1c0 .3.5 1.9 3.6 1.9s3.5-1.6 3.6-1.9v-.1c1-3.6 4.2-6.1 8-6.1h9.1c3.8 0 7.5-1.4 10.4-4 .7-.6 1-1.5 1-2.4-.3-.9-.7-1.7-1.5-2.3z"/></g>';

				return $svg_logo;
			}
		}
	});