<?php

/**
 * @link              https://wsform.com/knowledgebase/paypal_checkout/
 * @since             1.0.0
 * @package           WS_Form_PayPal_Checkout
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - PayPal Checkout
 * Plugin URI:        https://wsform.com/knowledgebase/paypal_checkout/
 * Description:       PayPal Checkout add-on for WS Form PRO
 * Version:           2.0.6
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-paypal-checkout
 */

	Class WS_Form_Add_On_PayPal_Checkout {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.138';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_PayPal_Checkout();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-paypal-checkout'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_paypal_checkout = new WS_Form_Add_On_PayPal_Checkout();	

	register_activation_hook(__FILE__, array($wsf_add_on_paypal_checkout, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_PayPal_Checkout extends WS_Form_Action {

			public $id = 'paypalcheckout';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 10;
			public $can_repost = false;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			public $environment;
			public $sandbox_client_id;
			public $live_client_id;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 3018;
			const WS_FORM_LICENSE_NAME = 'PayPal Checkout add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '2.0.6';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('PayPal Checkout', 'ws-form-paypal-checkout');

				// Set label for actions pull down
				$this->label_action = __('Pay with PayPal Checkout', 'ws-form-paypal-checkout');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_config_frameworks', array($this, 'config_frameworks'), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();

				// If configured, register the PayPal field type
				if($this->configured) {

					add_filter('wsf_config_field_types', array($this, 'config_field_types'), 10, 2);
				}
			}

			// Config - Frameworks
			public function config_frameworks($frameworks) {

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['paypal_checkout']['class_field'] = array();
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['paypal_checkout']['class_field'] = array();
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['paypal_checkout']['class_field'] = array();

				return $frameworks;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API keys <a href="%s">here</a>.', 'ws-form-'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-paypal-checkout')));

				return $links;
			}

			// Plug-in field types for this action
			public function config_field_types($field_types) {

				$field_types['ecommerce']['types']['paypal_checkout'] = array (

					'label'						=>	__('PayPal', 'ws-form'),
					'pro_required'				=>	!WS_Form_Common::is_edition('pro'),
// @if NODE_ENV='license_pro'
					'label_default'				=>	__('Checkout', 'ws-form'),
					'label_position_force'		=>	'top',
					'mask_field'				=>	'<div id="#id" data-ecommerce-payment#attributes></div>#help<script>' . file_get_contents(__DIR__ . '/js/paypal-checkout.js') . '</script>',
					'mask_field_attributes'		=>	array('class'),
					'label_disabled'			=>	true,
					'submit_save'				=>	false,
					'submit_edit'				=>	false,
					'progress'					=>	false,
					'multiple'					=>	false,
					'icon'						=>	'<svg height="16" width="16" viewBox="0 0 16 16"><path d="M12.5 1.2C11.8.4 10.5 0 8.8 0h-5c-.4 0-.7.3-.7.6L1 13.7c0 .3.2.5.4.5h3.1l.8-4.9v.2c.1-.3.4-.6.7-.6h1.5c2.9 0 5.1-1.2 5.8-4.5 0-.1.1-.3.1-.3 0-1.3-.2-2.1-.9-2.9z"/><path d="M14 4.7c-.7 3.3-3 5.1-6.6 5.1H6.1l-1 6.2h2.1c.3 0 .6-.2.6-.5v-.1l.5-3.1v-.2c.1-.3.3-.5.6-.5h.4c2.5 0 4.5-1 5-4 .4-1.2.3-2.2-.3-2.9z"/></svg>',
					'conditional'				=>	array(

						'exclude_condition'	=>	true,
						'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper')
					),
					'meta_key_parse_variables'	=>	array(

						'action_' . $this->id . '_button_color',
						'action_' . $this->id . '_button_shape',
						'action_' . $this->id . '_button_height',
						'action_' . $this->id . '_button_label',
						'action_' . $this->id . '_button_tagline',
						'action_' . $this->id . '_button_layout',
						'action_' . $this->id . '_transaction_total',
						'action_' . $this->id . '_transaction_description',
						'action_' . $this->id . '_transaction_detail_mapping',
						'action_' . $this->id . '_transaction_zero',
						'action_' . $this->id . '_transaction_zero_message',
						'action_' . $this->id . '_message_method',
						'action_' . $this->id . '_message_clear',
						'action_' . $this->id . '_message_scroll_top',
						'action_' . $this->id . '_message_form_hide',
						'action_' . $this->id . '_message_duration',
						'action_' . $this->id . '_shipping_recipient_name',
						'action_' . $this->id . '_shipping_address_mapping',
						'action_' . $this->id . '_payer_given_name',
						'action_' . $this->id . '_payer_surname',
						'action_' . $this->id . '_payer_email_address',
						'action_' . $this->id . '_payer_phone_type',
						'action_' . $this->id . '_payer_phone',
						'action_' . $this->id . '_payer_address_mapping',
					),

					'fieldsets'				=> array(

						// Tab: Basic
						'basic'	=> array(

							'label'			=>	__('Basic', 'ws-form'),
							'meta_keys'		=>	array('hidden', 'help'),

							'fieldsets'		=>	array(

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_total',
										'action_' . $this->id . '_transaction_description'
									)
								),

								array(

									'label'		=>	__('Billing Address', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_payer_given_name',
										'action_' . $this->id . '_payer_surname',
										'action_' . $this->id . '_payer_email_address',
										'action_' . $this->id . '_payer_phone_type',
										'action_' . $this->id . '_payer_phone',
										'action_' . $this->id . '_payer_address_mapping'
									)
								),

								array(

									'label'		=>	__('Shipping Address', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_shipping_recipient_name',
										'action_' . $this->id . '_shipping_address_mapping'
									)
								)
							)
						),

						// Tab: Advanced
						'advanced'	=>	array(

							'label'			=>	__('Advanced', 'ws-form'),

							'fieldsets'	=>	array(

								array(
									'label'		=>	__('Styles', 'ws-form'),
									'meta_keys'	=>	array('class_single_vertical_align_bottom', 'action_' . $this->id . '_button_layout', 'action_' . $this->id . '_button_color',
										'action_' . $this->id . '_button_shape', 'action_' . $this->id . '_button_height', 'action_' . $this->id . '_button_label', 'action_' . $this->id . '_button_tagline')
								),

								array(
									'label'		=>	__('Classes', 'ws-form'),
									'meta_keys'	=>	array('class_field_wrapper')
								),

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_detail_mapping',
										'action_' . $this->id . '_transaction_zero',
										'action_' . $this->id . '_transaction_zero_message'
									)
								),

								array(

									'label'		=>	__('Messages', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_message_method',
										'action_' . $this->id . '_message_clear',
										'action_' . $this->id . '_message_scroll_top',
										'action_' . $this->id . '_message_form_hide',
										'action_' . $this->id . '_message_duration'
									)
								),

								array(
									'label'		=>	__('Breakpoints', 'ws-form'),
									'meta_keys'	=> array('breakpoint_sizes'),
									'class'		=>	array('wsf-fieldset-panel')
								)
							)
						)
					)
				);

				// Parse JS
				$mask_field = $field_types['ecommerce']['types']['paypal_checkout']['mask_field'];

				$mask_fields = array(

					'action_' . $this->id . '_client_id' => (($this->environment == 'production') ? $this->live_client_id : $this->sandbox_client_id)
				);

				$field_types['ecommerce']['types']['paypal_checkout']['mask_field'] = WS_Form_Common::mask_parse($mask_field, $mask_fields);

				return $field_types;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-paypal-checkout') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-paypal-checkout') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { self::error(sprintf(__('Action not configured (%s)', 'ws-form-paypal-checkout'), $this->label)); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Transaction - Amount
					'action_' . $this->id . '_transaction_total'	=> array(

						'label'						=>	__('Amount To Charge', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('Enter the amount to charge the user. Leave blank to use cart total.', 'ws-form-paypal-checkout'),
						'select_list'				=>	true,
						'placeholder'				=>	'#ecommerce_cart_total'
					),

					// Transaction - Description
					'action_' . $this->id . '_transaction_description'	=> array(

						'label'						=>	__('Description', 'ws-form-paypal-checkout'),
						'type'						=>	'textarea',
						'help'						=>	__('Enter a description for this transaction.', 'ws-form-paypal-checkout'),
						'default'					=>	'Payment to #blog_name'
					),

					// Transaction - Detail Mapping
					'action_' . $this->id . '_transaction_detail_mapping'	=> array(

						'label'						=>	__('Breakdown Mapping', 'ws-form-paypal-checkout'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to PayPal transaction amount breakdown fields. The mapped fields MUST add up to the cart total.', 'ws-form-paypal-checkout'),
						'meta_keys'					=>	array(

							'ws_form_field_ecommerce_price_cart',
							'action_' . $this->id . '_price_type'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_price_type'
						),
						'default'					=>	array()
					),

					// Transaction - Zero Amount
					'action_' . $this->id . '_transaction_zero'	=> array(

						'label'						=>	__('Zero Amount', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-paypal-checkout')),
							array('value' => 'submit', 'text' => __('Submit form', 'ws-form-paypal-checkout')),
							array('value' => 'message', 'text' => __('Show error message', 'ws-form-paypal-checkout')),
						),
						'default'					=>	'message',
						'help'						=>	__('What should happen if the cart total is zero and the PayPal button is clicked?', 'ws-form-paypal-checkout'),
					),

					// Transaction - Zero Amount - Message
					'action_' . $this->id . '_transaction_zero_message'	=> array(

						'label'						=>	__('Zero Amount Message', 'ws-form-paypal-checkout'),
						'type'						=>	'textarea',
						'help'						=>	__('Enter a message to show if the PayPal button is clicked and the cart amount is zero.', 'ws-form-paypal-checkout'),
						'default'					=>	__('The total cart amount cannot be zero. Please review the form and try again.', 'ws-form-paypal-checkout'),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_zero',
								'meta_value'		=>	'message'
							)
						)
					),

					// Shipping - Recipient name
					'action_' . $this->id . '_shipping_recipient_name'	=> array(

						'label'						=>	__('Recipient Name', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'help'						=>	__('Use variables to set the recipients name, for example: #field(1) #field(2). Leave blank to use PayPal accounts default name.', 'ws-form-paypal-checkout'),
					),

					// Shipping - Address mapping
					'action_' . $this->id . '_shipping_address_mapping'	=> array(

						'label'						=>	__('Shipping Address Mapping', 'ws-form-paypal-checkout'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to PayPal shipping address fields. If you intend on passing the shipping address to PayPal, address line 1, city, state, postal code and country code are mandatory.', 'ws-form-paypal-checkout'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_shipping_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_shipping_address_field'
						),
						'default'					=>	array()
					),

					// Shipping - Address mapping - Fields
					'action_' . $this->id . '_shipping_address_field' => array(

						'label'						=>	__('PayPal Field', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'address_line_1', 'text' => __('Address Line 1', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'address_line_2', 'text' => __('Address Line 2', 'ws-form-paypal-checkout')),
							array('value' => 'admin_area_2', 'text' => __('City', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'admin_area_1', 'text' => __('State', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'postal_code', 'text' => __('Postal Code', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'country_code', 'text' => __('Country Code', 'ws-form-paypal-checkout') . ' *')
						)
					),

					// Billing - First Name
					'action_' . $this->id . '_payer_given_name'	=> array(

						'label'						=>	__('First Name', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'help'						=>	__('Use variables to set the billing first name, for example: #field(1). Leave blank to use PayPal accounts default name.', 'ws-form-paypal-checkout'),
					),

					// Billing - Last Name
					'action_' . $this->id . '_payer_surname'	=> array(

						'label'						=>	__('Last Name', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'help'						=>	__('Use variables to set the billing last name, for example: #field(2). Leave blank to use PayPal accounts default name.', 'ws-form-paypal-checkout'),
					),

					// Billing - Email
					'action_' . $this->id . '_payer_email_address'	=> array(

						'label'						=>	__('Email', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'help'						=>	__('Use variables to set the billing email address, for example: #field(3). Leave blank to use PayPal accounts default email address.', 'ws-form-paypal-checkout'),
					),

					// Billing - Phone Type
					'action_' . $this->id . '_payer_phone_type'	=> array(

						'label'						=>	__('Phone Type', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => '', 'text' => __('None', 'ws-form-paypal-checkout')),
							array('value' => 'FAX', 'text' => __('Fax', 'ws-form-paypal-checkout')),
							array('value' => 'HOME', 'text' => __('Home', 'ws-form-paypal-checkout')),
							array('value' => 'MOBILE', 'text' => __('Mobile', 'ws-form-paypal-checkout')),
							array('value' => 'OTHER', 'text' => __('Other', 'ws-form-paypal-checkout')),
							array('value' => 'PAGER', 'text' => __('Pager', 'ws-form-paypal-checkout'))
						)
					),

					// Billing - Phone
					'action_' . $this->id . '_payer_phone'	=> array(

						'label'						=>	__('Phone', 'ws-form-paypal-checkout'),
						'type'						=>	'text',
						'help'						=>	__('Use variables to set the billing phone number, for example: #field(4).', 'ws-form-paypal-checkout'),
						'condition'					=>	array(

							array(

								'logic'				=>	'!=',
								'meta_key'			=>	'action_' . $this->id . '_payer_phone_type',
								'meta_value'		=>	''
							)
						)
					),

					// Billing - Address mapping
					'action_' . $this->id . '_payer_address_mapping'	=> array(

						'label'						=>	__('Billing Address Mapping', 'ws-form-paypal-checkout'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to PayPal billing address fields. If you intend on passing the billing address to PayPal, address line 1, city, state, postal code and country code are mandatory.', 'ws-form-paypal-checkout'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_payer_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_payer_address_field'
						),
						'default'					=>	array()
					),

					// Billing - Address mapping - Fields
					'action_' . $this->id . '_payer_address_field' => array(

						'label'						=>	__('PayPal Field', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'address_line_1', 'text' => __('Address Line 1', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'address_line_2', 'text' => __('Address Line 2', 'ws-form-paypal-checkout')),
							array('value' => 'admin_area_2', 'text' => __('City', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'admin_area_1', 'text' => __('State', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'postal_code', 'text' => __('Postal Code', 'ws-form-paypal-checkout') . ' *'),
							array('value' => 'country_code', 'text' => __('Country Code', 'ws-form-paypal-checkout') . ' *')
						)
					),

					// Button - Color
					'action_' . $this->id . '_button_color'	=> array(

						'label'						=>	__('Button Color', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'gold', 'text' => __('Gold (Recommended)', 'ws-form-paypal-checkout')),
							array('value' => 'blue', 'text' => __('Blue (First Recommended)', 'ws-form-paypal-checkout')),
							array('value' => 'silver', 'text' => __('Silver (Second Alternatives)', 'ws-form-paypal-checkout')),
							array('value' => 'white', 'text' => __('White (Second Alternatives)', 'ws-form-paypal-checkout')),
							array('value' => 'black', 'text' => __('Black (Third Alternative)', 'ws-form-paypal-checkout')),
						),
						'default'					=>	'gold'
					),

					// Button - Color
					'action_' . $this->id . '_button_shape'	=> array(

						'label'						=>	__('Button Shape', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'rect', 'text' => __('Rectangle (Recommended)', 'ws-form-paypal-checkout')),
							array('value' => 'pill', 'text' => __('Pill', 'ws-form-paypal-checkout')),
						),
						'default'					=>	'rect'
					),

					// Button - Height
					'action_' . $this->id . '_button_height'	=> array(

						'label'						=>	__('Button Height (pixels)', 'ws-form-paypal-checkout'),
						'type'						=>	'number',
						'default'					=>	'',
						'min'						=>	25,
						'max'						=>	55,
						'help'						=>	__('Leave blank for default height', 'ws-form-paypal-checkout')
					),

					// Button - Label
					'action_' . $this->id . '_button_label'	=> array(

						'label'						=>	__('Button Label', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'paypal', 'text' => __('PayPal (Recommended)', 'ws-form-paypal-checkout')),
							array('value' => 'checkout', 'text' => __('PayPal Checkout', 'ws-form-paypal-checkout')),
							array('value' => 'buynow', 'text' => __('Buy Now', 'ws-form-paypal-checkout')),
							array('value' => 'pay', 'text' => __('Pay with PayPal', 'ws-form-paypal-checkout')),
							array('value' => 'installment', 'text' => __('PayPal Installment', 'ws-form-paypal-checkout')),
						),
						'default'					=>	'paypal'
					),

					// Button - Tagline
					'action_' . $this->id . '_button_tagline'	=> array(

						'label'						=>	__('Button Tagline', 'ws-form-paypal-checkout'),
						'type'						=>	'checkbox',
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_button_layout',
								'meta_value'		=>	'horizontal'
							)
						)
					),

					// Button - Layout
					'action_' . $this->id . '_button_layout'	=> array(

						'label'						=>	__('Button Layout', 'ws-form-paypal-checkout'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'vertical', 'text' => __('Vertical (Recommended)', 'ws-form-paypal-checkout')),
							array('value' => 'horizontal', 'text' => __('Horizontal', 'ws-form-paypal-checkout')),
						),
						'default'					=>	'vertical'
					),

					// Price Types
					'action_' . $this->id . '_price_type'	=> array(

						'label'							=>	__('PayPal Breakdown', 'ws-form-paypal-checkout'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => 'item_total', 'text' => __('Subtotal')),
							array('value' => 'shipping', 'text' => __('Shipping')),
							array('value' => 'handling', 'text' => __('Handling Fee')),
							array('value' => 'tax_total', 'text' => __('Total Tax')),
							array('value' => 'insurance', 'text' => __('Insurance')),
							array('value' => 'shipping_discount', 'text' => __('Shipping Discount')),
							array('value' => 'discount', 'text' => __('Discount'))
						),
						'options_blank'					=>	__('Select...', 'ws-form-paypal-checkout'),
					),

					// Message - Method
					'action_' . $this->id . '_message_method'	=> array(

						'label'						=>	__('Error Messages', 'ws-form-post'),
						'type'						=>	'select',
						'help'						=>	__('Where should error messages be added?', 'ws-form-post'),
						'options'					=>	array(

							array('value' => 'before', 'text' => __('Before Form', 'ws-form-post')),
							array('value' => 'after', 'text' => __('After Form', 'ws-form-post'))
						),
						'default'					=>	'before',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Duration
					'action_' . $this->id . '_message_duration'	=> array(

						'label'						=>	__('Error Message Duration (ms)', 'ws-form-post'),
						'type'						=>	'number',
						'help'						=>	__('Duration in milliseconds error message shown.', 'ws-form-post'),
						'default'					=>	'4000',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Clear other messages
					'action_' . $this->id . '_message_clear'	=> array(

						'label'						=>	__('Clear Other Messages On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Clear any other messages when error message shown?', 'ws-form'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Scroll to top
					'action_' . $this->id . '_message_scroll_top'	=> array(

						'label'						=>	__('Scroll To Top On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Scroll to top of page when error message shown?', 'ws-form'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Form Hide
					'action_' . $this->id . '_message_form_hide'	=> array(

						'label'						=>	__('Hide Form On Error', 'ws-form-post'),
						'type'						=>	'checkbox',
						'help'						=>	__('Hide form when error message shown?', 'ws-form-post'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-paypal-checkout'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-paypal-checkout'),
							'type'		=>	'text',
							'help'		=>	__('Enter your PayPal Checkout add-on for WS Form PRO license key here.', 'ws-form-paypal-checkout'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-paypal-checkout'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_environment'	=>	array(

							'label'		=>	__('Environment', 'ws-form-paypal-checkout'),
							'type'		=>	'select',
							'help'		=>	__('Select which PayPal environment to use.', 'ws-form'),
							'default'	=>	'sandbox',
							'options'	=>	array(

								'sandbox'		=>	array('text' => __('Sandbox', 'ws-form')),
								'production'	=>	array('text' => __('Live', 'ws-form')),
							)				
						),

						'action_' . $this->id . '_sandbox_client_id'	=>	array(

							'label'			=>	__('Sandbox Client ID', 'ws-form-paypal-checkout'),
							'type'			=>	'text',
							'help'			=>	__('Enter your app sandbox client ID. If you do not have these details, <a href="https://developer.paypal.com/developer/applications/create" target="_blank">create a PayPal app</a>.', 'ws-form-paypal-checkout')
						),

						'action_' . $this->id . '_live_client_id'	=>	array(

							'label'			=>	__('Live Client ID', 'ws-form-paypal-checkout'),
							'type'			=>	'text',
							'help'			=>	__('Enter your app live client ID. If you do not have these details, <a href="https://developer.paypal.com/developer/applications/create" target="_blank">create a PayPal app</a>.', 'ws-form-paypal-checkout')
						)
					)
				);

				return $options;
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->environment  		= WS_Form_Common::option_get('action_' . $this->id . '_environment', '');
				$this->sandbox_client_id  	= WS_Form_Common::option_get('action_' . $this->id . '_sandbox_client_id', '');
				$this->live_client_id  		= WS_Form_Common::option_get('action_' . $this->id . '_live_client_id', '');

				// Build endpoint using API key
				if(
					empty($this->sandbox_client_id) && 
					empty($this->live_client_id)
				) {

					$this->configured = false;
				}

				return $this->configured;
			}
		}
	});