(function($) {

	'use strict';

	$.WS_Form_PayTrace = function(atts) {

		this.form_instance_id = false;
		this.transaction_amount = 0;
		this.transaction_zero = false;
		this.transaction_zero_message = '';
		this.message_method = false;
		this.message_clear = false;
		this.message_scroll_top = false;
		this.message_form_hide = false;
		this.message_duration = false;
		this.ws_form = false;
		this.pem = false;
	}

	$.WS_Form_PayTrace.prototype.init = function() {

		var ws_this = this;

		this.ws_form = window.wsf_form_instances[this.form_instance_id];

		// Set pt-encrypt classes
		$('[data-pt-encrypt]', this.ws_form.form_canvas_obj).addClass('pt-encrypt');

		// Set PEM key
		if(!this.pem) { return; }
		paytrace.setKeyAjax(this.pem) ;

		// Button click
		$('[data-wsf-paytrace-submit]').click(function(e) {

			// Reset validation
			ws_this.invalidate();

			if(ws_this.click()) {

				// Log
				ws_this.ws_form.log('log_payment', 'PayTrace - Payment Initiated', 'payment');

				// Set status
				ws_this.ws_form.form_ecommerce_set_status('pending_payment');

				// Set payment method
				ws_this.ws_form.form_ecommerce_set_payment_method('PayTrace');

				// Set payment amount (Overrides calculated amount in case the payment button has a fixed value, we always want to show the processed amount in submissions)
				ws_this.ws_form.form_ecommerce_set_payment_amount(ws_this.get_amount_total());

				// Encrypt fields with pt-encrypt class
				$('[data-pt-encrypt]', ws_this.ws_form.form_canvas_obj).each(function() {

					// Get value
					var value = $(this).val();
					if(value == '') { return; }

					// Get ID
					var id = $(this).attr('id');

					// Disable field so it is not submitted
					$(this).attr('disabled', '').attr('data-required-bypass', '');

					// Encrypt value
					var encrypted_value = paytrace.encryptValue(value);

					// Set encrypted hidden field value
					$('#' + id + '_encrypted').val(encrypted_value);
				})

				// Migratae fields with pt-noencrypt class
				$('[data-pt-no-encrypt]', ws_this.ws_form.form_canvas_obj).each(function() {

					// Get value
					var value = $(this).val();
					if(value == '') { return; }

					// Get ID
					var id = $(this).attr('id');

					// Disable field so it is not submitted
					$(this).attr('disabled', '').attr('data-required-bypass', '');

					// Set encrypted hidden field value
					$('#' + id + '_not_encrypted').val(value);
				})

				// Submit form
				ws_this.ws_form.form_obj.submit();
			}

			e.preventDefault();
		});
	}

	$.WS_Form_PayTrace.prototype.error_message = function(message) {

		this.ws_form.action_message(message, 'danger', this.message_method, this.message_duration, this.message_form_hide, this.message_clear, this.message_scroll_top, true, true);
	}

	$.WS_Form_PayTrace.prototype.invalidate = function() {

		// Get classes
		var class_paytrace_invalid_label_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_label', []);
		var class_paytrace_invalid_field_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_field', []);
		var class_paytrace_invalid_invalid_feedback_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_invalid_feedback', []);
		var class_paytrace_valid_label_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_label', []);
		var class_paytrace_valid_field_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_field', []);
		var class_paytrace_valid_invalid_feedback_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_invalid_feedback', []);

		var paytrace_obj_field = $('[data-wsf-paytrace-cc]', this.ws_form.form_canvas_obj);
		var paytrace_obj_wrapper = paytrace_obj_field.closest('[data-id]');
		var paytrace_id = paytrace_obj_wrapper.attr('data-id');
		var paytrace_obj_label = $('label', paytrace_obj_wrapper);
		var paytrace_obj_invalid_feedback = $('#' + this.ws_form.form_id_prefix + 'invalid-feedback-' + paytrace_id);

		// Empty - Label
		paytrace_obj_label.addClass(class_paytrace_valid_label_array.join(' '));
		paytrace_obj_label.removeClass(class_paytrace_invalid_label_array.join(' '));

		// Empty - Field
		paytrace_obj_field.addClass(class_paytrace_valid_field_array.join(' '));
		paytrace_obj_field.removeClass(class_paytrace_invalid_field_array.join(' '));

		// Empty - Feedback
		paytrace_obj_invalid_feedback.addClass(class_paytrace_valid_invalid_feedback_array.join(' '));
		paytrace_obj_invalid_feedback.removeClass(class_paytrace_invalid_invalid_feedback_array.join(' '));

		// Unmark as validated
		this.ws_form.form_obj.removeClass(this.ws_form.class_validated);
	}

	$.WS_Form_PayTrace.prototype.invalid_feedback = function(message) {

		// Get classes
		var class_paytrace_invalid_label_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_label', []);
		var class_paytrace_invalid_field_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_field', []);
		var class_paytrace_invalid_invalid_feedback_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_invalid_invalid_feedback', []);
		var class_paytrace_valid_label_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_label', []);
		var class_paytrace_valid_field_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_field', []);
		var class_paytrace_valid_invalid_feedback_array = this.ws_form.get_field_value_fallback('payt_checkout', false, 'class_valid_invalid_feedback', []);

		var paytrace_obj_field = $('[data-wsf-paytrace-cc]', this.ws_form.form_canvas_obj);
		var paytrace_obj_wrapper = paytrace_obj_field.closest('[data-id]');
		var paytrace_id = paytrace_obj_wrapper.attr('data-id');
		var paytrace_obj_label = $('label', paytrace_obj_wrapper);
		var paytrace_obj_invalid_feedback = $('#' + this.ws_form.form_id_prefix + 'invalid-feedback-' + paytrace_id);

		// Set invalid feedback string
		paytrace_obj_invalid_feedback.html(message);

		// Empty - Label
		paytrace_obj_label.addClass(class_paytrace_invalid_label_array.join(' '));
		paytrace_obj_label.removeClass(class_paytrace_valid_label_array.join(' '));

		// Empty - Field
		paytrace_obj_field.addClass(class_paytrace_invalid_field_array.join(' '));
		paytrace_obj_field.removeClass(class_paytrace_valid_field_array.join(' '));

		// Empty - Feedback
		paytrace_obj_invalid_feedback.addClass(class_paytrace_invalid_invalid_feedback_array.join(' '));
		paytrace_obj_invalid_feedback.removeClass(class_paytrace_valid_invalid_feedback_array.join(' '));

		// Mark form as validated
		this.ws_form.form_obj.addClass(this.ws_form.class_validated);
	}

	$.WS_Form_PayTrace.prototype.click = function() {

		this.ws_form.log('log_payment', 'PayTrace - Clicked', 'payment');

		// Check if form is valid
		if(!this.ws_form.form_valid) {

			// Form is not valid, so submit it to simulate validation
			this.ws_form.form_obj.submit();
			return false;
		}

		// Check amount
		var paytrace_total = this.get_amount_total();

		// Zero amount check
		if(paytrace_total == 0) {

			// Process as zero amount
			switch(this.transaction_zero) {

				// Submit form
				case 'submit' :

					this.ws_form.form_obj.submit();
					break;

				// Show message
				case 'message' :

					var message = this.transaction_zero_message;
					this.error_message(message);
					break;
			}

			return false;
		}

		// Limit amount check
		var paytrace_limit = this.get_paytrace_limit();
		if(paytrace_limit > 0) {

			if(paytrace_total > paytrace_limit) {

				var message = this.transaction_limit_message;
				this.error_message(message);
				return false;
			}
		}

		return true;
	}

	$.WS_Form_PayTrace.prototype.get_paytrace_limit = function() {

		// Get Stripe total
		var paytrace_limit = this.ws_form.parse_variables_process(this.transaction_limit).output;

		// Convert to a number
		paytrace_limit = this.ws_form.get_number(paytrace_limit);

		return paytrace_limit;
	}

	$.WS_Form_PayTrace.prototype.get_amount_total = function() {

		var transaction_total = this.transaction_amount;
		if(transaction_total == '') { transaction_total = '#ecommerce_cart_total'; }

		// Get Stripe total
		var paytrace_total = this.ws_form.parse_variables_process(transaction_total).output;

		// Convert to a number
		paytrace_total = this.ws_form.get_number(paytrace_total);

		return paytrace_total;
	}

})(jQuery);
