<?php

	/**
	 * @link              https://wsform.com/knowledgebase/paytrace/
	 * @since             1.0.0
	 * @package           WS_Form_PayTrace
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - PayTrace
	 * Plugin URI:        https://wsform.com/knowledgebase/paytrace/
	 * Description:       PayTrace add-on for WS Form PRO
	 * Version:           1.0.6
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-paytrace-v1
	 */

	Class WS_Form_Add_On_PayTrace {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.8.31';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_PayTrace();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-paytrace-v1'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_paytrace = new WS_Form_Add_On_PayTrace();	

	register_activation_hook(__FILE__, array($wsf_add_on_paytrace, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_PayTrace extends WS_Form_Action {

			public $id = 'paytracev1';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = false;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_public_key;
			private $username;
			private $password;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 8696;
			const WS_FORM_LICENSE_NAME = 'PayTrace add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.6';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';
			const WS_FORM_INTEGRATOR_ID = '923983yld5o';

			public function __construct() {

				// Set label
				$this->label = __('PayTrace', 'ws-form-paytrace-v1');

				// Set label for actions pull down
				$this->label_action = __('Pay with PayTrace', 'ws-form-paytrace-v1');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('wsf_actions_post_submit', array($this, 'actions_post_submit'), 10, 1);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_config_frameworks', array($this, 'config_frameworks'), 10, 1);
				add_filter('wsf_config_icon_16_svg', array($this, 'config_icon_16_svg'), 10, 2);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);
				add_action('wp_enqueue_scripts', array($this, 'enqueue_scripts'), 10);
				add_filter('wsf_wizard_config_files', array($this, 'wizard_config_files'), 10, 1);
				add_filter('wsf_wizard_svg_buttons', array($this, 'wizard_svg_buttons'), 10, 1);
				add_filter('wsf_enqueue_js_dependencies', array($this, 'enqueue_js_dependencies'), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();

				// If configured, register the PayTrace field type
				if($this->configured) {

					add_filter('wsf_config_field_types', array($this, 'config_field_types'), 10, 2);
				}
			}

			// Icon
			public function config_icon_16_svg($icon, $id) {

				if($id == 'action_' . $this->id) {

					return '<path d="M12.6 2.1c.7 0 1.2.5 1.2 1.2v3.9h-.6l1.4 2.3L16 7.2h-.7V3.3c0-1.5-1.2-2.8-2.8-2.8H3.4C1.8.5.6 1.8.6 3.3v2.1h1.6V3.3c0-.7.5-1.2 1.2-1.2h9.2M3.4 13.6c-.7 0-1.2-.5-1.2-1.2V8.5h.6L1.4 6.2 0 8.5h.7v3.9c0 1.5 1.3 2.8 2.8 2.8h9.1c1.5 0 2.8-1.3 2.8-2.8v-2.1h-1.6v2.1c0 .7-.5 1.2-1.2 1.2H3.4"/><path d="M8 11.5c-2.2 0-4.1-1.6-4.1-3.7h1.5C5.4 9 6.6 10 8 10s2.6-1 2.6-2.2h1.5c0 2.1-1.8 3.7-4.1 3.7z"/>';
				}

				return $icon;
			}

			// Wizard - Config files
			public function wizard_config_files($config_files) {

				$config_files[] = plugin_dir_path(__FILE__) . 'includes/wizard/config.json';

				return $config_files;
			}

			// Wizard - SVG buttons
			public function wizard_svg_buttons($field_type_buttons) {

				$color_default_inverted = WS_Form_Common::option_get('skin_color_default_inverted');
				$color_primary          = WS_Form_Common::option_get('skin_color_primary');

				$field_type_buttons['payt_checkout'] = array('fill' => $color_primary, 'color' => $color_default_inverted);

				return $field_type_buttons;
			}

			// WS Form common JS library dependency
			public function enqueue_js_dependencies($dependencies) {

				$dependencies[] = 'wsform-' . $this->id . '-paytrace-e2ee';

				return $dependencies;
			}

			// Enqueue scripts (Public)
			public function enqueue_scripts() {

				// Enqueue in footer?
				$jquery_footer = (WS_Form_Common::option_get('jquery_footer', '') == 'on');

				// Enqueue PayTrace End-2-End Encryption (Has to be queued after input mask or it breaks it)
				wp_enqueue_script('wsform-' . $this->id . '-paytrace-e2ee', 'https://api.paytrace.com/assets/e2ee/paytrace-e2ee.js', array('jquery'), self::WS_FORM_LICENSE_VERSION, $jquery_footer);

				// Enqueue WS Form PayTrace script
				wp_enqueue_script('wsform-' . $this->id, plugin_dir_url(__FILE__) . 'js/paytrace.js', array('wsform-' . $this->id . '-paytrace-e2ee'), self::WS_FORM_LICENSE_VERSION, $jquery_footer);
			}

			// Actions post filter
			public function actions_post_submit($actions) {

				// Get PayTrace cc num
				$paytrace_cc_num = WS_Form_Common::get_query_var('paytrace_cc_num', false);
				$paytrace_cc_em = WS_Form_Common::get_query_var('paytrace_cc_em', false);
				$paytrace_cc_ey = WS_Form_Common::get_query_var('paytrace_cc_ey', false);
				$paytrace_cc_csc = WS_Form_Common::get_query_var('paytrace_cc_csc', false);

				// Check form data
				if(empty($paytrace_cc_num)) { return $actions; }
				if(empty($paytrace_cc_em)) { return $actions; }
				if(empty($paytrace_cc_ey)) { return $actions; }
				if(empty($paytrace_cc_csc)) { return $actions; }

				// Prepend this action so it runs first
				$action_prepend = array(

					'id' => $this->id,
					'meta' => array(),
					'events' => array(
						'0' => 'submit'
					),
					'label' => $this->label_action,
					'priority' => 0,
					'row_index' => 0
				);
				array_unshift($actions, $action_prepend);

				return $actions;
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check access token
				self::access_token_check();

				// Get field ID
				$paytrace_field_id = WS_Form_Common::get_query_var('paytrace_field_id', false);
				if(empty($paytrace_field_id)) { return false; }

				// Get field
				$fields = ( false === $form ) ? array() : WS_Form_Common::get_fields_from_form($form);
				$field = isset($fields[$paytrace_field_id]) ? $fields[$paytrace_field_id] : false;

				// Check field found
				if(!$field) { return false; }

				// Convert to object
				$field = (object)$field;

				// Read config from field
				$transaction_amount = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_amount', '');
				$transaction_zero = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero', '');
				$transaction_zero_message = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero_message', ''), $form, $submit);
				$transaction_limit = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_limit', '');
				$transaction_limit_message = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_limit_message', ''), $form, $submit);

				// Process transaction amount
				switch($transaction_amount) {

					case '' :
					case '#ecommerce_cart_total' :

						// Get calculated transaction amount from query var
						$ecommerce_cart = WS_Form_Common::get_query_var('wsf_ecommerce_cart', false);
						$amount = WS_Form_Common::get_number($ecommerce_cart['total']);
						break;

					default :

						// Use amount specified in field config
						$amount = WS_Form_Common::get_number($transaction_amount);
				}

				// Check zero amount
				if($amount <= 0) {

					switch($transaction_zero) {

						case 'message' :

							self::error($transaction_zero_message);
							break;
					}

					return false;
				}

				// Check limit amount
				$limit = WS_Form_Common::get_number($transaction_limit);
				if(
					($limit > 0) &&
					($amount > $limit)
				) {

					self::error($transaction_limit_message);
					return false;
				}

				// Get transaction method
				$transaction_type = 'sale_keyed';

				switch($transaction_type) {

					case 'sale_keyed' :

						return self::sale_keyed($form, $submit, $field, $amount);
				}

				return true;
			}

			// Charge
			public function sale_keyed($form, $submit, $field, $amount) {

				// Get PayTrace values
				$paytrace_cc_num = WS_Form_Common::get_query_var('paytrace_cc_num', false);
				if(empty($paytrace_cc_num)) { return false; }

				$paytrace_cc_em = WS_Form_Common::get_query_var('paytrace_cc_em', false);
				if(empty($paytrace_cc_em)) { return false; }
				$paytrace_cc_em = intval($paytrace_cc_em);
				$paytrace_cc_em_processed = (($paytrace_cc_em < 10) ? '0' : '') . strval($paytrace_cc_em);

				$paytrace_cc_ey = WS_Form_Common::get_query_var('paytrace_cc_ey', false);
				if(empty($paytrace_cc_ey)) { return false; }
				$paytrace_cc_ey_processed = strval(intval($paytrace_cc_ey));

				$paytrace_cc_csc = WS_Form_Common::get_query_var('paytrace_cc_csc', false);

				$paytrace_field_id = WS_Form_Common::get_query_var('paytrace_field_id', false);
				if(empty($paytrace_field_id)) { return false; }

				// Build keyed sale data packet

				// Required fields
				$data = array(

					'amount' => round($amount, 2),
					'credit_card' => array(

						'encrypted_number' => $paytrace_cc_num,
						'expiration_month' => $paytrace_cc_em_processed,
						'expiration_year' =>$paytrace_cc_ey_processed,
					),
					'integrator_id' => self::WS_FORM_INTEGRATOR_ID,
				);

				// CSC
				if(!empty($paytrace_cc_csc)) {

					$data['encrypted_csc'] = $paytrace_cc_csc;
				}

				// Billing address
				$billing_address = array();

				// Billing - Name
				$billing_name = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_billing_name', ''), $form, $submit));
				if(!empty($billing_name)) { $billing_address['name'] = $billing_name; }

				// Billing address mapping
				$billing_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_billing_address_mapping', '');

				foreach($billing_address_mapping as $field_map) {

					$field_id = $field_map->ws_form_field;
					$paytrace_parameter = $field_map->{'action_' . $this->id . '_billing_address_field'};

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// PayTrace cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					$billing_address[$paytrace_parameter] = $submit_value;
				}

				if(count($billing_address) > 0) {

					$data['billing_address'] = $billing_address;
				}

				// Shipping address
				$shipping_address = array();

				// Shipping - Name
				$shipping_name = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_name', ''), $form, $submit));
				if(!empty($shipping_name)) { $shipping_address['name'] = $shipping_name; }

				// Shipping address mapping
				$shipping_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_address_mapping', '');

				foreach($shipping_address_mapping as $field_map) {

					$field_id = $field_map->ws_form_field;
					$paytrace_parameter = $field_map->{'action_' . $this->id . '_shipping_address_field'};

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// PayTrace cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					$shipping_address[$paytrace_parameter] = $submit_value;
				}

				if(count($shipping_address) > 0) {

					$data['shipping_address'] = $shipping_address;
				}

				// Invoice ID
				$invoice_id = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_invoice_id', ''), $form, $submit));
				if(!empty($invoice_id)) { $data['invoice_id'] = $invoice_id; }

				// Customer reference ID
				$customer_reference_id = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_reference_id', ''), $form, $submit));
				if(!empty($customer_reference_id)) { $data['customer_reference_id'] = $customer_reference_id; }

				// Email
				$email_field_id = intval(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_email', ''));
				$email = ($email_field_id > 0) ? parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $email_field_id, false) : '';
				if(!empty($email)) {

					// Check user email address is valid
					if(!filter_var($email, FILTER_VALIDATE_EMAIL)) { 

						self::error(sprintf(__('Invalid user email address: %s', 'ws-form-paytrace-v1'), $email));
						return 'halt';
					}

					$data['email'] = $email;
				}

				// Transaction description
				$description = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_description', ''), $form, $submit));
				if(!empty($description)) { $data['description'] = $description; }


				// JSON encode
				$data_json = json_encode($data);

				// Call API
				$api_response = parent::api_call($this->api_endpoint, 'transactions/sale/keyed', 'POST', $data_json, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/json');

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) {

					self::error(__('Invalid API response data', 'ws-form-paytrace-v1'));
					return 'halt';
				}

				// Check status code
				if($api_response['http_code'] != 200) { 

					// Set error message
					$error_message = self::get_api_error_detail($api_response_decoded);
					self::error($error_message);

					// Stop process actions
					return 'halt';
				}

				// Check for errors
				if($api_response_decoded->success) {

					// Read transaction data
					$transaction_id = $api_response_decoded->transaction_id;

					// Set transaction ID
					$submit->meta['ecommerce_transaction_id'] = $transaction_id;

					// Set transaction
					$transaction = array(

						'amount' => $amount,
						'id' => $transaction_id,
						'transaction' => array($api_response_decoded)
					);
					$submit->meta['ecommerce_transaction'] = json_encode($transaction);

					// Set status
					$submit->meta['ecommerce_status'] = 'completed';

					self::success(__('Payment successfully processed with transaction ID: ', 'ws-form-paytrace-v1') . $transaction_id);
					return $submit;

				} else {

					// Set error message
					$error_message = $api_response_decoded->status_message;
					self::error($error_message);
				}

				// Stop process actions
				return 'halt';
			}

			// Get API error detail
			public function get_api_error_detail($api_response_decoded) {

				if(!isset($api_response_decoded->status_message)) { return __('Unknown PayTrace error', 'ws-form-paytrace-v1'); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->status_message);

				if(isset($api_response_decoded->errors)) {

					foreach($api_response_decoded->errors as $error_id => $error) {

						if(is_array($error)) {

							$error_detail_array[] = implode(', ', $error);
						}
					}
				}

				return $error_detail_array;
			}

			// Config - Frameworks
			public function config_frameworks($frameworks) {

				// WS Form
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('wsf-button');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('wsf-button-full');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_num']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_em']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_ey']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['payt_cc_csc']['class_valid_field'] = array('wsf-valid');

				// Bootstrap
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_num']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_em']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_ey']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['payt_cc_csc']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_num']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_em']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_ey']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['payt_cc_csc']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_num']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_em']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_ey']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['payt_cc_csc']['class_valid_field'] = array('is-valid');

				// Foundation
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('expand');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_num']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_num']['class_field'] = array('panel');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_num']['class_invalid_label'] = array('error');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array();
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_num']['class_invalid_invalid_feedback'] = array('error');

				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_em']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_em']['class_field'] = array('panel');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_em']['class_invalid_label'] = array('error');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array();
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_em']['class_invalid_invalid_feedback'] = array('error');

				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_ey']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_ey']['class_field'] = array('panel');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_label'] = array('error');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array();
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_invalid_feedback'] = array('error');

				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_csc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_csc']['class_field'] = array('panel');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_label'] = array('error');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array();
				$frameworks['types']['foundation5']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_invalid_feedback'] = array('error');

				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_num']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_num']['class_field'] = array('callout');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_num']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_num']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_em']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_em']['class_field'] = array('callout');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_em']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_em']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_ey']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_ey']['class_field'] = array('callout');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_csc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_csc']['class_field'] = array('callout');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_num']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_num']['class_field'] = array('callout');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_num']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_num']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_num']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_em']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_em']['class_field'] = array('callout');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_em']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_em']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_em']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_ey']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_ey']['class_field'] = array('callout');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_ey']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_csc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_csc']['class_field'] = array('callout');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['payt_cc_csc']['class_invalid_invalid_feedback'] = array('is-visible');

				return $frameworks;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-paytrace-v1')));

				return $links;
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your username and password <a href="%s">here</a> and upload your public key.', 'ws-form-paytrace'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Plug-in field types for this action
			public function config_field_types($field_types) {

				$field_types['paytrace'] = array(

					'label' => __('PayTrace', 'ws-form-paytrace-v1'),
					'types' => array(

						'payt_cc_num' => array (

							'label'				=>	__('CC Number', 'ws-form'),
							'pro_required'		=>	!WS_Form_Common::is_edition('pro'),
							'label_default'		=>	__('Credit Card Number', 'ws-form'),
							'submit_save'		=>	false,
							'submit_edit'		=>	false,
							'progress'			=>	false,
							'multiple'			=>	false,
							'icon'				=>	'action_' . $this->id,
							'credit_card'		=>	true,
							'conditional'		=>	array(

								'exclude_condition'	=>	true,
								'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field')
							),
							'events'			=>	array(

								'event'				=>	'keyup',
								'event_category'	=>	'Field'
							),
							'mask_field'					=>	'#pre_label<input type="tel" id="#id" name="#name" inputmode="tel" data-credit-card data-pt-encrypt data-post-lock data-required-bypass#attributes />#post_label#invalid_feedback#help<input type="hidden" id="#id_encrypted" name="paytrace_cc_num" />',
							'mask_field_attributes'			=>	array('class', 'disabled', 'required', 'placeholder', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
							'mask_field_label'				=>	'<label id="#label_id" for="#id"#attributes>#label</label>',
							'mask_field_label_attributes'	=>	array('class'),

							'fieldsets'	=>	array(

								// Tab: Basic
								'basic'	=>	array(

									'label'		=>	__('Basic', 'ws-form'),
									'meta_keys'	=>	array('label_render', 'label_position', 'label_column_width', 'required_on', 'hidden', 'placeholder'),

									'fieldsets'	=>	array(

										array(
											'label'		=>	__('Accessibility', 'ws-form'),
											'meta_keys'	=>	array('aria_label')
										)
									)
								),

								// Tab: Advanced
								'advanced'	=>	array(

									'label'	=>	__('Advanced', 'ws-form'),

									'fieldsets'		=>	array(

										array(
											'label'		=>	__('Styles', 'ws-form'),
											'meta_keys'	=>	array('class_single_vertical_align')
										),

										array(
											'label'		=>	__('Classes', 'ws-form'),
											'meta_keys'	=>	array('class_field_wrapper', 'class_field')
										),

										array(
											'label'		=>	__('Restrictions', 'ws-form'),
											'meta_keys'	=>	array('disabled')
										),

										array(
											'label'		=>	__('Validation', 'ws-form'),
											'meta_keys'	=>	array('invalid_feedback_render', 'invalid_feedback')
										),

										array(
											'label'		=>	__('Custom Attributes', 'ws-form'),
											'meta_keys'	=>	array('custom_attributes')
										),

										array(
											'label'		=>	__('Breakpoints', 'ws-form'),
											'meta_keys'	=> array('breakpoint_sizes'),
											'class'		=>	array('wsf-fieldset-panel')
										)
									)
								)
							)
						),

						'payt_cc_em' => array (

							'label'				=>	__('CC Exp Month', 'ws-form'),
							'pro_required'		=>	!WS_Form_Common::is_edition('pro'),
							'label_default'		=>	__('Expiry Month', 'ws-form'),
							'data_source'		=>	array('type' => 'data_grid', 'id' => 'action_' . $this->id . '_data_grid_select_cc_expiry_month'),
							'submit_save'		=>	false,
							'submit_edit'		=>	false,
							'submit_array'		=>	true,
							'progress'			=>	false,
							'multiple'			=>	false,
							'icon'				=>	'action_' . $this->id,
							'conditional'		=>	array(

								'exclude_condition'	=>	true,
								'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field')
							),
							'events'	=>	array(

								'event'						=>	'change',
								'event_category'			=>	'Field'
							),

							// Groups
							'mask_group'					=>	'<optgroup label="#group_label"#disabled>#group</optgroup>',
							'mask_group_label'				=>	'#group_label',

							// Rows
							'mask_row'						=>	'<option id="#row_id" data-id="#data_id" value="#action_' . $this->id . '_select_cc_expiry_month_field_value"#attributes>#action_' . $this->id . '_select_cc_expiry_month_field_label</option>',
							'mask_row_placeholder'			=>	'<option data-id="0" value="" data-placeholder>#value</option>',
							'mask_row_attributes'			=>	array('default', 'disabled'),
							'mask_row_lookups'				=>	array('action_' . $this->id . '_select_cc_expiry_month_field_value', 'action_' . $this->id . '_select_cc_expiry_month_field_label'),
							'datagrid_column_value'			=>	'action_' . $this->id . '_select_cc_expiry_month_field_value',
							'mask_row_default' 				=>	' selected',

							// Fields
							'mask_field'					=>	'#pre_label<select id="#id" name="#name" data-pt-no-encrypt data-post-lock data-required-bypass#attributes>#datalist/select>#post_label#invalid_feedback#help<input type="hidden" id="#id_not_encrypted" name="paytrace_cc_em" />',
							'mask_field_attributes'			=>	array('class', 'size', 'multiple', 'required', 'disabled', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
							'mask_field_label'				=>	'<label id="#label_id" for="#id"#attributes>#label</label>',
							'mask_field_label_attributes'	=>	array('class'),

							'fieldsets'	=> array(

								// Tab: Basic
								'basic'	=> array(

									'label'			=>	__('Basic', 'ws-form'),
									'meta_keys'		=> array('label_render', 'label_position', 'label_column_width', 'required_on', 'hidden', 'size', 'multiple', 'placeholder_row', 'help'),

									'fieldsets'		=>	array(

										array(
											'label'		=>	__('Accessibility', 'ws-form'),
											'meta_keys'	=>	array('aria_label')
										)
									)
								),

								// Tab: Advanced
								'advanced'	=>	array(

									'label'	=>	__('Advanced', 'ws-form'),

									'fieldsets'	=>	array(

										array(
											'label'		=>	__('Styles', 'ws-form'),
											'meta_keys'	=>	array('class_single_vertical_align')
										),

										array(
											'label'		=>	__('Classes', 'ws-form'),
											'meta_keys'	=> array('class_field_wrapper', 'class_field')
										),

										array(
											'label'		=>	__('Restrictions', 'ws-form'),
											'meta_keys'	=> array('disabled')
										),

										array(
											'label'		=>	__('Validation', 'ws-form'),
											'meta_keys'	=>	array('invalid_feedback_render', 'invalid_feedback')
										),

										array(
											'label'		=>	__('Custom Attributes', 'ws-form'),
											'meta_keys'	=>	array('custom_attributes')
										),

										array(
											'label'		=>	__('Breakpoints', 'ws-form'),
											'meta_keys'	=> array('breakpoint_sizes'),
											'class'		=>	array('wsf-fieldset-panel')
										)
									)
								),

								// Tab: Options
								'options'	=> array(

									'label'			=>	__('Options', 'ws-form'),
									'meta_keys'		=> array('action_' . $this->id . '_data_grid_select_cc_expiry_month', 'action_' . $this->id . '_select_cc_expiry_month_field_value', 'action_' . $this->id . '_select_cc_expiry_month_field_label')
								)
							)
						),

						'payt_cc_ey' => array (

							'label'				=>	__('CC Exp Year', 'ws-form'),
							'pro_required'		=>	!WS_Form_Common::is_edition('pro'),
							'label_default'		=>	__('Expiry Year', 'ws-form'),
							'data_source'		=>	array('type' => 'data_grid', 'id' => 'action_' . $this->id . '_data_grid_select_cc_expiry_year'),
							'submit_save'		=>	false,
							'submit_edit'		=>	false,
							'submit_array'		=>	true,
							'progress'			=>	false,
							'multiple'			=>	false,
							'icon'				=>	'action_' . $this->id,
							'conditional'		=>	array(

								'exclude_condition'	=>	true,
								'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field')
							),
							'events'	=>	array(

								'event'						=>	'change',
								'event_category'			=>	'Field'
							),

							// Groups
							'mask_group'					=>	'<optgroup label="#group_label"#disabled>#group</optgroup>',
							'mask_group_label'				=>	'#group_label',

							// Rows
							'mask_row'						=>	'<option id="#row_id" data-id="#data_id" value="#action_' . $this->id . '_select_cc_expiry_year_field_value"#attributes>#action_' . $this->id . '_select_cc_expiry_year_field_label</option>',
							'mask_row_placeholder'			=>	'<option data-id="0" value="" data-placeholder>#value</option>',
							'mask_row_attributes'			=>	array('default', 'disabled'),
							'mask_row_lookups'				=>	array('action_' . $this->id . '_select_cc_expiry_year_field_value', 'action_' . $this->id . '_select_cc_expiry_year_field_label'),
							'datagrid_column_value'			=>	'action_' . $this->id . '_select_cc_expiry_year_field_value',
							'mask_row_default' 				=>	' selected',

							// Fields
							'mask_field'					=>	'#pre_label<select id="#id" name="#name" data-pt-no-encrypt data-post-lock data-required-bypass#attributes>#datalist</select>#post_label#invalid_feedback#help<input type="hidden" id="#id_not_encrypted" name="paytrace_cc_ey" />',
							'mask_field_attributes'			=>	array('class', 'size', 'multiple', 'required', 'disabled', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
							'mask_field_label'				=>	'<label id="#label_id" for="#id"#attributes>#label</label>',
							'mask_field_label_attributes'	=>	array('class'),

							'fieldsets'	=> array(

								// Tab: Basic
								'basic'	=> array(

									'label'			=>	__('Basic', 'ws-form'),
									'meta_keys'		=> array('label_render', 'label_position', 'label_column_width', 'required_on', 'hidden', 'size', 'multiple', 'placeholder_row', 'help'),

									'fieldsets'		=>	array(

										array(
											'label'		=>	__('Accessibility', 'ws-form'),
											'meta_keys'	=>	array('aria_label')
										)
									)
								),

								// Tab: Advanced
								'advanced'	=>	array(

									'label'	=>	__('Advanced', 'ws-form'),

									'fieldsets'	=>	array(

										array(
											'label'		=>	__('Styles', 'ws-form'),
											'meta_keys'	=>	array('class_single_vertical_align')
										),

										array(
											'label'		=>	__('Classes', 'ws-form'),
											'meta_keys'	=> array('class_field_wrapper', 'class_field')
										),

										array(
											'label'		=>	__('Restrictions', 'ws-form'),
											'meta_keys'	=> array('disabled')
										),

										array(
											'label'		=>	__('Validation', 'ws-form'),
											'meta_keys'	=>	array('invalid_feedback_render', 'invalid_feedback')
										),

										array(
											'label'		=>	__('Custom Attributes', 'ws-form'),
											'meta_keys'	=>	array('custom_attributes')
										),

										array(
											'label'		=>	__('Breakpoints', 'ws-form'),
											'meta_keys'	=> array('breakpoint_sizes'),
											'class'		=>	array('wsf-fieldset-panel')
										)
									)
								),

								// Tab: Options
								'options'	=> array(

									'label'			=>	__('Options', 'ws-form'),
									'meta_keys'		=> array('action_' . $this->id . '_data_grid_select_cc_expiry_year', 'action_' . $this->id . '_select_cc_expiry_year_field_value', 'action_' . $this->id . '_select_cc_expiry_year_field_label')
								)
							)
						),

						'payt_cc_csc' => array (

							'label'				=>	__('CC CVV', 'ws-form'),
							'pro_required'		=>	!WS_Form_Common::is_edition('pro'),
							'label_default'		=>	__('CVV', 'ws-form'),
							'submit_save'		=>	false,
							'submit_edit'		=>	false,
							'progress'			=>	false,
							'multiple'			=>	false,
							'icon'				=>	'action_' . $this->id,
							'conditional'		=>	array(

								'exclude_condition'	=>	true,
								'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field')
							),
							'compatibility_id'	=>	'input-number',
							'events'			=>	array(

								'event'				=>	'input',
								'event_category'	=>	'Field'
							),

							// Fields
							'mask_field'					=>	'#pre_label<input type="tel" id="#id" name="#name" value="#value" data-pt-encrypt data-post-lock data-required-bypass#attributes />#post_label#invalid_feedback#help<input type="hidden" id="#id_encrypted" name="paytrace_cc_csc" />',
							'mask_field_attributes'			=>	array('class', 'min', 'max', 'step', 'disabled', 'required', 'placeholder', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
							'mask_field_label'				=>	'<label id="#label_id" for="#id"#attributes>#label</label>',
							'mask_field_label_attributes'	=>	array('class'),

							'fieldsets'	=> array(

								// Tab: Basic
								'basic'	=> array(

									'label'		=>	__('Basic', 'ws-form'),
									'meta_keys'	=>	array('label_render', 'label_position', 'label_column_width', 'required_on', 'hidden', 'placeholder', 'help'),

									'fieldsets'	=>	array(

										array(
											'label'		=>	__('Accessibility', 'ws-form'),
											'meta_keys'	=>	array('aria_label')
										)
									)
								),

								// Tab: Advanced
								'advanced'	=>	array(

									'label'			=>	__('Advanced', 'ws-form'),

									'fieldsets'		=>	array(

										array(
											'label'		=>	__('Styles', 'ws-form'),
											'meta_keys'	=>	array('class_single_vertical_align')
										),

										array(
											'label'		=>	__('Classes', 'ws-form'),
											'meta_keys'	=> array('class_field_wrapper', 'class_field')
										),

										array(
											'label'		=>	__('Restrictions', 'ws-form'),
											'meta_keys'	=> array('disabled', 'min', 'max', 'step')
										),

										array(
											'label'		=>	__('Validation', 'ws-form'),
											'meta_keys'	=>	array('invalid_feedback_render', 'invalid_feedback')
										),

										array(
											'label'		=>	__('Custom Attributes', 'ws-form'),
											'meta_keys'	=>	array('custom_attributes')
										),

										array(
											'label'		=>	__('Breakpoints', 'ws-form'),
											'meta_keys'	=> array('breakpoint_sizes'),
											'class'		=>	array('wsf-fieldset-panel')
										)
									)
								)
							)
						),

						'payt_checkout' => array(

							'label'						=>	__('PayTrace Submit', 'ws-form'),
							'pro_required'				=>	!WS_Form_Common::is_edition('pro'),
							'label_default'				=>	__('Submit Payment', 'ws-form'),
							'label_position_force'		=>	'top',
							'mask_field'				=>	'<button type="button" data-ecommerce-payment data-wsf-paytrace-submit id="#id" name="#name"#attributes>#label</button>#help<input type="hidden" id="#id_field_id" name="paytrace_field_id" value="#field_id" /><script>

(function($) {

	if(typeof(window.wsf_paytrace_functions) === \'undefined\') {
	
		window.wsf_paytrace_functions = [];
	}

	window.wsf_paytrace_functions[#form_instance_id] = function() {

		var wsf_paytrace = new $.WS_Form_PayTrace();
		wsf_paytrace.form_instance_id = \'#form_instance_id\';
		wsf_paytrace.transaction_amount = \'#action_' . $this->id . '_transaction_amount\';
		wsf_paytrace.transaction_zero = \'#action_' . $this->id . '_transaction_zero\';
		wsf_paytrace.transaction_zero_message = \'#action_' . $this->id . '_transaction_zero_message\';
		wsf_paytrace.transaction_limit = \'#action_' . $this->id . '_transaction_limit\';
		wsf_paytrace.transaction_limit_message = \'#action_' . $this->id . '_transaction_limit_message\';
		wsf_paytrace.message_method = \'#action_' . $this->id . '_message_method\';
		wsf_paytrace.message_clear = \'#action_' . $this->id . '_message_clear\';
		wsf_paytrace.message_scroll_top = \'#action_' . $this->id . '_message_scroll_top\';
		wsf_paytrace.message_form_hide = \'#action_' . $this->id . '_message_form_hide\';
		wsf_paytrace.message_duration = \'#action_' . $this->id . '_message_duration\';
		wsf_paytrace.pem = \'' . WS_Form_Common::get_api_path('action/' . $this->id . '/pem/') . '\';

		wsf_paytrace.init();
	};

	$(function() {

		window.wsf_paytrace_functions[#form_instance_id]();
	});

})(jQuery);

					</script>',
							'mask_field_attributes'		=>	array('class', 'disabled', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
							'mask_field_label'			=>	'#label',
							'submit_save'				=>	false,
							'submit_edit'				=>	false,
							'progress'					=>	false,
							'icon'						=>	'action_' . $this->id,
							'conditional'				=>	array(

								'logics_enabled'	=>	array('click', 'mouseover', 'mouseout', 'focus', 'blur'),
								'actions_enabled'	=>	array('visibility', 'focus', 'button_html', 'click', 'disabled', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field'),
								'condition_event'	=>	'click',
							),
							'events'					=>	array(

								'event'				=>	'click',
								'event_category'	=>	'Button'
							),
		
							'meta_key_parse_variables'	=>	array(

								'action_' . $this->id . '_transaction_amount',
								'action_' . $this->id . '_transaction_limit',
								'action_' . $this->id . '_transaction_limit_message',
								'action_' . $this->id . '_transaction_zero',
								'action_' . $this->id . '_transaction_zero_message',
								'action_' . $this->id . '_message_method',
								'action_' . $this->id . '_message_clear',
								'action_' . $this->id . '_message_scroll_top',
								'action_' . $this->id . '_message_form_hide',
								'action_' . $this->id . '_message_duration'
							),

							'fieldsets'				=> array(

								// Tab: Basic
								'basic'	=> array(

									'label'			=>	__('Basic', 'ws-form'),
									'meta_keys'		=>	array('hidden', 'help'),

									'fieldsets'		=>	array(

										array(

											'label'		=>	__('Transaction', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_transaction_amount',
												'action_' . $this->id . '_transaction_description'
											)
										),

										array(

											'label'		=>	__('Invoice', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_invoice_id'
											)
										),

										array(

											'label'		=>	__('Customer', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_email',
												'action_' . $this->id . '_customer_reference_id'
											)
										),

										array(

											'label'		=>	__('Billing Address', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_billing_name',
												'action_' . $this->id . '_billing_address_mapping'
											)
										),

										array(

											'label'		=>	__('Shipping Address', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_shipping_name',
												'action_' . $this->id . '_shipping_address_mapping'
											)
										),

										array(
											'label'		=>	__('Accessibility', 'ws-form'),
											'meta_keys'	=>	array('aria_label')
										),
									)
								),

								// Tab: Advanced
								'advanced'	=>	array(

									'label'			=>	__('Advanced', 'ws-form'),

									'fieldsets'	=>	array(

										array(
											'label'		=>	__('Styles', 'ws-form'),
											'meta_keys'	=>	array('class_single_vertical_align_bottom', 'class_field_button_type_primary', 'class_field_full_button_remove')
										),

										array(
											'label'		=>	__('Classes', 'ws-form'),
											'meta_keys'	=>	array('class_field_wrapper', 'class_field')
										),

										array(

											'label'		=>	__('Transaction', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_transaction_zero',
												'action_' . $this->id . '_transaction_zero_message',
												'action_' . $this->id . '_transaction_limit',
												'action_' . $this->id . '_transaction_limit_message'
											)
										),

										array(

											'label'		=>	__('Messages', 'ws-form'),
											'meta_keys'	=> array(

												'action_' . $this->id . '_message_method',
												'action_' . $this->id . '_message_clear',
												'action_' . $this->id . '_message_scroll_top',
												'action_' . $this->id . '_message_form_hide',
												'action_' . $this->id . '_message_duration'
											)
										),

										array(
											'label'		=>	__('Restrictions', 'ws-form'),
											'meta_keys'	=>	array('disabled')
										),

										array(
											'label'		=>	__('Custom Attributes', 'ws-form'),
											'meta_keys'	=>	array('custom_attributes')
										),

										array(
											'label'		=>	__('Breakpoints', 'ws-form'),
											'meta_keys'	=> array('breakpoint_sizes'),
											'class'		=>	array('wsf-fieldset-panel')
										)
									)
								)
							)
						)
					)
				);

				return $field_types;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = !empty($this->api_access_token) ? 'Connected' : 'Disconnected';
						break;

					case 'action_' . $this->id . '_api_public_key' :

						$value = '<div class="wsf-settings-code">' . str_replace("\n", '<br />', $this->api_public_key) . '</div>';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-paytrace-v1') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-paytrace-v1') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-paytrace-v1') . '" />';

						} else {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_get" value="' . __('Connect', 'ws-form-paytrace-v1') . '" />';
						}

						break;

					case 'pem_action_' . $this->id :

						if($this->api_public_key) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="pem_delete" value="' . __('Delete', 'ws-form-paytrace-v1') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="pem_upload" value="' . __('Upload', 'ws-form-paytrace-v1') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_get' :

								self::access_token_get();
								break;

							case 'access_token_delete' :

								WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
								break;

							case 'pem_upload' :

								self::pem_upload();
								break;

							case 'pem_delete' :

								WS_Form_Common::option_set('action_' . $this->id . '_api_public_key', '');
								break;
						}

						break;
				}
			}

			// Upload public_key.pem file
			public function pem_upload() {

				// Get file
				if(!isset($_FILES)) { return false; }
				if(!isset($_FILES['action_' . $this->id . '_api_public_key'])) { return false; }
				$file = $_FILES['action_' . $this->id . '_api_public_key'];

				// File checks
				if(
					(!isset($file['type']) || ($file['type'] != 'application/x-x509-ca-cert')) ||
					(!isset($file['error']) || ($file['error'] != 0)) ||
					(!isset($file['size']) || ($file['size'] == 0)) ||
					(!isset($file['tmp_name']) || (!file_exists($file['tmp_name'])))
				) {

					WS_Form_Common::admin_message_push(__('Invalid certificate file.', 'ws-form-paytrace-v1'), 'notice-error', false);
					return false;
				}

				// Read file
				$pem = file_get_contents($file['tmp_name']);
				if(substr($pem,0,3) == (chr(hexdec('EF')) . chr(hexdec('BB')) . chr(hexdec('BF')))) { $pem = substr($pem,3); }

				// Check file
				$rid = openssl_pkey_get_public($pem);
				if($rid === false) {

					WS_Form_Common::admin_message_push(__('Invalid certificate file.', 'ws-form-paytrace-v1'), 'notice-error', false);
					return false;
				}

				// Set value
				WS_Form_Common::option_set('action_' . $this->id . '_api_public_key', $pem);

				return true;
			}

			// Get access token
			public function access_token_get() {

				// Get username
				$this->username = WS_Form_Common::get_query_var('action_' . $this->id . '_username');
				if(empty($this->username)) {

					WS_Form_Common::admin_message_push(__('Please enter a valid PayTrace username.', 'ws-form-paytrace-v1'), 'notice-error', false);
					return false;
				}

				// Get password
				$this->password = WS_Form_Common::get_query_var('action_' . $this->id . '_password');
				if(empty($this->password)) {

					WS_Form_Common::admin_message_push(__('Please enter a valid PayTrace password.', 'ws-form-paytrace-v1'), 'notice-error', false);
					return false;
				}

				// Get access token
				if(!self::get_access_token()) {

					WS_Form_Common::admin_message_push(__('Username or password is incorrect.', 'ws-form-paytrace-v1'), 'notice-error', false);
					return false;
				}

				return true;
			}

			// Get auth token
			public function get_access_token() {

				// Build JSON packet
				$params = array(

					'username' => $this->username,
					'password' => $this->password,
					'grant_type' => 'password'
				);
				$post_fields = http_build_query($params);

				// Get contact
				$api_response = parent::api_call('https://api.paytrace.com/', 'oauth/token', 'POST', $post_fields, array(), false, false, false, '*/*', 'application/x-www-form-urlencoded');

				// Process response
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }

				// Check status code
				if($api_response['http_code'] != 200) { return false; }

				// Process response
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->access_token)) { return false; }
				if(!isset($api_response_decoded->expires_in)) { return false; }

				// Set access token
				$this->api_access_token = $api_response_decoded->access_token;
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				$this->api_access_token_expiry = time() + intval($api_response_decoded->expires_in);
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				return true;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				if(!self::get_access_token()) {

					self::error(__('Auth token error', 'ws-form-paytrace-v1'));

					self::reset_api();
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_username', '');
				WS_Form_Common::option_set('action_' . $this->id . '_password', '');
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { self::error(sprintf(__('Action not configured (%s)', 'ws-form-paytrace-v1'), $this->label)); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Transaction - Amount
					'action_' . $this->id . '_transaction_amount'	=> array(

						'label'						=>	__('Amount To Charge', 'ws-form-paytrace-v1'),
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('Enter the amount to charge the user. Leave blank to use cart total.', 'ws-form-paytrace-v1'),
						'select_list'				=>	true,
						'placeholder'				=>	'#ecommerce_cart_total'
					),

					// Transaction - Limit
					'action_' . $this->id . '_transaction_limit'	=> array(

						'label'						=>	__('Limit Amount', 'ws-form-paytrace-v1'),
						'type'						=>	'number',
						'step'						=>	'0.01',
						'default'					=>	'',
						'help'						=>	__('Maximum amount of transaction. Leave blank for no limit.', 'ws-form-paytrace-v1'),
					),

					// Transaction - Limit - Message
					'action_' . $this->id . '_transaction_limit_message'	=> array(

						'label'						=>	__('Limit Amount Message', 'ws-form-paytrace-v1'),
						'type'						=>	'textarea',
						'help'						=>	__('Enter a message to show if the PayTrace button is clicked and the cart amount exceeds the max amount.', 'ws-form-paytrace-v1'),
						'default'					=>	__('The transaction exceeds our specified limit. Please review the form and try again.', 'ws-form-paytrace-v1')
					),

					// Transaction - Description
					'action_' . $this->id . '_transaction_description'	=> array(

						'label'						=>	__('Description', 'ws-form-paytrace-v1'),
						'type'						=>	'textarea',
						'help'						=>	__('Optional 255-character max text describing the transaction, products, customers, or other attributes of the transaction..', 'ws-form-paytrace-v1'),
						'default'					=>	'Payment to #blog_name',
						'maxlength'					=>	'255'
					),

					// Invoice ID
					'action_' . $this->id . '_invoice_id'	=> array(

						'label'						=>	__('Invoice ID', 'ws-form-paytrace-v1'),
						'type'						=>	'text',
						'help'						=>	__('A unique identifier for this transaction in your accounting or inventory management system, e.g. #field(123).', 'ws-form-paytrace-v1'),
						'select_list'				=>	true,
						'default'					=>	''
					),

					// Email
					'action_' . $this->id . '_email'	=> array(

						'label'						=>	__('Email', 'ws-form-paytrace-v1'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form-paytrace-v1'),
						'fields_filter_type'		=>	array('email'),
						'help'						=>	__('Select the email field.', 'ws-form-paytrace-v1'),
						'default'					=>	''
					),

					// Customer reference ID
					'action_' . $this->id . '_customer_reference_id'	=> array(

						'label'						=>	__('P.O. Number', 'ws-form-paytrace-v1'),
						'type'						=>	'text',
						'help'						=>	__('The customer reference ID is an identifier that your customer may ask you to provide in order to reference the transaction to their credit card statement. Also known as a PO Number.', 'ws-form-paytrace-v1'),
						'select_list'				=>	true,
						'default'					=>	''
					),

					// Billing - Name
					'action_' . $this->id . '_billing_name'	=> array(

						'label'						=>	__('Billing Name', 'ws-form-paytrace-v1'),
						'type'						=>	'text',
						'help'						=>	__('Name that appears on the credit card, e.g. #field(123).', 'ws-form-paytrace-v1'),
						'select_list'				=>	true,
						'default'					=>	''
					),

					// Billing - Address mapping
					'action_' . $this->id . '_billing_address_mapping'	=> array(

						'label'						=>	__('Billing Address Mapping', 'ws-form-paytrace-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to PayTrace mailing address fields.', 'ws-form-paytrace-v1'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_billing_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_billing_address_field'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							)
						)
					),

					// Billing - Address mapping - Fields
					'action_' . $this->id . '_billing_address_field' => array(

						'label'						=>	__('PayTrace Field', 'ws-form-paytrace-v1'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'street_address', 'text' => __('Address Line 1', 'ws-form-paytrace-v1')),
							array('value' => 'street_address2', 'text' => __('Address Line 2', 'ws-form-paytrace-v1')),
							array('value' => 'city', 'text' => __('City', 'ws-form-paytrace-v1')),
							array('value' => 'state', 'text' => __('State', 'ws-form-paytrace-v1')),
							array('value' => 'zip', 'text' => __('Zip', 'ws-form-paytrace-v1')),
							array('value' => 'country', 'text' => __('Country', 'ws-form-paytrace-v1')),
						)
					),

					// Shipping - Name
					'action_' . $this->id . '_shipping_name'	=> array(

						'label'						=>	__('Shipping Name', 'ws-form-paytrace-v1'),
						'type'						=>	'text',
						'help'						=>	__('Name of the person where the product is to be delivered, e.g. #field(123).', 'ws-form-paytrace-v1'),
						'select_list'				=>	true,
						'default'					=>	''
					),


					// Shipping - Address mapping
					'action_' . $this->id . '_shipping_address_mapping'	=> array(

						'label'						=>	__('Shipping Address Mapping', 'ws-form-paytrace-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to PayTrace shipping address fields.', 'ws-form-paytrace-v1'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_shipping_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_shipping_address_field'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							)
						)
					),

					// Shipping - Address mapping - Fields
					'action_' . $this->id . '_shipping_address_field' => array(

						'label'						=>	__('PayTrace Field', 'ws-form-paytrace-v1'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'street_address', 'text' => __('Address Line 1', 'ws-form-paytrace-v1')),
							array('value' => 'street_address2', 'text' => __('Address Line 2', 'ws-form-paytrace-v1')),
							array('value' => 'city', 'text' => __('City', 'ws-form-paytrace-v1')),
							array('value' => 'state', 'text' => __('State', 'ws-form-paytrace-v1')),
							array('value' => 'zip', 'text' => __('Zip', 'ws-form-paytrace-v1')),
							array('value' => 'country', 'text' => __('Country', 'ws-form-paytrace-v1')),
						)
					),

					// Transaction - Zero Amount
					'action_' . $this->id . '_transaction_zero'	=> array(

						'label'						=>	__('Zero Amount', 'ws-form-paytrace-v1'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-paytrace-v1')),
							array('value' => 'submit', 'text' => __('Submit form', 'ws-form-paytrace-v1')),
							array('value' => 'message', 'text' => __('Show error message', 'ws-form-paytrace-v1')),
						),
						'default'					=>	'message',
						'help'						=>	__('What should happen if the cart total is zero and the PayTrace button is clicked?', 'ws-form-paytrace-v1'),
					),

					// Transaction - Zero Amount - Message
					'action_' . $this->id . '_transaction_zero_message'	=> array(

						'label'						=>	__('Zero Amount Message', 'ws-form-paytrace-v1'),
						'type'						=>	'textarea',
						'help'						=>	__('Enter a message to show if the PayTrace button is clicked and the cart amount is zero.', 'ws-form-paytrace-v1'),
						'default'					=>	__('The total cart amount cannot be zero. Please review the form and try again.', 'ws-form-paytrace-v1'),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_zero',
								'meta_value'		=>	'message'
							)
						)
					),

					// Message - Method
					'action_' . $this->id . '_message_method'	=> array(

						'label'						=>	__('Error Messages', 'ws-form-post'),
						'type'						=>	'select',
						'help'						=>	__('Where should error messages be added?', 'ws-form-post'),
						'options'					=>	array(

							array('value' => 'before', 'text' => __('Before Form', 'ws-form-post')),
							array('value' => 'after', 'text' => __('After Form', 'ws-form-post'))
						),
						'default'					=>	'after',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Duration
					'action_' . $this->id . '_message_duration'	=> array(

						'label'						=>	__('Error Message Duration (ms)', 'ws-form-post'),
						'type'						=>	'number',
						'help'						=>	__('Duration in milliseconds error message shown.', 'ws-form-post'),
						'default'					=>	'4000',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Clear other messages
					'action_' . $this->id . '_message_clear'	=> array(

						'label'						=>	__('Clear Other Messages On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Clear any other messages when error message shown?', 'ws-form'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Scroll to top
					'action_' . $this->id . '_message_scroll_top'	=> array(

						'label'						=>	__('Scroll To Top On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Scroll to top of page when error message shown?', 'ws-form'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Form Hide
					'action_' . $this->id . '_message_form_hide' => array(

						'label'						=>	__('Hide Form On Error', 'ws-form-post'),
						'type'						=>	'checkbox',
						'help'						=>	__('Hide form when error message shown?', 'ws-form-post'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					'action_' . $this->id . '_data_grid_select_cc_expiry_month'	=>	array(

						'label'					=>	__('Months', 'ws-form'),
						'type'					=>	'data_grid',
						'row_default'			=>	true,	// Is the default attribute supported on rows?
						'row_disabled'			=>	true,	// Is the disabled attribute supported on rows?
						'row_required'			=>	false,	// Is the required attribute supported on rows?
						'row_hidden'			=>	true,	// Is the hidden supported on rows?
						'groups_label'			=>	true,	// Is the group label feature enabled?
						'groups_label_label'	=>	__('Label', 'ws-form'),
						'groups_label_render'	=>	false,	// Is the group label render feature enabled?
						'groups_label_render_label'	=>	__('Render Label', 'ws-form'),
						'groups_auto_group'		=>	true,	// Is auto group feature enabled?
						'groups_disabled'		=>	true,	// Is the group disabled attribute?
						'groups_group'			=>	true,	// Is the group mask supported?
						'groups_group_label'	=>	__('Wrap In Optgroup', 'ws-form'),
						'upload_download'		=>	true,

						'default'			=>	array(

							// Config
							'rows_per_page'		=>	25,
							'group_index'		=>	0,
							'default'			=>	array(),

							// Columns
							'columns' => array(

								array('id' => 0, 'label' => __('Month ID', 'ws-form')),
								array('id' => 1, 'label' => __('Month Name', 'ws-form')),
							),

							// Group
							'groups' => array(

								array(

									'label' 		=> __('Months', 'ws-form'),
									'page'			=> 0,
									'disabled'		=> '',
									'mask_group'	=> '',

									// Rows (Only injected for a new data grid, blank for new groups)
									'rows' 		=> array()
								)
							)
						)
					),

					'action_' . $this->id . '_select_cc_expiry_month_field_value'	=> array(

						'label'						=>	__('Month ID', 'ws-form'),
						'type'						=>	'data_grid_field',
						'data_grid'					=>	'action_' . $this->id . '_data_grid_select_cc_expiry_month',
						'default'					=>	0,
						'html_encode'				=>	true
					),

					'action_' . $this->id . '_select_cc_expiry_month_field_label'		=> array(

						'label'						=>	__('Month Name', 'ws-form'),
						'type'						=>	'data_grid_field',
						'data_grid'					=>	'action_' . $this->id . '_data_grid_select_cc_expiry_month',
						'default'					=>	1
					),

					'action_' . $this->id . '_data_grid_select_cc_expiry_year'	=>	array(

						'label'					=>	__('Years', 'ws-form'),
						'type'					=>	'data_grid',
						'row_default'			=>	true,	// Is the default attribute supported on rows?
						'row_disabled'			=>	true,	// Is the disabled attribute supported on rows?
						'row_required'			=>	false,	// Is the required attribute supported on rows?
						'row_hidden'			=>	true,	// Is the hidden supported on rows?
						'groups_label'			=>	true,	// Is the group label feature enabled?
						'groups_label_label'	=>	__('Label', 'ws-form'),
						'groups_label_render'	=>	false,	// Is the group label render feature enabled?
						'groups_label_render_label'	=>	__('Render Label', 'ws-form'),
						'groups_auto_group'		=>	true,	// Is auto group feature enabled?
						'groups_disabled'		=>	true,	// Is the group disabled attribute?
						'groups_group'			=>	true,	// Is the group mask supported?
						'groups_group_label'	=>	__('Wrap In Optgroup', 'ws-form'),
						'upload_download'		=>	true,

						'default'			=>	array(

							// Config
							'rows_per_page'		=>	25,
							'group_index'		=>	0,
							'default'			=>	array(),

							// Columns
							'columns' => array(

								array('id' => 0, 'label' => __('Year ID', 'ws-form')),
								array('id' => 1, 'label' => __('Year Name', 'ws-form')),
							),

							// Group
							'groups' => array(

								array(

									'label' 		=> __('Years', 'ws-form'),
									'page'			=> 0,
									'disabled'		=> '',
									'mask_group'	=> '',

									// Rows (Only injected for a new data grid, blank for new groups)
									'rows' 		=> array()
								)
							)
						)
					),

					'action_' . $this->id . '_select_cc_expiry_year_field_value'	=> array(

						'label'						=>	__('Year ID', 'ws-form'),
						'type'						=>	'data_grid_field',
						'data_grid'					=>	'action_' . $this->id . '_data_grid_select_cc_expiry_year',
						'default'					=>	0,
						'html_encode'				=>	true
					),

					'action_' . $this->id . '_select_cc_expiry_year_field_label'		=> array(

						'label'						=>	__('Year Name', 'ws-form'),
						'type'						=>	'data_grid_field',
						'data_grid'					=>	'action_' . $this->id . '_data_grid_select_cc_expiry_year',
						'default'					=>	1
					)
				);

				// Months
				for($month_index = 1; $month_index <= 12; $month_index++) {

					$month = strval($month_index);

					$config_meta_keys['action_' . $this->id . '_data_grid_select_cc_expiry_month']['default']['groups'][0]['rows'][] = array(

						'id' => $month_index,
						'default' => '',
						'required' => '',
						'disabled' => '',
						'hidden' => '',
						'data' => array($month, (($month_index < 10) ? '0' : '') . $month)
					);
				}

				// Years
				for($year_index = 0; $year_index < 20; $year_index++) {

					$year = strval(intval(date('Y')) + $year_index);

					$config_meta_keys['action_' . $this->id . '_data_grid_select_cc_expiry_year']['default']['groups'][0]['rows'][] = array(

						'id' => ($year_index + 1),
						'default' => '',
						'required' => '',
						'disabled' => '',
						'hidden' => '',
						'data' => array($year, substr($year, -2))
					);
				}

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-paytrace-v1'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-paytrace-v1'),
							'type'		=>	'text',
							'help'		=>	__('Enter your PayTrace add-on for WS Form PRO license key here.', 'ws-form-paytrace-v1'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-paytrace-v1'),
							'type'		=>	'static'
						)
					)
				);

				// Add username and password if access token not retrieved
				if(empty($this->api_access_token)) {

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_username'] = array(

						'label'			=>	__('Username', 'ws-form-paytrace-v1'),
						'type'			=>	'text',
						'help'			=>	__('Enter your PayTrace username.', 'ws-form-paytrace-v1')
					);

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_password'] = array(

						'label'			=>	__('Password', 'ws-form-paytrace-v1'),
						'type'			=>	'password',
						'help'			=>	__('Enter your PayTrace password.', 'ws-form-paytrace-v1')
					);
				}

				// Add status
				$options['action_' . $this->id]['fields']['action_' . $this->id . '_status'] = array(

					'label'			=>	__('PayTrace Status', 'ws-form-paytrace-v1'),
					'type'			=>	'static',
					'help'			=>	empty($this->api_access_token) ? __("Enter your username and password, then click the 'Connect' button to connect WS Form to your PayTrace account, or <a href=\"https://www.paytrace.net/\" target=\"_blank\">create new account</a>.", 'ws-form-paytrace-v1') : '',
					'button'		=>	'oauth_action_' . $this->id,
				);

				// Add public key
				if(empty($this->api_public_key)) {

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_public_key'] = array(

						'label'			=>	__('Public Key File', 'ws-form-paytrace-v1'),
						'type'			=>	'file',
						'help'			=>	__('Upload your PayTrace public_key.pem file. You can obtain this by clicking on \'Integration\' &gt; \'Download Public Key\' from your PayTrace dashboard.', 'ws-form-paytrace-v1'),
						'button'		=>	'pem_action_' . $this->id,
					);

				} else {

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_public_key'] = array(

						'label'			=>	__('Public Key File', 'ws-form-paytrace-v1'),
						'type'			=>	'static',
						'button'		=>	'pem_action_' . $this->id,
					);
				}

				return $options;
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', 0);
				$this->api_public_key = WS_Form_Common::option_get('action_' . $this->id . '_api_public_key', '');
				$this->username = WS_Form_Common::option_get('action_' . $this->id . '_username', '');
				$this->password = WS_Form_Common::option_get('action_' . $this->id . '_password', '');

				// Build endpoint using API key
				if(
					!empty($this->api_access_token) &&
					!empty($this->api_public_key) &&
					!empty($this->username) &&
					!empty($this->password)
				) {

					// Set endpoint URL
					$this->api_endpoint = 'https://api.paytrace.com/v1/';

				} else {

					$this->api_endpoint = false;
					$this->api_public_key = false;
					$this->configured = false;
				}

				return $this->configured;
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/pem/', array('methods' => 'GET', 'callback' => array($this, 'api_pem'), 'permission_callback' => function () { return '__return_true'; }));
			}

			// Output PEM file
			public function api_pem() {

				self::load_config_plugin();

				if(!$this->configured) { exit; }

				// Output HTTP header
				header("Content-type: application/x-x509-ca-cert; charset: UTF-8");
				header('Content-Disposition: filename="public_key.pem"');

				// Output key
				echo $this->api_public_key;

				exit;
			}
		}
	});