<?php

/**
 * @link              https://wsform.com/knowledgebase/pdf/
 * @since             1.0.1
 * @package           WS_Form_PDF
 *
 * @wordpress-plugin
 * Plugin Name:       WS Form PRO - PDF
 * Plugin URI:        https://wsform.com/knowledgebase/pdf/
 * Description:       PDF add-on for WS Form PRO
 * Version:           1.0.10
 * Author:            Westguard Solutions
 * Author URI:        https://www.westguardsolutions.com/
 * Text Domain:       ws-form-pdf
 */
	use Dompdf\Dompdf;
	use Dompdf\Options;

	Class WS_Form_Add_On_PDF {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.145';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_PDF();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-pdf'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_pdf = new WS_Form_Add_On_PDF();	

	register_activation_hook(__FILE__, array($wsf_add_on_pdf, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_PDF extends WS_Form_Action {

			public $id = 'pdf';
			public $label;

			public $paper_size = 'Letter';
			public $paper_orientation = 'portrait';

			public $pdf_attach;
			public $pdf_filename;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 2855;
			const WS_FORM_LICENSE_NAME = 'PDF add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.10';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('PDF', 'ws-form-pdf');

				// Filters
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_table_submit_column_actions', array($this, 'table_submit_column_actions'), 10, 3);
				add_filter('wsf_action_email_settings', array($this, 'action_email_settings'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('wsf_action_email_email_attachments', array($this, 'action_email_email_attachments'), 10, 5);
				add_filter('wsf_config_field_types', array($this, 'config_field_types'), 10, 2);

				// Actions
				add_action('wsf_table_submit_action', array($this, 'table_submit_action'), 10, 2);

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Action - Email - Settings
			public function action_email_settings($settings) {

				$settings->fieldsets['email']['meta_keys'][] = 'action_email_pdf_attach';
				$settings->fieldsets['email']['meta_keys'][] = 'action_email_pdf_filename';

				return $settings;
			}

			// Action - Email - Email attachments
			public function action_email_email_attachments($email_attachments, $form, $submit, $config, $temp_path) {

				// Load config
				self::load_config($config);

				// Check for PDF attachment
				if($this->pdf_attach) {

					// Convert to object
					$form_object = json_decode(json_encode($form));

					// Get PDF data
					$pdf_data = self::pdf_get_data($form_object, $submit);

					// Build PDF filename
					if($this->pdf_filename == '') {

						// Use default
						$pdf_filename = self::pdf_get_filename();

					} else {

						// Use filename specified by user
						$pdf_filename = WS_Form_Common::parse_variables_process($this->pdf_filename, $form, $submit, 'text/plain');

						// Sanitize
						$pdf_filename = sanitize_file_name($pdf_filename);

						// Check extension is pdf
						$pdf_filename_extension = strtolower(pathinfo($pdf_filename, PATHINFO_EXTENSION));
						if($pdf_filename_extension !== 'pdf') { $pdf_filename .= '.pdf'; }
					}

					// Create temporary directory for storing attachment files
					if(!file_exists($temp_path)) { mkdir($temp_path); }

					// Get destination filename
					$filename_destination = $temp_path . '/' . $pdf_filename;

					// Write PDF data to destination
					file_put_contents($filename_destination, $pdf_data);

					// Add to attachments
					$email_attachments[] = array(

						'path' => $filename_destination,
						'unlink_after_use' => true
					);
				}

				return $email_attachments;
			}

			// Load config
			public function load_config($config) {

				// Get configuration
				$this->pdf_attach = parent::get_config($config, 'action_email_pdf_attach');
				$this->pdf_filename = parent::get_config($config, 'action_email_pdf_filename');
			}

			// Config - Meta keys
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Attach PDF
				$meta_keys['action_email_pdf_attach']	= array(

					'label'						=>	__('Attach PDF', 'ws-form'),
					'type'						=>	'checkbox',
					'default'					=>	'',
					'help'						=>	__('Attach a PDF file of this submission to the email?', 'ws-form-pdf')
				);

				// PDF filename
				$meta_keys['action_email_pdf_filename']	= array(

					'label'						=>	__('Custom PDF Filename', 'ws-form'),
					'type'						=>	'text',
					'default'					=>	'',
					'help'						=>	__('Enter a custom filename for the PDF attachment. Leave blank to use default.', 'ws-form-pdf'),
					'placeholder'				=>	self::pdf_get_filename(),
					'condition'					=>	array(

						array(

							'logic'			=>	'==',
							'meta_key'		=>	'action_email_pdf_attach',
							'meta_value'	=>	'text/html'
						)
					)
				);

				return $meta_keys;
			}

			// Submissions page - Table actions
			public function table_submit_column_actions($actions, $item, $status) {

				// Get submissions ID
				$id = absint($item['id']);

				// Add PDF
				$actions['pdf'] = sprintf('<a href="#" data-action="wsf-pdf" data-id="%u">%s</a>', $id, __('Export PDF', 'ws-form'));

				return $actions;
			}

			// Submissions page - Action
			public function table_submit_action($action, $submit_id) {

				switch($action) {

					case 'wsf-pdf' :	self::pdf_export($submit_id); break;
				}
			}

			// PDF - Export
			public function pdf_export($submit_id) {

				// Build filename
				$filename = self::pdf_get_filename();

				// HTTP headers
				WS_Form_Common::file_download_headers($filename, 'application/pdf');

				// Start CSV output
				$pdf_out = fopen('php://output', 'w');

				// Get submit
				$ws_form_submit = New WS_Form_Submit();
				$ws_form_submit->id = $submit_id;
				$ws_form_submit->db_read();

				// Get published form data
				$ws_form_form = New WS_Form_Form();
				$ws_form_form->id = $ws_form_submit->form_id;
				$form_object = $ws_form_form->db_read_published();

				// If no published data...
				if(!$form_object) {

					// Read draft form data
					$form_object = $ws_form_form->db_read(true, true);
				}

				// Get data
				$pdf_data = self::pdf_get_data($form_object, $ws_form_submit);

				// Write data
				fwrite($pdf_out, $pdf_data);

				// Close
				fclose($pdf_out);

				exit;
			}

			// PDF - Get filename
			public function pdf_get_filename() {

				return WS_Form_Common::filename_datestamp('ws-form-submit', 'pdf');
			}

			// PDF - Get data
			public function pdf_get_data($form, $ws_form_submit) {

				$plugin_dir = dirname(__FILE__) . '/';

				// Get form ID
				$form_id = $form->id;

				// Load DOMPDF
				require_once $plugin_dir . 'lib/dompdf/autoload.inc.php';

				// Build file upload path
				$file_upload_path = 'pdf/templates';

				// Apply filter
				$file_upload_path = apply_filters('wsf_file_upload_path_pdf', $file_upload_path);

				$upload_dir = WS_Form_Common::upload_dir_create($file_upload_path);
				if($upload_dir['error']) { self::api_throw_error($upload_dir['error']); }
				$file_upload_dir = $upload_dir['dir'];

				// Move uploaded file to WordPress uploads folder
				$file_name = $file_upload_dir . '/' . $form_id . '.html';

				// Get PDF template
				$pdf_template = file_get_contents(file_exists($file_name) ? $file_name : $plugin_dir . 'templates/standard.html');

				// Get PDF HTML
				$pdf_html = WS_Form_Common::parse_variables_process($pdf_template, $form, $ws_form_submit, 'text/html');

				// Initialize DOMPDF object
				$dompdf = new Dompdf();

				// Load from HTML
				$dompdf->loadHtml($pdf_html);

				// Setup the paper size and orientation
				$dompdf->setPaper($this->paper_size, $this->paper_orientation);

				// Options
				$options = new Options();
				$options->setIsRemoteEnabled(true);
				$dompdf->setOptions($options);

				// Render
				$dompdf->render();

				// Get output
				return $dompdf->output();
			}

			// Plug-in field types for this action
			public function config_field_types($field_types) {

				$field_types[$this->id] = array(

					'label' =>	__('PDF', 'ws-form-pdf'),
					'types' =>	array(

						'pdf_page_break' => array (

							'label'					=>	__('Page Break', 'ws-form'),
							'pro_required'			=>	!WS_Form_Common::is_edition('pro'),
							'label_default'			=>	__('PDF Page Break', 'ws-form'),
							'mask_field'			=>	'<div style="page-break-after: always;"></div>',
							'label_disabled'		=>	true,
							'submit_save'			=>	false,
							'submit_edit'			=>	false,
							'progress'				=>	false,
							'conditional'			=>	array(

								'exclude_condition'		=>	true,
								'exclude_then'			=>	true,
								'exclude_else'			=>	true
							),
							'icon'					=>	'<svg width="14" height="18" viewBox="0 0 14 18"><path fill="#444" d="M13,12 L1,12 L1,18 L13,18 L13,12 Z M12,13 L12,18 L2,18 L2,13 L12,13 Z" id="Shape"></path><path fill="#444" d="M0 8.5 0 9.5 2 9.5 2 8.5z"></path><path fill="#444" d="M3 8.5 3 9.5 5 9.5 5 8.5z"></path><path fill="#444" d="M6 8.5 6 9.5 8 9.5 8 8.5z"></path><path fill="#444" d="M9 8.5 9 9.5 11 9.5 11 8.5z"></path><path fill="#444" d="M12 8.5 12 9.5 14 9.5 14 8.5z"></path><path fill="#444" d="M1,0 L1,6 L13,6 L13,0 L1,0 Z M12,0 L12,5 L2,5 L2,0 L12,0 Z" id="Shape"></path></svg>',
							'static'				=>	true,
							'mask_wrappers_drop'	=>	true,

							'fieldsets'	=> array(

								// Tab: Basic
								'basic'	=> array(

									'label'			=>	__('Basic', 'ws-form')
								)
							)
						)
					)
				);

				return $field_types;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-pdf')));

				return $links;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-pdf'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-pdf'),
							'type'		=>	'text',
							'help'		=>	__('Enter your PDF add-on for WS Form PRO license key here.', 'ws-form-pdf'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-pdf'),
							'type'		=>	'static'
						)
					)
				);

				return $options;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-pdf') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-pdf') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}
		}
	});