<?php

	/**
	 * @link              https://wsform.com/knowledgebase/pipedrive/
	 * @since             1.0.0
	 * @package           WS_Form_Pipedrive_V1
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Pipedrive
	 * Plugin URI:        https://wsform.com/knowledgebase/pipedrive/
	 * Description:       Pipedrive add-on for WS Form PRO
	 * Version:           1.0.4
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-pipedrive-v1
	 */

	Class WS_Form_Add_On_Pipedrive_V1 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.152';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Pipedrive_V1();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-pipedrive-v1'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_pipedrive_v1 = new WS_Form_Add_On_Pipedrive_V1();	

	register_activation_hook(__FILE__, array($wsf_add_on_pipedrive_v1, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Pipedrive_V1 extends WS_Form_Action {

			public $id = 'pipedrivev1';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Deals';
			public $list_sub_modal_label = 'Select Stage';

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_oauth = false;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_refresh_token;
			private $api_key = false;

			public $list_id = false;
			public $list_sub_id = false;
			public $field_mapping = false;
			public $custom_mapping = false;

			public $update_organization = false;
			public $dedupe_organization = false;

			public $update_person = false;
			public $dedupe_person = false;
			public $dedupe_person_email = false;

			public $update_deal = false;
			public $dedupe_deal = false;
			public $dedupe_deal_emai = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 7057;
			const WS_FORM_LICENSE_NAME = 'Pipedrive add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.4';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const LISTS_FETCH_COUNT = 500;

			public function __construct() {

				// Set label
				$this->label = __('Pipedrive', 'ws-form-pipedrive-v1');

				// Set label for actions pull down
				$this->label_action = __('Add to Pipedrive', 'ws-form-pipedrive-v1');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save access token expiry
				$this->api_access_token_expiry = time() + intval(WS_Form_Common::get_query_var($this->id . '_expires_in', false)) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// If manual key being used, do not check
				if(!$this->api_oauth) { return true; }

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = '/pipedrive-v1?refresh_token=' . rawurlencode($this->api_refresh_token);
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

					// Save access token
					$this->api_access_token = parent::get_object_value($api_response_decoded, $this->id . '_access_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save refresh token
					$this->api_refresh_token = parent::get_object_value($api_response_decoded, $this->id . '_refresh_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

					// Save access token expiry
					$this->api_access_token_expiry = time() + intval(parent::get_object_value($api_response_decoded, $this->id . '_expires_in')) - self::ACCESS_TOKEN_EXPIRY_WINDOW;
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-pipedrive-v1')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-pipedrive-v1') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-pipedrive-v1') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-pipedrive-v1') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/pipedrive-v1';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-pipedrive-v1'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured && $this->api_oauth) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Pipedrive account <a href="%s">here</a>.', 'ws-form-pipedrive-v1'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Defaults
				$deal_id = false;
				$organization_id = false;
				$person_id = false;

				// Process field mapping
				$data_deals = array();
				$data_deals_update_rules = array();
				$data_deals_api = false;

				$data_organizations = array();
				$data_organizations_update_rules = array();
				$data_organizations_api = false;

				$data_persons = array();
				$data_persons_update_rules = array();
				$data_persons_api = false;

				foreach($this->field_mapping as $field_map) {

					// Get submit value
					$field_id = $field_map['ws_form_field'];
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($submit_value === false) { continue; }
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					// Get API key
					$api_field_string = $field_map['action_' . $this->id . '_list_fields'];
					if(strpos($api_field_string, ':') === false) { continue; }
					$api_field_array = explode(':', $api_field_string);
					if(count($api_field_array) != 2) { continue; }

					// Get update rule
					$api_update_rule = isset($field_map['action_' . $this->id . '_update_rule']) ? $field_map['action_' . $this->id . '_update_rule'] : '';

					// Get API field type
					$api_field_type = $api_field_array[0];
					if(!in_array($api_field_type, array('deals', 'organizations', 'persons'))) { continue; }

					// Get API field
					$api_field = $api_field_array[1];

					// Process API fields
					switch($api_field) {

						case 'value' :

							$submit_value = floatval(WS_Form_Common::get_number($submit_value));
							break;
					}

					// Assign to array
					$array_name = sprintf('data_%s', $api_field_type);
					$$array_name[$api_field] = $submit_value;
					$array_name_update_rules = sprintf('data_%s_update_rules', $api_field_type);
					$$array_name_update_rules[$api_field] = $api_update_rule;
				}

				// Process custom mapping
				foreach($this->custom_mapping as $custom_map) {

					// Get field value
					$field_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$field_value = WS_Form_Common::parse_variables_process($field_value, $form, $submit, 'text/plain');

					// Get API key
					$api_field_string = $custom_map['action_' . $this->id . '_list_fields'];
					if(strpos($api_field_string, ':') === false) { continue; }
					$api_field_array = explode(':', $api_field_string);
					if(count($api_field_array) != 2) { continue; }

					// Get update rule
					$api_update_rule = isset($custom_map['action_' . $this->id . '_update_rule']) ? $custom_map['action_' . $this->id . '_update_rule'] : '';

					// Get API field type
					$api_field_type = $api_field_array[0];
					if(!in_array($api_field_type, array('deals', 'organizations', 'persons'))) { continue; }

					// Get API field
					$api_field = $api_field_array[1];

					// Assign to array
					$array_name = sprintf('data_%s', $api_field_type);
					$$array_name[$api_field] = $field_value;
					$array_name_update_rules = sprintf('data_%s_update_rules', $api_field_type);
					$$array_name_update_rules[$api_field] = $api_update_rule;
				}

				// Organization - Find
				if($this->dedupe_organization) { $organization_id = self::find_object($data_organizations, 'organizations'); }
				if($organization_id === false) {

					// Organization - Add
					$organization_id = self::add_object($data_organizations, 'organizations');

				} else {

					if($this->update_organization == 'update') {

						// organization - Update
						$data_organizations = self::data_apply_update_rules($data_organizations, 'organizationFields', 'organizations', $data_organizations_update_rules, $data_organizations_api);
						if(count($data_organizations) > 0) {

							self::update_object($data_organizations, 'organizations', $organization_id);
						}
					}
				}

				// Person - Find
				if($this->dedupe_person) { $person_id = self::find_object($data_persons, 'persons'); }
				if($this->dedupe_person_email && ($person_id === false)) { $person_id = self::find_object($data_persons, 'persons', 'email', true); }
				if($person_id === false) {

					// Person - Add
					if($organization_id !== false) { $data_persons['org_id'] = $organization_id; }
					$person_id = self::add_object($data_persons, 'persons');

				} else {

					if($this->update_person == 'update') {

						// person - Update
						$data_persons = self::data_apply_update_rules($data_persons, 'personFields', 'persons', $data_persons_update_rules, $data_persons_api);
						if(count($data_persons) > 0) {

							self::update_object($data_persons, 'persons', $person_id);
						}
					}
				}

				// Deal - Find - Title
				if($this->dedupe_deal) { $deal_id = self::find_object($data_deals, 'deals', 'title'); }
				if($this->dedupe_deal_email && ($deal_id == false)) {

					// Find person
					$person_id = self::find_object($data_persons, 'persons', 'email');

					// Find associated deals
					if($person_id !== false) {

						$deals = self::find_objects_by_person('deals', $person_id);

						// If there is one deal, we can update it. If there is more than one we should not update.
						if($deals && (count($deals) == 1)) {

							$data_deals_api = $deals[0];
							$deal_id = $data_deals_api->id;
						}
					}
				}

				if($deal_id === false) {

					// Deal - Add
					if($organization_id !== false) { $data_deals['org_id'] = $organization_id; }
					if($person_id !== false) { $data_deals['person_id'] = $person_id; }
					if(!empty($this->list_sub_id)) { $data_deals['stage_id'] = $this->list_sub_id; }
					$deal_id = self::add_object($data_deals, 'deals', 'title');

				} else {

					if($this->update_deal) {

						// Deal - Update
						$data_deals = self::data_apply_update_rules($data_deals, 'dealFields', 'deals', $data_deals_update_rules, $data_deals_api);
						if(count($data_deals) > 0) {

							self::update_object($data_deals, 'deals', $deal_id);

						} else {

							self::success(__('Nothing to update for deal ID: ', 'ws-form-pipedrive-v1') . $deal_id);	
						}
					}
				}
			}

			// Data apply update rules
			public function data_apply_update_rules($data, $method, $object, $update_rules, $data_api) {

				// Get fields
				$sort_index = 0;
				$fields_api = false;
				$return_array = array();

				foreach($data as $field => $value) {

					$update_rule = isset($update_rules[$field]) ? $update_rules[$field] : '';

					switch($update_rule) {

						case 'ignore' :

							break;

						case 'merge' :

							if($data_api === false) { $return_array[$field] = $value; break; }

							// Load API fields
							if($fields_api === false) {

								$fields_api = self::get_custom_fields_raw($method);

								if(!isset($fields_api[$field])) { $return_array[$field] = $value; break; }
							}

							// Get field type
							$field_api = $fields_api[$field];
							$field_type = isset($field_api->field_type) ? $field_api->field_type : 'varchar';

							// Get existing value
							$value_existing = isset($data_api->{$field}) ? $data_api->{$field} : false;
							if($value_existing === false) { $return_array[$field] = $value; break; }

							// Process according to field type
							switch($field_type) {

								// Checkboxes
								case 'set' :

									// Get array of existing values
									$value_existing_array = explode(',', $value_existing);

									// Get new values as array
									$value_array = explode(',', $value);

									// Look up existing values and create lookup table
									$options_lookup = array();
									$options = isset($field_api->options) ? $field_api->options : array();
									foreach($options as $option) {

										$options_lookup[$option->id] = $option->label;
									}

									// Add existing values to new value
									foreach($value_existing_array as $value_existing) {

										if(isset($options_lookup[$value_existing])) {

											$value_array[] = $options_lookup[$value_existing];
										}
									}

									// Merge
									$value_array = array_unique($value_array);

									// Turn into value
									$value = implode(',', $value_array);

									// Assign value
									$return_array[$field] = $value;

									break;

								// Other
								default :

									$value = $value_existing . $value;
							}

							break;

						default :

							$return_array[$field] = $value;
					}
				}

				return $return_array;
			}

			// Get API response ID
			public function get_api_response_id($api_response) {

				// Get organization ID
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->data)) { return false; }
				if(!isset($api_response_decoded->data->id)) { return false; }

				return $api_response_decoded->data->id;
			}

			// Get API find result
			public function get_api_find_result($api_response, $search_term, $identifier = 'name') {

				// Get results
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->data)) { return false; }
				if(is_null($api_response_decoded->data)) { return false; }
				if(!is_array($api_response_decoded->data)) { return false; }
				if(count($api_response_decoded->data) == 0) { return false; }

				// Get result (must match name exactly)
				$search_term_lower = strtolower($search_term);
				foreach($api_response_decoded->data as $result) {

					if(strtolower($result->{$identifier}) == $search_term_lower) {

						self::success(__('Duplicate found: ', 'ws-form-pipedrive-v1') . $search_term);
						return $result->id;
					}
				}

				return false;
			}

			// Get deals by person
			public function find_objects_by_person($object, $person_id) {

				// Check access token
				self::access_token_check();

				// Find organization
				$params_find = array();

				// Add authentication parameters
				self::params_extra_get($params_find);

				$api_response = parent::api_call($this->api_endpoint, 'persons/' . $person_id . '/' . $object, 'GET', $params_find, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

				// Get results
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->data)) { return false; }
				if(is_null($api_response_decoded->data)) { return false; }
				if(!is_array($api_response_decoded->data)) { return false; }
				if(count($api_response_decoded->data) == 0) { return false; }

				self::success(__('Duplicate found for person ID: ', 'ws-form-pipedrive-v1') . $person_id);

				return $api_response_decoded->data;
			}

			// Find object
			public function find_object($data, $object, $identifier = 'name', $search_by_email = false) {

				if(
					(count($data) == 0) ||
					!isset($data[$identifier])||
					empty($data[$identifier])
				) {

					return false;
				}

				// Check access token
				self::access_token_check();

				// Find organization
				$params_find = array(

					'term' => $data[$identifier]
				);

				// Add authentication parameters
				self::params_extra_get($params_find);

				if($search_by_email) { $params_find['search_by_email'] = true; }

				$api_response = parent::api_call($this->api_endpoint, $object . '/find', 'GET', $params_find, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

				return self::get_api_find_result($api_response, $data[$identifier], $identifier);
			}

			// Add object
			public function add_object($data, $object, $identifier = 'name') {

				if(
					(count($data) == 0) ||
					!isset($data[$identifier])||
					empty($data[$identifier])
				) {

					return false;
				}

				// Check access token
				self::access_token_check();

				// Add organization
				$api_response = parent::api_call($this->api_endpoint, $object . self::params_extra_post(), 'POST', $data, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

				// Process response
				switch($api_response['http_code']) {

					case 201 :

						self::success(__('Successfully added: ', 'ws-form-pipedrive-v1') . $data[$identifier]);					
						return self::get_api_response_id($api_response);

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Add object
			public function update_object($data, $object, $object_id) {

				// Check access token
				self::access_token_check();

				// Add organization
				$api_response = parent::api_call($this->api_endpoint, $object . '/' . $object_id. self::params_extra_post(), 'PUT', $data, self::get_api_headers(), 'basic', false, false, 'application/json', 'application/x-www-form-urlencoded');

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						self::success(__('Successfully updated: ', 'ws-form-pipedrive-v1') . $object_id);					
						return self::get_api_response_id($api_response);

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$start = 0;
					$more_items_in_collection = false;

					do {

						// Check access token
						self::access_token_check();

						// Params
						$params = array(

							'limit'			=> self::LISTS_FETCH_COUNT,
							'start'			=> $start,
							'sort[name]'	=> 'asc',
							'fields'		=> 'campaignId,name',
						);

						// Add authentication parameters
						self::params_extra_get($params);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'pipelines', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }
						if(!isset($api_response_decoded->data)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

						// Get total pages
						if(
							isset($api_response_decoded->additional_data) &&
							isset($api_response_decoded->additional_data->pagination)
						) {

							$more_items_in_collection = $api_response_decoded->additional_data->pagination->more_items_in_collection;
							$start = intval($api_response_decoded->additional_data->pagination->start);
						}

						foreach($api_response_decoded->data as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	false, 
								'record_count' => 	false,
								'list_sub' =>		true
							);
						}

					} while ($more_items_in_collection);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Add authentication parameters
					$params = array();
					self::params_extra_get($params);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'pipelines/' . $this->list_id, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(sprintf(__('HTTP Status Code: %s', 'ws-form-pipedrive-v1'), $api_response['http_code'])); return false; }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); return false; }
					if(!isset($api_response_decoded->data)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

					$data = $api_response_decoded->data;

					// Build list
					$list = array(

						'label' => parent::get_object_value($data, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list subs
			public function get_list_subs($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check access token
				self::access_token_check();

				// Check to see if list_subs are cached
				$list_subs = WS_Form_Common::option_get('action_' . $this->id . '_list_subs_' . $this->list_id);

				// Retried if fetch is requested or list_subs are not cached
				if($fetch || ($list_subs === false)) {

					$list_subs = array();

					// Params
					$params = array(

						'pipeline_id'		=> $this->list_id,
					);

					// Add authentication parameters
					self::params_extra_get($params);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'stages', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }
					if(!isset($api_response_decoded->data)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

					$stages = $api_response_decoded->data;

					foreach($stages as $stage) {

						$active_flag = isset($stage->active_flag) ? $stage->active_flag : false;
						if(!$active_flag) { continue; }

						$list_subs[] = array(

							'id' => 			$stage->id, 
							'label' => 			$stage->name,
							'record_count' =>	false
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_subs_' . '_' . $this->list_id, $list_subs);
				}

				return $list_subs;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id . '_deal');

				$sort_index = 0;

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Deals - Custom Fields
					$list_fields = array_merge($list_fields, self::get_custom_fields('dealFields', 'deals', 'Deal', $sort_index, array('title', 'value')));

					// Build currency meta
					$meta_keys = WS_Form_Config::get_meta_keys();
					$data_grid_select = $meta_keys['data_grid_select']['default'];

					// Build new rows
					$rows = array();
					$id = 1;
					$currencies = WS_Form_Config::get_currencies();
					foreach($currencies as $currency_code => $currency) {

						$rows[] = array(

							'id'		=> $id,
							'default'	=> ($currency_code == WS_Form_Common::get_currency_default()) ? 'on' : '',
							'required'	=> '',
							'disabled'	=> '',
							'hidden'	=> '',
							'data'		=> array($currency_code, sprintf('%s (%s)', $currency['n'], $currency['s']))
						);

						$id++;
					}

					// Modify meta
					$data_grid_select['groups'][0]['rows'] = $rows;

					// Columns
					$data_grid_select['columns'] = array(

						array('id' => 0, 'label' => __('Value', 'ws-form')),
						array('id' => 1, 'label' => __('Label', 'ws-form'))
					);

					// Deals - Currency
					$list_fields[] = array(

						'id' => 			'deals:currency', 
						'id_sub' =>			false,
						'label' => 			'Deal: Currency', 
						'label_field' => 	'Deal Currency', 
						'label_sub' =>		false,
						'type' => 			'select', 
						'required' => 		false, 
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	$sort_index++,
						'visible' =>		true,
						'meta' =>			array(

							'data_grid_select' => $data_grid_select,
							'select_field_label' => 1
						)
					);

					// Persons - Custom Fields
					$list_fields = array_merge($list_fields, self::get_custom_fields('personFields', 'persons', 'Person', $sort_index, array('name')));

					// Organizations - Custom Fields
					$list_fields = array_merge($list_fields, self::get_custom_fields('organizationFields', 'organizations', 'Organization', $sort_index, array('name')));

					// Sort by display_order
					usort($list_fields, function ($a, $b) { return $a['label'] > $b['label']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id . '_' . $this->list_sub_id, $list_fields);
				}

				return $list_fields;
			}

			// Get custom fields
			public function get_custom_fields($endpoint, $prefix, $label, &$sort_index, $no_add_false_array = array()) {

				// Initialize
				$start = 0;
				$more_items_in_collection = false;
				$list_fields = array();
				$api_list_fields = array();

				do {

					// Check access token
					self::access_token_check();

					// Params
					$params = array(

						'limit'			=> self::LISTS_FETCH_COUNT,
						'start'			=> $start,
						'sort[name]'	=> 'asc',
						'fields'		=> 'campaignId,name',
					);

					// Add authentication parameters
					self::params_extra_get($params);

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, $endpoint, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

					// Get total pages
					if(
						isset($api_response_decoded->additional_data) &&
						isset($api_response_decoded->additional_data->pagination)
					) {

						$more_items_in_collection = $api_response_decoded->additional_data->pagination->more_items_in_collection;
						$start = intval($api_response_decoded->additional_data->pagination->start);
					}

					$api_list_fields = array_merge($api_list_fields, $api_response_decoded->data);

				} while ($more_items_in_collection);

				// Process fields
				foreach($api_list_fields as $field) {

//					$add_visible_flag = parent::get_object_value($field, 'add_visible_flag');
//					if(!$add_visible_flag) { continue; }

					$is_subfield = parent::get_object_value($field, 'is_subfield');
					if($is_subfield) { continue; }

					$id = parent::get_object_value($field, 'key');
					if(in_array($id, array('id', 'label'))) { continue; }

					$field_type = parent::get_object_value($field, 'fieldType');

					// WS Form type
					$type = self::action_field_type_to_ws_form_field_type($field);

					// Meta data
					$meta = self::action_field_to_ws_form_meta_keys($field);

					$list_fields[] = array(

						'id' => 			sprintf('%s:%s', $prefix, $id),
						'label' => 			sprintf('%s: %s', $label, parent::get_object_value($field, 'name')),
						'label_field' => 	sprintf('%s %s', $label, parent::get_object_value($field, 'name')),
						'type' => 			$type,
						'action_type' =>	$field_type,
						'required' => 		parent::get_object_value($field, 'mandatory_flag'),
						'default_value' => 	'',
						'pattern' => 		'',
						'placeholder' => 	'',
						'input_mask' => 	'',
						'help' => 			'', 
						'sort_index' => 	$sort_index++,
						'visible' =>		true,
						'meta' => 			$meta,
						'no_add' =>			!in_array($id, $no_add_false_array)
					);
				}

				// Sort by display_order
				usort($list_fields, function ($a, $b) { return $a['label'] > $b['label']; });

				return $list_fields;
			}

			// Get custom fields - Raw (For POST method)
			public function get_custom_fields_raw($endpoint) {

				// Initialize
				$start = 0;
				$more_items_in_collection = false;
				$list_fields = array();
				$api_list_fields = array();

				do {

					// Check access token
					self::access_token_check();

					// Params
					$params = array(

						'limit'			=> self::LISTS_FETCH_COUNT,
						'start'			=> $start,
						'sort[name]'	=> 'asc',
						'fields'		=> 'campaignId,name',
					);

					// Add authentication parameters
					self::params_extra_get($params);

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, $endpoint, 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

					// Get total pages
					if(
						isset($api_response_decoded->additional_data) &&
						isset($api_response_decoded->additional_data->pagination)
					) {

						$more_items_in_collection = $api_response_decoded->additional_data->pagination->more_items_in_collection;
						$start = intval($api_response_decoded->additional_data->pagination->start);
					}

					$api_list_fields = array_merge($api_list_fields, $api_response_decoded->data);

				} while ($more_items_in_collection);

				// Change keys to api_list_field keys
				$return_array = array();
				foreach($api_list_fields as $api_list_field) {

					$return_array[$api_list_field->key] = $api_list_field;
				}

				return $return_array;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'submit' => array(

						'type' => 'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_list_sub_id'		=>	$this->list_sub_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-pipedrive-v1')); }

				$error_message = $api_response_decoded->error;
				if(isset($api_response_decode->error_info)) {

					$error_message .= ' - ' . $api_response_decoded->error_info;
				}

				return array($error_message);
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_list_sub_id',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping'
					),

					'fieldsets'	=>	array(

						array(
							'label'		=>	__('Deduplication', 'ws-form'),
							'meta_keys'	=>	array(

								'action_' . $this->id . '_update_organization', 
								'action_' . $this->id . '_dedupe_organization', 

								'action_' . $this->id . '_update_person', 
								'action_' . $this->id . '_dedupe_person',
								'action_' . $this->id . '_dedupe_person_email',

								'action_' . $this->id . '_update_deal', 
								'action_' . $this->id . '_dedupe_deal',
								'action_' . $this->id . '_dedupe_deal_email'
							)
						),
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-pipedrive-v1'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-pipedrive-v1')); }

				return ($this->list_id !== false);
			}

			// Check list sub ID is set
			public function check_list_sub_id() {

				if($this->list_sub_id === false) { return self::error(__('List sub ID is not set', 'ws-form-pipedrive-v1')); }

				return ($this->list_sub_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('Pipeline', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'help'							=>	__('Which pipeline should this deal be added to?', 'ws-form-pipedrive-v1'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-pipedrive-v1'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// List sub ID
					'action_' . $this->id . '_list_sub_id'	=> array(

						'label'						=>	__('Stage', 'ws-form-pipedrive-v1'),
						'type'						=>	'select',
						'help'						=>	__('Select the pipeline stage to associate this deal with.', 'ws-form-pipedrive-v1'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-pipedrive-v1'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_subs',
						'reload'						=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_subs_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
					),

					// Update - Organization
					'action_' . $this->id . '_update_organization'	=> array(

						'label'							=>	__('If duplicate organization found', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-pipedrive-v1')),
							array('value' => 'update', 'text' => __('Update', 'ws-form-pipedrive-v1'))
						)
					),

					// Dedupe - Organization
					'action_' . $this->id . '_dedupe_organization'	=> array(

						'label'						=>	__('Dedupe organizations by name', 'ws-form-pipedrive-v1'),
						'type'						=>	'checkbox',
						'help'						=>	__('Attempt to dedupe organizations by their name.', 'ws-form-pipedrive-v1'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Update - Person
					'action_' . $this->id . '_update_person'	=> array(

						'label'							=>	__('If duplicate person found', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-pipedrive-v1')),
							array('value' => 'update', 'text' => __('Update', 'ws-form-pipedrive-v1'))
						)
					),

					// Dedupe - Person
					'action_' . $this->id . '_dedupe_person'	=> array(

						'label'						=>	__('Dedupe person by name', 'ws-form-pipedrive-v1'),
						'type'						=>	'checkbox',
						'help'						=>	__('Attempt to dedupe person by their name.', 'ws-form-pipedrive-v1'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Dedupe - Person (Email)
					'action_' . $this->id . '_dedupe_person_email'	=> array(

						'label'						=>	__('Dedupe person by email', 'ws-form-pipedrive-v1'),
						'type'						=>	'checkbox',
						'help'						=>	__('Attempt to dedupe person by their email address. Email address must be mapped.', 'ws-form-pipedrive-v1'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Update - Deals
					'action_' . $this->id . '_update_deal'	=> array(

						'label'							=>	__('If duplicate deal found', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-pipedrive-v1')),
							array('value' => 'update', 'text' => __('Update', 'ws-form-pipedrive-v1'))
						)
					),

					// Dedupe - Deal - Title
					'action_' . $this->id . '_dedupe_deal'	=> array(

						'label'						=>	__('Dedupe deals by title', 'ws-form-pipedrive-v1'),
						'type'						=>	'checkbox',
						'help'						=>	__('Attempt to dedupe deals by their title.', 'ws-form-pipedrive-v1'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Dedupe - Deal - Email
					'action_' . $this->id . '_dedupe_deal_email'	=> array(

						'label'						=>	__('Dedupe deals by email', 'ws-form-pipedrive-v1'),
						'type'						=>	'checkbox',
						'help'						=>	__('Attempt to dedupe deals by persons email address. Email address must be mapped.', 'ws-form-pipedrive-v1'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-pipedrive-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Pipedrive fields.', 'ws-form-pipedrive-v1'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-pipedrive-v1'),
						'type'						=>	'repeater',
						'help'						=>	__('Map values or variables to Pipedrive fields.', 'ws-form-pipedrive-v1'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'

						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Pipedrive', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-pipedrive-v1'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// Update rules
					'action_' . $this->id . '_update_rule'	=> array(

						'label'							=>	__('On Update', 'ws-form-pipedrive-v1'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Update', 'ws-form-pipedrive-v1')),
							array('value' => 'merge', 'text' => __('Merge', 'ws-form-pipedrive-v1')),
							array('value' => 'ignore', 'text' => __('Ignore', 'ws-form-pipedrive-v1'))
						)
					),

					// Custom mapping value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'							=>	__('Value', 'ws-form-pipedrive-v1'),
						'type'							=>	'text'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-pipedrive-v1'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-pipedrive-v1'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Pipedrive add-on for WS Form PRO license key here.', 'ws-form-pipedrive-v1'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-pipedrive-v1'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Pipedrive Status', 'ws-form-pipedrive-v1'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Pipedrive account, or <a href=\"https://www.pipedrive.com/start-free-trial\" target=\"blank\">create a new account</a>.", 'ws-form-pipedrive-v1'),
							'button'		=>	'oauth_action_' . $this->id,
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'			=>	__('API Key (Optional)', 'ws-form-pipedrive-v1'),
							'type'			=>	'text',
							'help'			=>	__('If you want to use your API key instead of oAuth, enter it here.', 'ws-form-pipedrive-v1')
						),

						'action_' . $this->id . '_api_domain'	=>	array(

							'label'			=>	__('API Domain (Optional)', 'ws-form-pipedrive-v1'),
							'type'			=>	'text',
							'help'			=>	__('If you want to use your API key instead of oAuth, enter your company domain here. e.g. companydomain.pipedrive.com', 'ws-form-pipedrive-v1')
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($field) {

				$field_type = parent::get_object_value($field, 'field_type');

				// Field type
				switch($field_type) {

					case 'varchar' : return 'text';
					case 'varchar_auto' : return 'text';
					case 'enum' : return 'radio';
					case 'set' : return 'checkbox';
					case 'text' : return 'textarea';
					case 'monetary' : return 'price';
					case 'double' : return 'number';
					case 'int' : return 'number';
					case 'date' : return 'datetime';
					case 'time' : return 'datetime';
					case 'phone' : return 'tel';

					case 'user' : return 'text';		// Unsupported
					case 'org' : return 'text';			// Unsupported
					case 'people' : return 'text';		// Unsupported
					case 'stage' : return 'text';		// Unsupported
					case 'address' : return 'text';		// Unsupported
					case 'daterange' : return 'text';	// Unsupported
					case 'timerange' : return 'text';	// Unsupported
					case 'status' : return 'text';		// Unsupported
				}

				return 'text';
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$field_type = parent::get_object_value($field, 'field_type');

				// Get WS Form meta configurations for action field types
				switch($field_type) {

					// Build data grids for radio and select
					case 'enum' :
					case 'set' :

						switch($field_type) {

							case 'enum' : $meta_key = 'data_grid_radio'; break;
							case 'set' : $meta_key = 'data_grid_checkbox'; break;
						}

						// Get options
						$options = parent::get_object_value($field, 'options');

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys[$meta_key])) { return false; }
						if(!isset($meta_keys[$meta_key]['default'])) { return false; }

						$meta = $meta_keys[$meta_key]['default'];

						// Columns
						$meta['columns'] = array(

							array('id' => 0, 'label' => __('Value', 'ws-form')),
							array('id' => 1, 'label' => __('Label', 'ws-form'))
						);

						// Build new rows
						$rows = array();
						$id = 1;
						foreach($options as $option) {

							$value = isset($option->id) ? $option->id : '';
							$label = isset($option->label) ? $option->label : '';

							$rows[] = array(

								'id'		=> $id,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array($value, $label)
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						$return_array = array(

							$meta_key => $meta,
							($field_type == 'enum') ? 'radio_field_label' : 'checkbox_field_label' => 1
						);

						return($return_array);

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				if($this->list_sub_id === false) { $this->list_sub_id = parent::get_config($config, 'action_' . $this->id . '_list_sub_id'); }
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }

				$this->update_organization = parent::get_config($config, 'action_' . $this->id . '_update_organization', false);
				$this->dedupe_organization = parent::get_config($config, 'action_' . $this->id . '_dedupe_organization', false);

				$this->update_person = parent::get_config($config, 'action_' . $this->id . '_update_person', false);
				$this->dedupe_person = parent::get_config($config, 'action_' . $this->id . '_dedupe_person', false);
				$this->dedupe_person_email = parent::get_config($config, 'action_' . $this->id . '_dedupe_person_email', false);

				$this->update_deal = parent::get_config($config, 'action_' . $this->id . '_update_deal', false);
				$this->dedupe_deal = parent::get_config($config, 'action_' . $this->id . '_dedupe_deal', false);
				$this->dedupe_deal_email = parent::get_config($config, 'action_' . $this->id . '_dedupe_deal_email', false);

				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry  = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token  		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');

				// Set endpoint URL
				$api_key_manual = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');
				$api_domain_manual = WS_Form_Common::option_get('action_' . $this->id . '_api_domain', '');
				if(
					($api_key_manual != '') &&
					($api_domain_manual != '')
				) {

					// Manual
					$this->api_endpoint = sprintf('https://%s/v1/', $api_domain_manual);
					$this->api_oauth = false;

				} else {

					// oAuth
					$this->api_endpoint = 'https://api-proxy.pipedrive.com/';
					$this->api_oauth = true;
				}

				// Check access token
				if($this->api_oauth && empty($this->api_access_token)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Get extra params for GET requests
			public function params_extra_get(&$params_find) {

				if(!$this->api_oauth) {

					$api_key_manual = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');
					$params_find['api_token'] = $api_key_manual;
				}
			}

			// Get extra params for POST requests
			public function params_extra_post() {

				if($this->api_oauth) {

					return '';

				} else {

					$api_key_manual = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');
					return sprintf('?api_token=%s', urlencode($api_key_manual));
				}
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_subs'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/(?P<list_sub_id>[a-zA-Z0-9-_]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_subs'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/(?P<list_sub_id>[a-zA-Z0-9-_]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List Subs
			public function api_get_list_subs($parameters) {

				// Get list subs
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_subs = self::get_list_subs();

				// Process response
				self::api_response($list_subs);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List subs with fetch
			public function api_fetch_list_subs($parameters) {

				// Get list subs
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_subs = self::get_list_subs(true);

				// Process response
				self::api_response($list_subs);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(39.000000, 60.000000)"><path d="M5.5 0h53.1c3 0 5.5 2.5 5.5 5.5v53.1c0 3-2.5 5.5-5.5 5.5H5.5c-3 0-5.5-2.5-5.5-5.5V5.5C0 2.5 2.5 0 5.5 0z" fill-rule="evenodd" clip-rule="evenodd" fill="#26292c"/><path d="M26.7 26.4c0 4.3 2.2 8.9 7 8.9 3.5 0 7.1-2.8 7.1-9 0-5.4-2.8-9.1-7-9.1-3.5 0-7.1 2.4-7.1 9.2zm8.7-17.1c8.6 0 14.3 6.8 14.3 16.9 0 10-6.1 16.9-14.8 16.9-4.1 0-6.8-1.8-8.2-3.1v14.5H18v-36c0-.5-.2-.7-.7-.7h-3.1V10h7.5c3.4 0 4.3 1.7 4.5 3.1 1.4-1.6 4.2-3.8 9.2-3.8z" fill-rule="evenodd" clip-rule="evenodd" fill="#fff"/></g>';

				return $svg_logo;
			}
		}
	});