<?php

	/**
	 * @link              https://wsform.com/knowledgebase/salesforce/
	 * @since             1.0.0
	 * @package           WS_Form_Salesforce
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Salesforce
	 * Plugin URI:        https://wsform.com/knowledgebase/salesforce/
	 * Description:       Salesforce add-on for WS Form PRO
	 * Version:           1.1.4
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-salesforce
	 */

	Class WS_Form_Add_On_Salesforce {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.95';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Salesforce();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-salesforce'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_salesforce = new WS_Form_Add_On_Salesforce();	

	register_activation_hook(__FILE__, array($wsf_add_on_salesforce, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Salesforce extends WS_Form_Action {

			public $id = 'salesforce';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Objects';

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $hostname;
			private $api_endpoint = false;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_refresh_token;
			private $api_instance_url;
			private $api_objects;

			private $objects_default;

			public $list_id = false;
			public $opt_in_field = false;
			public $field_mapping = false;
			public $custom_mapping = false;
			public $dedupe = false;
			public $dedupe_field = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 5480;
			const WS_FORM_LICENSE_NAME = 'Salesforce add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.1.4';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const SLACK_FIELD_SHORT_THRESHOLD = 20;
			const EMAIL_MERGE_FIELD = 'EmailAddress';
			const API_PATH = '/services/data/v46.0/';
			const API_MAX_RECORDS = 50;

			public function __construct() {

				// Set label
				$this->label = __('Salesforce', 'ws-form-salesforce');

				// Set label for actions pull down
				$this->label_action = __('Add to Salesforce', 'ws-form-salesforce');

				// Default objects
				$this->objects_default = array(

					'Account',
					'Contact',
					'Lead'
				);

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Error checking
				$error = WS_Form_Common::get_query_var($this->id . '_error', false);
				if($error !== false) {

					$error_description = WS_Form_Common::get_query_var($this->id . '_error_description', __('Unknown error', 'ws-form-salesforce'));

					WS_Form_Common::admin_message_push(sprintf(__('An error occurred when connecting your account: %s.', 'ws-form-'), $error_description), 'notice-error', false);

					self::reset_api();

					return false;
				}

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save access token expiry
				$this->api_access_token_expiry = WS_Form_Common::get_query_var($this->id . '_access_token_expiry', false);
				if($this->api_access_token_expiry === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				if($this->api_refresh_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save instance URL
				$this->api_instance_url = WS_Form_Common::get_query_var($this->id . '_instance_url', false);
				WS_Form_Common::option_set('action_' . $this->id . '_api_instance_url', $this->api_instance_url);

				// Reset objects
				WS_Form_Common::option_set('action_' . $this->id . '_api_objects', $this->objects_default);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = sprintf('/salesforce?state=%s&refresh_token=%s', rawurlencode(json_encode(array('hostname' => $this->hostname))), rawurlencode($this->api_refresh_token));
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

					// Save access token
					$this->api_access_token = parent::get_object_value($api_response_decoded, $this->id . '_access_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save access token expiry
					$this->api_access_token_expiry = parent::get_object_value($api_response_decoded, $this->id . '_access_token_expiry');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-salesforce')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;

					case 'action_' . $this->id . '_api_instance_url' :

						$value = $this->api_instance_url;
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-salesforce') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-salesforce') . '" />';
						}

						break;

					case 'hostname_apply_' . $this->id :

						$value = '<input class="wsf-button" type="submit" data-mode="deactivate" value="' . __('Apply', 'ws-form-salesforce') . '" />';

						break;

					case 'oauth-action-' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access-token-delete" value="' . __('Disconnect', 'ws-form-salesforce') . '" />';

						} else {

							// Build query string
							$button_url = sprintf('https://wsform.com/salesforce/?state=%s', rawurlencode(json_encode(array('hostname' => $this->hostname))));
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-salesforce'));
						}

						break;

					case 'entity-definitions' :

						$value = '<div><input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="entity-definitions" value="' . __('Refresh', 'ws-form-salesforce') . '" /></div>';

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access-token-delete' :

								self::reset_api();
								break;

							case 'entity-definitions' :

								self::get_entity_definitions(true, true);
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_instance_url', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_objects', $this->objects_default);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Salesforce account <a href="%s">here</a>.', 'ws-form-salesforce'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-salesforce'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Get list fields
				$list_fields = self::get_list_fields();
				$list_fields_type_lookup = array();
				foreach($list_fields as $list_field) {

					$list_fields_type_lookup[$list_field['id']] = $list_field['action_type'];
				}

				// Build API payload
				$api_payload = array();
				$api_payload_update_rules = array();

				// Process field mapping
				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					$field_key = $field_map['action_' . $this->id . '_list_fields'];
					$field_update_rule = isset($field_map['action_' . $this->id . '_update_rule']) ? $field_map['action_' . $this->id . '_update_rule'] : '';

					// Get submit value
					$field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Process according to Salesforce type
					if(!isset($list_fields_type_lookup[$field_key])) { continue; }
					$action_type = $list_fields_type_lookup[$field_key];

					// Process field value
					$field_value = self::field_value_process($field_value, $action_type);

					// Add to API payload
					$api_payload[$field_key] = $field_value;

					// Add to update rules
					$api_payload_update_rules[$field_key] = $field_update_rule;
				}

				// Process custom mapping
				foreach($this->custom_mapping as $custom_map) {

					$field_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$field_key = $custom_map['action_' . $this->id . '_list_fields'];
					$field_update_rule = isset($custom_map['action_' . $this->id . '_update_rule']) ? $custom_map['action_' . $this->id . '_update_rule'] : '';

					// Process according to Salesforce type
					if(!isset($list_fields_type_lookup[$field_key])) { continue; }
					$action_type = $list_fields_type_lookup[$field_key];

					// Parse field value
					$field_value = WS_Form_Common::parse_variables_process($field_value, $form, $submit, 'text/plain');

					// Process field value
					$field_value = self::field_value_process($field_value, $action_type);

					// Add to API payload
					$api_payload[$field_key] = $field_value;

					// Add to update rules
					$api_payload_update_rules[$field_key] = $field_update_rule;
				}

				// Check payload
				if(count($api_payload) == 0) {

					self::error(__('No data to add.', 'ws-form-salesforce'));
					return true;
				}

				// Check access token
				self::access_token_check();

				// Object - Find
				$object_id = false;
				$object = false;
				if(
					($this->dedupe_field != '') &&
					isset($api_payload[$this->dedupe_field]) &&
					($api_payload[$this->dedupe_field] != '')
				) {

					// Get dedupe field
					$dedupe_field = $this->dedupe_field;

					// Get dedupe field value
					$dedupe_value = $api_payload[$dedupe_field];

					// Get Salesforce fields
					$salesforce_fields_array = array_keys($api_payload);
					$salesforce_fields = implode(', ', $salesforce_fields_array);

					// Build query
					$params = array(

						'q' => sprintf("SELECT Id, %s FROM %s WHERE %s = '%s' LIMIT 1", esc_sql($salesforce_fields), esc_sql($this->list_id), esc_sql($dedupe_field), esc_sql($dedupe_value))
					);

					// Call API
					$api_response = parent::api_call($this->api_endpoint, 'query/', 'GET', $params, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] == 200) { 

						// Get API response
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid GET API response data', 'ws-form-salesforce')); }

						// Get records
						$records = isset($api_response_decoded->records) ? $api_response_decoded->records : array();

						// Get record
						$object = isset($records[0]) ? $records[0] : false;

						// Get object ID
						$object_id = isset($object->Id) ? $object->Id : false;

						self::success(__('Duplicate object found: ', 'ws-form-pipedrive-v1') . $object_id);
					}
				}

				if($object_id === false) {

					// Create new record

					// JSON encode payload
					$api_payload_json = json_encode($api_payload);

					// Call API
					$api_response = parent::api_call($this->api_endpoint, 'sobjects/' . $this->list_id, 'POST', $api_payload_json, self::get_api_headers());

					// Process response
					switch($api_response['http_code']) {

						case 201 :

							self::success(__('Successfully added object: ', 'ws-form-salesforce') . $this->list_id);
							return true;

						default :

							// Get invalid merge fields
							return self::error(self::get_api_error_detail($api_response['response']));
					}

				} else {

					if($this->dedupe == 'update') {

						// Update
						$api_payload = self::api_payload_apply_update_rules($api_payload, $api_payload_update_rules, $object);
						if(count($api_payload) > 0) {

							// JSON encode payload
							$api_payload_json = json_encode($api_payload);

							// Call API
							$api_response = parent::api_call($this->api_endpoint, 'sobjects/' . $this->list_id . '/' . $object_id, 'PATCH', $api_payload_json, self::get_api_headers());

							// Process response
							switch($api_response['http_code']) {

								case 204 :

									self::success(__('Successfully updated object: ', 'ws-form-salesforce') . $object_id);
									return true;

								default :

									// Get invalid merge fields
									return self::error(self::get_api_error_detail($api_response['response']));
							}
						}
					}
				}
			}

			// Data apply update rules
			public function api_payload_apply_update_rules($api_payload, $update_rules, $object) {

				// Get fields
				$sort_index = 0;
				$fields_api = false;
				$return_array = array();

				foreach($api_payload as $field => $value) {

					$update_rule = isset($update_rules[$field]) ? $update_rules[$field] : '';

					switch($update_rule) {

						case 'ignore' :

							break;

						case 'merge' :

							if($object === false) { $return_array[$field] = $value; break; }

							// Load API fields
							if($fields_api === false) {

								$fields_api = self::object_describe();

								if(!isset($fields_api[$field])) { $return_array[$field] = $value; break; }
							}

							// Get field type
							$field_type = $fields_api[$field];

							// Get existing value
							$value_existing = isset($object->{$field}) ? $object->{$field} : false;
							if($value_existing === false) { $return_array[$field] = $value; break; }

							// Process according to field type
							switch($field_type) {

								// Checkboxes
								case 'Picklist (Multi-Select)' :

									// Get array of existing values
									$value_existing_array = explode(';', $value_existing);

									// Get new values as array
									$value_array = explode(',', $value);

									// Merge
									$value_array = array_merge($value_existing_array, $value_array);
									$value_array = array_unique($value_array);

									// Turn into value
									$value = implode(';', $value_array);

									// Assign value
									$return_array[$field] = $value;

									break;

								// Other
								default :

									$value = $value_existing . $value;
							}

							break;

						default :

							$return_array[$field] = $value;
					}
				}

				return $return_array;
			}

			// Get object fields
			public function object_describe() {

				$return_array = array();

				// Build query
				$params = array(

					'q' => sprintf("SELECT QualifiedApiName, DataType FROM FieldDefinition WHERE EntityDefinition.QualifiedApiName = '%s'", esc_sql($this->list_id))
				);

				// Call API
				$api_response = parent::api_call($this->api_endpoint, 'query/', 'GET', $params, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Get API response
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid GET API response data', 'ws-form-salesforce')); }

					// Get fields
					$fields = isset($api_response_decoded->records) ? $api_response_decoded->records : array();

					foreach($fields as $field) {

						$return_array[$field->QualifiedApiName] = $field->DataType;
					}
				}

				return $return_array;
			}

			// Get API find result
			public function get_api_find_result($api_response, $search_term, $identifier = 'name') {

				// Get results
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return false; }
				if(!isset($api_response_decoded->data)) { return false; }
				if(is_null($api_response_decoded->data)) { return false; }
				if(!is_array($api_response_decoded->data)) { return false; }
				if(count($api_response_decoded->data) == 0) { return false; }

				// Get result (must match name exactly)
				$search_term_lower = strtolower($search_term);
				foreach($api_response_decoded->data as $result) {

					if(strtolower($result->{$identifier}) == $search_term_lower) {

						self::success(__('Duplicate found: ', 'ws-form-salesforce') . $search_term);
						return $result->id;
					}
				}

				return false;
			}

			// Process field value
			public function field_value_process($field_value, $action_type) {

				// Salesforce cannot accept arrays
				if(is_array($field_value)) { $field_value = implode(',', $field_value); }

				switch($action_type) {

					case 'boolean' :

						$field_value = ($field_value == 'on') ? 'true' : 'false';
						break;

					case 'reference' :

						$field_value = ($field_value != '') ? $field_value : NULL;
						break;

					case 'currency' :

						$field_value = filter_var($field_value, FILTER_SANITIZE_NUMBER_FLOAT, FILTER_FLAG_ALLOW_FRACTION);;
						break;

					case 'date' :

						$field_value = date('Y-m-d', strtotime($field_value));
						break;

					case 'datetime' :

						$field_value = date('Y-m-d', strtotime($field_value)) . 'T' . date('H:i:s', strtotime($field_value));
						break;

					case 'time' :

						$field_value = date('H:i:s', strtotime($field_value));
						break;

					case 'multipicklist' :

						if(is_array($field_value)) {

							$field_value = implode(';', $field_value);
						}
						break;
				}

				return $field_value;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Get selected entity definitions
					$entity_definitions = self::get_entity_definitions(false, $fetch);
					if($entity_definitions === false) { return false; }

					foreach($entity_definitions as $id => $entity_definition) {

						$lists[] = array(

							'id' => 			$id, 
							'label' => 			$entity_definition['text'], 
							'field_count' => 	false, 
							'record_count' => 	false
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'sobjects/' . $this->list_id . '/', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

					// Get entity definitions
					if(!isset($api_response_decoded->objectDescribe) || !isset($api_response_decoded->objectDescribe->label)) { return self::error(__('Unable to retrieve entity definition', 'ws-form-salesforce')); }

					// Build list
					$list = array(

						'label' => $api_response_decoded->objectDescribe->label
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();
					$total_fields;
					$total_fields_retrieved = 0;

					// Check access token
					self::access_token_check();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'sobjects/' . $this->list_id . '/describe/', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

					// Checks
					if(!isset($api_response_decoded->fields)) { return self::error(__('Unable to read object field definitions', 'ws-form-salesforce')); }

					$fields = $api_response_decoded->fields;

					// Process fields
					$sort_index = 0;

					foreach($fields as $field) {

						// Skip non createble fields
						$createable = parent::get_object_value($field, 'createable');
						if(!$createable) { continue;}

						// Type
						$action_type = parent::get_object_value($field, 'type');
						$type = self::action_field_type_to_ws_form_field_type($action_type);

						// ID
						$id = parent::get_object_value($field, 'name');

						// Default value
						$default_value = parent::get_object_value($field, 'defaultValue');

						$list_fields[] = array(

							'id' => 			$id,
							'label' => 			parent::get_object_value($field, 'label'), 
							'label_field' => 	parent::get_object_value($field, 'label'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		false,	// API does not return this parameter 
							'default_value' => 	$default_value,
							'pattern' => 		'',
							'placeholder' => 	'',
							'input_mask' => 	'',
							'help' => 			'', 
							'sort_index' => 	$sort_index,
							'visible' =>		true,
							'meta' => 			self::action_field_to_ws_form_meta_keys($field)
						);

						$sort_index++;
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit',
						'width_factor'	=>	0.5
					),

					'reset' => array(

						'type'			=>	'reset',
						'width_factor'	=>	0.5
					),
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field'
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$error_detail_array = array();

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

				if(!is_array($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

				// Get error detail from JSON response
				foreach($api_response_decoded as $error) {

					$error_string_array = array();
					if(isset($error->errorCode)) { $error_string_array[] = $error->errorCode; }
					if(isset($error->message)) { $error_string_array[] = $error->message; }
					if(isset($error->fields)) {

						$field_string_array = array();
						foreach($error->fields as $field) {

							$field_string_array[] = $field;
						}
						$field_string = implode(', ', $field_string_array);

						$error_string_array[] = $field_string;
					}
					$error_string = implode(' - ', $error_string_array);

					$error_detail_array[] = $error_string;
				}

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_dedupe_field',
						'action_' . $this->id . '_dedupe'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-salesforce'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-salesforce')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('Salesforce Object', 'ws-form-salesforce'),
						'type'							=>	'select',
						'help'							=>	__('Which Salesforce object should this form data be added to?', 'ws-form-salesforce'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-salesforce'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-salesforce'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-salesforce'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-salesforce')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-salesforce'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Salesforce fields.', 'ws-form-salesforce'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-salesforce'),
						'type'						=>	'repeater',
						'help'						=>	__('Map values or variables to Salesforce fields.', 'ws-form-salesforce'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields',
							'action_' . $this->id . '_update_rule'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Dedupe 
					'action_' . $this->id . '_dedupe'	=> array(

						'label'							=>	__('If Duplicate Found', 'ws-form-salesforce'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-salesforce')),
							array('value' => 'update', 'text' => __('Update', 'ws-form-salesforce'))
						)
					),

					// Dedupe - Field
					'action_' . $this->id . '_dedupe_field'	=> array(

						'label'							=>	__('Find Duplicates By', 'ws-form-salesforce'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-salesforce'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Salesforce Field', 'ws-form-salesforce'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-salesforce'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// Update rules
					'action_' . $this->id . '_update_rule'	=> array(

						'label'							=>	__('On Update', 'ws-form-salesforce'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => '', 'text' => __('Update', 'ws-form-salesforce')),
							array('value' => 'merge', 'text' => __('Merge', 'ws-form-salesforce')),
							array('value' => 'ignore', 'text' => __('Ignore', 'ws-form-salesforce'))
						)
					),

					// Custom mapping value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'							=>	__('Value', 'ws-form-salesforce'),
						'type'							=>	'text'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-salesforce'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-salesforce'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Salesforce add-on for WS Form PRO license key here.', 'ws-form-salesforce'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-salesforce'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_hostname'	=>	array(

							'label'		=>	__('Custom Endpoint Hostname', 'ws-form-salesforce'),
							'type'		=>	'text',
							'help'		=>	__('Enter a custom API endpoint hostname. For example: test.salesforce.com (Sandbox). Defaults to login.salesforce.com', 'ws-form'),
							'default'	=>	'',
							'button'	=>	'hostname_apply_' . $this->id
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Salesforce Status', 'ws-form-salesforce'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Salesforce account, or <a href=\"https://www.salesforce.com/\" target=\"blank\">create a new account</a>.", 'ws-form-salesforce'),
							'button'		=>	'oauth-action-' . $this->id,
						)
					)
				);

				if($this->configured) {

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_instance_url'] =	array(

						'label'			=>	__('Instance URL', 'ws-form-salesforce'),
						'type'			=>	'static'
					);

					// Get objects data
					$entity_definitions = self::get_entity_definitions();

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_objects'] =	array(

						'label'			=>	__('Working Objects', 'ws-form-salesforce'),
						'type'			=>	'select',
						'size'			=>	20,
						'multiple'		=>	true,
						'help'			=>	__('Select which SalesForce objects you would like to work with.', 'ws-form-salesforce'),
						'options'		=>	$entity_definitions,
						'button'		=>	'entity-definitions'
					);
				}

				return $options;
			}

			// Get objects
			public function get_entity_definitions($include_unselected = true, $fetch = false) {

				$entity_definitions_raw = WS_Form_Common::option_get('action_' . $this->id . '_entity_definitions');

				if($fetch || ($entity_definitions_raw === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get lists
					$api_response = parent::api_call($this->api_endpoint, 'sobjects/', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-salesforce')); }

					// Get entity definitions
					if(!is_array($api_response_decoded->sobjects)) { return self::error(__('Unable to retrieve entity definitions', 'ws-form-salesforce')); }

					// Build raw data
					$entity_definitions_raw = array();

					foreach($api_response_decoded->sobjects as $sobject) {

						if(strpos($sobject->label, '__MISSING LABEL__') !== false) { continue; }

						$entity_definitions_raw[$sobject->name] = array('text' => $sobject->label);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_entity_definitions', $entity_definitions_raw);
				}

				// Process entity definitions
				$entity_definitions_selected = array();
				$entity_definitions_unselected = array();

				foreach($entity_definitions_raw as $name => $entity_definition) {

					if(in_array($name, $this->api_objects)) {

						$entity_definitions_selected[$name] = array('text' => $entity_definition['text']);

					} else {

						if($include_unselected) {

							$entity_definitions_unselected[$name] = array('text' => $entity_definition['text']);
						}
					}
				}

				// Sort arrays
				$text = array_column($entity_definitions_selected, 'text');
				array_multisort($text, SORT_ASC, $entity_definitions_selected);

				$text = array_column($entity_definitions_unselected, 'text');
				array_multisort($text, SORT_ASC, $entity_definitions_unselected);

				$entity_definitions = array_merge($entity_definitions_selected, $entity_definitions_unselected);

				return $entity_definitions;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'string' : 			return 'text';		// Text
					case 'reference' : 			return 'text';		// Lookup relationship
					case 'boolean' :			return 'checkbox';	// Checkbox
					case 'currency' :			return 'price';		// Currency
					case 'date' :				return 'datetime';	// Date
					case 'datetime' :			return 'datetime';	// Date/Time
					case 'email' :				return 'email';		// Email
					case 'double' :				return 'number';	// Double (Geo Location Lat / Lon)
					case 'location' :			return 'text';		// Geo Location
					case 'percent' :			return 'number';	// Percent
					case 'phone' :				return 'tel';		// Phone
					case 'picklist' :			return 'select';		// Picklist
					case 'multipicklist' : 		return 'select';	// Picklist multi
					case 'textarea' :			return 'textarea';	// Textarea
					case 'encryptedstring' :	return 'text';		// Encrypted string
					case 'time' :				return 'datetime';	// Time
					case 'url' :				return 'url';		// URL

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$meta_keys = array();
				$data_grid_meta_key = false;

				// Type
				$type = parent::get_object_value($field, 'type');

				switch($type) {

					case 'boolean' :

						// Select data grid type
						$data_grid_meta_key = 'data_grid_checkbox';

						// Label
						$label = parent::get_object_value($field, 'label');

						// Default value
						$default_value = parent::get_object_value($field, 'defaultValue');

						$picklist_object = new stdClass;
						$picklist_object->defaultValue = $default_value;
						$picklist_object->label = $label;
						$picklist_object->value = $label;

						// Set picklist values
						$picklist_values = array($picklist_object);

						// Settings (Support 2 columns)
						$meta_keys['checkbox_field_label'] = 1;

						break;

					case 'datetime' :

						$meta_keys['input_type_datetime'] = 'datetime-local';
						break;

					case 'date' :

						$meta_keys['input_type_datetime'] = 'date';
						break;

					case 'time' :

						$meta_keys['input_type_datetime'] = 'time';
						break;

					case 'picklist' :
					case 'multipicklist' :

						// Select data grid type
						$data_grid_meta_key = 'data_grid_select';

						// Get picklist values
						$picklist_values = parent::get_object_value($field, 'picklistValues');

						// Settings (Support 2 columns)
						$meta_keys['select_field_label'] = 1;

						// Settings for multi picklist
						if($type == 'multipicklist') {

							$meta_keys['multiple'] = 'on';
							$meta_keys['size'] = 10;
						}

						break;

					case 'string' :
					case 'textarea' :
					case 'encryptedstring' :

						// Length
						$length = intval(parent::get_object_value($field, 'length'));
						if($length > 0) {

							$meta_keys['max_length'] = $length;
						}

						break;

					case 'currency' :
					case 'double' :
					case 'percent' :

						// Scale (Precision)
						$scale = intval(parent::get_object_value($field, 'scale'));
						if($scale > 0) {

							$meta_keys['step'] = floatval('0.' . str_repeat('0', $scale - 1) . '1');
						}

						break;
				}

				if($data_grid_meta_key !== false) {

					// Get base meta
					$meta_keys_config = WS_Form_Config::get_meta_keys();
					if(!isset($meta_keys_config[$data_grid_meta_key])) { return false; }
					if(!isset($meta_keys_config[$data_grid_meta_key]['default'])) { return false; }

					$meta = $meta_keys_config[$data_grid_meta_key]['default'];

					$meta['columns'] = array(

						array('id' => 0, 'label' => __('Value', 'ws-form')),
						array('id' => 1, 'label' => __('Label', 'ws-form'))
					);

					// Build new rows
					$rows = array();
					$id = 1;
					foreach($picklist_values as $picklist_object) {

						$rows[] = array(

							'id'		=> $id,
							'default'	=> $picklist_object->defaultValue,
							'required'	=> '',
							'disabled'	=> '',
							'hidden'	=> '',
							'data'		=> array($picklist_object->value, $picklist_object->label)
						);

						$id++;
					}

					// Modify meta
					$meta['groups'][0]['rows'] = $rows;

					$meta_keys[$data_grid_meta_key] = $meta;
				}

				// Extra Type Info
				$extra_type_info = parent::get_object_value($field, 'extraTypeInfo');

				switch($extra_type_info) {

					case 'richtextarea' :

						global $wp_version;

						if(version_compare($wp_version, '4.8', '>=')) {

							$meta_keys['input_type_textarea'] = 'tinymce';
						}

						break;
				}

				return $meta_keys;

			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->dedupe = parent::get_config($config, 'action_' . $this->id . '_dedupe', '');
				$this->dedupe_field = parent::get_config($config, 'action_' . $this->id . '_dedupe_field', '');
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->hostname  					= WS_Form_Common::option_get('action_' . $this->id . '_hostname', 'login');
				$this->api_access_token 		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry 	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token 		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');
				$this->api_instance_url 		= WS_Form_Common::option_get('action_' . $this->id . '_api_instance_url', '');
				$this->api_objects 				= WS_Form_Common::option_get('action_' . $this->id . '_api_objects', $this->objects_default);
				if(!is_array($this->api_objects)) { $this->api_objects = array(); }

				// Set endpoint URL
				$this->api_endpoint = $this->api_instance_url . self::API_PATH;

				// Build endpoint using API key
				if(empty($this->api_access_token)) {

					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array(

					'Authorization' => 'OAuth ' . $this->api_access_token
				);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(39.000000, 69.000000)"><path d="M26.6 4.9c2.1-2.1 4.9-3.5 8.1-3.5 4.2 0 7.9 2.4 9.9 5.9 1.7-.8 3.6-1.2 5.6-1.2C57.8 6.1 64 12.3 64 20s-6.2 13.9-13.8 13.9c-.9 0-1.8-.1-2.7-.3-1.7 3.1-5 5.2-8.8 5.2-1.6 0-3.1-.4-4.4-1-1.7 4.1-5.8 7-10.6 7-4.9 0-9.2-3.1-10.8-7.5-.7.1-1.5.2-2.2.2C4.8 37.5 0 32.7 0 26.7c0-4 2.1-7.5 5.3-9.4-.7-1.5-1-3.2-1-4.9C4.3 5.6 9.9 0 16.7 0c4.1 0 7.7 1.9 9.9 4.9" fill="#00a1e0"/><path d="M9.3 23.2l.3-.7c0-.1.1-.1.2 0 .1 0 .1.1.2.1.8.5 1.5.5 1.7.5.6 0 .9-.3.9-.7 0-.4-.5-.6-1.2-.8h-.1c-.9-.2-1.8-.6-1.8-1.7 0-1 .8-1.8 2-1.8h.1c.7 0 1.4.2 1.9.5l.1.1c0 .1-.2.6-.3.7 0 .1-.2 0-.2 0-.4-.2-1.1-.4-1.7-.4-.5 0-.8.3-.8.6 0 .4.6.6 1.2.8h.1c.9.3 1.8.6 1.8 1.7s-.8 1.8-2.1 1.8c-.6 0-1.3-.1-1.9-.4-.1-.1-.2-.1-.4-.2 0 .1-.1 0 0-.1zm19.2 0l.3-.7c0-.1.1-.1.2 0 .1 0 .1.1.2.1.8.5 1.5.5 1.7.5.6 0 .9-.3.9-.7 0-.4-.5-.6-1.2-.8h-.1c-.9-.2-1.8-.6-1.8-1.7 0-1 .8-1.8 2-1.8h.1c.7 0 1.4.2 1.9.5l.1.1c0 .1-.2.6-.3.7 0 .1-.2 0-.2 0-.4-.2-1.1-.4-1.7-.4-.5 0-.8.3-.8.6 0 .4.6.6 1.2.8h.1c.9.3 1.8.6 1.8 1.7s-.8 1.8-2.1 1.8c-.6 0-1.3-.1-1.9-.4-.1-.1-.2-.1-.4-.2 0 .1-.1 0 0-.1zm14.2-3.3c.1.4.2.8.2 1.2 0 .4 0 .8-.2 1.2-.1.4-.3.7-.5 1-.2.3-.5.5-.8.6-.3.1-.7.2-1.2.2-.4 0-.8-.1-1.2-.2-.3-.1-.6-.4-.8-.6-.2-.3-.4-.6-.5-1-.1-.4-.2-.8-.2-1.2 0-.4 0-.8.2-1.2.1-.4.3-.7.5-1 .2-.3.5-.5.8-.6.3-.1.7-.2 1.2-.2.4 0 .8.1 1.2.2.3.1.6.4.8.6.2.3.4.6.5 1zM41.6 21c0-.6-.1-1.1-.3-1.5-.3-.3-.6-.5-1.1-.5-.5 0-.8.2-1.1.5-.2.3-.3.8-.3 1.5s.1 1.1.3 1.5c.2.3.6.5 1.1.5.5 0 .8-.2 1.1-.5.2-.3.3-.8.3-1.5zm10.1 1.9l.3.7v.1c-.4.2-1 .3-1.6.3-1 0-1.7-.3-2.2-.8-.5-.5-.7-1.3-.7-2.2 0-.4 0-.8.2-1.2.1-.4.3-.7.5-1 .2-.3.5-.5.9-.6.3-.1.8-.2 1.2-.2h.8c.2 0 .6.1.7.2 0 0 .1 0 .1.1-.1.3-.2.5-.3.7 0 .1-.1.1-.1.1-.3-.1-.7-.1-1.1-.1-.5 0-1 .2-1.2.5-.3.3-.4.8-.4 1.5s.2 1.2.5 1.5c.3.3.7.5 1.3.5h.6c.2 0 .3-.1.5-.1-.1-.1 0-.1 0 0zm5.6-3.2c.2.8.1 1.6.1 1.6 0 .1-.1.1-.1.1h-3.7c0 .6.1 1 .4 1.2.3.3.7.4 1.3.4.9 0 1.3-.2 1.5-.3 0 0 .1 0 .1.1l.2.7v.2c-.2.1-.8.4-1.9.4-.5 0-1-.1-1.4-.2s-.7-.3-.9-.6c-.2-.3-.4-.6-.5-.9-.1-.4-.2-.8-.2-1.2 0-.4 0-.8.2-1.2.1-.4.3-.7.5-1 .2-.3.5-.5.9-.6.3-.2.8-.2 1.2-.2.4 0 .8.1 1.1.2.2.1.5.3.7.5.2 0 .4.4.5.8zm-3.7.7h2.6c0-.3-.1-.6-.2-.9-.2-.3-.5-.5-1-.5s-.8.2-1.1.5l-.3.9zm-26-.7c.2.8.1 1.6.1 1.6 0 .1-.1.1-.1.1h-3.7c0 .6.1 1 .4 1.2.3.3.7.4 1.3.4.9 0 1.3-.2 1.5-.3 0 0 .1 0 .1.1l.2.7v.2c-.2.1-.8.4-1.9.4-.5 0-1-.1-1.4-.2s-.7-.3-.9-.6c-.2-.3-.4-.6-.5-.9-.1-.4-.2-.8-.2-1.2 0-.4 0-.8.2-1.2.1-.4.3-.7.5-1 .2-.3.5-.5.9-.6.3-.2.8-.2 1.2-.2.4 0 .8.1 1.1.2.2.1.5.3.7.5.2 0 .4.4.5.8zm-3.7.7h2.7c0-.3-.1-.6-.2-.9-.2-.3-.5-.5-1-.5s-.8.2-1.1.5c-.3.3-.3.6-.4.9zm-6.5-.2s.3 0 .6.1v-.1c0-.5-.1-.7-.3-.9s-.5-.3-.9-.3c0 0-.9 0-1.6.4s-.1 0-.1 0l-.3-.7v-.1c.3-.3 1.1-.4 1.1-.4.3 0 .7-.1 1-.1.7 0 1.3.2 1.7.5.4.3.6.9.6 1.6v3.4s0 .1-.1.1c0 0-.1 0-.3.1-.1 0-.6.1-.9.2h-1.2c-.4 0-.7 0-1-.1s-.5-.2-.8-.3c-.2-.1-.4-.3-.5-.6-.1-.2-.2-.5-.2-.8s.1-.6.2-.9c.1-.2.3-.4.5-.6.2-.2.5-.3.8-.4.3-.1.6-.1.9-.1h.8zm-1.5 2.7s.3.3 1.1.2c.5 0 1-.1 1-.1v-1.7s-.5-.1-1-.1c-.8 0-1.1.3-1.1.3-.2.1-.3.4-.3.7 0 .2 0 .4.1.5 0 0 0 .1.2.2zm31.4-4.6c0 .1-.2.6-.3.8 0 .1-.1.1-.1.1h-.4c-.1 0-.3 0-.5.1-.2 0-.3.1-.5.3-.1.1-.2.3-.3.5-.1.2-.1.6-.1 1v2.8c0 .1 0 .1-.1.1h-1c-.1 0-.1 0-.1-.1v-5.5c0-.1 0-.1.1-.1h1c.1 0 .1 0 .1.1v.4c.1-.2.4-.4.6-.5.2-.1.5-.2 1-.1.2 0 .6.1.6.1 0-.1 0-.1 0 0zm-9.2-2.5s.1 0 .1.1l-.3.8s0 .1-.2 0h-.9c-.1 0-.2.1-.3.2-.1.1-.2.2-.3.4l-.2.8h1.2c.1 0 .1 0 .1.1l-.1.8c0 .1-.1.1-.1.1h-1.2l-.8 4.8c-.1.5-.2.9-.3 1.3-.1.3-.3.6-.5.8-.2.2-.4.4-.7.5-.2.1-.6.1-.9.1h-.5c-.1 0-.2 0-.3-.1 0 0-.1-.1 0-.1 0-.1.2-.7.3-.8 0-.1.1 0 .1 0 .1 0 .1 0 .2.1h.3c.2 0 .3 0 .4-.1.1 0 .2-.1.3-.3.1-.1.2-.3.3-.5.1-.2.1-.5.2-.9l.8-4.7H34c-.1 0-.1 0-.1-.1l.1-.8c0-.1.1-.1.1-.1h.8V18c.1-.7.4-1.3.7-1.7.4-.4.9-.6 1.6-.6h.5c.2 0 .3 0 .4.1zm-16.6 8c0 .1 0 .1-.1.1h-1c-.1 0-.1 0-.1-.1v-7.9s0-.1.1-.1h1c.1 0 .1 0 .1.1v7.9z" fill="#fff"/></g>';

				return $svg_logo;
			}
		}
	});