<?php

	/**
	 * @link              https://wsform.com/knowledgebase/sendinblue/
	 * @since             1.0.0
	 * @package           WS_Form_Sendinblue
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Sendinblue
	 * Plugin URI:        https://wsform.com/knowledgebase/sendinblue/
	 * Description:       Sendinblue add-on for WS Form PRO
	 * Version:           1.0.1
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-sendinblue-v3
	 */

	Class WS_Form_Add_On_Sendinblue_V3 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.136';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Sendinblue_V3();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-sendinblue-v3'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_sendinblue = new WS_Form_Add_On_Sendinblue_V3();	

	register_activation_hook(__FILE__, array($wsf_add_on_sendinblue, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Sendinblue_V3 extends WS_Form_Action {

			public $id = 'sendinbluev3';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Contacts';

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_key;
			public $list_id = false;
			public $status;
			public $update_enabled;
			public $opt_in_field;
			public $field_mapping;
			public $custom_mapping;

			// Constants
			const DEFAULT_STATUS = 'subscribed';
			const DEFAULT_EMAIL_TYPE = 'html';
			const EMAIL_ATTRIBUTE = 'email';
			const SMS_ATTRIBUTE = 'SMS';
			const LISTS_FETCH_COUNT = 50;
			const TAG_CATEGORIES_FETCH_COUNT = 10;
			const TAGS_FETCH_COUNT = 10;
			const WS_FORM_LICENSE_ITEM_ID = 28548;
			const WS_FORM_LICENSE_NAME = 'Sendinblue add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.1';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Sendinblue', 'ws-form-sendinblue-v3');

				// Set label for actions pull down
				$this->label_action = __('Add to Sendinblue', 'ws-form-sendinblue-v3');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-sendinblue-v3') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-sendinblue-v3') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

					break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_key', '');
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Sendinblue account  <a href="%s">here</a>.', 'ws-form-sendinblue-v3'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Reset list ID (For multiple actions)
				$this->list_id = false;

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-sendinblue-v3'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Get attribute to action_type array
				$list_fields = self::get_list_fields();
				$attribute_to_action_type = array();
				foreach($list_fields as $list_field) {

					$attribute_to_action_type[$list_field['id']] = (isset($list_field['action_type']) ? $list_field['action_type'] : false);
				}

				// Process field mapping
				$email_address = false;
				$email_address_mandatory = true;
				$attributes = array();

				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					if($field_id == '') { continue; }

					$attribute = $field_map['action_' . $this->id . '_list_fields'];
					if($attribute == '') { continue; }

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($submit_value === false) { continue; }

					// Sendinblue cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($attribute) {

						case self::EMAIL_ATTRIBUTE :

							$email_address = $submit_value;
							continue 2;

						case self::SMS_ATTRIBUTE :

							$email_address_mandatory = false;
							break;
					}

					// Process by type
					$submit_value = self::attribute_to_action_type($submit_value, $attribute_to_action_type, $attribute);

					// Map attribute
					$attributes[$attribute] = $submit_value;
				}

				foreach($this->custom_mapping as $custom_map) {

					// Get value
					$custom_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$custom_value = WS_Form_Common::parse_variables_process($custom_value, $form, $submit, 'text/plain');
					if($custom_value == '') { continue; }

					// Get attribute
					$attribute = $custom_map['action_' . $this->id . '_list_fields'];
					if($attribute == '') { continue; }

					// Array conversion
					if(is_array($custom_value)) { $custom_value = implode(',', $custom_value); }

					// Process by type
					$custom_value = self::attribute_to_action_type($custom_value, $attribute_to_action_type, $attribute);

					// Map attribute
					$attributes[$attribute] = $custom_value;
				}

				// Check email address is configured
				if(
					($email_address === false) &&

					(
						($this->status == 'unsubscribed') ||

						(
							($this->status == 'subscribed') &&
							$email_address_mandatory
						)
					)

				) { return self::error(__('Email address not mapped', 'ws-form-sendinblue-v3')); }

				// Build post data
				$data = array();
				if(count($attributes) > 0) { $data['attributes'] = $attributes; }

				// Process based upon status
				switch($this->status) {

					case 'unsubscribed' :

						$api_path = sprintf('contacts/%s', urlencode($email_address));
						$api_method = 'PUT';
						$data['unlinkListIds'] = array(intval($this->list_id));

						break;

					default :

						if($email_address !== false) { $data['email'] = $email_address; }
						$data['listIds'] = array(intval($this->list_id));
						$data['updateEnabled'] = ($this->update_enabled == 'on');
						$api_path = 'contacts';
						$api_method = 'POST';

						break;
				}

				// Data for API call
				$data_json = json_encode($data);

				// Make API call
				$api_response = parent::api_call($this->api_endpoint, $api_path, $api_method, $data_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 201 :

						self::success(sprintf(__('Successfully added to list: %s', 'ws-form-sendinblue-v3'), $this->list_id));
						return true;

					case 204 :

						self::success(__('Successfully updated contact', 'ws-form-sendinblue-v3'));
						return true;

					default :

						// Error
						return self::error(self::get_api_error_detail($api_response['response']));
				}
			}

			// Process by type
			public function attribute_to_action_type($value, $attribute_to_action_type, $attribute) {

				if(!isset($attribute_to_action_type[$attribute])) { return $value; }

				switch($attribute_to_action_type[$attribute]) {

					case 'boolean' :

						$value = ($value == '1');
						break;

					case 'date' :

						$value = date('Y-m-d', strtotime($value));
						break;
				}

				return $value;
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-sendinblue-v3')); }

				// Get member
				$api_response = parent::api_call($this->api_endpoint, sprintf('contacts/%s', urlencode($user_email)), 'GET', null, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-sendinblue-v3')); }

				// Get attribute to action_type array
				$list_fields = self::get_list_fields();
				$attribute_to_action_type = array();
				foreach($list_fields as $list_field) {

					$attribute_to_action_type[$list_field['id']] = (isset($list_field['action_type']) ? $list_field['action_type'] : false);
				}

				// Process fields
				$fields_return = array(

					self::EMAIL_ATTRIBUTE	=>	parent::get_object_value($api_response_decoded, self::EMAIL_ATTRIBUTE),
				);

				$attributes = parent::get_object_value($api_response_decoded, 'attributes');
				if($attributes !== false) {

					$attributes = get_object_vars($attributes);
					foreach($attributes as $attribute => $value) {

						if(isset($attribute_to_action_type[$attribute])) {

							switch($attribute_to_action_type[$attribute]) {

								case 'date' :

									$value = $value . ' 00:00:00';
									break;
							}
						}

						if(is_null($value)) { continue; }

						$fields_return[$attribute] = (string) $value;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;

					do {

						// Params
						$params = array(

							'limit'		=> self::LISTS_FETCH_COUNT,
							'offset'	=> $total_lists_retrieved
						);

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'contacts/lists', 'GET', $params, self::get_api_headers());

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-sendinblue-v3')); }
						if(!isset($api_response_decoded->lists)) { return self::error(__('Lists not found in API response data', 'ws-form-sendinblue-v3')); }

						// Get total field count
						if(isset($api_response_decoded->count)) {

							$total_lists = $api_response_decoded->count;
						}

						// Increment read counter
						$total_lists_retrieved += count($api_response_decoded->lists);

						foreach($api_response_decoded->lists as $list) {

							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'), 
								'label' => 			parent::get_object_value($list, 'name'), 
								'field_count' => 	false, 
								'record_count' => 	parent::get_object_value($list, 'totalSubscribers')
							);
						}

					} while ($total_lists_retrieved < $total_lists);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'contacts/lists/' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_list = json_decode($api_response['response']);
					if(is_null($api_response_list)) { return self::error(__('Invalid API response data', 'ws-form-sendinblue-v3')); }

					// Build list
					$list = array(

						'label' => parent::get_object_value($api_response_list, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields');

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();

					// Get list fields
					$api_response = parent::api_call($this->api_endpoint, 'contacts/attributes', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-sendinblue-v3')); }
					if(!isset($api_response_decoded->attributes)) { return self::error(__('Attributes not found in API response data', 'ws-form-sendinblue-v3')); }

					// Get the fields
					$api_list_fields = $api_response_decoded->attributes;

					// Process fields
					$sort_index = 1;
					foreach($api_list_fields as $field) {

						$no_add = false;

						$action_category = parent::get_object_value($field, 'category');
						$action_type = parent::get_object_value($field, 'type');
						$field_name = parent::get_object_value($field, 'name');

						switch($action_category) {

							case 'normal':

								if($field_name == self::SMS_ATTRIBUTE) { $no_add = true; }
								break;

							case 'category':

								$action_type = $field->type = 'category';
								break;

							case 'transactional' :

								$no_add = true;
								break;

							default:

								continue 2;
						}

						$type = self::action_field_type_to_ws_form_field_type($action_type);

						$list_fields[] = array(

							'id' => 			$field_name,
							'label' => 			$field_name, 
							'label_field' => 	$field_name, 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		false, 
							'default_value' => 	false,
							'pattern' => 		false,
							'placeholder' => 	false,
							'input_mask' => 	false,
							'help' => 			false, 
							'sort_index' => 	$sort_index++,
							'visible' =>		true,
							'meta' => 			self::action_field_to_ws_form_meta_keys($field),
							'no_add' =>			$no_add
						);
					}

					// Add email
					$list_fields[] = array(

						'id' => 			self::EMAIL_ATTRIBUTE, 
						'label' => 			'EMAIL ADDRESS', 
						'label_field' => 	'EMAIL ADDRESS', 
						'type' => 			'email',
						'action_type' => 	'email',
						'required' => 		true,
						'default_value' => 	'', 
						'pattern' => 		false, 
						'placeholder' => 	false, 
						'input_mask' =>		false,
						'help' => 			'', 
						'sort_index' => 	0,
						'visible' =>		true
					);

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields', $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field',
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-sendinblue-v3'); }

				if(!isset($api_response_decoded->message)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-sendinblue-v3'); }

				return $api_response_decoded->message;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_status',
						'action_' . $this->id . '_update_enabled',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-sendinblue-v3') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-sendinblue-v3')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'						=>	__('Sendinblue List', 'ws-form-sendinblue-v3'),
						'type'						=>	'select',
						'help'						=>	__('Select the Sendinblue list to associate this form with.', 'ws-form-sendinblue-v3'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-sendinblue-v3'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Status (subscribed, unsubscribed)
					'action_' . $this->id . '_status'	=> array(

						'label'						=>	__('Status', 'ws-form-sendinblue-v3'),
						'type'						=>	'select',
						'help'						=>	__("Set contact's list status to this.", 'ws-form-sendinblue-v3'),
						'options'					=>	array(

							array('value' => 'subscribed', 'text' => __('Subscribed', 'ws-form-sendinblue-v3')),
							array('value' => 'unsubscribed', 'text' => __('Unsubscribed', 'ws-form-sendinblue-v3'))
						),
						'default'					=>	self::DEFAULT_STATUS
					),

					// VIP
					'action_' . $this->id . '_update_enabled'	=> array(

						'label'						=>	__('Enable Update', 'ws-form-sendinblue-v3'),
						'type'						=>	'checkbox',
						'help'						=>	__('If checked, existing contacts will be updated.', 'ws-form-sendinblue-v3'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'==',
								'meta_key'		=>	'action_' . $this->id . '_status',
								'meta_value'	=>	'subscribed'
							)
						)
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-sendinblue-v3'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-sendinblue-v3'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-sendinblue-v3')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-sendinblue-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Sendinblue attributes.', 'ws-form-sendinblue-v3'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Sendinblue Field', 'ws-form-sendinblue-v3'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-sendinblue-v3'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// Custom field mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-sendinblue-v3'),
						'type'						=>	'repeater',
						'help'						=>	__('Map custom values to Sendinblue attributes.', 'ws-form-sendinblue-v3'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_fields_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom field mapping - Value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'						=>	__('Value', 'ws-form-sendinblue-v3'),
						'type'						=>	'text'
					)
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-sendinblue-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-sendinblue-v3'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Sendinblue add-on for WS Form PRO license key here.', 'ws-form-sendinblue-v3'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-sendinblue-v3'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'		=>	__('Sendinblue API Key', 'ws-form-sendinblue-v3'),
							'type'		=>	'text',
							'default'	=>	'',
							'help'		=>	__('<a href="https://wsform.com/knowledgebase/sendinblue/" target="_blank">Learn more</a>', 'ws-form-sendinblue-v3')
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Sendinblue Status', 'ws-form-sendinblue-v3'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Enter your <a href=\"https://account.sendinblue.com/advanced/api\" target=\"blank\">Sendinblue API key</a> above, or <a href=\"https://app.sendinblue.com/account/register\" target=\"blank\">register for a new account</a>.", 'ws-form-sendinblue-v3')
						)
					)
				);

				return $options;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'text' : return 'text';
					case 'date' : return 'datetime';
					case 'float' : return 'number';
					case 'category' : return 'select';
					case 'boolean' : return 'select';

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$type = parent::get_object_value($field, 'type');

				// Get WS Form meta configurations for action field types
				switch($type) {

					// Build data grids for select
					case 'category' :
					case 'boolean' :

						switch($type) {

							case 'category' :

								// Get choices
								$choices = parent::get_object_value($field, 'enumeration');
								if($choices === false) { return false; }

								break;

							case 'boolean' :

								$choices = array(

									(object) array(

										'value' => 1,
										'label' => 'Yes'
									),

									(object) array(

										'value' => 0,
										'label' => 'No'
									)
								);

								break;
						}

						// Get base meta
						$meta_keys = WS_Form_Config::get_meta_keys();
						if(!isset($meta_keys['data_grid_select'])) { return false; }
						if(!isset($meta_keys['data_grid_select']['default'])) { return false; }

						$meta = $meta_keys['data_grid_select']['default'];

						// Build new rows
						$rows = array();
						$id = 1;
						foreach($choices as $choice) {

							$rows[] = array(

								'id'		=> $id,
								'default'	=> '',
								'required'	=> '',
								'disabled'	=> '',
								'hidden'	=> '',
								'data'		=> array(parent::get_object_value($choice, 'value'), parent::get_object_value($choice, 'label'))
							);

							$id++;
						}

						// Modify meta
						$meta['groups'][0]['rows'] = $rows;

						// Columns
						$meta['columns'] = array(

							array('id' => 0, 'label' => __('Value', 'ws-form')),
							array('id' => 1, 'label' => __('Label', 'ws-form'))
						);

						return(array('data_grid_select' => $meta, 'select_field_label' => 1));

					default :

						return false;
				}
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->status = parent::get_config($config, 'action_' . $this->id . '_status', self::DEFAULT_STATUS);
				$this->update_enabled = parent::get_config($config, 'action_' . $this->id . '_update_enabled');
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_key = WS_Form_Common::option_get('action_' . $this->id . '_api_key', '');

				if(empty($this->api_key)) {

					$this->api_endpoint = false;
					$this->configured = false;

				} else {

					$this->api_endpoint = 'https://api.sendinblue.com/v3/';
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Convert email to action record ID
			public function ws_form_email_to_action_record_id($email_address) {

				return md5($email_address);
			}
		
			// Get API headers
			public function get_api_headers() {

				$headers = array('api-key' => $this->api_key);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(44.000000, 62.000000)"><path d="M43.6 35.8c-1.3 2.2-3.4 3.9-5.8 4.6.4-1.2.5-2.5.5-3.7 0-6.3-4.3-11.5-10.1-13 1.2-1.2 2.7-2 4.3-2.5 2.5-.7 5.2-.3 7.5 1 4.7 2.8 6.3 8.8 3.6 13.6M25 46.6c-2.6 0-5-1-6.9-2.8 1.2-.3 2.4-.8 3.5-1.4 5.4-3.1 7.8-9.5 6.1-15.3 4.1 1.2 7.2 5 7.2 9.5 0 5.5-4.4 10-9.9 10M6.4 35.8c-1.3-2.2-1.6-4.9-1-7.4.9.9 1.9 1.7 2.9 2.3 2 1.2 4.3 1.8 6.6 1.8 3.6 0 7.1-1.5 9.6-4.1 1 4.2-.7 8.7-4.6 11-1.5.9-3.2 1.3-4.9 1.3-3.5.1-6.8-1.8-8.6-4.9m0-21.6c1.3-2.2 3.4-3.9 5.8-4.6-.4 1.2-.5 2.5-.5 3.7 0 6.3 4.3 11.5 10.1 13-3.1 3-7.9 3.7-11.8 1.5-2.3-1.3-3.9-3.5-4.6-6-.6-2.6-.3-5.3 1-7.6M25 3.4c2.6 0 5 1 6.9 2.8-1.2.3-2.4.8-3.5 1.4-3.1 1.8-5.3 4.7-6.2 8.1-.6 2.4-.6 4.8.1 7.2-4.1-1.2-7.2-5-7.2-9.5 0-5.5 4.5-10 9.9-10m18.6 10.8c1.3 2.2 1.6 4.9 1 7.4-.9-.9-1.8-1.7-2.9-2.3C38.6 17.5 35 17 31.6 18c-2.3.6-4.5 1.9-6.1 3.7-.4-1.6-.4-3.3 0-5 .7-2.6 2.3-4.7 4.6-6 2.3-1.3 4.9-1.7 7.5-1 2.6.6 4.7 2.2 6 4.5m2.9-1.7c-1.8-3.1-4.6-5.3-8.1-6.2-.8-.2-1.6-.4-2.5-.4C33.6 2.2 29.4 0 25 0c-4.6 0-8.6 2.3-11 5.9-4.3.3-8.3 2.8-10.5 6.6-1.8 3.1-2.2 6.7-1.3 10.1.2.8.5 1.6.8 2.4-1.9 4-1.8 8.7.4 12.5 2.3 4 6.3 6.3 10.5 6.6 2.5 3.6 6.6 5.9 11 5.9 4.6 0 8.6-2.3 11-5.9 4.4-.3 8.3-2.8 10.5-6.6 2.3-4 2.3-8.6.4-12.5 2.1-4 2-8.7-.3-12.5" fill-rule="evenodd" clip-rule="evenodd" fill="#4888c8"/></g>';

				return $svg_logo;
			}
		}
	});