<?php

	/**
	 * @link              https://wsform.com/knowledgebase/user-management/
	 * @since             1.0.0
	 * @package           WS_Form_Slack
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Slack
	 * Plugin URI:        https://wsform.com/knowledgebase/user-management/
	 * Description:       Slack add-on for WS Form PRO
	 * Version:           1.1.2
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-slack
	 */

	Class WS_Form_Add_On_Slack {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.5.13';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Slack();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-slack'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_slack = new WS_Form_Add_On_Slack();	

	register_activation_hook(__FILE__, array($wsf_add_on_slack, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Slack extends WS_Form_Action {

			public $id = 'slack';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Users';

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_access_token_user;
			private $api_access_token_bot;

			public $team_name;
			public $team_id;

			public $list_id = false;
			public $attachment = false;
			public $attachment_title = '';
			public $attachment_title_link = '';
			public $attachment_fallback = '';
			public $attachment_pretext = '';
			public $attachment_author_name = '';
			public $attachment_author_link = '';
			public $attachment_author_icon = '';
			public $attachment_text = '';
			public $attachment_color = '';
			public $attachment_buttons;
			public $attachment_field_mapping;
			public $attachment_image_field_id;
			public $attachment_thumb_field_id;
			public $attachment_footer = '';
			public $attachment_footer_icon = '';

			// Constants
			const LISTS_FETCH_COUNT = 100;
			const LISTS_FETCH_COUNT_MAX = 1000;
			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const WS_FORM_LICENSE_ITEM_ID = 2519;
			const WS_FORM_LICENSE_NAME = 'Slack add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.1.2';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';
			const SLACK_FIELD_SHORT_THRESHOLD = 20;

			public function __construct() {

				// Set label
				$this->label = __('Slack', 'ws-form-slack');

				// Set label for actions pull down
				$this->label_action = __('Add to Slack', 'ws-form-slack');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token - Bot
				$this->api_access_token_bot = WS_Form_Common::get_query_var($this->id . '_access_token_bot', false);
				if($this->api_access_token_bot === false) { return false; }
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_bot', $this->api_access_token_bot);

				// Save access token - User
				$this->api_access_token_user = WS_Form_Common::get_query_var($this->id . '_access_token_user', false);
				if($this->api_access_token_user === false) { return false; }
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_user', $this->api_access_token_user);

				// Read account info
				self::account_info();

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-slack')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;

					case 'action_' . $this->id . '_team_name' :

						$value = $this->team_name;
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-slack') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-slack') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token_bot && $this->api_access_token_user) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-slack') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/slack';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-slack'));
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_bot', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_user', '');
				WS_Form_Common::option_set('action_' . $this->id . '_team_id', '');
				WS_Form_Common::option_set('action_' . $this->id . '_team_name', '');
				WS_Form_Common::option_set('action_' . $this->id . '_lists', false);
			}

			// Get account info
			public function account_info() {

				self::load_config_plugin();

				$params = array(

					'token' => $this->api_access_token_user
				);

				// Get account info
				$api_response = parent::api_call($this->api_endpoint, 'team.info', 'GET', $params);

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-slack')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					$message = sprintf(__('%s API Error: %s', 'ws-form-slack'), $this->label, $api_response_decoded->message);

					WS_Form_Common::admin_message_push($message, 'notice-error', false);

					return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Check that we got properties
				if(
					!isset($api_response_decoded->ok) ||
					!$api_response_decoded->ok

				) { return false; }

				// Save settings
				WS_Form_Common::option_set('action_' . $this->id . '_team_id', $api_response_decoded->team->id);
				WS_Form_Common::option_set('action_' . $this->id . '_team_name', $api_response_decoded->team->name);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Slack account <a href="%s">here</a>.', 'ws-form-'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Text
				$text = WS_Form_Common::parse_variables_process(trim($this->text), $form, $submit, 'text/plain');

				// Build params
				$params = array(

					'channel'	=> $this->list_id,
					'text'		=> $text
				);

				// Add button to submission in admin
				if($this->attachment) {

					$attachment = array();

					// Title
					if($this->attachment_title) {

						$attachment['title'] = WS_Form_Common::parse_variables_process($this->attachment_title, $form, $submit, 'text/plain');
						if($this->attachment_title_link) { $attachment['title_link'] = WS_Form_Common::parse_variables_process($this->attachment_title_link, $form, $submit, 'text/plain'); }
					}

					// Text
					if($this->attachment_fallback) { $attachment['fallback'] = WS_Form_Common::parse_variables_process($this->attachment_fallback, $form, $submit, 'text/plain'); }
					if($this->attachment_pretext) { $attachment['pretext'] = WS_Form_Common::parse_variables_process($this->attachment_pretext, $form, $submit, 'text/plain'); }
					if($this->attachment_text) { $attachment['text'] = WS_Form_Common::parse_variables_process($this->attachment_text, $form, $submit, 'text/plain'); }

					// Color
					if($this->attachment_color) { $attachment['color'] = WS_Form_Common::parse_variables_process($this->attachment_color, $form, $submit, 'text/plain'); }

					// Author
					if($this->attachment_author_name) {

						$attachment['author_name'] = WS_Form_Common::parse_variables_process($this->attachment_author_name, $form, $submit, 'text/plain');
						if($this->attachment_author_link) { $attachment['author_link'] = WS_Form_Common::parse_variables_process($this->attachment_author_link, $form, $submit, 'text/plain'); }
						if($this->attachment_author_icon) { $attachment['author_icon'] = WS_Form_Common::parse_variables_process($this->attachment_author_icon, $form, $submit, 'text/plain'); }
					}

					// Footer
					if($this->attachment_footer) {

						$attachment['footer'] = WS_Form_Common::parse_variables_process($this->attachment_footer, $form, $submit, 'text/plain');
						if($this->attachment_footer_icon) { $attachment['footer_icon'] = WS_Form_Common::parse_variables_process($this->attachment_footer_icon, $form, $submit, 'text/plain'); }
					}

					// Actions
					$actions = array();

					// Actions - Buttons
					if(count($this->attachment_buttons) > 0) {

						foreach($this->attachment_buttons as $action) {

							$action_text = WS_Form_Common::parse_variables_process($action['action_' . $this->id . '_attachment_action_text'], $form, $submit, 'text/plain');
							$action_style = WS_Form_Common::parse_variables_process($action['action_' . $this->id . '_attachment_action_style'], $form, $submit, 'text/plain');
							$action_url = WS_Form_Common::parse_variables_process($action['action_' . $this->id . '_attachment_action_url'], $form, $submit, 'text/plain');

							$actions[] = array(

								'type'	=>	'button',
								'text'	=>	$action_text,
								'style'	=>	$action_style,
								'url'	=>	$action_url
							);
						}
					}
					if(count($actions) > 0) { $attachment['actions'] = $actions; }

					// Fields
					$fields = array();
					if(count($this->attachment_field_mapping) > 0) {

						$field_lookup = WS_Form_Common::get_fields_from_form($form);

						foreach($this->attachment_field_mapping as $field_map) {

							$field_id = $field_map['ws_form_field'];

							// Get submit value
							$field_title = isset($field_lookup[$field_id]) ? $field_lookup[$field_id]->label : '';
							$field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, '');
							if(is_array($field_value)) { $field_value = implode(', ', $field_value); }
							$field_short = (strlen($field_value) <= self::SLACK_FIELD_SHORT_THRESHOLD);

							if(
								($field_title !== '') &&
								($field_value !== '')
							) {

								$fields[] = array(

									'title'	=> $field_title,
									'value'	=> $field_value,
									'short'	=> $field_short
								);
							}
						}
					}
					if(count($fields) > 0) { $attachment['fields'] = $fields; }

					// Build files
					if(
						(intval($this->attachment_image_field_id) > 0) ||
						(intval($this->attachment_thumb_field_id) > 0)
					) {

						// Create directory for storing files
						$upload_dir = wp_upload_dir();
						$upload_path = $this->id . $upload_dir['subdir'] . (($submit->id > 0) ? '/' . $submit->id : '');
						$upload_dir_create = WS_Form_Common::upload_dir_create($upload_path);
						if($upload_dir_create['error']) { return self::error($upload_dir_create['message']); }

						// Set all paths
						$upload_url = $upload_dir['baseurl'] . '/' . $upload_dir_create['path'];
						$upload_dir = $upload_dir_create['dir'];
					}

					// image_url
					$image_url = false;
					if(intval($this->attachment_image_field_id) > 0) {

						// Process first file
						$files = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->attachment_image_field_id, '');

						if(is_array($files) && isset($files[0])) {

							$file = $files[0];

							$image_url = self::process_attachment_file($file, $upload_dir, $upload_url);
							if($image_url !== false) { $attachment['image_url'] = $image_url; }
						}
					}

					// thumb_url
					if(intval($this->attachment_thumb_field_id) > 0) {

						if($this->attachment_image_field_id != $this->attachment_thumb_field_id) {

							// Process first file
							$files = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->attachment_thumb_field_id, '');

							if(is_array($files) && isset($files[0])) {

								$file = $files[0];

								$thumb_url = self::process_attachment_file($file, $upload_dir, $upload_url);
								if($thumb_url !== false) { $attachment['thumb_url'] = $thumb_url; }
							}

						} else {

							if($image_url !== false) {

								$attachment['thumb_url'] = $image_url;
							}
						}
					}

					if(count($attachment) > 0) { $params['attachments'] = array($attachment); }
				}

				// POST
				$api_response = parent::api_call($this->api_endpoint, 'chat.postMessage', 'POST', json_encode($params), self::get_api_headers_bot(), 'basic', false, false, 'application/json', 'application/json; charset=utf-8');

				// Process response
				switch($api_response['http_code']) {

					case 200 :

						// Process API response
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-slack')); }

						// Check ok variable
						if(!$api_response_decoded->ok) {

							// Check error
							switch($api_response_decoded->error) {

								case 'channel_not_found' :

									return self::error(__("The WS Form app has not been added to this private channel. Please add this app to the private channel by clicking on 'Conversation settings' (gear icon at the top of the channel), then 'Add App', then add WS Form.", 'ws-form')); 

								default :

									return self::error(self::get_api_error_detail($api_response_decoded)); 
							}

						}

						self::success(__('Successfully posted chat message to channel: ', 'ws-form-slack') . $this->list_id);
						return true;

					case 400 :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));

					case 500 :

						// Server error
						return self::error('500 Server Error');
				}
			}

			// Process attachment file
			public function process_attachment_file($file, $upload_dir, $upload_url) {

				// Check submit file data
				if(
					!is_array($file) ||
					!isset($file['name']) ||
					!isset($file['hash']) ||
					!isset($file['type']) ||
					!isset($file['size']) ||
					!isset($file['path'])

				) { return false; }

				// Check MIME type is compatible with allowable image file MIME types
				if(!in_array($file['type'], array(

					'image/jpg',
					'image/jpeg',
					'image/png',
					'image/gif',
					'image/bmp'

				))) { return false; }

				// File name of submitted file
				$file_name_source = wp_upload_dir()['basedir'] . '/' . $file['path'];

				// Build destinateion
				$file_name_destination = $upload_dir . '/' . $file['name'];

				// Create public version of file for Slack
				if(!copy($file_name_source, $file_name_destination)) { return false; }

				// Add file to email_attachments
				$image_url = $upload_url . '/' . $file['name'];

				return $image_url;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Load configuration
					self::load_config();

					// Initialize
					$total_lists_retrieved = 0;
					$cursor = false;

					do {

						// Params
						$params = array(

							'token'				=> $this->api_access_token_user,
							'limit'				=> self::LISTS_FETCH_COUNT,
							'exclude_members'	=> true,
							'exclude_archived'	=> true,
							'types'				=> 'public_channel,private_channel'
						);

						// Cursor
						if($cursor !== false) { $params['cursor'] = $cursor; }

						// Get lists
						$api_response = parent::api_call($this->api_endpoint, 'conversations.list', 'GET', $params);

						// Check status code
						if($api_response['http_code'] != 200) { return self::error(sprintf(__('HTTP Status Code: %s', 'ws-form-slack'), $api_response['http_code'])); }

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-slack')); }

						// Check ok variable
						if(!$api_response_decoded->ok) { return self::error(self::get_api_error_detail($api_response_decoded)); }

						// Get total lists
						$total_lists = count($api_response_decoded->channels);

						if(
							isset($api_response_decoded->response_metadata->next_cursor) &&
							$api_response_decoded->response_metadata->next_cursor != ''

						) { $cursor = $api_response_decoded->response_metadata->next_cursor; } else { $cursor = false; }

						// Increment read counter
						$total_lists_retrieved += $total_lists;

						foreach($api_response_decoded->channels as $list) {

							// Build list
							$lists[] = array(

								'id' => 			parent::get_object_value($list, 'id'),
								'label' => 			'#' . parent::get_object_value($list, 'name'),
								'field_count' => 	false,
								'record_count' => 	parent::get_object_value($list, 'num_members')
							);
						}

					} while (

						($cursor !== false) &&
						($total_lists_retrieved < self::LISTS_FETCH_COUNT_MAX)
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Params
					$params = array(

						'token'		=> $this->api_access_token_user,
						'channel'	=> $this->list_id
					);

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'conversations.info', 'GET', $params);

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(sprintf(__('HTTP Status Code: %s', 'ws-form-slack'), $api_response['http_code'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-slack')); }

					// Check ok variable
					if(!$api_response_decoded->ok) { return self::error(self::get_api_error_detail($api_response_decoded)); }

					// Build list
					$list = array(

						'label' => '#' . parent::get_object_value($api_response_decoded->channel, 'name')
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Fixed fields for slack form
					$fields = array(

						(object) array('id' => 'text', 'name' => 'Text', 'action_type' => 'textarea', 'type' => 'textarea', 'required' => true, 'display_order' => 1)
					);

					// Process fields
					$sort_index_offset = 0;
					foreach($fields as $field) {

						$type = parent::get_object_value($field, 'type');
						$action_type = parent::get_object_value($field, 'action_type');

						$list_fields[] = array(

							'id' => 			parent::get_object_value($field, 'id'),
							'label' => 			parent::get_object_value($field, 'name'), 
							'label_field' => 	parent::get_object_value($field, 'name'), 
							'type' => 			$type,
							'action_type' =>	$action_type,
							'required' => 		parent::get_object_value($field, 'required'), 
							'default_value' => 	parent::get_object_value($field, 'default_value'),
							'pattern' => 		'',
							'placeholder' => 	'',
							'help' => 			parent::get_object_value($field, 'help_text'), 
							'sort_index' => 	intval(parent::get_object_value($field, 'display_order')) + $sort_index_offset,
							'visible' =>		true,
							'meta' => 			false
						);

						$sort_index_offset ++;
					}

					// Sort by sort_index
					usort($list_fields, function ($a, $b) { return $a['sort_index'] - $b['sort_index']; });

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id, $list_fields);
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				switch($this->list_id) {

					default :

						$form_fields = array(

							'submit' => array(

								'type'			=>	'submit',
								'label'			=>	__('Submit', 'ws-form-slack')
							)
						);

						break;
				}

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config();

				switch($this->list_id) {

					default :

						$form_actions = array(

							$this->id => array(

								'meta'	=> array(

									'action_' . $this->id . '_list_id'				=>	$this->list_id,
									'action_' . $this->id . '_attachment_text'		=>	'#field(#text)',
									'action_' . $this->id . '_attachment_buttons'	=>	array(

										array(

											'action_' . $this->id . '_attachment_action_text' => __('View', 'ws-form-slack'),
											'action_' . $this->id . '_attachment_action_url' => '#submit_admin_url',
											'action_' . $this->id . '_attachment_action_style' => 'default'
										)
									)
								)
							),

							'message' => array(

								'meta'	=> array(

									'action_message_message' => __('Thanks for submitting the form.', 'ws-form-slack')
								)
							),

							'database'
						);

						break;
				}

				return $form_actions;
			}

			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($api_response_decoded) {

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->error);

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_text',
						'action_' . $this->id . '_attachment',
						'action_' . $this->id . '_attachment_title',
						'action_' . $this->id . '_attachment_title_link',
						'action_' . $this->id . '_attachment_fallback',
						'action_' . $this->id . '_attachment_pretext',
						'action_' . $this->id . '_attachment_author_name',
						'action_' . $this->id . '_attachment_author_link',
						'action_' . $this->id . '_attachment_author_icon',
						'action_' . $this->id . '_attachment_text',
						'action_' . $this->id . '_attachment_color',
						'action_' . $this->id . '_attachment_buttons',
						'action_' . $this->id . '_attachment_field_mapping',
						'action_' . $this->id . '_attachment_image_field_id',
						'action_' . $this->id . '_attachment_thumb_field_id',
						'action_' . $this->id . '_attachment_footer',
						'action_' . $this->id . '_attachment_footer_icon',
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-slack'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-slack')); }

				return ($this->list_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('Slack Channel', 'ws-form-slack'),
						'type'							=>	'select',
						'help'							=>	__('Which Slack channel should this message be pushed to?', 'ws-form-slack'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-slack'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// Text
					'action_' . $this->id . '_text'	=> array(

						'label'							=>	__('Text', 'ws-form-slack'),
						'type'							=>	'textarea',
						'help'							=>	__('Text of the message to send. Max 40,000 characters.', 'ws-form-slack'),
						'default'						=>	__('You have received a new form submission on #blog_name.', 'ws-form-slack'),
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Attachment
					'action_' . $this->id . '_attachment'	=> array(

						'label'							=>	__('Attachment', 'ws-form-slack'),
						'type'							=>	'checkbox',
						'help'							=>	__('Add attachment to message?', 'ws-form-slack'),
						'default'						=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Attachment - Title
					'action_' . $this->id . '_attachment_title'	=> array(

						'label'							=>	__('Attachment Title', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the title of the attachment.', 'ws-form-slack'),
						'default'						=>	__('Submission Details', 'ws-form-slack'),
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=> '&&'
							)
						)
					),

					// Attachment - Title Link
					'action_' . $this->id . '_attachment_title_link'	=> array(

						'label'							=>	__('Attachment Title Link', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the title link for the attachment.', 'ws-form-slack'),
						'default'						=>	'#submit_admin_url',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Fallback
					'action_' . $this->id . '_attachment_fallback'	=> array(

						'label'							=>	__('Attachment Fallback', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the fallback text.', 'ws-form-slack'),
						'default'						=>	__('Unable to preview submission', 'ws-form-slack'),
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Pretext
					'action_' . $this->id . '_attachment_pretext'	=> array(

						'label'							=>	__('Attachment Pretext', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the pretext of the attachment.', 'ws-form-slack'),
						'default'						=>	'',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Author - Name
					'action_' . $this->id . '_attachment_author_name'	=> array(

						'label'							=>	__('Author Name', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Attachment author name (Leave blank for none)', 'ws-form-slack'),
						'default'						=>	'#blog_name',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Author - Link
					'action_' . $this->id . '_attachment_author_link'	=> array(

						'label'							=>	__('Author Link', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Attachment author URL', 'ws-form-slack'),
						'default'						=>	'#blog_url',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Author - Icon
					'action_' . $this->id . '_attachment_author_icon'	=> array(

						'label'							=>	__('Author Icon', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Attachement author icon URL (16px x 16px)', 'ws-form-slack'),
						'default'						=>	'',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Title
					'action_' . $this->id . '_attachment_text'	=> array(

						'label'							=>	__('Attachment Text', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the text of the attachment.', 'ws-form-slack'),
						'default'						=>	'',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Color
					'action_' . $this->id . '_attachment_color'	=> array(

						'label'							=>	__('Attachment Color', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter the hex color of the attachment. You can also enter: good, warning or danger.', 'ws-form-slack'),
						'default'						=>	'#skin_color_information',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Actions
					'action_' . $this->id . '_attachment_buttons'	=> array(

						'label'						=>	__('Buttons', 'ws-form-slack'),
						'type'						=>	'repeater',
						'meta_keys'					=>	array(

							'action_' . $this->id . '_attachment_action_text',
							'action_' . $this->id . '_attachment_action_url',
							'action_' . $this->id . '_attachment_action_style',
						),
						'help'						=>	__('Buttons to add to the attachment.', 'ws-form-slack'),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Actions - Text
					'action_' . $this->id . '_attachment_action_text'	=> array(

						'label'						=>	__('Text', 'ws-form-slack'),
						'type'						=>	'text'
					),

					// Attachment - Actions - URL
					'action_' . $this->id . '_attachment_action_url'	=> array(

						'label'						=>	__('URL', 'ws-form-slack'),
						'type'						=>	'text'
					),

					// Attachment - Actions - Style
					'action_' . $this->id . '_attachment_action_style'	=> array(

						'label'						=>	__('Style', 'ws-form-slack'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'default', 'text' => __('Default', 'ws-form-slack')),
							array('value' => 'primary', 'text' => __('Primary', 'ws-form-slack')),
							array('value' => 'danger', 'text' => __('Danger', 'ws-form-slack')),
						),
						'default'					=>	'default'
					),

					// Field mapping
					'action_' . $this->id . '_attachment_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-slack'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to attachment fields.', 'ws-form-slack'),
						'meta_keys'					=>	array(

							'ws_form_field'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Image
					'action_' . $this->id . '_attachment_image_field_id'	=> array(

						'label'						=>	__('Image', 'ws-form'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form'),
						'fields_filter_type'		=>	array('signature', 'file'),
						'help'						=>	__('File or signature field to use as attachment image.', 'ws-form'),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Image
					'action_' . $this->id . '_attachment_thumb_field_id'	=> array(

						'label'						=>	__('Thumbnail', 'ws-form'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form'),
						'fields_filter_type'		=>	array('signature', 'file'),
						'help'						=>	__('File or signature field to use as attachment thumbnail.', 'ws-form'),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Footer
					'action_' . $this->id . '_attachment_footer'	=> array(

						'label'							=>	__('Footer Text', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter text for the attachment footer.', 'ws-form-slack'),
						'default'						=>	self::WS_FORM_LICENSE_NAME,
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),

					// Attachment - Footer - Icon
					'action_' . $this->id . '_attachment_footer_icon'	=> array(

						'label'							=>	__('Footer Icon URL', 'ws-form-slack'),
						'type'							=>	'text',
						'help'							=>	__('Enter icon URL for the attachment footer.', 'ws-form-slack'),
						'default'						=>	plugin_dir_url( __FILE__ ) . 'images/wsform.png',
						'select_list'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_attachment',
								'meta_value'		=>	'on',
								'logic_previous'	=>	'&&'
							)
						)
					),
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-slack'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-slack'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Slack add-on for WS Form PRO license key here.', 'ws-form-slack'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-slack'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Slack Status', 'ws-form-slack'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Slack account, or <a href=\"https://slack.com/get-started/\" target=\"blank\">create a new account</a>.", 'ws-form-slack'),
							'button'		=>	'oauth_action_' . $this->id,
						),

						'action_' . $this->id . '_team_name'	=>	array(

							'label'		=>	__('Slack Team Name', 'ws-form-slack'),
							'type'		=>	'static'
						)
					)
				);

				return $options;
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				$this->text = parent::get_config($config, 'action_' . $this->id . '_text');

				// Attachment
				$this->attachment = parent::get_config($config, 'action_' . $this->id . '_attachment');
				$this->attachment_title = parent::get_config($config, 'action_' . $this->id . '_attachment_title');
				$this->attachment_title_link = parent::get_config($config, 'action_' . $this->id . '_attachment_title_link');
				$this->attachment_fallback = parent::get_config($config, 'action_' . $this->id . '_attachment_fallback');
				$this->attachment_pretext = parent::get_config($config, 'action_' . $this->id . '_attachment_pretext');
				$this->attachment_author_name = parent::get_config($config, 'action_' . $this->id . '_attachment_author_name');
				$this->attachment_author_link = parent::get_config($config, 'action_' . $this->id . '_attachment_author_link');
				$this->attachment_author_icon = parent::get_config($config, 'action_' . $this->id . '_attachment_author_icon');
				$this->attachment_text = parent::get_config($config, 'action_' . $this->id . '_attachment_text');
				$this->attachment_color = parent::get_config($config, 'action_' . $this->id . '_attachment_color');
				$this->attachment_buttons = parent::get_config($config, 'action_' . $this->id . '_attachment_buttons');
				if(!is_array($this->attachment_buttons)) { $this->attachment_buttons = array(); }
				$this->attachment_field_mapping = parent::get_config($config, 'action_' . $this->id . '_attachment_field_mapping');
				if(!is_array($this->attachment_field_mapping)) { $this->attachment_field_mapping = array(); }
				$this->attachment_image_field_id = parent::get_config($config, 'action_' . $this->id . '_attachment_image_field_id');
				$this->attachment_thumb_field_id = parent::get_config($config, 'action_' . $this->id . '_attachment_thumb_field_id');
				$this->attachment_footer = parent::get_config($config, 'action_' . $this->id . '_attachment_footer');
				$this->attachment_footer_icon = parent::get_config($config, 'action_' . $this->id . '_attachment_footer_icon');
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token_bot  	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_bot', '');
				$this->api_access_token_user 	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_user', '');
				$this->team_id  				= WS_Form_Common::option_get('action_' . $this->id . '_team_id', '');
				$this->team_name 				= WS_Form_Common::option_get('action_' . $this->id . '_team_name', '');

				// Build endpoint using API key
				if(!empty($this->api_access_token_bot) && !empty($this->api_access_token_user)) {

					// Set endpoint URL
					$this->api_endpoint = 'https://slack.com/api/';

				} else {

					$this->api_endpoint = false;
					$this->configured = false;
				}

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers - Bot
			public function get_api_headers_bot() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token_bot);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(42.000000, 62.000000)" fill-rule="nonzero"><path d="M12.7741935,37.8387097 C12.3813973,40.9872353 9.70519084,43.3503156 6.53225806,43.3503156 C3.35932531,43.3503156 0.68311889,40.9872353 0.290322581,37.8387097 C0.31665599,34.4023507 3.09589905,31.6231076 6.53225806,31.5967742 L12.7741935,31.5967742 L12.7741935,37.8387097 L12.7741935,37.8387097 Z" fill="#E01E5A"></path><path d="M15.9193548,37.8387097 C16.3121511,34.6901841 18.9883575,32.3271037 22.1612903,32.3271037 C25.3342231,32.3271037 28.0104295,34.6901841 28.4032258,37.8387097 L28.4032258,53.4677419 C28.0104295,56.6162676 25.3342231,58.9793477 22.1612903,58.9793477 C18.9883575,58.9793477 16.3121511,56.6162676 15.9193548,53.4677419 L15.9193548,37.8387097 L15.9193548,37.8387097 Z" fill="#E01E5A"></path><path d="M22.1612903,12.7741935 C19.0127647,12.3813973 16.6496844,9.70519084 16.6496844,6.53225806 C16.6496844,3.35932531 19.0127647,0.68311889 22.1612903,0.290322581 C25.5976493,0.31665599 28.3768924,3.09589905 28.4032258,6.53225806 L28.4032258,12.7741935 L22.1612903,12.7741935 L22.1612903,12.7741935 Z" fill="#36C5F0"></path><path d="M22.1612903,15.9193548 C25.3098159,16.3121511 27.6728963,18.9883575 27.6728963,22.1612903 C27.6728963,25.3342231 25.3098159,28.0104295 22.1612903,28.4032258 L6.53225806,28.4032258 C3.38373249,28.0104295 1.02065211,25.3342231 1.02065211,22.1612903 C1.02065211,18.9883575 3.38373249,16.3121511 6.53225806,15.9193548 L22.1612903,15.9193548 L22.1612903,15.9193548 Z" fill="#36C5F0"></path><g transform="translate(31.451613, 0.967742)"><path d="M15.7741935,21.1935484 C16.1669898,18.0450228 18.8431963,15.6819424 22.016129,15.6819424 C25.1890618,15.6819424 27.8652682,18.0450228 28.2580645,21.1935484 C28.2317311,24.6299074 25.452488,27.4091505 22.016129,27.4354839 L15.7741935,27.4354839 L15.7741935,21.1935484 L15.7741935,21.1935484 Z" fill="#2EB67D"></path><path d="M12.6290323,21.1935484 C12.236236,24.342074 9.56002955,26.7051543 6.38709677,26.7051543 C3.21416402,26.7051543 0.5379576,24.342074 0.14516129,21.1935484 L0.14516129,5.56451613 C0.5379576,2.41599055 3.21416402,0.0529101745 6.38709677,0.0529101745 C9.56002955,0.0529101745 12.236236,2.41599055 12.6290323,5.56451613 L12.6290323,21.1935484 Z" fill="#2EB67D"></path></g><g transform="translate(31.451613, 31.451613)"><path d="M6.38709677,15.7741935 C9.53562237,16.1669898 11.8987027,18.8431963 11.8987027,22.016129 C11.8987027,25.1890618 9.53562237,27.8652682 6.38709677,28.2580645 C2.95073775,28.2317311 0.1714947,25.452488 0.14516129,22.016129 L0.14516129,15.7741935 L6.38709677,15.7741935 L6.38709677,15.7741935 Z" fill="#ECB22E"></path><path d="M6.38709677,12.6290323 C3.2385712,12.236236 0.875490818,9.56002955 0.875490818,6.38709677 C0.875490818,3.21416402 3.2385712,0.5379576 6.38709677,0.14516129 L22.016129,0.14516129 C25.1646546,0.5379576 27.527735,3.21416402 27.527735,6.38709677 C27.527735,9.56002955 25.1646546,12.236236 22.016129,12.6290323 L6.38709677,12.6290323 L6.38709677,12.6290323 Z" fill="#ECB22E"></path></g></g>';

				return $svg_logo;
			}
		}
	});