(function($) {

	'use strict';

	$(document).on('wsf-rendered', function(e, form, form_id, instance_id) {

		if(
			(typeof(window.wsf_stripe_elements_functions) !== 'undefined') &&
			(typeof(window.wsf_stripe_elements_functions[instance_id]) !== 'undefined')
		) {
		
			window.wsf_stripe_elements_functions[instance_id]();
		}
	});

	$.WS_Form_Stripe_Elements = function(atts) {

		this.id = false;
		this.id_mode = '';
		this.id_json = '';
		this.form_instance_id = false;
		this.transaction_type = '';
		this.transaction_amount = 0;
		this.transaction_zero = false;
		this.transaction_zero_message = '';
		this.options_icon_style = false;
		this.options_icon_hidden = false;
		this.message_method = false;
		this.message_clear = false;
		this.message_scroll_top = false;
		this.message_form_hide = false;
		this.message_duration = false;
		this.publishable_key = '';
		this.ws_form = false;
	}

	$.WS_Form_Stripe_Elements.prototype.init = function() {

		var ws_this = this;

		this.ws_form = window.wsf_form_instances[this.form_instance_id];

		var stripe = Stripe(this.publishable_key);
		var elements = stripe.elements();

		var card = elements.create('card', this.get_options());

		if(!$(this.id).length) { return; }

		card.mount(this.id);

		// Handle real-time validation errors from the card Element.
		card.addEventListener('change', function(event) {

			if(event.error) {

				// Unlock form
				ws_this.ws_form.form_post_unlock();

				ws_this.invalid_feedback(event.error.message);

			} else {

				ws_this.invalidate();
			}
		});

		// Button click
		$('[data-wsf-stripe-submit]').click(function(e) {

			// Reset validation
			ws_this.invalidate();

			if(ws_this.click()) {

				// Log
				ws_this.ws_form.log('log_payment', 'Stripe Elements - Payment initiated', 'payment');

				// Set payment method
				ws_this.ws_form.form_ecommerce_set_payment_method('Stripe Elements');

				// Set payment amount (Overrides calculated amount in case the payment button has a fixed value, we always want to show the processed amount in submissions)
				ws_this.ws_form.form_ecommerce_set_payment_amount(ws_this.get_amount_total());

				// Set status
				ws_this.ws_form.form_ecommerce_set_status('pending_payment');

				switch(ws_this.transaction_type) {

					// Subscription / Setup intent
					case 'subscription' :
					case 'setup_intent' :

						// Lock form
						ws_this.ws_form.form_post_lock('progress', true, true);

						// Create token
						stripe.createToken(card).then(function(result) {

							if(result.error) {

								// Unlock form
								ws_this.ws_form.form_post_unlock('progress', false, true);

								// Set invalid feedback
								ws_this.invalid_feedback(result.error.message);

							} else {

								// Get token ID
								var token_id = result.token.id;

								// Set token
								$('#' + ws_this.id_token).val(token_id);

								// Log
								ws_this.ws_form.log('log_payment', 'Stripe Elements - Token response received', 'payment');

								// Initiate payment
								ws_this.ws_form.form_post('save', 'stripeelements');
							}
						});

						break;

					// One-time
					default :

						// Initiate payment
						ws_this.ws_form.form_post('save', 'stripeelements');
				}
			}

			e.preventDefault();
		});

		// Success
		$(document).on('action_stripeelements_success', function(e, params) {

			// Set mode
			$('#' + ws_this.id_mode).val('success');

			// Set JSON
			$('#' + ws_this.id_json).val(JSON.stringify(params.subscription));

			// Unlock form
			ws_this.ws_form.form_post_unlock('progress', false, true);

			// Submit the form and run actions
			ws_this.ws_form.form_obj.submit();
		});

		// Authentication required
		$(document).on('action_stripeelements_confirm_card_payment', function(e, params) {

			// Lock form
			ws_this.ws_form.form_post_lock('progress', true, true);

			// Get client secret
			var client_secret = params.client_secret;

			// Confirm card payment
			stripe.confirmCardPayment(

				client_secret,
				{
					payment_method: {card: card}
				}

			).then(function(result) {

				if(result.error) {

					// Unlock form
					ws_this.ws_form.form_post_unlock('progress', true, true);

					// Set mode
					$('#' + ws_this.id_mode).val('error');

					// Set invalid feedback
					ws_this.invalid_feedback(result.error.message);

				} else {

					// Set mode
					$('#' + ws_this.id_mode).val('success');

					switch(ws_this.transaction_type) {

						case 'subscription' :

							// Get subscription
							var return_data = params.subscription;
							break;

						default :

							// Get payment intent
							var return_data = result.paymentIntent;
					}

					// Set JSON
					$('#' + ws_this.id_json).val(JSON.stringify(return_data));

					// Unlock form
					ws_this.ws_form.form_post_unlock('progress', false, true);

					// Submit the form and run actions
					ws_this.ws_form.form_obj.submit();
				}
			});
		});

		// Authentication required
		$(document).on('action_stripeelements_confirm_card_setup', function(e, params) {

			// Lock form
			ws_this.ws_form.form_post_lock('progress', true, true);

			// Get client secret
			var client_secret = params.client_secret;

			// Confirm card payment
			stripe.confirmCardSetup(

				client_secret,
				{
					payment_method: {card: card}
				}

			).then(function(result) {

				if(result.error) {

					// Unlock form
					ws_this.ws_form.form_post_unlock('progress', true, true);

					// Set mode
					$('#' + ws_this.id_mode).val('error');

					// Set invalid feedback
					ws_this.invalid_feedback(result.error.message);

				} else {

					// Set mode
					$('#' + ws_this.id_mode).val('success');

					// Get setup intent
					var return_data = result.setupIntent;

					// Set JSON
					$('#' + ws_this.id_json).val(JSON.stringify(return_data));

					// Unlock form
					ws_this.ws_form.form_post_unlock('progress', false, true);

					// Submit the form and run actions
					ws_this.ws_form.form_obj.submit();
				}
			});
		});
	}

	$.WS_Form_Stripe_Elements.prototype.error_message = function(message) {

		this.ws_form.action_message(message, 'danger', this.message_method, this.message_duration, this.message_form_hide, this.message_clear, this.message_scroll_top, true, true);
	}

	$.WS_Form_Stripe_Elements.prototype.invalidate = function() {

		// Get classes
		var class_stripe_elements_invalid_label_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_label', []);
		var class_stripe_elements_invalid_field_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_field', []);
		var class_stripe_elements_invalid_invalid_feedback_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_invalid_feedback', []);
		var class_stripe_elements_valid_label_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_label', []);
		var class_stripe_elements_valid_field_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_field', []);
		var class_stripe_elements_valid_invalid_feedback_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_invalid_feedback', []);

		var stripe_elements_obj_field = $('[data-wsf-stripe-cc]', this.ws_form.form_canvas_obj);
		var stripe_elements_obj_wrapper = stripe_elements_obj_field.closest('[data-id]');
		var stripe_elements_id = stripe_elements_obj_wrapper.attr('data-id');
		var stripe_elements_obj_label = $('label', stripe_elements_obj_wrapper);
		var stripe_elements_obj_invalid_feedback = $('#' + this.ws_form.form_id_prefix + 'invalid-feedback-' + stripe_elements_id);

		// Empty - Label
		stripe_elements_obj_label.addClass(class_stripe_elements_valid_label_array.join(' '));
		stripe_elements_obj_label.removeClass(class_stripe_elements_invalid_label_array.join(' '));

		// Empty - Field
		stripe_elements_obj_field.addClass(class_stripe_elements_valid_field_array.join(' '));
		stripe_elements_obj_field.removeClass(class_stripe_elements_invalid_field_array.join(' '));

		// Empty - Feedback
		stripe_elements_obj_invalid_feedback.addClass(class_stripe_elements_valid_invalid_feedback_array.join(' '));
		stripe_elements_obj_invalid_feedback.removeClass(class_stripe_elements_invalid_invalid_feedback_array.join(' '));

		// Unmark as validated
		this.ws_form.form_obj.removeClass(this.ws_form.class_validated);
	}

	$.WS_Form_Stripe_Elements.prototype.invalid_feedback = function(message) {

		// Get classes
		var class_stripe_elements_invalid_label_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_label', []);
		var class_stripe_elements_invalid_field_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_field', []);
		var class_stripe_elements_invalid_invalid_feedback_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_invalid_invalid_feedback', []);
		var class_stripe_elements_valid_label_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_label', []);
		var class_stripe_elements_valid_field_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_field', []);
		var class_stripe_elements_valid_invalid_feedback_array = this.ws_form.get_field_value_fallback('stri_el_cc', false, 'class_valid_invalid_feedback', []);

		var stripe_elements_obj_field = $('[data-wsf-stripe-cc]', this.ws_form.form_canvas_obj);
		var stripe_elements_obj_wrapper = stripe_elements_obj_field.closest('[data-id]');
		var stripe_elements_id = stripe_elements_obj_wrapper.attr('data-id');
		var stripe_elements_obj_label = $('label', stripe_elements_obj_wrapper);
		var stripe_elements_obj_invalid_feedback = $('#' + this.ws_form.form_id_prefix + 'invalid-feedback-' + stripe_elements_id);

		// Set invalid feedback string
		stripe_elements_obj_invalid_feedback.html(message);

		// Empty - Label
		stripe_elements_obj_label.addClass(class_stripe_elements_invalid_label_array.join(' '));
		stripe_elements_obj_label.removeClass(class_stripe_elements_valid_label_array.join(' '));

		// Empty - Field
		stripe_elements_obj_field.addClass(class_stripe_elements_invalid_field_array.join(' '));
		stripe_elements_obj_field.removeClass(class_stripe_elements_valid_field_array.join(' '));

		// Empty - Feedback
		stripe_elements_obj_invalid_feedback.addClass(class_stripe_elements_invalid_invalid_feedback_array.join(' '));
		stripe_elements_obj_invalid_feedback.removeClass(class_stripe_elements_valid_invalid_feedback_array.join(' '));

		// Mark form as validated
		this.ws_form.form_obj.addClass(this.ws_form.class_validated);
	}

	$.WS_Form_Stripe_Elements.prototype.click = function() {

		// Set mode
		$('#' + this.id_mode).val('init');

		this.ws_form.log('log_payment', 'Stripe Elements - Clicked', 'payment');

		// Check if form is valid
		if(!this.ws_form.form_valid) {

			// Form is not valid, so submit it to simulate validation
			this.ws_form.form_obj.submit();
			return false;
		}

		// Check amount
		if(this.transaction_type == '') {

			var stripe_total = this.get_amount_total();

			if(stripe_total == 0) {

				// Process as zero amount
				switch(this.transaction_zero) {

					// Submit form
					case 'submit' :

						this.ws_form.form_obj.submit();
						break;

					// Show message
					case 'message' :

						var message = this.transaction_zero_message;
						this.error_message(message);
						break;
				}

				return false;
			}
		}

		return true;
	}

	$.WS_Form_Stripe_Elements.prototype.get_amount_total = function() {

		var transaction_total = this.transaction_amount;
		if(transaction_total == '') { transaction_total = '#ecommerce_cart_total'; }

		// Get Stripe total
		var stripe_total = this.ws_form.parse_variables_process(transaction_total).output;

		// Convert to a number
		stripe_total = this.ws_form.get_number(stripe_total);

		return stripe_total;
	}

	$.WS_Form_Stripe_Elements.prototype.get_options = function() {

		var font_family = $('[data-wsf-stripe-cc]').css('font-family');
		var font_size = $('[data-wsf-stripe-cc]').css('font-size');
		var line_height = $('[data-wsf-stripe-cc]').css('line-height');
		var color = $('[data-wsf-stripe-cc]').css('color');

		var options = {

			iconStyle: this.options_icon_style,

			hideIcon: (this.options_icon_hidden == 'on'),

			style: {

				base: {

					color: color,
					fontFamily: font_family,
					fontSize: font_size,
					'::placeholder': {

						color: this.ws_form.parse_variables_process('#skin_color_default_lighter').output
					},
					lineHeight: line_height
				},

				invalid: {

					color: this.ws_form.parse_variables_process('#skin_color_danger').output,
					iconColor: this.ws_form.parse_variables_process('#skin_color_danger').output
				}
			}
		};

		// Mapping
		// value.postalCodeiconStyle
		// hidePostalCode true

		return options;
	}

})(jQuery);
