<?php

	/**
	 * @link              https://wsform.com/knowledgebase/stripe-elements/
	 * @since             1.0.0
	 * @package           WS_Form_Stripe_Elements
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Stripe Elements
	 * Plugin URI:        https://wsform.com/knowledgebase/stripe-elements/
	 * Description:       Stripe Elements add-on for WS Form PRO
	 * Version:           1.3.5
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-stripe-elements
	 */

	Class WS_Form_Add_On_Stripe_Elements {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.155';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Stripe_Elements();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-stripe-elements'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_stripe_elements = new WS_Form_Add_On_Stripe_Elements();	

	register_activation_hook(__FILE__, array($wsf_add_on_stripe_elements, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Stripe_Elements extends WS_Form_Action {

			public $id = 'stripeelements';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 10;
			public $can_repost = false;
			public $form_add = false;

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			public $environment;
			public $test_publishable_key;
			public $test_secret_key;
			public $live_publishable_key;
			public $live_secret_key;
			public $publishable_key;
			public $secret_key;

			// Variables
			public $customer_id = '';
			public $transaction_id = '';
			public $subscription_id = '';

			// Enqueueing
			public $enqueue_js_stripe_elements = false;
			public $enqueued_js_stripe_elements = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 6237;
			const WS_FORM_LICENSE_NAME = 'Stripe Elements add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.3.5';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Stripe Elements', 'ws-form-stripe-elements');

				// Set label for actions pull down
				$this->label_action = __('Pay with Stripe Elements', 'ws-form-stripe-elements');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('wsf_actions_post_save', array($this, 'actions_post_save'), 10, 1);
				add_filter('wsf_actions_post_submit', array($this, 'actions_post_submit'), 10, 1);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_config_frameworks', array($this, 'config_frameworks'), 10, 1);
				add_action('wsf_enqueue', array( $this, 'enqueue_scripts'), 10, 1);
				add_filter('wsf_wizard_config_files', array( $this, 'wizard_config_files' ), 10, 1);
				add_filter('wsf_wizard_svg_buttons', array( $this, 'wizard_svg_buttons' ), 10, 1);
				add_action('wsf_form_pre_process_field', array($this, 'form_pre_process_field'), 10, 1);
				add_filter('wsf_get_skin', array($this, 'get_skin'), 10, 1);
				add_filter('wsf_config_parse_variables', array( $this, 'config_parse_variables' ), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();

				// If configured, register the Stripe field type
				if($this->configured) {

					add_filter('wsf_config_field_types', array($this, 'config_field_types'), 10, 2);
				}
			}

			// Parse variables
			public function config_parse_variables($parse_variables) {

				$parse_variables['stripe'] = array(

					'label' => __('Stripe Elements', 'ws-form-stripe-elements'),

					'variables'	=> array(

						'stripe_customer_id' =>	array('label' => __('Customer ID', 'ws-form-stripe-elements'), 'value' => $this->customer_id, 'description' => __('Returns the customer ID used or created during the transaction.', 'ws-form-stripe-elements')),
						'stripe_transaction_id' =>	array('label' => __('Transaction ID', 'ws-form-stripe-elements'), 'value' => $this->transaction_id, 'description' => __('Returns the transaction ID for one-time transactions.', 'ws-form-stripe-elements')),
						'stripe_subscription_id' =>	array('label' => __('Subscription ID', 'ws-form-stripe-elements'), 'value' => $this->subscription_id, 'description' => __('Returns the subscription ID for subscription transactions.', 'ws-form-stripe-elements'))
					)
				);

				return $parse_variables;
			}

			// Wizard - Config files
			public function wizard_config_files($config_files) {

				$config_files[] = plugin_dir_path(__FILE__) . 'includes/wizard/config.json';

				return $config_files;
			}

			// Wizard - SVG buttons
			public function wizard_svg_buttons($field_type_buttons) {

				$color_default_inverted = WS_Form_Common::option_get('skin_color_default_inverted');
				$color_primary          = WS_Form_Common::option_get('skin_color_primary');

				$field_type_buttons['stri_el_checkout'] = array('fill' => $color_primary, 'color' => $color_default_inverted);

				return $field_type_buttons;
			}

			// Script enqueuing
			public function form_pre_process_field($field) {

//				echo "*";

				// Get field type
				$field_type = $field->type;

				if($field_type === 'stri_el_cc') {

					$this->enqueue_js_stripe_elements = true;
				}
			}

			// Enqueue scripts (Public)
			public function enqueue_scripts() {

				// Enqueue in footer?
				$jquery_footer = (WS_Form_Common::option_get('jquery_footer', '') == 'on');

				if(!$this->enqueued_js_stripe_elements && apply_filters('wsf_enqueue_js_stripe_elements', $this->enqueue_js_stripe_elements)) {

					wp_enqueue_script('ws-form-' . $this->id . '-stripe-js', 'https://js.stripe.com/v3/', array('ws-form-form-public'), self::WS_FORM_LICENSE_VERSION, $jquery_footer);
					wp_enqueue_script('ws-form-' . $this->id . '-js', plugin_dir_url(__FILE__) . 'js/stripe-elements.js', array('ws-form-' . $this->id . '-stripe-js'), self::WS_FORM_LICENSE_VERSION, $jquery_footer);

					$this->enqueued_js_stripe_elements = true;
				}
			}

			// Actions post filter (save)
			public function actions_post_save($actions) {

				// Get Stripe mode
				$mode = WS_Form_Common::get_query_var('wsf_stripe_elements_mode', false);
				if(empty($mode)) { return $actions; }

				// Get field ID
				$field_id = WS_Form_Common::get_query_var('wsf_stripe_elements_field_id', false);
				if($field_id === 0) { return $actions; }

				// Only run this action
				$actions = array(

					array(

						'id' => $this->id,
						'meta' => array(
							'action_' . $this->id . '_mode' => $mode,
							'action_' . $this->id . '_field_id' => $field_id
						),
						'events' => array(
							'0' => 'save'
						),
						'label' => $this->label_action,
						'priority' => $this->priority,
						'row_index' => 0
					)
				);

				return $actions;
			}

			// Actions post filter
			public function actions_post_submit($actions) {

				// Get Stripe mode
				$mode = WS_Form_Common::get_query_var('wsf_stripe_elements_mode', false);
				if(empty($mode)) { return $actions; }

				// Get field ID
				$field_id = WS_Form_Common::get_query_var('wsf_stripe_elements_field_id', false);
				if($field_id === 0) { return $actions; }

				// Prepend this action so it runs first
				$actions[] = array(

					'id' => $this->id,
					'meta' => array(
						'action_' . $this->id . '_mode' => $mode,
						'action_' . $this->id . '_field_id' => $field_id
					),
					'events' => array(
						'0' => 'submit'
					),
					'label' => $this->label_action,
					'priority' => $this->priority,
					'row_index' => 0
				);

				return $actions;
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Get Stripe mode
				$mode = parent::get_config($config, 'action_' . $this->id . '_mode');
				if(empty($mode)) { return false; }

				// Get field ID
				$field_id = intval(parent::get_config($config, 'action_' . $this->id . '_field_id'));
				if($field_id === 0) { return false; }

				// Get field
				$fields = ( false === $form ) ? array() : WS_Form_Common::get_fields_from_form($form);
				$field = isset($fields[$field_id]) ? $fields[$field_id] : false;

				// Check field found
				if(!$field) { return false; }

				// Convert to object
				$field = (object)$field;

				// Read config from field
				$transaction_type = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_type', '');

				// Process by transaction type
				switch($transaction_type) {

					// Subscription
					case 'subscription' :

						// Get token mode
						$token = WS_Form_Common::get_query_var('wsf_stripe_elements_token', false);
						if(empty($token)) { return $actions; }

						switch($mode) {

							case 'init' :

								return self::subscription_create($form, $submit, $field, $token);

							case 'success' :

								return self::subscription_success($form, $submit, $field);
						}

					// Payment Method
					case 'setup_intent' :

						// Get token mode
						$token = WS_Form_Common::get_query_var('wsf_stripe_elements_token', false);
						if(empty($token)) { return $actions; }

						switch($mode) {

							case 'init' :

								return self::setup_intent_create($form, $submit, $field, $token);

							case 'success' :

								return self::setup_intent_success($form, $submit, $field);
						}

					// One time
					default :

						switch($mode) {

							case 'init' :

								return self::payment_intent($form, $submit, $field);

							case 'success' :

								return self::payment_success($form, $submit, $field);
						}
				}

				return true;
			}

			// Payment intent
			public function payment_intent($form, $submit, $field) {

				// Get transaction data
				$transaction_amount = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_amount', ''));
				$transaction_currency = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_currency', '');
				if(empty($transaction_currency)) { $transaction_currency = WS_Form_Common::option_get('currency', 'USD'); }
				$transaction_zero = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero', '');
				$transaction_zero_message = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_zero_message', ''), $form, $submit);

				// Process transaction amount
				switch($transaction_amount) {

					case '' :
					case '#ecommerce_cart_total' :

						// Get calculated transaction amount from query var
						$ecommerce_cart = WS_Form_Common::get_query_var('wsf_ecommerce_cart', false);
						$amount = floatval($ecommerce_cart['total']);
						break;

					default :

						// Use amount specified in field config
						$amount = floatval($transaction_amount);
				}

				// Check for a zero amount
				if($amount <= 0) {

					switch($transaction_zero) {

						case 'message' :

							self::error($transaction_zero_message);
							break;

						case '' :

							return false;
					}
				}

				// Should a customer be created?
				$customer_create = (WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_create', '') == 'on');

				if($customer_create) {

					// Create customer
					$customer_id = self::create_customer($form, $submit, $field);
					if($customer_id === false) {

						self::error(__('Unable to create customer. Ensure customer fields are correctly mapped.', 'ws-form-stripe-elements'));
						return 'halt';
					}

				} else {

					// No customer created (Use token for payment)
					$customer_id = false;
				}

				// Amount
				$amount_cents = $amount * 100;

				// Transaction description
				$transaction_description = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_description', ''), $form, $submit);

				// Create payment intent
				$data = array(

					'amount' 		=> $amount_cents,
					'currency' 		=> $transaction_currency,
					'description'	=> $transaction_description
				);

				// Add customer
				if($customer_id !== false) {

					$data['customer'] = $customer_id;
				}

				$data_query = http_build_query($data);

				$api_response = parent::api_call('https://api.stripe.com/', 'v1/payment_intents', 'POST', $data_query, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

				// Process list data
				$payment_intent = json_decode($api_response['response']);
				if(is_null($payment_intent)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					// Set error message
					$error_message = self::get_api_error_detail($payment_intent);
					self::error($error_message);

					// Stop process actions
					return 'halt';
				}

				// Check for client secret
				if(
					isset($payment_intent->client_secret)
				) {

					return self::api_trigger('action_' . $this->id . '_confirm_card_payment', array('client_secret' => $payment_intent->client_secret));

				} else {

					// Set error message
					self::error(__('Unable to retrieve payment intent data', 'ws-form'));

					// Stop process actions
					return 'halt';
				}
			}

			// Payment - Success
			public function payment_success($form, $submit, $field) {

				// Get payment intent json
				$payment_intent_json = WS_Form_Common::get_query_var('wsf_stripe_elements_json', false);

				// Decode
				$payment_intent = json_decode($payment_intent_json);
				if(is_null($payment_intent)) {

					// Set error message
					self::error(__('Payment intent JSON data error', 'ws-form'));

					// Stop process actions
					return 'halt';						
				}

				//  Read full payment intent record
				$api_response = parent::api_call('https://api.stripe.com/', 'v1/payment_intents/' . $payment_intent->id, 'GET', null, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

				// Process response data
				$payment_intent = json_decode($api_response['response']);
				if(is_null($payment_intent)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

				// Read transaction data
				$transaction_id = $payment_intent->id;
				$this->transaction_id = $transaction_id;
				$this->customer_id = isset($payment_intent->customer) ? $payment_intent->customer : '';
				$transaction_currency = $payment_intent->currency;
				$transaction_amount = floatval($payment_intent->amount) / 100;

				// Set transaction ID
				$submit->meta['ecommerce_transaction_id'] = $transaction_id;

				// Set transaction
				$transaction = array(

					'environment' => $this->environment,
					'amount' => $transaction_amount,
					'id' => $transaction_id,
					'transaction' => array($payment_intent)
				);
				$submit->meta['ecommerce_transaction'] = json_encode($transaction);

				// Set status
				$submit->meta['ecommerce_status'] = 'completed';

				self::success(__('Payment successfully processed with transaction ID: ', 'ws-form-stripe-elements') . $transaction_id);

				return $submit;
			}

			// Setup intent
			public function setup_intent_create($form, $submit, $field, $token) {

				// Create customer
				$customer_id = self::create_customer($form, $submit, $field, false);
				if($customer_id === false) {

					self::error(__('Unable to create customer. Ensure customer fields are correctly mapped.', 'ws-form-stripe-elements'));
					return 'halt';
				}

				// Transaction description
				$description = WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_transaction_description', ''), $form, $submit);

				// Setup intent
				$setup_intent = array(

					'customer' => $customer_id,
					'description' => $description,
					'payment_method_data' => array(

						'type' => 'card',
						'card' => array(

							'token' => $token
						)
					)
				);

				// Build data query for API call
				$data_query = http_build_query($setup_intent);

				//  Make API call
				$api_response = parent::api_call('https://api.stripe.com/', 'v1/setup_intents', 'POST', $data_query, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

				// Process response data
				$setup_intent = json_decode($api_response['response']);
				if(is_null($setup_intent)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					// Set error message
					$error_message = self::get_api_error_detail($setup_intent);
					self::error($error_message);

					// Stop process actions
					return 'halt';
				}

				// Check for client secret
				if(
					isset($setup_intent->client_secret)
				) {

					return self::api_trigger('action_' . $this->id . '_confirm_card_setup', array('client_secret' => $setup_intent->client_secret));

				} else {

					// Set error message
					self::error(__('Unable to retrieve setup intent data', 'ws-form'));

					// Stop process actions
					return 'halt';
				}

				// Stop process actions
				return 'halt';
			}

			// Setup intent - Success
			public function setup_intent_success($form, $submit, $field) {

				// Get setup intent json
				$setup_intent_json = WS_Form_Common::get_query_var('wsf_stripe_elements_json', false);

				// Decode
				$setup_intent = json_decode($setup_intent_json);
				if(is_null($setup_intent)) {

					// Set error message
					self::error(__('Setup intent JSON data error', 'ws-form'));

					// Stop process actions
					return 'halt';						
				}

				//  Read full payment method record
				$api_response = parent::api_call('https://api.stripe.com/', 'v1/setup_intents/' . $setup_intent->id, 'GET', null, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

				// Process response data
				$setup_intent = json_decode($api_response['response']);
				if(is_null($setup_intent)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

				// Read transaction data
				$transaction_id = $setup_intent->id;
				$this->transaction_id = $transaction_id;
				$this->customer_id = isset($setup_intent->customer) ? $setup_intent->customer : '';

				// Set transaction ID
				$submit->meta['ecommerce_transaction_id'] = $transaction_id;

				// Set transaction
				$transaction = array(

					'environment' => $this->environment,
					'id' => $transaction_id,
					'transaction' => array($setup_intent)
				);
				$submit->meta['ecommerce_transaction'] = json_encode($transaction);

				// Set status
				$submit->meta['ecommerce_status'] = 'completed';

				self::success(__('Payment method processed with transaction ID: ', 'ws-form-stripe-elements') . $transaction_id);

				return $submit;
			}

			// Create subscription
			public function subscription_create($form, $submit, $field, $token) {

				// Create customer
				$customer_id = self::create_customer($form, $submit, $field, $token);
				if($customer_id === false) {

					self::error(__('Unable to create customer. Ensure customer fields are correctly mapped.', 'ws-form-stripe-elements'));
					return 'halt';
				}

				// Subscription price mapping
				$subscription_price_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_subscription_price_mapping', '');
				if(!is_array($subscription_price_mapping)) { $subscription_price_mapping = array(); }

				$subscription_id = false;

				$subscription = array(

					'payment_behavior' => 'allow_incomplete',
					'customer' => $customer_id,
					'items' => array()
				);

				$price_ids = array();

				foreach($subscription_price_mapping as $subscription_price_map) {

					$subscription_price_id = $subscription_price_map->{'action_' . $this->id . '_subscription_price_id'};
					if($subscription_price_id == '') { continue; }

					$price_id = WS_Form_Common::parse_variables_process($subscription_price_id, $form, $submit, 'column_value');

					$price_id_array = explode("\n", $price_id);

					foreach($price_id_array as $price_id) {

						if($price_id == '') { continue; }

						$price_ids[] = $price_id;
					}
				}

				$price_ids = array_unique($price_ids);

				foreach($price_ids as $price_id) {

					$subscription['items'][] = array('price' => $price_id);
				}

				if(count($subscription['items']) == 0) {

					self::error(__('No subscription API price ID specified', 'ws-form-stripe-elements'));
					return 'halt';
				}

				// Free trial
				$subscription_trial = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_subscription_trial', '');
				switch($subscription_trial) {

					case 'trial_period_days' :

						$subscription_trial_period_days = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_subscription_trial_period_days', '');
						$subscription_trial_period_days = intval(WS_Form_Common::parse_variables_process($subscription_trial_period_days, $form, $submit, 'column_value'));

						if($subscription_trial_period_days > 0) {

							$subscription['trial_period_days'] = $subscription_trial_period_days;
						}

						break;

					case 'trial_from_plan' :

						$subscription['trial_from_plan'] = 'true';

						break;
				}

				// Invoice item mapping
				$subscription_invoice_item_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_subscription_invoice_item_mapping', '');
				if(!is_array($subscription_invoice_item_mapping)) { $subscription_invoice_item_mapping = array(); }

				$price_ids = array();

				foreach($subscription_invoice_item_mapping as $subscription_invoice_item_map) {

					$subscription_price_id = $subscription_invoice_item_map->{'action_' . $this->id . '_subscription_price_id'};
					if($subscription_price_id == '') { continue; }

					$price_id = WS_Form_Common::parse_variables_process($subscription_price_id, $form, $submit, 'column_value');

					$price_id_array = explode("\n", $price_id);

					foreach($price_id_array as $price_id) {

						if($price_id == '') { continue; }

						$price_ids[] = $price_id;
					}
				}

				if(count($price_ids) > 0) {

					$price_ids = array_unique($price_ids);

					$subscription['add_invoice_items'] = array();

					foreach($price_ids as $price_id) {

						$subscription['add_invoice_items'][] = array('price' => $price_id);
					}
				}

				// Build data query for API call
				$data_query = http_build_query($subscription);

				//  Make API call
				$api_response = parent::api_call('https://api.stripe.com/', 'v1/subscriptions', 'POST', $data_query, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

				// Process response data
				$subscription = json_decode($api_response['response']);
				if(is_null($subscription)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

				// Check status code
				if($api_response['http_code'] != 200) { 

					// Set error message
					$error_message = self::get_api_error_detail($subscription);
					self::error($error_message);

					// Stop process actions
					return 'halt';
				}

				if(
					isset($subscription->status)
				) {
					
					switch($subscription->status) {
							
						case 'active' :
						case 'trialing' :

							// Process subscription
							return self::api_trigger('action_' . $this->id . '_success', array('subscription' => $subscription));
							
						case 'incomplete' :

							// Get latest invoice
							$latest_invoice = $subscription->latest_invoice;
							
							//  Make API call
							$api_response = parent::api_call('https://api.stripe.com/', 'v1/invoices/' . $latest_invoice, 'GET', null, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

							// Process response data
							$invoice = json_decode($api_response['response']);
							if(is_null($invoice)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

							// Get payment intent
							$payment_intent = $invoice->payment_intent;

							//  Make API call
							$api_response = parent::api_call('https://api.stripe.com/', 'v1/payment_intents/' . $payment_intent, 'GET', null, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

							// Process response data
							$payment_intent = json_decode($api_response['response']);
							if(is_null($payment_intent)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

							// Process client secret
							return self::api_trigger('action_' . $this->id . '_confirm_card_payment', array('subscription' => $subscription, 'client_secret' => $payment_intent->client_secret));

						default :
							
							self::error(__('Unknown subscription error.'));
							
							return 'halt';
}

				} else {

					// Set error message
					$error_message = self::get_api_error_detail($subscription);
					self::error($error_message);
				}

				// Stop process actions
				return 'halt';
			}

			// Subscription success
			public function subscription_success($form, $submit, $field) {

				// Get subscription json
				$subscription_json = WS_Form_Common::get_query_var('wsf_stripe_elements_json', false);

				// Decode
				$subscription = json_decode($subscription_json);
				if(is_null($subscription)) {

					// Set error message
					self::error(__('Subscription JSON data error', 'ws-form'));

					// Stop process actions
					return 'halt';						
				}

				// Read transaction data
				$subscription_id = $subscription->id;
				$this->subscription_id = $subscription_id;
				$this->customer_id = isset($subscription->customer) ? $subscription->customer : '';

				// Set transaction ID
				$submit->meta['ecommerce_transaction_id'] = $subscription_id;

				// Calculate total value of initial subscription payment
				$amount = 0;
				foreach($subscription->items->data as $item) {

					$amount += floatval($item->plan->amount) / 100;
				}

				// Set cart total
				$submit->meta['ecommerce_cart_total'] = $amount;

				// Set transaction
				$transaction = array(

					'environment' => $this->environment,
					'amount' => $amount,
					'id' => $subscription_id,
					'transaction' => array($subscription)
				);
				$submit->meta['ecommerce_transaction'] = json_encode($transaction);

				// Set status
				$submit->meta['ecommerce_status'] = 'active';

				self::success(__('Subscription successfully created with subscription ID: ', 'ws-form-stripe-elements') . $subscription_id);
				return $submit;
			}

			// Create customer
			public function create_customer($form, $submit, $field, $token = false) {

				// Create customer record
				$customer = array();
				$customer_id = false;

				// Customer - Name
				$customer_name = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_name', ''), $form, $submit));
				if(!empty($customer_name)) { $customer['name'] = $customer_name; }

				// Customer - Email
				$customer_email_field_id = intval(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_email', ''));
				$customer_email = ($customer_email_field_id > 0) ? parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $customer_email_field_id, false) : '';
				if(!empty($customer_email)) { $customer['email'] = $customer_email; }

				// Customer - Phone
				$customer_phone_field_id = intval(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_phone', ''));
				$customer_phone = ($customer_phone_field_id > 0) ? parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $customer_phone_field_id, false) : '';
				if(!empty($customer_phone)) { $customer['phone'] = $customer_phone; }

				// Customer - Description
				$customer_description = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_customer_description', ''), $form, $submit));
				if(!empty($customer_description)) { $customer['description'] = $customer_description; }

				// Customer - Billing address mapping
				$billing_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_billing_address_mapping', '');
				if(!is_array($billing_address_mapping)) { $billing_address_mapping = array(); }

				$customer_billing_address = array();

				foreach($billing_address_mapping as $field_map) {

					$field_id = $field_map->ws_form_field;
					$stripe_parameter = $field_map->{'action_' . $this->id . '_billing_address_field'};

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Stripe cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					$customer_billing_address[$stripe_parameter] = $submit_value;
				}

				if(count($customer_billing_address) > 0) { $customer['address'] = $customer_billing_address; }

				$customer_shipping_address = array();

				// Customer - Shipping Name
				$customer_shipping_name = trim(WS_Form_Common::parse_variables_process(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_name', ''), $form, $submit));
				if(!empty($customer_shipping_name)) { $customer_shipping_address['name'] = $customer_shipping_name; }

				// Customer - Shipping Phone
				$customer_shipping_phone_field_id = intval(WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_phone', ''));
				$customer_shipping_phone = ($customer_shipping_phone_field_id > 0) ? parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $customer_shipping_phone_field_id, false) : '';
				if(!empty($customer_shipping_phone)) { $customer_shipping_address['phone'] = $customer_shipping_phone; }

				// Customer - Shipping address mapping
				$shipping_address_mapping = WS_Form_Common::get_object_meta_value($field, 'action_' . $this->id . '_shipping_address_mapping', '');
				if(!is_array($shipping_address_mapping)) { $shipping_address_mapping = array(); }

				foreach($shipping_address_mapping as $field_map) {

					$field_id = $field_map->ws_form_field;
					$stripe_parameter = $field_map->{'action_' . $this->id . '_shipping_address_field'};

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Stripe cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					if(!isset($customer_shipping_address['address'])) { $customer_shipping_address['address'] = array(); }
					$customer_shipping_address['address'][$stripe_parameter] = $submit_value;
				}

				if(count($customer_shipping_address) > 0) { $customer['shipping'] = $customer_shipping_address; }

				// Create customer
				if(count($customer) > 0) {

					if($token !== false) {

						$customer['source'] = $token;
					}

					$customer_id = false;

					if(isset($customer['email'])) {

						$data_query = array(

							'email' => $customer['email']
						);

						// Find existing customer
						$api_response = parent::api_call('https://api.stripe.com/', 'v1/customers', 'GET', $data_query, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

						// Process list data
						$customers = json_decode($api_response['response']);
						if(is_null($customers)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

						// Check status code
						if($api_response['http_code'] == 200) { 

							$customers = isset($customers->data) ? $customers->data : array();

							if(count($customers) > 0) {

								$customer_single = $customers[0];

								$customer_id = isset($customer_single->id) ? $customer_single->id : false;
							}

						} else {

							// Set error message
							$error_message = self::get_api_error_detail($customers);
							self::error($error_message);

							// Stop process actions
							return false;
						}
					}

					if($customer_id == false) {

						$data_query = http_build_query($customer);

						$api_response = parent::api_call('https://api.stripe.com/', 'v1/customers', 'POST', $data_query, self::get_api_headers(), 'basic', false, false, 'application/x-www-form-urlencoded', 'application/x-www-form-urlencoded');

						// Process list data
						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { self::error(__('Invalid API response data', 'ws-form-stripe-elements')); }

						// Check status code
						if($api_response['http_code'] == 200) { 

							if(isset($api_response_decoded->id)) {

								$customer_id = $api_response_decoded->id;
							}
		
						} else {

							// Set error message
							$error_message = self::get_api_error_detail($api_response_decoded);
							self::error($error_message);

							// Stop process actions
							return false;
						}
					}

					return $customer_id;
				}

				return false;
			}

			// Get API error detail
			public function get_api_error_detail($api_response_decoded) {

				if(!isset($api_response_decoded->error)) { return __('Unknown Stripe elements error', 'ws-form-stripe-elements'); }
				if(!isset($api_response_decoded->error->message)) { return __('Unknown Stripe elements error', 'ws-form-stripe-elements'); }

				// Get error detail from JSON response
				$error_detail_array = array($api_response_decoded->error->message);

				return $error_detail_array;
			}

			// Config - Frameworks
			public function config_frameworks($frameworks) {

				// WS Form
				$frameworks['types']['ws-form']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('wsf-button');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('wsf-button-full');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['ws-form']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['ws-form']['fields']['public']['field_types']['stri_el_cc']['class_valid_field'] = array('wsf-valid');

				// Bootstrap
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('wsf-invalid');
				$frameworks['types']['bootstrap3']['fields']['public']['field_types']['stri_el_cc']['class_valid_field'] = array('wsf-valid');

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap4']['fields']['public']['field_types']['stri_el_cc']['class_valid_field'] = array('is-valid');

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('btn');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('btn-block');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('is-invalid');
				$frameworks['types']['bootstrap41']['fields']['public']['field_types']['stri_el_cc']['class_valid_field'] = array('is-valid');

				// Foundation
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('expand');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_cc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_cc']['class_field'] = array('panel');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_cc']['class_invalid_label'] = array('error');
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array();
				$frameworks['types']['foundation5']['fields']['public']['field_types']['stri_el_cc']['class_invalid_invalid_feedback'] = array('error');

				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_cc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_cc']['class_field'] = array('callout');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_cc']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation6']['fields']['public']['field_types']['stri_el_cc']['class_invalid_invalid_feedback'] = array('is-visible');

				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_checkout']['mask_field_label'] = '#label';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_checkout']['class_field'] = array('button');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_checkout']['class_field_full_button'] = array('expanded');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_checkout']['class_field_button_type_fallback'] = 'primary';

				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_cc']['mask_field_label'] = '<label id="#label_id" for="#id"#attributes>#label</label>';
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_cc']['class_field'] = array('callout');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_cc']['class_invalid_label'] = array('is-invalid-label');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_cc']['class_invalid_field'] = array('is-invalid-input');
				$frameworks['types']['foundation64']['fields']['public']['field_types']['stri_el_cc']['class_invalid_invalid_feedback'] = array('is-visible');

				return $frameworks;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-stripe-elements')));

				return $links;
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please enter your API keys <a href="%s">here</a>.', 'ws-form-'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Plug-in field types for this action
			public function config_field_types($field_types) {

				$field_types['ecommerce']['types']['stri_el_cc'] = array (

					'label'						=>	__('Stripe Card', 'ws-form'),
					'pro_required'				=>	!WS_Form_Common::is_edition('pro'),
// @if NODE_ENV='license_pro'
					'label_default'				=>	__('Credit or Debit Card', 'ws-form'),
					'mask_field'				=>	'#pre_label<div id="#id" data-wsf-stripe-cc#attributes></div>#post_label#invalid_feedback#help<input type="hidden" id="#id-mode" name="wsf_stripe_elements_mode" /><input type="hidden" id="#id-json" name="wsf_stripe_elements_json" /><input type="hidden" id="#id-token" name="wsf_stripe_elements_token" /><input type="hidden" id="#id-field-id" name="wsf_stripe_elements_field_id" value="#field_id" /><script>

(function($) {

	if(typeof(window.wsf_stripe_elements_functions) === \'undefined\') {
	
		window.wsf_stripe_elements_functions = [];
	}

	window.wsf_stripe_elements_functions[#form_instance_id] = function() {

		var wsf_stripe_elements = new $.WS_Form_Stripe_Elements();
		wsf_stripe_elements.id = \'##id\';
		wsf_stripe_elements.id_mode = \'#id-mode\';
		wsf_stripe_elements.id_json = \'#id-json\';
		wsf_stripe_elements.id_token = \'#id-token\';
		wsf_stripe_elements.form_instance_id = \'#form_instance_id\';
		wsf_stripe_elements.transaction_type = \'#action_' . $this->id . '_transaction_type\';
		wsf_stripe_elements.transaction_amount = \'#action_' . $this->id . '_transaction_amount\';
		wsf_stripe_elements.transaction_zero = \'#action_' . $this->id . '_transaction_zero\';
		wsf_stripe_elements.transaction_zero_message = \'#action_' . $this->id . '_transaction_zero_message\';
		wsf_stripe_elements.options_icon_style = \'#action_' . $this->id . '_options_icon_style\';
		wsf_stripe_elements.options_icon_hidden = \'#action_' . $this->id . '_options_icon_hidden\';
		wsf_stripe_elements.message_method = \'#action_' . $this->id . '_message_method\';
		wsf_stripe_elements.message_clear = \'#action_' . $this->id . '_message_clear\';
		wsf_stripe_elements.message_scroll_top = \'#action_' . $this->id . '_message_scroll_top\';
		wsf_stripe_elements.message_form_hide = \'#action_' . $this->id . '_message_form_hide\';
		wsf_stripe_elements.message_duration = \'#action_' . $this->id . '_message_duration\';
		wsf_stripe_elements.publishable_key = \'' . $this->publishable_key . '\';

		wsf_stripe_elements.init();
	};

})(jQuery);

					</script>',
					'mask_field_attributes'		=>	array('class'),
					'mask_field_label'			=>	'<label id="#label_id" for="#id"#attributes>#label</label>',
					'mask_field_label_attributes'	=>	array('class'),
					'submit_save'				=>	false,
					'submit_edit'				=>	false,
					'progress'					=>	false,
					'multiple'					=>	false,
					'icon'						=>	'<svg height="16" width="16" viewBox="0 0 16 16"><path d="M6.9 4.8c0-.7.6-.9 1.5-.9 1.3 0 3 .4 4.4 1.1V.8C11.3.2 9.8 0 8.4 0c-3.6 0-6 1.9-6 5 0 4.9 6.7 4.1 6.7 6.2 0 .8-.7 1.1-1.7 1.1-1.5 0-3.3-.6-4.8-1.4V15c1.6.7 3.3 1 4.8 1 3.7 0 6.2-1.8 6.2-5 0-5.2-6.7-4.3-6.7-6.2z"/></svg>',
					'conditional'				=>	array(

						'exclude_condition'	=>	true,
						'actions_enabled'	=>	array('visibility', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field')
					),

					'meta_key_parse_variables'	=>	array(

						'action_' . $this->id . '_transaction_type',
						'action_' . $this->id . '_transaction_amount',
						'action_' . $this->id . '_transaction_currency',
						'action_' . $this->id . '_transaction_zero',
						'action_' . $this->id . '_transaction_zero_message',
						'action_' . $this->id . '_options_icon_style',
						'action_' . $this->id . '_options_icon_hidden',
						'action_' . $this->id . '_message_method',
						'action_' . $this->id . '_message_clear',
						'action_' . $this->id . '_message_scroll_top',
						'action_' . $this->id . '_message_form_hide',
						'action_' . $this->id . '_message_duration'
					),

					'fieldsets'				=> array(

						// Tab: Basic
						'basic'	=> array(

							'label'			=>	__('Basic', 'ws-form'),
							'meta_keys'	=>	array('label_render', 'label_position', 'label_column_width', 'hidden', 'help'),

							'fieldsets'		=>	array(

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_type',
										'action_' . $this->id . '_transaction_amount',
										'action_' . $this->id . '_transaction_currency',
										'action_' . $this->id . '_transaction_description',
										'action_' . $this->id . '_subscription_price_mapping',
										'action_' . $this->id . '_subscription_trial',
										'action_' . $this->id . '_subscription_trial_period_days',
										'action_' . $this->id . '_subscription_invoice_item_mapping'
									)
								),

								array(

									'label'		=>	__('Customer', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_customer_create',
										'action_' . $this->id . '_customer_name',
										'action_' . $this->id . '_customer_email',
										'action_' . $this->id . '_customer_phone',
										'action_' . $this->id . '_customer_description',
										'action_' . $this->id . '_billing_address_mapping',
										'action_' . $this->id . '_shipping_name',
										'action_' . $this->id . '_shipping_phone',
										'action_' . $this->id . '_shipping_address_mapping'
									)
								),

								array(
									'label'		=>	__('Accessibility', 'ws-form'),
									'meta_keys'	=>	array('aria_label')
								)
							)
						),

						// Tab: Advanced
						'advanced'	=>	array(

							'label'			=>	__('Advanced', 'ws-form'),

							'fieldsets'	=>	array(

								array(
									'label'		=>	__('Styles', 'ws-form'),
									'meta_keys'	=>	array('class_single_vertical_align_bottom', 'action_' . $this->id . '_options_icon_hidden', 'action_' . $this->id . '_options_icon_style',)
								),

								array(
									'label'		=>	__('Classes', 'ws-form'),
									'meta_keys'	=>	array('class_field_wrapper', 'class_field')
								),

								array(

									'label'		=>	__('Transaction', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_transaction_zero',
										'action_' . $this->id . '_transaction_zero_message'
									)
								),

								array(

									'label'		=>	__('Messages', 'ws-form'),
									'meta_keys'	=> array(

										'action_' . $this->id . '_message_method',
										'action_' . $this->id . '_message_clear',
										'action_' . $this->id . '_message_scroll_top',
										'action_' . $this->id . '_message_form_hide',
										'action_' . $this->id . '_message_duration'
									)
								),

								array(
									'label'		=>	__('Validation', 'ws-form'),
									'meta_keys'	=>	array('invalid_feedback_render')
								),

								array(
									'label'		=>	__('Custom Attributes', 'ws-form'),
									'meta_keys'	=>	array('custom_attributes')
								),

								array(
									'label'		=>	__('Breakpoints', 'ws-form'),
									'meta_keys'	=> array('breakpoint_sizes'),
									'class'		=>	array('wsf-fieldset-panel')
								)
							)
						)
					)
				);

				$field_types['ecommerce']['types']['stri_el_checkout'] = array(

					'label'						=>	__('Stripe Submit', 'ws-form'),
					'pro_required'				=>	!WS_Form_Common::is_edition('pro'),
					'label_default'				=>	__('Submit Payment', 'ws-form'),
					'label_position_force'		=>	'top',
					'mask_field'				=>	'<button type="button" data-ecommerce-payment data-wsf-stripe-submit id="#id" name="#name"#attributes>#label</button>#help',
					'mask_field_attributes'		=>	array('class', 'disabled', 'aria_describedby', 'aria_labelledby', 'aria_label', 'custom_attributes'),
					'mask_field_label'			=>	'#label',
					'submit_save'				=>	false,
					'submit_edit'				=>	false,
					'progress'					=>	false,
					'icon'						=>	'<svg height="16" width="16" viewBox="0 0 16 16"><path d="M6.9 4.8c0-.7.6-.9 1.5-.9 1.3 0 3 .4 4.4 1.1V.8C11.3.2 9.8 0 8.4 0c-3.6 0-6 1.9-6 5 0 4.9 6.7 4.1 6.7 6.2 0 .8-.7 1.1-1.7 1.1-1.5 0-3.3-.6-4.8-1.4V15c1.6.7 3.3 1 4.8 1 3.7 0 6.2-1.8 6.2-5 0-5.2-6.7-4.3-6.7-6.2z"/></svg>',
					'conditional'				=>	array(

						'logics_enabled'	=>	array('click', 'mouseover', 'mouseout', 'focus', 'blur'),
						'actions_enabled'	=>	array('visibility', 'focus', 'button_html', 'click', 'disabled', 'class_add_wrapper', 'class_remove_wrapper', 'class_add_field', 'class_remove_field'),
						'condition_event'	=>	'click',
					),
					'events'					=>	array(

						'event'				=>	'click',
						'event_category'	=>	'Button'
					),

					'fieldsets'				=> array(

						// Tab: Basic
						'basic'	=> array(

							'label'			=>	__('Basic', 'ws-form'),
							'meta_keys'		=>	array('hidden', 'help')
						),

						// Tab: Advanced
						'advanced'	=>	array(

							'label'			=>	__('Advanced', 'ws-form'),

							'fieldsets'	=>	array(

								array(
									'label'		=>	__('Styles', 'ws-form'),
									'meta_keys'	=>	array('class_single_vertical_align_bottom', 'class_field_button_type_primary', 'class_field_full_button_remove')
								),

								array(
									'label'		=>	__('Classes', 'ws-form'),
									'meta_keys'	=>	array('class_field_wrapper', 'class_field')
								),

								array(
									'label'		=>	__('Restrictions', 'ws-form'),
									'meta_keys'	=>	array('disabled')
								),

								array(
									'label'		=>	__('Custom Attributes', 'ws-form'),
									'meta_keys'	=>	array('custom_attributes')
								),

								array(
									'label'		=>	__('Breakpoints', 'ws-form'),
									'meta_keys'	=> array('breakpoint_sizes'),
									'class'		=>	array('wsf-fieldset-panel')
								)
							)
						)
					)
				);

				// Parse JS
				$mask_field = $field_types['ecommerce']['types']['stri_el_cc']['mask_field'];

				$mask_fields = array(

					'skin_font_size' => WS_Form_Common::option_get('skin_font_size') . 'px',
					'skin_font_family' => WS_Form_Common::option_get('skin_font_family'),
					'skin_color_default_lighter' => WS_Form_Common::option_get('skin_color_default_lighter'),
					'skin_color_default' => WS_Form_Common::option_get('skin_color_default'),
					'skin_color_danger' => WS_Form_Common::option_get('skin_color_danger'),
					'action_' . $this->id . '_environment' => $this->environment,
					'action_' . $this->id . '_test_publishable_key' => $this->test_publishable_key,
					'action_' . $this->id . '_live_publishable_key' => $this->live_publishable_key,
				);

				// Parse with mask fields
				$field_types['ecommerce']['types']['stri_el_cc']['mask_field'] = WS_Form_Common::mask_parse($mask_field, $mask_fields);

				return $field_types;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-stripe-elements') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-stripe-elements') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

						break;
				}
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { self::error(sprintf(__('Action not configured (%s)', 'ws-form-stripe-elements'), $this->label)); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// Transaction - Type
					'action_' . $this->id . '_transaction_type'	=> array(

						'label'						=>	__('Type', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'default'					=>	'',
						'options'					=>	array(

							array('value' => '', 'text' => __('One Time', 'ws-form-stripe-elements')),
							array('value' => 'subscription', 'text' => __('Subscription', 'ws-form-stripe-elements')),
							array('value' => 'setup_intent', 'text' => __('Setup Intent', 'ws-form-stripe-elements'))
						)
					),

					// Transaction - Amount
					'action_' . $this->id . '_transaction_amount'	=> array(

						'label'						=>	__('Amount to Charge', 'ws-form-stripe-elements'),
						'type'						=>	'text',
						'default'					=>	'',
						'help'						=>	__('Enter the amount to charge the user. Leave blank to use cart total.', 'ws-form-stripe-elements'),
						'select_list'				=>	true,
						'placeholder'				=>	'#ecommerce_cart_total',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	''
							)
						)
					),

					// Transaction - Currency
					'action_' . $this->id . '_transaction_currency'	=> array(

						'label'						=>	__('Currency', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'default'					=>	WS_Form_Common::option_get('currency', 'USD'),
						'help'						=>	__('Choose the currency for the transaction.', 'ws-form-stripe-elements'),
						'options'					=> array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	''
							)
						)
					),

					// Transaction - Description
					'action_' . $this->id . '_transaction_description'	=> array(

						'label'						=>	__('Description', 'ws-form-stripe-elements'),
						'type'						=>	'textarea',
						'help'						=>	__('A description of the transaction shown to the user.', 'ws-form-stripe-elements'),
						'default'					=>	'Payment to #blog_name',
						'condition'					=>	array(

							array(

								'logic'				=>	'!=',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Subscription - Prices
					'action_' . $this->id . '_subscription_price_mapping'	=> array(

						'label'						=>	__('Subscription Price API ID(s)', 'ws-form-stripe-elements'),
						'type'						=>	'repeater',
						'help'						=>	sprintf(__('Enter one or more Price API IDs to associate with this transaction. <a href="%s" target="_blank">Learn more</a>', 'ws-form-stripe-elements'), WS_Form_Common::get_plugin_website_url('/knowledgebase/stripe-elements/')),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_subscription_price_id'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Subscription - Price ID
					'action_' . $this->id . '_subscription_price_id'	=> array(

						'label'						=>	__('Price API ID', 'ws-form-post'),
						'type'						=>	'text'
					),

					// Subscription - Trial
					'action_' . $this->id . '_subscription_trial'	=> array(

						'label'						=>	__('Trial', 'ws-form-post'),
						'type'						=>	'select',
						'help'						=>	__('Choose how to apply a trial to this subscription.', 'ws-form'),
						'options'					=>	array(

							array('value' => '', 'text' => __('No Trial', 'ws-form-stripe-elements')),
							array('value' => 'trial_period_days', 'text' => __('Number of Days', 'ws-form-stripe-elements')),
							array('value' => 'trial_from_plan', 'text' => __('Use Plan Setting', 'ws-form-stripe-elements'))
						),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Subscription - Trial Period Days
					'action_' . $this->id . '_subscription_trial_period_days'	=> array(

						'label'						=>	__('Number of Days', 'ws-form-post'),
						'type'						=>	'text',
						'help'						=>	__('Enter the number of trial period days before the customer is charged for the first time. This will always overwrite any trials that might apply via a subscribed plan.', 'ws-form'),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							),

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_subscription_trial',
								'meta_value'		=>	'trial_period_days'
							)
						)
					),

					// Subscription - Invoice Items
					'action_' . $this->id . '_subscription_invoice_item_mapping'	=> array(

						'label'						=>	__('Initial Invoice Item Price API ID(s)', 'ws-form-stripe-elements'),
						'type'						=>	'repeater',
						'help'						=>	sprintf(__('Enter one or more Price API IDs to add as invoice items to the initial transaction. <a href="%s" target="_blank">Learn more</a>', 'ws-form-stripe-elements'), WS_Form_Common::get_plugin_website_url('/knowledgebase/stripe-elements/')),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_subscription_price_id'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Options - Icon Hidden
					'action_' . $this->id . '_options_icon_hidden'	=> array(

						'label'						=>	__('Icon Hidden', 'ws-form-stripe-elements'),
						'type'						=>	'checkbox',
						'help'						=>	__('Should the payment icon be hidden?.', 'ws-form-stripe-elements'),
						'default'					=>	''
					),

					// Options - Icon Style
					'action_' . $this->id . '_options_icon_style'	=> array(

						'label'						=>	__('Icon Style', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'help'						=>	__('Appearance of the icon.', 'ws-form-stripe-elements'),
						'options'					=>	array(

							array('value' => 'default', 'text' => __('Default', 'ws-form-stripe-elements')),
							array('value' => 'solid', 'text' => __('Solid', 'ws-form-stripe-elements'))
						),
						'default'					=>	'default',
						'condition'					=>	array(

							array(

								'logic'				=>	'!=',
								'meta_key'			=>	'action_' . $this->id . '_options_icon_hidden',
								'meta_value'		=>	'on'
							)
						)
					),

					// Customer - Create
					'action_' . $this->id . '_customer_create'	=> array(

						'label'						=>	__('Create Customer', 'ws-form-stripe-elements'),
						'type'						=>	'checkbox',
						'help'						=>	__('Creating a customer can improve your ability to cross reference transactions in Stripe.', 'ws-form-stripe-elements'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	''
							)
						)
					),

					// Customer - Name
					'action_' . $this->id . '_customer_name'	=> array(

						'label'						=>	__('Name', 'ws-form-stripe-elements'),
						'type'						=>	'text',
						'help'						=>	__('The name of the customer e.g. #field(123).', 'ws-form-stripe-elements'),
						'select_list'				=>	true,
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Customer - Email
					'action_' . $this->id . '_customer_email'	=> array(

						'label'						=>	__('Email', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form-stripe-elements'),
						'fields_filter_type'		=>	array('email'),
						'help'						=>	__('Select the customer email field.', 'ws-form-stripe-elements'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Customer - Phone
					'action_' . $this->id . '_customer_phone'	=> array(

						'label'						=>	__('Phone', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form-stripe-elements'),
						'fields_filter_type'		=>	array('tel', 'text'),
						'help'						=>	__('Select the customer phone field.', 'ws-form-stripe-elements'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Customer - Description
					'action_' . $this->id . '_customer_description'	=> array(

						'label'						=>	__('Description', 'ws-form-stripe-elements'),
						'type'						=>	'text',
						'help'						=>	__('A description to include with the customer.', 'ws-form-stripe-elements'),
						'select_list'				=>	true,
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Billing - Address mapping
					'action_' . $this->id . '_billing_address_mapping'	=> array(

						'label'						=>	__('Billing Address Mapping', 'ws-form-stripe-elements'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Stripe customer mailing address fields.', 'ws-form-stripe-elements'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_billing_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_billing_address_field'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Billing - Address mapping - Fields
					'action_' . $this->id . '_billing_address_field' => array(

						'label'						=>	__('Stripe Field', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'line1', 'text' => __('Address Line 1', 'ws-form-stripe-elements')),
							array('value' => 'line2', 'text' => __('Address Line 2', 'ws-form-stripe-elements')),
							array('value' => 'city', 'text' => __('City', 'ws-form-stripe-elements')),
							array('value' => 'state', 'text' => __('State', 'ws-form-stripe-elements')),
							array('value' => 'postal_code', 'text' => __('Zip', 'ws-form-stripe-elements')),
							array('value' => 'country', 'text' => __('Country', 'ws-form-stripe-elements')),
						)
					),

					// Shipping - Name
					'action_' . $this->id . '_shipping_name'	=> array(

						'label'						=>	__('Shipping Name', 'ws-form-stripe-elements'),
						'type'						=>	'text',
						'help'						=>	__('The customers name. This setting is required if you are mapping shipping address fields.', 'ws-form-stripe-elements'),
						'select_list'				=>	true,
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Shipping - Phone
					'action_' . $this->id . '_shipping_phone'	=> array(

						'label'						=>	__('Shipping Phone', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	'fields',
						'options_blank'				=>	__('Select...', 'ws-form-stripe-elements'),
						'fields_filter_type'		=>	array('tel', 'text'),
						'help'						=>	__('Select the customer phone field.', 'ws-form-stripe-elements'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Shipping - Address mapping
					'action_' . $this->id . '_shipping_address_mapping'	=> array(

						'label'						=>	__('Shipping Address Mapping', 'ws-form-stripe-elements'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Stripe customer shipping address fields.', 'ws-form-stripe-elements'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_shipping_address_field'
						),
						'meta_keys_unique'			=>	array(

							'action_' . $this->id . '_shipping_address_field'
						),
						'default'					=>	array(),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_customer_create',
								'meta_value'		=>	'on'
							),

							array(

								'logic_previous'	=>	'||',
								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_type',
								'meta_value'		=>	'subscription'
							)
						)
					),

					// Shipping - Address mapping - Fields
					'action_' . $this->id . '_shipping_address_field' => array(

						'label'						=>	__('Stripe Field', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => 'line1', 'text' => __('Address Line 1', 'ws-form-stripe-elements')),
							array('value' => 'line2', 'text' => __('Address Line 2', 'ws-form-stripe-elements')),
							array('value' => 'city', 'text' => __('City', 'ws-form-stripe-elements')),
							array('value' => 'state', 'text' => __('State', 'ws-form-stripe-elements')),
							array('value' => 'postal_code', 'text' => __('Zip', 'ws-form-stripe-elements')),
							array('value' => 'country', 'text' => __('Country', 'ws-form-stripe-elements')),
						)
					),

					// Transaction - Zero Amount
					'action_' . $this->id . '_transaction_zero'	=> array(

						'label'						=>	__('Zero Amount', 'ws-form-stripe-elements'),
						'type'						=>	'select',
						'options'					=>	array(

							array('value' => '', 'text' => __('Do nothing', 'ws-form-stripe-elements')),
							array('value' => 'submit', 'text' => __('Submit form', 'ws-form-stripe-elements')),
							array('value' => 'message', 'text' => __('Show error message', 'ws-form-stripe-elements')),
						),
						'default'					=>	'message',
						'help'						=>	__('What should happen if the cart total is zero and the Stripe button is clicked?', 'ws-form-stripe-elements'),
					),

					// Transaction - Zero Amount - Message
					'action_' . $this->id . '_transaction_zero_message'	=> array(

						'label'						=>	__('Zero Amount Message', 'ws-form-stripe-elements'),
						'type'						=>	'textarea',
						'help'						=>	__('Enter a message to show if the Stripe button is clicked and the cart amount is zero.', 'ws-form-stripe-elements'),
						'default'					=>	__('The total cart amount cannot be zero. Please review the form and try again.', 'ws-form-stripe-elements'),
						'condition'					=>	array(

							array(

								'logic'				=>	'==',
								'meta_key'			=>	'action_' . $this->id . '_transaction_zero',
								'meta_value'		=>	'message'
							)
						)
					),

					// Message - Method
					'action_' . $this->id . '_message_method'	=> array(

						'label'						=>	__('Error Messages', 'ws-form-post'),
						'type'						=>	'select',
						'help'						=>	__('Where should error messages be added?', 'ws-form-post'),
						'options'					=>	array(

							array('value' => 'before', 'text' => __('Before Form', 'ws-form-post')),
							array('value' => 'after', 'text' => __('After Form', 'ws-form-post'))
						),
						'default'					=>	'after',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Duration
					'action_' . $this->id . '_message_duration'	=> array(

						'label'						=>	__('Error Message Duration (ms)', 'ws-form-post'),
						'type'						=>	'number',
						'help'						=>	__('Duration in milliseconds error message shown.', 'ws-form-post'),
						'default'					=>	'4000',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Clear other messages
					'action_' . $this->id . '_message_clear'	=> array(

						'label'						=>	__('Clear Other Messages On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Clear any other messages when error message shown?', 'ws-form'),
						'default'					=>	'on',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Form - Scroll to top
					'action_' . $this->id . '_message_scroll_top'	=> array(

						'label'						=>	__('Scroll To Top On Error', 'ws-form'),
						'type'						=>	'checkbox',
						'help'						=>	__('Scroll to top of page when error message shown?', 'ws-form'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					),

					// Message - Form Hide
					'action_' . $this->id . '_message_form_hide'	=> array(

						'label'						=>	__('Hide Form On Error', 'ws-form-post'),
						'type'						=>	'checkbox',
						'help'						=>	__('Hide form when error message shown?', 'ws-form-post'),
						'default'					=>	'',
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							)
						)
					)
				);

				// Add currencies
				$currencies = WS_Form_Config::get_currencies();
				foreach($currencies as $currency_code => $currency_config) {

					$config_meta_keys['action_' . $this->id . '_transaction_currency']['options'][] = array('value' => $currency_code, 'text' => sprintf('%s (%s)', $currency_config['n'], $currency_config['s']));
				}

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-stripe-elements'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-stripe-elements'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Stripe Elements add-on for WS Form PRO license key here.', 'ws-form-stripe-elements'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-stripe-elements'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_environment'	=>	array(

							'label'		=>	__('Environment', 'ws-form-stripe-elements'),
							'type'		=>	'select',
							'help'		=>	__('Select which Stripe environment to use.', 'ws-form'),
							'default'	=>	'sandbox',
							'options'	=>	array(

								'test'		=>	array('text' => __('Test', 'ws-form')),
								'live'		=>	array('text' => __('Live', 'ws-form')),
							)				
						),

						'action_' . $this->id . '_test_publishable_key'	=>	array(

							'label'			=>	__('Test publishable key', 'ws-form-stripe-elements'),
							'type'			=>	'text',
							'help'			=>	__('Enter your test publishable key. You can find it by <a href="https://dashboard.stripe.com/account/apikeys" target="_blank">clicking here</a>. If you do not yet have a Stripe account <a href="https://dashboard.stripe.com/register" target="_blank">click here</a>.', 'ws-form-stripe-elements')
						),

						'action_' . $this->id . '_test_secret_key'	=>	array(

							'label'			=>	__('Test secret key', 'ws-form-stripe-elements'),
							'type'			=>	'text',
							'help'			=>	__('Enter your test secret key. You can find it by <a href="https://dashboard.stripe.com/account/apikeys" target="_blank">clicking here</a>. If you do not yet have a Stripe account <a href="https://dashboard.stripe.com/register" target="_blank">click here</a>.', 'ws-form-stripe-elements')
						),

						'action_' . $this->id . '_live_publishable_key'	=>	array(

							'label'			=>	__('Live publishable key', 'ws-form-stripe-elements'),
							'type'			=>	'text',
							'help'			=>	__('Enter your live publishable key. You can find it by <a href="https://dashboard.stripe.com/account/apikeys" target="_blank">clicking here</a>. If you do not yet have a Stripe account <a href="https://dashboard.stripe.com/register" target="_blank">click here</a>.', 'ws-form-stripe-elements')
						),

						'action_' . $this->id . '_live_secret_key'	=>	array(

							'label'			=>	__('Live secret key', 'ws-form-stripe-elements'),
							'type'			=>	'text',
							'help'			=>	__('Enter your live secret key. You can find it by <a href="https://dashboard.stripe.com/account/apikeys" target="_blank">clicking here</a>. If you do not yet have a Stripe account <a href="https://dashboard.stripe.com/register" target="_blank">click here</a>.', 'ws-form-stripe-elements')
						)
					)
				);

				return $options;
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->environment  			= WS_Form_Common::option_get('action_' . $this->id . '_environment', '');
				$this->test_publishable_key 	= WS_Form_Common::option_get('action_' . $this->id . '_test_publishable_key', '');
				$this->test_secret_key  		= WS_Form_Common::option_get('action_' . $this->id . '_test_secret_key', '');
				$this->live_publishable_key  	= WS_Form_Common::option_get('action_' . $this->id . '_live_publishable_key', '');
				$this->live_secret_key  		= WS_Form_Common::option_get('action_' . $this->id . '_live_secret_key', '');

				switch($this->environment) {

					case 'live' :

						$this->publishable_key = $this->live_publishable_key;
						$this->secret_key = $this->live_secret_key;
						break;

					default :

						$this->publishable_key = $this->test_publishable_key;
						$this->secret_key = $this->test_secret_key;
				}

				// Check configuration
				if(
					(empty($this->test_publishable_key) || empty($this->test_secret_key)) &&
					(empty($this->live_publishable_key) || empty($this->live_secret_key))
				) {
					$this->configured = false;
				}

				return $this->configured;
			}

			// Get API headers
			public function get_api_headers() {

				$bearer_key = ($this->environment == 'live') ? $this->live_secret_key : $this->test_secret_key;

				$headers = array('Authorization' => 'Bearer ' . $bearer_key);

				return $headers;
			}

			// Escape CSS values
			public function e($css_value) {

				$css_value = wp_strip_all_tags($css_value);
				$css_value = str_replace(';', '', $css_value);
				echo $css_value;
			}

			public function get_skin($css_return) {

				ob_start();
				self::render_skin();
				$css_return .= ob_get_contents();
				ob_end_clean();

				return $css_return;
			}

			// Stripe Elements CSS
			public function render_skin() {

				// Customizer
				$enable_cache = !(WS_Form_Common::get_query_var('customize_theme') !== '');

				$customize_groups = WS_Form_Config::get_customize();

				// Set skin variables
				foreach($customize_groups as $customize_group) {

					foreach($customize_group['fields'] as $meta_key => $config) {

						$variable = str_replace('skin_', '', $meta_key);

						$default = isset($config['default']) ? $config['default'] : '';

						$type = isset($config['type']) ? $config['type'] : '';

						$$variable = WS_Form_Common::option_get($meta_key, $default, false, $enable_cache, true);
					}
				}

				// Advanced
				$unit_of_measurement = 'px';

				// Spacing
				$spacing = 20;
				$spacing_extra_large = 80;
				$spacing_large = 40;
				$spacing_small = 10;
				$spacing_extra_small = 5;

				// Forms
				$form_background_color = $color_default_inverted;
				$form_border = $border; // true | false
				$form_border_color = $color_default_lighter;
				$form_border_style = $border_style;
				$form_border_width = $border_width;
				$form_border_radius = $border_radius;
				$form_box_shadow = true;
				$form_box_shadow_width = 2;
				$form_color = $color_default;
				$form_error_background_color = $form_background_color;
				$form_error_border_color = $color_danger;
				$form_error_color = $form_color;
				$form_focus = true; // true | false
				$form_focus_background_color = $form_background_color;
				$form_focus_border_color = $color_primary;
				$form_focus_color = $form_color;
				$form_font_size = $font_size;
				$form_hover = false; // true | false
				$form_hover_background_color = $form_background_color;
				$form_hover_border_color = $color_primary;
				$form_hover_color = $form_color;
				$form_spacing_horizontal = $spacing_small;
				$form_spacing_vertical = ($spacing_small * .85);
				$form_transition = $transition; // true | false
				$form_transition_timing_function = $transition_timing_function;
				$form_transition_speed = $transition_speed . 'ms ' . $form_transition_timing_function;
				$input_height = ((($form_font_size * $line_height) + ($form_spacing_vertical * 2)) + ($form_border_width * 2));
?>
.StripeElement.wsf-field {
	background-color: <?php echo $form_background_color; ?>;
<?php if ($form_border) { ?>
	border: <?php echo $form_border_width . $unit_of_measurement . ' ' . $form_border_style . ' ' . $form_border_color; ?>;
<?php } ?>
<?php if ($form_border_radius > 0) { ?>
	border-radius: <?php echo $form_border_radius . $unit_of_measurement; ?>;
<?php } ?>
	cursor: text;
	font-family: <?php echo $font_family; ?>;
	font-size: <?php echo $form_font_size . $unit_of_measurement; ?>;
	height: <?php echo $input_height . $unit_of_measurement; ?>;
	line-height: <?php echo $line_height; ?>;
	padding: <?php echo $form_spacing_vertical . $unit_of_measurement . ' ' . $form_spacing_horizontal . $unit_of_measurement; ?>;
<?php if ($form_transition) { ?>
	transition: background-color <?php self::e($form_transition_speed); ?>, border-color <?php self::e($form_transition_speed); ?>, box-shadow <?php self::e($form_transition_speed); ?>;
<?php } ?>
	width: 100%;
}

<?php if ($form_hover) { ?>
.StripeElement.wsf-field:hover {
<?php if ($form_hover_background_color != $form_background_color) { ?>
	background-color: <?php echo $form_hover_background_color; ?>;
<?php } ?>
<?php if ($form_hover_border_color != $form_border_color) { ?>
	border-color: <?php echo $form_hover_border_color; ?>;
<?php } ?>
<?php if ($form_hover_color != $form_color) { ?>
	color: <?php echo $form_hover_color; ?>;
<?php } ?>
}
<?php } ?>

.StripeElement.wsf-field.StripeElement--focus {
<?php if ($form_focus) { ?>
<?php if ($form_focus_background_color != $form_background_color) { ?>
	background-color: <?php echo $form_focus_background_color; ?>;
<?php } ?>
<?php if ($form_focus_border_color != $form_border_color) { ?>
	border-color: <?php echo $form_focus_border_color; ?>;
<?php } ?>
<?php if ($form_box_shadow) { ?>
	box-shadow: 0 0 0 <?php self::e($form_box_shadow_width . $unit_of_measurement); ?> <?php self::e(WS_Form_Common::hex_lighten_percentage($color_primary, 85)); ?>;
<?php } ?>
<?php if ($form_focus_color != $form_color) { ?>
	color: <?php echo $form_focus_color; ?>;
<?php } ?>
	outline: 0;
}
<?php } ?>

.StripeElement.wsf-field.StripeElement--invalid {
<?php if ($form_error_background_color != $form_background_color) { ?>
	background-color: <?php echo $form_error_background_color; ?>;
<?php } ?>
<?php if ($form_border) { ?>
<?php if ($form_error_border_color != $form_border_color) { ?>
	border-color: <?php echo $form_error_border_color; ?>;
<?php } ?>
<?php } ?>
}

<?php if ($form_focus) { ?>
<?php if ($form_box_shadow) { ?>
.StripeElement.wsf-field.StripeElement--invalid.StripeElement--focus {
	box-shadow: 0 0 0 <?php self::e($form_box_shadow_width . $unit_of_measurement); ?> <?php self::e(WS_Form_Common::hex_lighten_percentage($form_error_border_color, 85)); ?>;
}
<?php } ?>
<?php } ?>

.StripeElement--webkit-autofill {
	background-color: #fefde5 !important;
}

.StripeElement.callout {
	padding: 0.5rem;
}
<?php
			}
		}
	});