<?php

	/**
	 * @link              https://wsform.com/knowledgebase/zapier/
	 * @since             1.0.0
	 * @package           WS_Form_Zapier
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Zapier
	 * Plugin URI:        https://wsform.com/knowledgebase/zapier/
	 * Description:       Zapier add-on for WS Form PRO
	 * Version:           1.0.7
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-zapier
	 */

	Class WS_Form_Add_On_Zapier {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.7.115';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Zapier();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-zapier'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_zapier = new WS_Form_Add_On_Zapier();	

	register_activation_hook(__FILE__, array($wsf_add_on_zapier, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Zapier extends WS_Form_Action {

			public $id = 'zapier';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 50;
			public $can_repost = true;
			public $form_add = false;

			// Licensing
			private $licensing;

			// Config
			private $api_key = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 2672;
			const WS_FORM_LICENSE_NAME = 'Zapier add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.7';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Zapier', 'ws-form-zapier');

				// Set label for actions pull down
				$this->label_action = __('Trigger Zap', 'ws-form-zapier');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-zapier')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_api_url' :

						$value = get_site_url();
						break;

					case 'action_' . $this->id . '_api_key' :

						$value = $this->api_key;
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-zapier') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-zapier') . '" />';
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;
						}

					break;
				}
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-zapier') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-zapier'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-zapier'),
							'type'		=>	'text',
							'help'		=>	__('Enter your zapier add-on for WS Form PRO license key here.', 'ws-form-zapier'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-zapier'),
							'type'		=>	'static'

						),

						'action_' . $this->id . '_api_url'	=>	array(

							'label'		=>	__('API URL', 'ws-form-zapier'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_key'	=>	array(

							'label'		=>	__('API Key', 'ws-form-zapier'),
							'type'		=>	'text',
							'help'		=>	sprintf(__('<a href="%s" target="_blank">Learn more</a>', 'ws-form-zapier'), WS_Form_Common::get_plugin_website_url('/knowledgebase/zapier/'))
						),
					)
				);

				return $options;
			}

			// Load config for this action
			public function load_config($config = array()) {

			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = false;

				$this->api_key = WS_Form_Common::option_get('action_' . $this->id . '_api_key', false);
				if(empty($this->api_key)) {

					// Build random API key
					$this->api_key = md5(get_site_url() . $this->id . time());
					WS_Form_Common::option_set('action_' . $this->id . '_api_key', $this->api_key);
				}

				// Check API endpoint
				$this->configured = true;

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/test/', array('methods' => 'GET', 'callback' => array($this, 'api_test'), 'permission_callback' => '__return_true'));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/trigger/submission/list/(?P<form_id>[\d]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_trigger_submission'), 'permission_callback' => '__return_true'));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/trigger/submission/fields/input/', array('methods' => 'GET', 'callback' => array($this, 'api_trigger_submission_fields_input'), 'permission_callback' => '__return_true'));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/trigger/submission/fields/output/(?P<form_id>[\d]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_trigger_submission_fields_output'), 'permission_callback' => '__return_true'));
			}

			// API endpoint - Test
			public function api_test() {

				// Protect
				self::api_protect();

				// Process response
				self::api_zapier_response(false, array('api_key' => $this->api_key, 'connection_label' => get_bloginfo('name')));
			}

			// API endpoint - Submission input fields (These are the fields presented to the user in Zapier for configuring the submission trigger)
			public function api_trigger_submission_fields_input() {

				// Get forms from API
				$ws_form_form = New WS_Form_Form();
				$forms = $ws_form_form->db_read_all('', 'status="publish"', 'label ASC', '', '', false, true);

				// Build choices
				$choices = array();

				foreach($forms as $form) {

					$choices[$form['id']] = $form['label'];
				}

				$response_array = array(

					array(

						'label' => 		'Select Form',
						'key' => 		'form_id',
						'required' => 	true,
						'choices' => 	$choices,
						'helpText' =>	'Select which published form will trigger this zap. If you form does not appear in this list, please ensure you have published it first.'
					)
				);

				header('Content-Type: application/json');
				echo json_encode($response_array);
				exit;
			}

			// API endpoint - Submission output fields (Definition of the output fields)
			public function api_trigger_submission_fields_output($parameters) {

				// Form ID
				$form_id = intval(WS_Form_Common::get_query_var('form_id', 0, $parameters));
				if($form_id == 0) { exit; }

				// Add fixed fields
				$response_array = array(

					array('key' => 'id', 'label' => 'Submission ID'),
					array('key' => 'form_id', 'label' => 'Form ID'),
					array('key' => 'hash', 'label' => 'Hash'),
					array('key' => 'duration', 'label' => 'Duration'),
					array('key' => 'status', 'label' => 'Status'),
					array('key' => 'preview', 'label' => 'Preview Mode?'),
					array('key' => 'spam_level', 'label' => 'Spam Level'),
					array('key' => 'date_added', 'label' => 'Date Added'),
					array('key' => 'date_updated', 'label' => 'Date Updated'),
					array('key' => 'date_expire', 'label' => 'Date Expires'),
					array('key' => 'admin_url', 'label' => 'Admin URL')
				);

				// Create form object
				$ws_form_form = New WS_Form_Form();
				$ws_form_form->id = $form_id;

				// Read published form data
				$form_object = $ws_form_form->db_read_published();

				// Form not yet published
				if($form_object !== false) {

					// Get fields in single dimension array
					$fields = WS_Form_Common::get_fields_from_form($form_object);

					// Get field types in single dimension array
					$field_types = WS_Form_Config::get_field_types_flat();

					foreach($fields as $field_id => $field) {

						// If field type not specified, skip
						if(!isset($field->type)) { continue; };
						$field_type = $field->type;

						// If field type not known, skip
						if(!isset($field_types[$field_type])) { continue; };
						$field_type_config = $field_types[$field_type];

						// If field is not licensed, skip
						if(
							isset($field_type_config['pro_required']) &&
							$field_type_config['pro_required']

						) { continue; }

						// If field type not saved, skip
						$submit_save = isset($field_type_config['submit_save']) ? $field_type_config['submit_save'] : false;
						if(!$submit_save) { continue; }

						$required = (bool)WS_Form_Common::get_object_meta_value($field, 'required', false);

						// Add dynamic field
						$response_array[] = array('key' => WS_FORM_FIELD_PREFIX . $field_id, 'label' => $field->label, 'required' => $required, 'important' => true);
					}
				}

				header('Content-Type: application/json');
				echo json_encode($response_array);
				exit;
			}
			// API endpoint - Submission
			public function api_trigger_submission($parameters) {

				// Protect
				self::api_protect();

				// Form ID
				$form_id = intval(WS_Form_Common::get_query_var('form_id', 0, $parameters));
				if($form_id == 0) { exit; }

				// Pagination - Limit
				$limit = WS_Form_Common::get_query_var('limit', false);
				$limit = ($limit === false) ? '' : intval($limit);

				// Pagination - Offset
				$offset = WS_Form_Common::get_query_var('offset', false);
				$offset = ($offset === false) ? '' : intval($offset);

				// Get submissions in reverse chronological order
				$ws_form_submit = New WS_Form_Submit();
				$submissions = $ws_form_submit->db_read_all('', 'form_id = ' . $form_id, 'id DESC', $limit, $offset, true, true, true);

				// Create form object
				$ws_form_form = New WS_Form_Form();
				$ws_form_form->id = $form_id;

				// Read published form data
				$form_object = $ws_form_form->db_read_published();

				// Form not yet published
				$submission_fields = array();
				if($form_object !== false) {

					// Get fields in single dimension array
					$fields = WS_Form_Common::get_fields_from_form($form_object);
					// Get field types in single dimension array
					$field_types = WS_Form_Config::get_field_types_flat();

					foreach($fields as $field_id => $field) {

						// If field type not specified, skip
						if(!isset($field->type)) { continue; };
						$field_type = $field->type;

						// If field type not known, skip
						if(!isset($field_types[$field_type])) { continue; };
						$field_type_config = $field_types[$field_type];

						// If field is not licensed, skip
						if(
							isset($field_type_config['pro_required']) &&
							$field_type_config['pro_required']

						) { continue; }

						// If field type not saved, skip
						$submit_save = isset($field_type_config['submit_save']) ? $field_type_config['submit_save'] : false;
						if(!$submit_save) { continue; }

						$submission_fields[$field_id] = $field;
					}
				}

				$response_array = array();

				foreach($submissions as $submission) {

					// Add fixed fields
					$response_single = array(

						'id' =>				$submission->id,
						'form_id' =>		$form_id,
						'hash' =>			$submission->hash,
						'duration' =>		$submission->duration,
						'status' =>			$submission->status,
						'preview' =>		(bool)$submission->preview,
						'spam_level' =>		$submission->spam_level,
						'date_added' =>		date('c', strtotime($submission->date_added)),
						'date_updated' =>	date('c', strtotime($submission->date_updated)),
						'date_expire' =>	date('c', strtotime($submission->date_expire)),
						'admin_url' =>		get_admin_url(null, 'admin.php?page=ws-form-submit&id=' . $form_id . '#' . $submission->id)
					);

					// Add dynamic fields
					foreach($submission_fields as $field_id => $field) {

						// Default value
						$value = '';

						// Get submitted field
						$submitted_field = (isset($submission->meta[WS_FORM_FIELD_PREFIX . $field_id]) ? $submission->meta[WS_FORM_FIELD_PREFIX . $field_id] : '');

						if(isset($submitted_field['type'])) {

							// Get submitted field value
							$value = isset($submitted_field['value']) ? $submitted_field['value'] : '';

							// Field type processing
							switch($submitted_field['type']) {

								case 'file' :
								case 'signature' :

									if(is_array($value) && isset($value[0]) && isset($value[0]['url'])) {

										$value = $value[0]['url'];

									} else {

										$value = '';
									}

									break;
							}
						}

						$response_single[WS_FORM_FIELD_PREFIX . $field_id] = $value;
					}

					$response_array[] = $response_single;
				}

				header('Content-Type: application/json');
				echo json_encode($response_array);
				exit;
			}

			// API endpoint - Header test
			public function api_protect() {

				$api_key = WS_Form_Common::get_query_var('api_key', false);

				if(
					($api_key === false) ||
					($api_key !== $this->api_key)

				) { self::api_zapier_response(true); }

				return true;
			}

			// API endpoint - Response
			public function api_zapier_response($error = false, $data = false) {

				$response_array = array(

					'error'	=> $error
				);

				if($data !== false) { $response_array['data'] = $data; }

				header('Content-Type: application/json');
				echo json_encode($response_array);
				exit;
			}
		}
	});