<?php

	/**
	 * @link              https://wsform.com/knowledgebase/zendesk/
	 * @since             1.0.0
	 * @package           WS_Form_Zendesk
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Zendesk
	 * Plugin URI:        https://wsform.com/knowledgebase/zendesk/
	 * Description:       Zendesk add-on for WS Form PRO
	 * Version:           1.0.1
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-zendesk-v2
	 */

	Class WS_Form_Add_On_Zendesk_V2 {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.40';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Zendesk_V2();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-zendesk-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_zendesk = new WS_Form_Add_On_Zendesk_V2();	

	register_activation_hook(__FILE__, array($wsf_add_on_zendesk, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Zendesk_V2 extends WS_Form_Action {

			public $id = 'zendeskv2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 25;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Tickets';

			// Licensing
			private $licensing;

			// Config
			private $api_access_token;
			private $api_subdomain;
			private $api_endpoint;
			public $field_mapping;
			public $tag_mapping;
			public $attachments;

			// Ticket
			public $ticket_url;
			public $ticket_id;

			// Constants
			const NAME_FIELD = 'name';
			const EMAIL_FIELD = 'email';
			const DESCRIPTION_FIELD = 'description';
			const SUBJECT_FIELD = 'subject';
			const WS_FORM_LICENSE_ITEM_ID = 6035;
			const WS_FORM_LICENSE_NAME = 'Zendesk add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.1';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			public function __construct() {

				// Set label
				$this->label = __('Zendesk', 'ws-form-zendesk-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to Zendesk', 'ws-form-zendesk-v2');

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);
				add_filter('wsf_wizard_config_files', array( $this, 'wizard_config_files' ), 10, 1);
				add_filter('wsf_config_parse_variables', array( $this, 'config_parse_variables' ), 10, 1);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Parse variables
			public function config_parse_variables($parse_variables) {

				$parse_variables['zendesk'] = array(

					'label' => __('Zendesk', 'ws-form-zendesk-v2'),

					'variables'	=> array(

						'zendesk_ticket_id' =>	array('label' => __('Ticket ID', 'ws-form-zendesk-v2'), 'value' => $this->ticket_id, 'description' => __('Returns the ID of the created ticket.', 'ws-form-zendesk-v2')),
						'zendesk_ticket_url' =>	array('label' => __('Ticket URL', 'ws-form-zendesk-v2'), 'value' => $this->ticket_url, 'description' => __('Returns the URL of the created ticket in your Zendesk account. Do not use this URL in any communications to the person submitting the ticket.', 'ws-form-zendesk-v2'))
					)
				);

				return $parse_variables;
			}

			// Wizard - Config files
			public function wizard_config_files($config_files) {

				$config_files[] = plugin_dir_path(__FILE__) . 'includes/wizard/config.json';

				return $config_files;
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-zendesk-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-zendesk-v2') . '" />';
						}

						break;

					case 'oauth_action_' . $this->id :

						if($this->api_access_token) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access_token_delete" value="' . __('Disconnect', 'ws-form-zendesk-v2') . '" />';

						} else {

							// Check subdomain
							if(empty($this->api_subdomain)) {

								$value = '<p>Please enter your Zendesk subdomain above.</p>';

							} else {

								// Build query string
								$button_url = 'https://wsform.com/zendesk-v2?subdomain=' . $this->api_subdomain;
								$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
								$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

								$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-zendesk-v2'));
							}
						}

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access_token_delete' :

								self::reset_api();
								break;
						}

					break;
				}
			}

			// Reset API
			public function reset_api() {

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Zendesk account  <a href="%s">here</a>.', 'ws-form-zendesk-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Build attachments
				$upload_tokens = array();
				if(is_array($this->attachments) && isset($submit->meta)) {

					foreach($this->attachments as $attachment) {

						// Get field_id
						if(!isset($attachment['ws_form_field']) || empty($attachment['ws_form_field'])) { continue; }
						$field_id = $attachment['ws_form_field'];

						// Get field
						if(!isset($submit->meta[WS_FORM_FIELD_PREFIX . $field_id])) { continue; }
						$field = $submit->meta[WS_FORM_FIELD_PREFIX . $field_id];

						// Get value
						if(!isset($field['value'])) { continue; }
						$files = $field['value'];

						// Check files
						if(!is_array($files)) { continue; }

						// Get base upload_dir
						$upload_dir = wp_upload_dir()['basedir'];

						// Process each file
						foreach($files as $file) {

							// Source
							if(!isset($file['path'])) { continue; }
							$filename_source = $upload_dir . '/' . $file['path'];

							// Get binary file data
							$handle = fopen($filename_source, "rb");
							$contents = fread($handle, filesize($filename_source));
							fclose($handle);

							// Post request
							$api_response = parent::api_call($this->api_endpoint, 'uploads.json?filename=' . urlencode($file['name']), 'POST', $contents, self::get_api_headers(), false, false, false, 'application/json', 'application/binary');

							// Process response
							switch($api_response['http_code']) {

								case 201 :

									$api_response_decoded = json_decode($api_response['response']);
									if(is_null($api_response_decoded)) { return __('Unknown API error processing upload for action: ' . $this->label, 'ws-form-zendesk-v2'); }
									if(!isset($api_response_decoded->upload)) { return __('Unknown API error processing upload for action: ' . $this->label, 'ws-form-zendesk-v2'); }
									if(!isset($api_response_decoded->upload->token)) { return __('Token not found processing upload for action: ' . $this->label, 'ws-form-zendesk-v2'); }

									$upload_tokens[] = $api_response_decoded->upload->token;
									break;

								default :

									self::error(self::get_api_error_detail($api_response['response']));
									return false;
							}
						}
					}
				}

				// Process field mapping
				$request = array();
				$name = false;
				$email_address = false;
				$description = false;
				$subject = false;

				foreach($this->field_mapping as $field_map) {

					// Get request field
					$request_field = $field_map['action_' . $this->id . '_list_fields'];
					if($request_field == '') { continue; }

					// Get field ID
					$field_id = $field_map['ws_form_field'];

					// Get submit value
					$submit_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Zendesk cannot accept arrays
					if(is_array($submit_value)) { $submit_value = implode(',', $submit_value); }

					switch($request_field) {

						case self::NAME_FIELD :

							$name = $submit_value;
							break;

						case self::EMAIL_FIELD :

							$email_address = $submit_value;
							break;

						case self::DESCRIPTION_FIELD :

							$description = $submit_value;
							break;

						case self::SUBJECT_FIELD :

							$subject = $submit_value;
							break;

						default :

							// Normal merge field
							$request[$request_field] = $submit_value;
					}

					$request_fields_push = true;
				}

				// Check email address is configured
				if($email_address === false) { return self::error(__('Email address not mapped', 'ws-form-zendesk-v2')); }
				if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid email address', 'ws-form-zendesk-v2')); }
				$request['requester'] = array(

					'email' => $email_address
				);

				// Check name
				if($name !== false) { $request['requester']['name'] = $name; }

				// Check description
				if($description === false) { return self::error(__('Ticket description not mapped', 'ws-form-zendesk-v2')); }
				if($description == '') { return self::error(__('Ticket description empty', 'ws-form-zendesk-v2')); }
				$request['comment'] = array('body' => $description);

				// Check subject
				if($subject === false) { return self::error(__('Ticket subject not mapped', 'ws-form-zendesk-v2')); }
				if($subject == '') { return self::error(__('Ticket subject empty', 'ws-form-zendesk-v2')); }
				$request['subject'] = $subject;

				// Process tags
				$tags = array();
				foreach($this->tag_mapping as $tag_map) {

					// Get field ID
					$field_id = $tag_map['ws_form_field'];
					if($field_id == '') { continue; }

					// Read submit meta
					$tags_array = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);
					if($tags_array === false) { continue; }

					// Turn into array if it isn't already
					if(!is_array($tags_array)) { $tags_array = array($tags_array); }

					// Process each tag ready for API format
					foreach($tags_array as $tag) {

						if($tag == '') { continue; }

						$tags[] = $tag;
					}
				}
				if(count($tags) > 0) { $request['tags'] = $tags; }

				// Uploads
				if(count($upload_tokens) > 0) { $request['comment']['uploads'] = $upload_tokens; }

				$request_json = json_encode(array('ticket' => $request));

				// Post request
				$api_response = parent::api_call($this->api_endpoint, 'tickets.json', 'POST', $request_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 201 :

						$api_response_decoded = json_decode($api_response['response']);
						if(is_null($api_response_decoded)) { return __('Unknown API error adding ticket for action: ' . $this->label, 'ws-form-zendesk-v2'); }
						if(!isset($api_response_decoded->ticket)) { return __('Unknown API error adding ticket for action: ' . $this->label, 'ws-form-zendesk-v2'); }

						// Get ticket ID
						if(!isset($api_response_decoded->ticket->id)) { return __('Unable to obtain ticket ID for action: ' . $this->label, 'ws-form-zendesk-v2'); }
						$this->ticket_id = $api_response_decoded->ticket->id;

						// Get ticket URL
						if(!isset($api_response_decoded->ticket->url)) { return __('Unable to obtain ticket URL for action: ' . $this->label, 'ws-form-zendesk-v2'); }
						$this->ticket_url = $api_response_decoded->ticket->url;

						self::success(__('Successfully created ticket', 'ws-form-zendesk-v2'));
						return true;

					default :

						self::error(self::get_api_error_detail($api_response['response']));
						return false;
				}
			}

			// Get API error detail
			public function get_api_error_detail($api_response) {

				$api_response_decoded = json_decode($api_response);
				if(is_null($api_response_decoded)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-zendesk-v2'); }
				if(!isset($api_response_decoded->description)) { return __('Unknown API error for action: ' . $this->label, 'ws-form-zendesk-v2'); }

				return $api_response_decoded->description;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_tag_mapping',
						'action_' . $this->id . '_attachments'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(__('Action not configured', 'ws-form-zendesk-v2') . ' (' . $this->label . ''); }

				return $this->configured;
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(


					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-zendesk-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Zendesk ticket fields.', 'ws-form-zendesk-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						)
					),

					// Tag mapping
					'action_' . $this->id . '_tag_mapping'	=> array(

						'label'						=>	__('Tag Mapping', 'ws-form-zendesk-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Zendesk tags.', 'ws-form-zendesk-v2'),
						'meta_keys'					=>	array(

							'ws_form_field_choice'
						),
						'meta_keys_unique'			=>	array(
							'ws_form_field_choice'
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Zendesk Field', 'ws-form-zendesk-v2'),
						'type'							=>	'select',
						'options'						=>	array(

							array('value' => 'name', 'text' => __('Requester Name', 'ws-form-zendesk-v2') . ' *'),
							array('value' => 'email', 'text' => __('Requester Email', 'ws-form-zendesk-v2') . ' *'),
							array('value' => 'type', 'text' => __('Type', 'ws-form-zendesk-v2')),
							array('value' => 'priority', 'text' => __('Priority', 'ws-form-zendesk-v2')),
							array('value' => 'subject', 'text' => __('Subject', 'ws-form-zendesk-v2') . ' *'),
							array('value' => 'description', 'text' => __('Description', 'ws-form-zendesk-v2') . ' *'),
						),
						'options_blank'					=>	__('Select...', 'ws-form-zendesk-v2')
					),

					// Attachments
					'action_' . $this->id . '_attachments'	=> array(

						'label'						=>	__('Attachments', 'ws-form'),
						'type'						=>	'repeater',
						'meta_keys'					=>	array(

							'ws_form_field_file',
						),
						'meta_keys_unique'			=>	array(
							'ws_form_field_file'
						),
						'help'						=>	__('Email attachments.', 'ws-form')
					),
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-zendesk-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-zendesk-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Zendesk add-on for WS Form PRO license key here.', 'ws-form-zendesk-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-zendesk-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_api_subdomain'	=>	array(

							'label'		=>	__('Zendesk Subdomain', 'ws-form-zendesk-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Zendesk subdomain (e.g. your-company). To find your subdomain, <a href="https://support.zendesk.com/hc/en-us/articles/221682747-Where-can-I-find-my-Zendesk-subdomain-" target="_blank">click here</a>', 'ws-form-zendesk-v2')
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Zendesk Status', 'ws-form-zendesk-v2'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : (empty($this->api_subdomain) ? '' : __("Click the 'Connect' button to connect WS Form to your Zendesk account, or <a href=\"https://www.zendesk.com/register/\" target=\"blank\">register a new account</a>.", 'ws-form-zendesk-v2')),
							'button'		=>	'oauth_action_' . $this->id,
						)
					)
				);

				return $options;
			}

			// Load config for this action
			public function load_config($config = array()) {

				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->tag_mapping = parent::get_config($config, 'action_' . $this->id . '_tag_mapping', array());
				if(!is_array($this->tag_mapping)) { $this->tag_mapping = array(); }
				$this->attachments = parent::get_config($config, 'action_' . $this->id . '_attachments', array());
				if(!is_array($this->attachments)) { $this->attachments = array(); }
			}

			// Load config at plugin level
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_subdomain = WS_Form_Common::option_get('action_' . $this->id . '_api_subdomain', '');
				$this->api_access_token = WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');

				$this->api_endpoint = sprintf('https://%s.zendesk.com/api/v2/', $this->api_subdomain);

				if(
					empty($this->api_subdomain) ||
					empty($this->api_access_token)
				) {

					$this->configured = false;
				}

				return $this->configured;
			}
	
			// Get API headers
			public function get_api_headers() {

				$headers = array('Authorization' => 'Bearer ' . $this->api_access_token);

				return $headers;
			}
		}
	});