<?php

	/**
	 * @link              https://wsform.com/knowledgebase/zoho/
	 * @since             1.0.0
	 * @package           WS_Form_Zoho
	 *
	 * @wordpress-plugin
	 * Plugin Name:       WS Form PRO - Zoho
	 * Plugin URI:        https://wsform.com/knowledgebase/zoho/
	 * Description:       Zoho add-on for WS Form PRO
	 * Version:           1.0.3
	 * Author:            Westguard Solutions
	 * Author URI:        https://www.westguardsolutions.com/
	 * Text Domain:       ws-form-zoho-v2
	 */

	Class WS_Form_Add_On_Zoho {

		const WS_FORM_PRO_ID 			= 'ws-form-pro/ws-form.php';
		const WS_FORM_PRO_VERSION_MIN 	= '1.4.56';

		function __construct() {

			// Load plugin.php
			if(!function_exists('is_plugin_active')) {

				include_once(ABSPATH . 'wp-admin/includes/plugin.php');
			}

			// Admin init
			add_action('plugins_loaded', array($this, 'plugins_loaded'), 20);
		}

		function plugins_loaded() {

			if(self::is_dependency_ok()) {

				new WS_Form_Action_Zoho();

			} else {

				self::dependency_error();

				if(isset($_GET['activate'])) { unset($_GET['activate']); }
			}
		}

		function activate() {

			if (!self::is_dependency_ok()) {

				self::dependency_error();
			}
		}

		// Check dependencies
		function is_dependency_ok() {

			if(!defined('WS_FORM_VERSION')) { return false; }

			return(

				is_plugin_active(self::WS_FORM_PRO_ID) &&
				(version_compare(WS_FORM_VERSION, self::WS_FORM_PRO_VERSION_MIN) >= 0)
			);
		}

		// Add error notice action - Pro
		function dependency_error() {

			// Show error notification
			add_action('after_plugin_row_' . plugin_basename(__FILE__), array($this, 'dependency_error_notification'), 10, 2);
		}

		// Dependency error - Notification
		function dependency_error_notification($file, $plugin) {

			// Checks
			if(!current_user_can('update_plugins')) { return; }
			if($file != plugin_basename(__FILE__)) { return; }

			// Build notice
			$dependency_notice = sprintf('<tr class="plugin-update-tr"><td colspan="3" class="plugin-update colspanchange"><div class="update-message notice inline notice-error notice-alt"><p>%s</p></div></td></tr>', sprintf(__('This add-on requires %s (version %s or later) to be installed and activated.', 'ws-form-zoho-v2'), '<a href="https://wsform.com?utm_source=ws_form_pro&utm_medium=plugins" target="_blank">WS Form PRO</a>', self::WS_FORM_PRO_VERSION_MIN));

			// Show notice
			echo $dependency_notice;
		}
	}

	$wsf_add_on_zoho = new WS_Form_Add_On_Zoho();	

	register_activation_hook(__FILE__, array($wsf_add_on_zoho, 'activate'));

	// This gets fired by WS Form when it is ready to register add-ons
	add_action('wsf_plugins_loaded', function() {

		class WS_Form_Action_Zoho extends WS_Form_Action {

			public $id = 'zohov2';
			public $pro_required = true;
			public $label;
			public $label_action;
			public $events;
			public $multiple = true;
			public $configured = false;
			public $priority = 150;
			public $can_repost = true;
			public $form_add = false;
			public $record_label = 'Modules';
			public $list_sub_modal_label = 'Select Layout';

			// Add new features
			public $add_new_reload = true;

			// Licensing
			private $licensing;

			// Config
			private $api_endpoint = false;
			private $api_access_token;
			private $api_access_token_expiry;
			private $api_refresh_token;
			private $api_domain;
			private $api_modules;

			private $modules_default;

			public $list_id = false;
			public $list_sub_id = false;
			public $opt_in_field = false;
			public $field_mapping = false;
			public $custom_mapping = false;
			public $attachment_mapping = false;

			// Constants
			const WS_FORM_LICENSE_ITEM_ID = 6535;
			const WS_FORM_LICENSE_NAME = 'Zoho add-on for WS Form PRO';
			const WS_FORM_LICENSE_VERSION = '1.0.3';
			const WS_FORM_LICENSE_AUTHOR = 'Westguard Solutions';

			const ACCESS_TOKEN_EXPIRY_WINDOW = 600;
			const SLACK_FIELD_SHORT_THRESHOLD = 20;
			const EMAIL_MERGE_FIELD = 'EmailAddress';
			const API_PATH = '/crm/v2/';
			const API_MAX_RECORDS = 50;

			public function __construct() {

				// Set label
				$this->label = __('Zoho', 'ws-form-zoho-v2');

				// Set label for actions pull down
				$this->label_action = __('Add to Zoho', 'ws-form-zoho-v2');

				// Default objects
				$this->modules_default = array(

					'Accounts',
					'Contacts',
					'Leads'
				);

				// Events
				$this->events = array('submit');

				// Register config filters
				add_filter('wsf_config_options', array($this, 'config_options'), 10, 1);
				add_filter('wsf_config_meta_keys', array($this, 'config_meta_keys'), 10, 2);
				add_filter('plugin_action_links_' . plugin_basename(__FILE__), array($this, 'plugin_action_links'), 10, 1);
				add_action('wsf_settings', array($this, 'wsf_settings_access_token'), 10, 0);
				add_action('rest_api_init', array($this, 'rest_api_init'), 10, 0);

				// Add nag action
				add_action('wsf_nag', array($this, 'nag'));

				// Licensing
				$this->licensing = new WS_Form_Licensing(

					self::WS_FORM_LICENSE_ITEM_ID,
					$this->id,
					self::WS_FORM_LICENSE_NAME,
					self::WS_FORM_LICENSE_VERSION,
					self::WS_FORM_LICENSE_AUTHOR,
					__FILE__
				);
				$this->licensing->transient_check();
				add_action('admin_init', array($this->licensing, 'updater'));
				add_filter('wsf_settings_static', array($this, 'settings_static'), 10, 2);
				add_filter('wsf_settings_button', array($this, 'settings_button'), 10, 3);
				add_filter('wsf_settings_update_fields', array($this, 'settings_update_fields'), 10, 2);

				// Register action
				parent::register($this);

				// Load plugin level configuration
				self::load_config_plugin();
			}

			// Get license item ID
			public function get_license_item_id() {

				return self::WS_FORM_LICENSE_ITEM_ID;
			}

			// Settings page rendered, check for access token
			public function wsf_settings_access_token() {

				// Error checking
				$error = WS_Form_Common::get_query_var($this->id . '_error', false);
				if($error !== false) {

					$error_description = WS_Form_Common::get_query_var($this->id . '_error_description', __('Unknown error', 'ws-form-zoho-v2'));

					WS_Form_Common::admin_message_push(sprintf(__('An error occurred when connecting your account: %s.', 'ws-form-'), $error_description), 'notice-error', false);

					self::reset_api();

					return false;
				}

				// Save access token
				$this->api_access_token = WS_Form_Common::get_query_var($this->id . '_access_token', false);
				if($this->api_access_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

				// Save access token expiry
				$this->api_access_token_expiry = time() + intval(WS_Form_Common::get_query_var($this->id . '_access_token_expiry', false));
				if($this->api_access_token_expiry === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

				// Save refresh token
				$this->api_refresh_token = WS_Form_Common::get_query_var($this->id . '_refresh_token', false);
				if($this->api_refresh_token === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', $this->api_refresh_token);

				// Save API domain
				$this->api_domain = WS_Form_Common::get_query_var($this->id . '_api_domain', false);
				if($this->api_domain === false) { return false; }

				WS_Form_Common::option_set('action_' . $this->id . '_api_domain', $this->api_domain);

				// Reset objects
				WS_Form_Common::option_set('action_' . $this->id . '_api_modules', $this->modules_default);

				// Refresh
				$redirect_url = WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id);
				header('Location: ' . $redirect_url);
				exit;
			}

			// Check access token to see if it has expired
			public function access_token_check() {

				// Check if token has expired
				if($this->api_access_token_expiry > time()) { return true; }

				// Get new access token
				$path = '/zoho-v2?refresh_token=' . rawurlencode($this->api_refresh_token) . '&location=' . rawurlencode(WS_Form_Common::option_get('action_' . $this->id . '_location'));
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Check status code
				if($api_response['http_code'] == 200) { 

					// Process new access token
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

					// Save access token
					$this->api_access_token = parent::get_object_value($api_response_decoded, $this->id . '_access_token');
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', $this->api_access_token);

					// Save access token expiry
					$this->api_access_token_expiry = time() + intval(parent::get_object_value($api_response_decoded, $this->id . '_access_token_expiry'));
					WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', $this->api_access_token_expiry);

					// Save API domain
					$this->api_domain = parent::get_object_value($api_response_decoded, $this->id . '_api_domain');
					WS_Form_Common::option_set('action_' . $this->id . '_api_domain', $this->api_domain);

				} else {

					self::reset_api();
				}
			}

			// Plugin action link
			public function plugin_action_links($links) {

				// Settings
				array_unshift($links, sprintf('<a href="%s">%s</a>', WS_Form_Common::get_admin_url('ws-form-settings', false, 'tab=action_' . $this->id), __('Settings', 'ws-form-zoho-v2')));

				return $links;
			}

			// Settings - Static
			public function settings_static($value, $field) {

				switch ($field) {

					case 'action_' . $this->id . '_license_version' :

						$value = self::WS_FORM_LICENSE_VERSION;
						break;

					case 'action_' . $this->id . '_license_status' :

						$value = $this->licensing->license_status();
						break;

					case 'action_' . $this->id . '_status' :

						$value = $this->configured ? 'Connected' : 'Disconnected';
						break;

					case 'action_' . $this->id . '_api_domain' :

						$value = $this->api_domain;
						break;
				}

				return $value;
			}

			// Settings - Button
			public function settings_button($value, $field, $button) {

				switch($button) {

					case 'license_action_' . $this->id :

						$license_activated = WS_Form_Common::option_get('action_' . $this->id . '_license_activated', false);
						if($license_activated) {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="deactivate" value="' . __('Deactivate', 'ws-form-zoho-v2') . '" />';

						} else {

							$value = '<input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="activate" value="' . __('Activate', 'ws-form-zoho-v2') . '" />';
						}

						break;

					case 'oauth-action-' . $this->id :

						if($this->configured) {

							$value = '<input class="wsf-button wsf-button-inline" type="button" data-action="wsf-mode-submit" data-mode="access-token-delete" value="' . __('Disconnect', 'ws-form-zoho-v2') . '" />';

						} else {

							// Build query string
							$button_url = 'https://wsform.com/zoho-v2';
							$button_query_string = 'referrer=' . urlencode(WS_Form_Common::get_current_url());
							$button_query_string .= '&location=' . urlencode(WS_Form_Common::option_get('action_' . $this->id . '_location'));
							$button_url = WS_Form_Common::add_query_string($button_url, $button_query_string);

							$value = sprintf('<a href="%s" class="wsf-button wsf-button-inline">%s</a>', $button_url,  __('Connect', 'ws-form-zoho-v2'));
						}

						break;

					case 'modules' :

						$value = '<div><input class="wsf-button" type="button" data-action="wsf-mode-submit" data-mode="modules" value="' . __('Refresh', 'ws-form-zoho-v2') . '" /></div>';

						break;
				}
				
				return $value;
			}

			// Settings - Update fields
			public function settings_update_fields($field, $value) {

				switch ($field) {

					case 'action_' . $this->id . '_license_key' :

						$mode = WS_Form_Common::get_query_var('action_mode');

						switch($mode) {

							case 'activate' :

								$this->licensing->activate($value);
								break;

							case 'deactivate' :

								$this->licensing->deactivate($value);
								break;

							case 'access-token-delete' :

								self::reset_api();
								break;

							case 'modules' :

								self::get_modules(true, true);
								break;
						}

						break;
				}
			}

			// Reset API
			public function reset_api() {

				// Revoke access token
				self::load_config();

				// Check access token
				self::access_token_check();

				// Revoke refresh token
				$path = 'zoho-v2?refresh_token_revoke=' . rawurlencode($this->api_refresh_token) . '&location=' . rawurlencode(WS_Form_Common::option_get('action_' . $this->id . '_location'));
				$api_response = parent::api_call('https://wsform.com/', $path, 'GET');

				// Reset options
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_access_token_expiry', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_refresh_token', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_domain', '');
				WS_Form_Common::option_set('action_' . $this->id . '_api_modules', $this->modules_default);
			}

			// Nag
			public function nag() {

				// Load plugin level configuration
				self::load_config_plugin();

				if(!$this->configured) {

					WS_Form_Common::admin_message_push(sprintf(__('To complete the %s setup for %s, please connect your Zoho account <a href="%s">here</a>.', 'ws-form-zoho-v2'), $this->label, WS_FORM_NAME_PRESENTABLE, WS_Form_Common::get_admin_url('ws-form-settings', -1, 'tab=action_' . $this->id)), 'notice-warning', false);
				}
			}

			// Post to API
			public function post($form, &$submit, $config) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Load configuration
				self::load_config($config);

				// Check list ID is configured properly
				if(!self::check_list_id()) { return false; }

				// Opt-In Check
				$opted_in = false;

				// Get opt in value (False if field not submitted)
				$opt_in_field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $this->opt_in_field, false);

				if(($this->opt_in_field !== false) && ($this->opt_in_field !== '') && ($opt_in_field_value !== false)) {

					// End user did not opt in, exit gracefully
					if(empty($opt_in_field_value)) {

						self::success(__('User did not opt in, no data pushed to action', 'ws-form-zoho-v2'));
						return true;

					} else {

						$opted_in = true;
					}
				}

				// Get list fields
				$list_fields = self::get_list_fields();
				$list_fields_type_lookup = array();
				foreach($list_fields as $list_field) {

					$list_fields_type_lookup[$list_field['id']] = $list_field['action_type'];
				}

				// Build API payload
				$api_payload = array();

				// Process field mapping
				foreach($this->field_mapping as $field_map) {

					$field_id = $field_map['ws_form_field'];
					$field_key = $field_map['action_' . $this->id . '_list_fields'];

					// Get submit value
					$field_value = parent::get_submit_value($submit, WS_FORM_FIELD_PREFIX . $field_id, false);

					// Process according to Zoho type
					if(!isset($list_fields_type_lookup[$field_key])) { continue; }
					$action_type = $list_fields_type_lookup[$field_key];

					// Process field value
					$field_value = self::field_value_process($field_value, $action_type);

					// Add to API payload
					$api_payload[$field_key] = $field_value;
				}

				// Process custom mapping
				foreach($this->custom_mapping as $custom_map) {

					$field_value = $custom_map['action_' . $this->id . '_custom_mapping_value'];
					$field_key = $custom_map['action_' . $this->id . '_list_fields'];

					// Process according to Zoho type
					if(!isset($list_fields_type_lookup[$field_key])) { continue; }
					$action_type = $list_fields_type_lookup[$field_key];

					// Process field value
					$field_value = self::field_value_process($field_value, $action_type);

					// Parse field value
					$field_value = WS_Form_Common::parse_variables_process($field_value, $form, $submit, 'text/plain');

					// Add to API payload
					$api_payload[$field_key] = $field_value;
				}

				// Check API payload
				if(count($api_payload) == 0) {

					return self::error(__('No module fields mapped', 'ws-form-zoho-v2'));
				}

				// Check access token
				self::access_token_check();

				$data = array('data' => array($api_payload));

				// JSON encode payload
				$api_payload_json = json_encode($data);

				// Call API
				$api_response = parent::api_call($this->api_endpoint, $this->list_id, 'POST', $api_payload_json, self::get_api_headers());

				// Process response
				switch($api_response['http_code']) {

					case 201 :

						self::success(__('Successfully pushed to module: ', 'ws-form-zoho-v2') . $this->list_id);
						break;

					default :

						// Get invalid merge fields
						return self::error(self::get_api_error_detail($api_response['response']));
				}

				// Get record ID
				if(!isset($api_response['response'])) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }
				$response = $api_response['response'];

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

				if(!is_object($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

				if(
					!isset($api_response_decoded->data) ||
					!isset($api_response_decoded->data[0]) ||
					!isset($api_response_decoded->data[0]->details) ||
					!isset($api_response_decoded->data[0]->details->id)
				) {
					return self::error(__('Invalid API response data', 'ws-form-zoho-v2'));
				}

				$id = $api_response_decoded->data[0]->details->id;

				// Get base upload_dir
				$upload_dir = wp_upload_dir()['basedir'];

				// Process attachment mapping
				$files = array();
				foreach($this->attachment_mapping as $attachment_map) {

					$field_id = $attachment_map['ws_form_field'];

					// Get submit value
					$get_submit_value_repeatable_return = parent::get_submit_value_repeatable($submit, WS_FORM_FIELD_PREFIX . $field_id, array(), true);
					if(
						!is_array($get_submit_value_repeatable_return) ||
						!is_array($get_submit_value_repeatable_return['value']) ||
						!isset($get_submit_value_repeatable_return['value'][0])
					) { continue; }

					// Add each value
					foreach($get_submit_value_repeatable_return['value'] as $repeater_index => $meta_value) {

						$meta_value = $get_submit_value_repeatable_return['value'][$repeater_index];
						if(!is_array($meta_value)) { continue; }

						foreach($meta_value as $file) {

							// Check submit file data
							if(
								!isset($file['name']) ||
								!isset($file['type']) ||
								!isset($file['path'])

							) { continue; }

							// Source
							if(!isset($file['path'])) { continue; }
							$filename_source = $upload_dir . '/' . $file['path'];

							// Get binary file data
							$handle = fopen($filename_source, "rb");
							$contents = fread($handle, filesize($filename_source));
							fclose($handle);

							$boundary = hash('sha256', uniqid('', true));

							$body = '--' . $boundary . "\r\n";
							$body .= 'Content-Disposition: form-data; name="file"; filename="' . $file['name'] . "\"\r\n";
							$body .= 'Content-Type: ' . $file['type'] . "\r\n\r\n";
							$body .= $contents . "\r\n";
							$body .= '--' . $boundary . '--' . "\r\n";

							// Post request
							$api_response = parent::api_call($this->api_endpoint, $this->list_id . '/' . $id . '/Attachments', 'POST', $body, self::get_api_headers(), false, false, false, 'application/json', 'multipart/form; boundary=' . $boundary);
						}
					}
				}
			}

			// Process field value
			public function field_value_process($field_value, $action_type) {

				switch($action_type) {

					case 'boolean' :

						$field_value = is_array($field_value) ? implode(',', $field_value) : $field_value;
						$field_value = (!empty($field_value));
						break;

					case 'currency' :

						$field_value = filter_var($field_value, FILTER_SANITIZE_NUMBER_FLOAT, FILTER_FLAG_ALLOW_FRACTION);;
						break;

					case 'date' :

						$field_value = date('Y-m-d', strtotime($field_value));
						break;

					case 'datetime' :

						$field_value = date('Y-m-d', strtotime($field_value)) . 'T' . date('H:i:s', strtotime($field_value));
						break;

					case 'picklist' :

						$field_value = is_array($field_value) ? implode(',', $field_value) : $field_value;
						break;
				}

				return $field_value;
			}

			// Get from API
			public function get($form, $user) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Get user email address
				$user_email = $user->user_email;

				// Check user email address is valid
				if(!filter_var($user_email, FILTER_VALIDATE_EMAIL)) { return self::error(__('Invalid user email address', 'ws-form-zoho-v2')); }

				// Check list membership
				$params = array(

					'email' => $user_email
				);

				$api_response = parent::api_call($this->api_endpoint, $this->list_id . '/search', 'GET', $params, self::get_api_headers());

				// Check status code
				if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

				// Process list data
				$api_response_decoded = json_decode($api_response['response']);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

				// Get result
				if(!isset($api_response_decoded->data) || !isset($api_response_decoded->data[0])) { return self::error(__('No results found', 'ws-form-zoho-v2')); }
				$api_fields = (array) $api_response_decoded->data[0];

				// Process fields
				$fields_return = array();

				if($api_fields !== false) {

					foreach($api_fields as $key => $value) {

						$fields_return[$key] = $value;
					}
				}

				$return_array = array('fields' => $fields_return, 'tags' => false);

				return $return_array;
			}

			// Get lists
			public function get_lists($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check to see if lists are cached
				$lists = WS_Form_Common::option_get('action_' . $this->id . '_lists');

				// Retried if fetch is requested or lists are not cached
				if($fetch || ($lists === false)) {

					$lists = array();

					// Get selected entity definitions
					$modules = self::get_modules(false);
					if($modules === false) { return false; }

					foreach($modules as $id => $entity_definition) {

						$lists[] = array(

							'id' => 			$id, 
							'label' => 			$entity_definition['text'], 
							'field_count' => 	false, 
							'record_count' => 	false,
							'list_sub' =>		true
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_lists', $lists);
				}

				return $lists;
			}

			// Get list subs
			public function get_list_subs($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				// Check access token
				self::access_token_check();

				// Check to see if list_subs are cached
				$list_subs = WS_Form_Common::option_get('action_' . $this->id . '_list_subs_' . $this->list_id);

				// Retried if fetch is requested or list_subs are not cached
				if($fetch || ($list_subs === false)) {

					$list_subs = array();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'settings/layouts?module=' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

					// Checks
					if(!isset($api_response_decoded->layouts)) { return self::error(__('Unable to read fields', 'ws-form-zoho-v2')); }

					$layouts = $api_response_decoded->layouts;

					foreach($layouts as $layout) {

						$visible = isset($layout->visible) ? $layout->visible : false;
						if(!$visible) { continue; }

						$list_subs[] = array(

							'id' => 			$layout->id, 
							'label' => 			$layout->name,
							'record_count' =>	false
						);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_subs_' . '_' . $this->list_id, $list_subs);
				}

				return $list_subs;
			}

			// Get list
			public function get_list($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list = WS_Form_Common::option_get('action_' . $this->id . '_list_' . $this->list_id);

				if($fetch || ($list === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'settings/modules/' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

					// Checks
					if(!isset($api_response_decoded->modules[0])) { return self::error(__('Unable to read module', 'ws-form-zoho-v2')); }

					$module = $api_response_decoded->modules[0];

					if(!isset($module->singular_label)) { return self::error(__('Unable to read module label', 'ws-form-zoho-v2')); }

					// Build list
					$list = array(

						'label' => $module->singular_label
					);

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_list_' . $this->list_id, $list);
				}

				return $list;
			}

			// Get list fields
			public function get_list_fields($fetch = false) {

				// Check action is configured properly
				if(!self::check_configured()) { return false; }

				// Check list ID is set
				if(!self::check_list_id()) { return false; }

				$list_fields = WS_Form_Common::option_get('action_' . $this->id . '_list_fields_' . $this->list_id . '_' . $this->list_sub_id);

				if($fetch || ($list_fields === false)) {

					$list_fields = array();

					// Load configuration
					self::load_config();

					// Initialize
					$api_list_fields = array();
					$total_fields;
					$total_fields_retrieved = 0;

					// Check access token
					self::access_token_check();

					// Get list
					$api_response = parent::api_call($this->api_endpoint, 'settings/layouts?module=' . $this->list_id, 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);

					// Checks
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

					// Checks
					if(!isset($api_response_decoded->layouts)) { return self::error(__('Unable to read fields', 'ws-form-zoho-v2')); }

					$layouts = $api_response_decoded->layouts;
					$layout = false;

					if($this->list_sub_id !== false) {

						// Find selected layout
						foreach($layouts as $layout_single) {

							if($layout_single->id == $this->list_sub_id) { $layout = $layout_single; break; }
						}
					} else {

						$layout = $layouts[0];
					}

					if($layout === false) { return self::error(__('Layout not found', 'ws-form-zoho-v2')); }

					$sections = $layout->sections;

					// Process fields
					$sort_index = 0;

					foreach($sections as $section) {

						if(!isset($section->fields)) { continue; }

						foreach($section->fields as $field) {

							// Skip read only fields
							$field_read_only = parent::get_object_value($field, 'field_read_only');
							if($field_read_only) { continue;}

							// Skip invisible fields
							$field_visible = parent::get_object_value($field, 'visible');
							if(!$field_visible) { continue; }

							// Label
							$label = parent::get_object_value($field, 'display_label');
							if($label == 'Layout') { continue; }		// Skip layout, it won't map properly

							// Type
							$action_type = parent::get_object_value($field, 'data_type');
							$type = self::action_field_type_to_ws_form_field_type($action_type);
							if($type === false) { continue; }

							// Required
							$required = (parent::get_object_value($field, 'required') == 1);

							// Default value
							$default_value = parent::get_object_value($field, 'default_value');

							// ID
							$id = parent::get_object_value($field, 'api_name');

							$list_fields[] = array(

								'id' => 			$id,
								'label' => 			$label, 
								'label_field' => 	$label, 
								'type' => 			$type,
								'action_type' =>	$action_type,
								'required' => 		$required,
								'default_value' => 	$default_value,
								'pattern' => 		'',
								'placeholder' => 	'',
								'input_mask' => 	'',
								'help' => 			'', 
								'sort_index' => 	$sort_index,
								'visible' =>		true,
								'meta' => 			self::action_field_to_ws_form_meta_keys($field)
							);

							$sort_index++;
						}

						// Store to options
						WS_Form_Common::option_set('action_' . $this->id . '_list_fields_' . $this->list_id . '_' . $this->list_sub_id, $list_fields);
					}
				}

				return $list_fields;
			}

			// Get form fields
			public function get_fields() {

				$form_fields = array(

					'opt_in_field' => array(

						'type'	=>	'checkbox',
						'label'	=>	__('Opt-In', 'ws-form'),
						'meta'	=>	array(

							'data_grid_checkbox' => WS_Form_Common::build_data_grid_meta('data_grid_checkbox', false, false, array(

								array(

									'id'		=> 1,
									'default'	=> '',
									'required'	=> '',
									'disabled'	=> '',
									'hidden'	=> '',
									'data'		=> array(__('I would like to opt-in and receive emails from #blog_name', 'ws-form'))
								)
							))
						)
					),

					'submit' => array(

						'type'			=>	'submit'
					)
				);

				return $form_fields;
			}

			// Get form actions
			public function get_actions() {

				$form_actions = array(

					$this->id => array(

						'meta'	=> array(

							'action_' . $this->id . '_list_id'			=>	$this->list_id,
							'action_' . $this->id . '_list_sub_id'		=>	$this->list_sub_id,
							'action_' . $this->id . '_field_mapping'	=>	'field_mapping',
							'action_' . $this->id . '_opt_in_field'		=>	'opt_in_field'
						)
					),

					'message',

					'database'
				);

				return $form_actions;
			}


			// Get form meta
			public function get_meta() {

				return array('submit_reload' => '');
			}

			// Get API error detail
			public function get_api_error_detail($response) {

				$api_response_decoded = json_decode($response);
				if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

				if(!is_object($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

				$error_detail_array = array();
				if(isset($api_response_decoded->data)) {

					// Get error detail from JSON response
					foreach($api_response_decoded->data as $error) {

						$details_array = array();
						if(isset($error->details)) {

							foreach($error->details as $key => $value) {

								if(is_array($value)) { $value = implode(', ', $value); }
								$details_array[] = $key . ': ' . $value;
							}
						}
						$details = implode(', ' , $details_array);

						$error_detail_array[] = $error->status . ': ' . $error->message . ' (' . $error->code . ')' . (!empty($details) ? ' - ' . $details : '');
					}

				} else {

					if(isset($api_response_decoded->message)) {

						$error_detail_array[] = $api_response_decoded->message;
					}
				}

				return $error_detail_array;
			}

			// Get settings
			public function get_action_settings() {

				$settings = array(

					'meta_keys'		=> array(

						'action_' . $this->id . '_list_id',
						'action_' . $this->id . '_list_sub_id',
						'action_' . $this->id . '_opt_in_field',
						'action_' . $this->id . '_field_mapping',
						'action_' . $this->id . '_custom_mapping',
						'action_' . $this->id . '_attachment_mapping'
					)
				);

				// Wrap settings so they will work with sidebar_html function in admin.js
				$settings = parent::get_settings_wrapper($settings);

				// Add labels
				$settings->label = $this->label;
				$settings->label_action = $this->label_action;

				// Add multiple
				$settings->multiple = $this->multiple;

				// Add events
				$settings->events = $this->events;

				// Add can_repost
				$settings->can_repost = $this->can_repost;

				// Apply filter
				$settings = apply_filters('wsf_action_' . $this->id . '_settings', $settings);

				return $settings;
			}

			// Check action is configured properly
			public function check_configured() {

				if(!$this->configured) { return self::error(sprintf(__('Action not configured (%s)', 'ws-form-zoho-v2'), $this->label)); }

				return $this->configured;
			}

			// Check list ID is set
			public function check_list_id() {

				if($this->list_id === false) { return self::error(__('List ID is not set', 'ws-form-zoho-v2')); }

				return ($this->list_id !== false);
			}

			// Check list sub ID is set
			public function check_list_sub_id() {

				if($this->list_sub_id === false) { return self::error(__('List sub ID is not set', 'ws-form-zoho-v2')); }

				return ($this->list_sub_id !== false);
			}

			// Meta keys for this action
			public function config_meta_keys($meta_keys = array(), $form_id = 0) {

				// Build config_meta_keys
				$config_meta_keys = array(

					// List ID
					'action_' . $this->id . '_list_id'	=> array(

						'label'							=>	__('Zoho Module', 'ws-form-zoho-v2'),
						'type'							=>	'select',
						'help'							=>	__('Which Zoho module should this form data be added to?', 'ws-form-zoho-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-zoho-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_action_api_populate'	=>	'lists',
						'reload'					=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'lists_fetch'
						),
					),

					// List sub ID
					'action_' . $this->id . '_list_sub_id'	=> array(

						'label'						=>	__('Layout', 'ws-form-zoho-v2'),
						'type'						=>	'select',
						'help'						=>	__('Select the layout to associate this form with.', 'ws-form-zoho-v2'),
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-zoho-v2'),
						'options_action_id_meta_key'	=>	'action_id',
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_action_api_populate'	=>	'list_subs',
						'reload'						=>	array(

							'action_id'			=>	$this->id,
							'method'			=>	'list_subs_fetch',
							'list_id_meta_key'	=>	'action_' . $this->id . '_list_id'
						),
					),

					// Opt-In field
					'action_' . $this->id . '_opt_in_field'	=> array(

						'label'							=>	__('Opt-In Field', 'ws-form-zoho-v2'),
						'type'							=>	'select',
						'options'						=>	'fields',
						'options_blank'					=>	__('Select...', 'ws-form-zoho-v2'),
						'fields_filter_type'			=>	array('select', 'checkbox', 'radio'),
						'help'							=>	__('Checkbox recommended', 'ws-form-zoho-v2')
					),

					// Field mapping
					'action_' . $this->id . '_field_mapping'	=> array(

						'label'						=>	__('Field Mapping', 'ws-form-zoho-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Zoho fields.', 'ws-form-zoho-v2'),
						'meta_keys'					=>	array(

							'ws_form_field',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'reload'					=>	array(

							'action_id'				=>	$this->id,
							'method'				=>	'list_fields_fetch',
							'list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
							'list_sub_id_meta_key'	=>	'action_' . $this->id . '_list_sub_id'
						),
						'auto_map'					=>	true,
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Custom mapping
					'action_' . $this->id . '_custom_mapping'	=> array(

						'label'						=>	__('Custom Mapping', 'ws-form-zoho-v2'),
						'type'						=>	'repeater',
						'help'						=>	__('Map values or variables to Zoho fields.', 'ws-form-zoho-v2'),
						'meta_keys'					=>	array(

							'action_' . $this->id . '_custom_mapping_value',
							'action_' . $this->id . '_list_fields'
						),
						'meta_keys_unique'			=>	array(
							'action_' . $this->id . '_list_fields'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// Attachment mapping
					'action_' . $this->id . '_attachment_mapping'	=> array(

						'label'						=>	__('Attachment Mapping', 'ws-form-post'),
						'type'						=>	'repeater',
						'help'						=>	__('Map WS Form fields to Zoho attachments.', 'ws-form-post'),
						'meta_keys'					=>	array(

							'ws_form_field_file'
						),
						'meta_keys_unique'			=>	array(

							'ws_form_field_file'
						),
						'condition'					=>	array(

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_id',
								'meta_value'	=>	''
							),

							array(

								'logic'			=>	'!=',
								'meta_key'		=>	'action_' . $this->id . '_list_sub_id',
								'meta_value'	=>	''
							)
						)
					),

					// List fields
					'action_' . $this->id . '_list_fields'	=> array(

						'label'							=>	__('Zoho Field', 'ws-form-zoho-v2'),
						'type'							=>	'select',
						'options'						=>	'action_api_populate',
						'options_blank'					=>	__('Select...', 'ws-form-zoho-v2'),
						'options_action_id'				=>	$this->id,
						'options_list_id_meta_key'		=>	'action_' . $this->id . '_list_id',
						'options_list_sub_id_meta_key'	=>	'action_' . $this->id . '_list_sub_id',
						'options_action_api_populate'	=>	'list_fields'
					),

					// Custom mapping value
					'action_' . $this->id . '_custom_mapping_value'	=> array(

						'label'							=>	__('Value', 'ws-form-zoho-v2'),
						'type'							=>	'text'
					),
				);

				// Merge
				$meta_keys = array_merge($meta_keys, $config_meta_keys);

				return $meta_keys;
			}

			// Plug-in options for this action
			public function config_options($options) {

				$options['action_' . $this->id] = array(

					'label'		=>	$this->label,
					'fields'	=>	array(

						'action_' . $this->id . '_license_version'	=>	array(

							'label'		=>	__('Add-on Version', 'ws-form-zoho-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_license_key'	=>	array(

							'label'		=>	__('Add-on License Key', 'ws-form-zoho-v2'),
							'type'		=>	'text',
							'help'		=>	__('Enter your Zoho add-on for WS Form PRO license key here.', 'ws-form-zoho-v2'),
							'button'	=>	'license_action_' . $this->id,
							'action'	=>	$this->id
						),

						'action_' . $this->id . '_license_status'	=>	array(

							'label'		=>	__('Add-on License Status', 'ws-form-zoho-v2'),
							'type'		=>	'static'
						),

						'action_' . $this->id . '_location'	=>	array(

							'label'		=>	__('Domain', 'ws-form-zoho-v2'),
							'type'		=>	'select',
							'options'	=>	array(

								''			=>	array('text' => __('US', 'ws-form')),
								'cn'		=>	array('text' => __('Canada', 'ws-form')),
								'eu'		=>	array('text' => __('EU', 'ws-form')),
								'in'		=>	array('text' => __('India', 'ws-form')),
							),
							'default'	=>	''
						),

						'action_' . $this->id . '_status'	=>	array(

							'label'			=>	__('Zoho Status', 'ws-form-zoho-v2'),
							'type'			=>	'static',
							'help'			=>	$this->configured ? '' : __("Click the 'Connect' button to connect WS Form to your Zoho account, or <a href=\"https://www.zoho.com/signup.html\" target=\"blank\">create a new account</a>.", 'ws-form-zoho-v2'),
							'button'		=>	'oauth-action-' . $this->id,
						)
					)
				);

				if($this->configured) {

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_domain'] =	array(

						'label'			=>	__('API Domain', 'ws-form-zoho-v2'),
						'type'			=>	'static'
					);

					// Get objects data
					$modules = self::get_modules();

					$options['action_' . $this->id]['fields']['action_' . $this->id . '_api_modules'] =	array(

						'label'			=>	__('Working Modules', 'ws-form-zoho-v2'),
						'type'			=>	'select',
						'size'			=>	20,
						'multiple'		=>	true,
						'help'			=>	__('Select which Zoho modules you would like to work with.', 'ws-form-zoho-v2'),
						'options'		=>	$modules,
						'button'		=>	'modules'
					);
				}

				return $options;
			}

			// Get objects
			public function get_modules($include_unselected = true, $fetch = false) {

				$modules_raw = WS_Form_Common::option_get('action_' . $this->id . '_modules');

				if($fetch || ($modules_raw === false)) {

					// Load configuration
					self::load_config();

					// Check access token
					self::access_token_check();

					// Get modules
					$api_response = parent::api_call($this->api_endpoint, 'settings/modules', 'GET', null, self::get_api_headers());

					// Check status code
					if($api_response['http_code'] != 200) { return self::error(self::get_api_error_detail($api_response['response'])); }

					// Process list data
					$api_response_decoded = json_decode($api_response['response']);
					if(is_null($api_response_decoded)) { return self::error(__('Invalid API response data', 'ws-form-zoho-v2')); }

					// Get entity definitions
					if(!is_array($api_response_decoded->modules)) { return self::error(__('Unable to retrieve entity definitions', 'ws-form-zoho-v2')); }

					// Build raw data
					$modules_raw = array();

					foreach($api_response_decoded->modules as $module) {

						// Check creatable
						if(!$module->creatable) { continue; }

						// Check API supported
						if(!$module->api_supported) { continue; }

						if(strpos($module->singular_label, '__MISSING LABEL__') !== false) { continue; }

						$modules_raw[$module->api_name] = array('text' => $module->singular_label);
					}

					// Store to options
					WS_Form_Common::option_set('action_' . $this->id . '_modules', $modules_raw);
				}

				// Process entity definitions
				$modules_selected = array();
				$modules_unselected = array();

				foreach($modules_raw as $name => $entity_definition) {

					if(in_array($name, $this->api_modules)) {

						$modules_selected[$name] = array('text' => $entity_definition['text']);

					} else {

						if($include_unselected) {

							$modules_unselected[$name] = array('text' => $entity_definition['text']);
						}
					}
				}

				// Sort arrays
				$text = array_column($modules_selected, 'text');
				array_multisort($text, SORT_ASC, $modules_selected);

				$text = array_column($modules_unselected, 'text');
				array_multisort($text, SORT_ASC, $modules_unselected);

				$modules = array_merge($modules_selected, $modules_unselected);

				return $modules;
			}

			// Convert action field type to WS Form field type
			public function action_field_type_to_ws_form_field_type($type) {

				switch($type) {

					case 'text' : 					return 'text';
					case 'textarea' :				return 'textarea';
					case 'boolean' :				return 'checkbox';
					case 'email' : 					return 'email';
					case 'date' :					return 'datetime';
					case 'datetime' :				return 'datetime';
					case 'currency' :				return 'price';
					case 'picklist' :				return 'select';
					case 'website' :				return 'url';
//					case 'fileupload' :				return 'file';		// TO DO
					case 'integer' :				return 'number';
					case 'bigint' :					return 'number';
					case 'double' :					return 'number';
					case 'phone' :					return 'tel';
					case 'multiselectpicklist' :	return 'select';
					case 'profileimage'	:			return false;
					case 'ownerlookup' :			return false;
					case 'fileupload' :				return false;

					default :

						return 'text';
				}
			}

			// Convert action field to WS Form meta key
			public function action_field_to_ws_form_meta_keys($field) {

				$meta_keys = array();
				$data_grid_meta_key = false;

				// Type
				$type = parent::get_object_value($field, 'data_type');

				switch($type) {

					case 'boolean' :

						// Select data grid type
						$data_grid_meta_key = 'data_grid_checkbox';

						// Label
						$label = parent::get_object_value($field, 'display_label');

						// Default value
						$default_value = parent::get_object_value($field, 'default_value');

						$picklist_object = new stdClass;
						$picklist_object->default_value = $default_value;
						$picklist_object->actual_value = $label;
						$picklist_object->display_value = $label;

						// Set picklist values
						$picklist_values = array($picklist_object);

						// Settings (Support 2 columns)
						$meta_keys['checkbox_field_label'] = 1;

						break;

					case 'datetime' :

						$meta_keys['input_type_datetime'] = 'datetime-local';
						break;

					case 'date' :

						$meta_keys['input_type_datetime'] = 'date';
						break;

					case 'picklist' :
					case 'multiselectpicklist' :

						// Select data grid type
						$data_grid_meta_key = 'data_grid_select';

						// Get picklist values
						$picklist_values = parent::get_object_value($field, 'pick_list_values');

						// Settings (Support 2 columns)
						$meta_keys['select_field_label'] = 1;

						// Settings for multi picklist
						if($type == 'multiselectpicklist') {

							$meta_keys['multiple'] = 'on';
							$meta_keys['size'] = 10;
						}

						break;

					case 'text' :
					case 'textarea' :

						// Length
						$length = intval(parent::get_object_value($field, 'length'));
						if($length > 0) {

							$meta_keys['max_length'] = $length;
						}

						break;

					case 'currency' :
					case 'double' :
					case 'percent' :

						// Decimal places (Precision)
						$decimal_place = intval(parent::get_object_value($field, 'decimal_place'));
						if($decimal_place > 0) {

							$meta_keys['step'] = floatval('0.' . str_repeat('0', $decimal_place - 1) . '1');
						}

						break;
				}

				if($data_grid_meta_key !== false) {

					// Get base meta
					$meta_keys_config = WS_Form_Config::get_meta_keys();
					if(!isset($meta_keys_config[$data_grid_meta_key])) { return false; }
					if(!isset($meta_keys_config[$data_grid_meta_key]['default'])) { return false; }

					$meta = $meta_keys_config[$data_grid_meta_key]['default'];

					$meta['columns'] = array(

						array('id' => 0, 'label' => __('Value', 'ws-form')),
						array('id' => 1, 'label' => __('Label', 'ws-form'))
					);

					// Build new rows
					$rows = array();
					$id = 1;
					foreach($picklist_values as $picklist_object) {

						$rows[] = array(

							'id'		=> $id,
							'default'	=> isset($picklist_object->default_value) ? $picklist_object->default_value : false,
							'required'	=> '',
							'disabled'	=> '',
							'hidden'	=> '',
							'data'		=> array($picklist_object->actual_value, $picklist_object->display_value)
						);

						$id++;
					}

					// Modify meta
					$meta['groups'][0]['rows'] = $rows;

					$meta_keys[$data_grid_meta_key] = $meta;
				}

				return $meta_keys;
			}

			// Load config for this action
			public function load_config($config = array()) {

				if($this->list_id === false) { $this->list_id = parent::get_config($config, 'action_' . $this->id . '_list_id'); }
				if($this->list_sub_id === false) { $this->list_sub_id = parent::get_config($config, 'action_' . $this->id . '_list_sub_id'); }
				$this->opt_in_field = parent::get_config($config, 'action_' . $this->id . '_opt_in_field');
				$this->field_mapping = parent::get_config($config, 'action_' . $this->id . '_field_mapping', array());
				if(!is_array($this->field_mapping)) { $this->field_mapping = array(); }
				$this->custom_mapping = parent::get_config($config, 'action_' . $this->id . '_custom_mapping', array());
				if(!is_array($this->custom_mapping)) { $this->custom_mapping = array(); }
				$this->attachment_mapping = parent::get_config($config, 'action_' . $this->id . '_attachment_mapping', array());
				if(!is_array($this->attachment_mapping)) { $this->attachment_mapping = array(); }
			}

			// Load config at plugin leveloun
			public function load_config_plugin() {

				$this->configured = true;

				$this->api_access_token 		= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token', '');
				$this->api_access_token_expiry 	= WS_Form_Common::option_get('action_' . $this->id . '_api_access_token_expiry', '');
				$this->api_refresh_token 		= WS_Form_Common::option_get('action_' . $this->id . '_api_refresh_token', '');
				$this->api_domain 				= WS_Form_Common::option_get('action_' . $this->id . '_api_domain', '');
				$this->api_modules 				= WS_Form_Common::option_get('action_' . $this->id . '_api_modules', $this->modules_default);
				if(!is_array($this->api_modules)) { $this->api_modules = array(); }

				// Set endpoint URL
				$this->api_endpoint = $this->api_domain . self::API_PATH;

				if(empty($this->api_access_token)) { $this->configured = false; }
				if(empty($this->api_refresh_token)) { $this->configured = false; }

				return $this->configured;
			}

			// Build REST API endpoints
			public function rest_api_init() {

				// API routes - get_* (Use cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/', array('methods' => 'GET', 'callback' => array($this, 'api_get_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_subs'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/(?P<list_sub_id>[a-zA-Z0-9-_]+)/fields/', array('methods' => 'GET', 'callback' => array($this, 'api_get_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));

				// API routes - fetch_* (Pull from API and update cache)
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/lists/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_lists'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_subs'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
				register_rest_route(WS_FORM_RESTFUL_NAMESPACE, '/action/' . $this->id . '/list/(?P<list_id>[a-zA-Z0-9-_]+)/subs/(?P<list_sub_id>[a-zA-Z0-9-_]+)/fields/fetch/', array('methods' => 'GET', 'callback' => array($this, 'api_fetch_list_fields'), 'permission_callback' => function () { return WS_Form_Common::can_user('create_form'); }));
			}

			// API endpoint - Lists
			public function api_get_lists() {

				// Get lists
				$lists = self::get_lists();

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List
			public function api_get_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list();

				// Process response
				self::api_response($list);
			}

			// API endpoint - List Subs
			public function api_get_list_subs($parameters) {

				// Get list subs
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_subs = self::get_list_subs();

				// Process response
				self::api_response($list_subs);
			}

			// API endpoint - List fields
			public function api_get_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$this->list_sub_id = WS_Form_Common::get_query_var('list_sub_id', false, $parameters);
				$list_fields = self::get_list_fields();

				// Process response
				self::api_response($list_fields);
			}

			// API endpoint - Lists with fetch
			public function api_fetch_lists() {

				// Get lists
				$lists = self::get_lists(true);

				// Process response
				self::api_response($lists);
			}

			// API endpoint - List with fetch
			public function api_fetch_list($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list = self::get_list(true);

				// Process response
				self::api_response($list);
			}

			// API endpoint - List subs with fetch
			public function api_fetch_list_subs($parameters) {

				// Get list subs
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$list_subs = self::get_list_subs(true);

				// Process response
				self::api_response($list_subs);
			}

			// API endpoint - List fields with fetch
			public function api_fetch_list_fields($parameters) {

				// Get lists
				$this->list_id = WS_Form_Common::get_query_var('list_id', false, $parameters);
				$this->list_sub_id = WS_Form_Common::get_query_var('list_sub_id', false, $parameters);
				$list_fields = self::get_list_fields(true);

				// Process response
				self::api_response($list_fields);
			}

			// Get API headers
			public function get_api_headers() {

				$headers = array(

					'Authorization' => 'Zoho-oauthtoken ' . $this->api_access_token
				);

				return $headers;
			}

			// SVG Logo - Color (Used for the 'Add Form' page)
			public function get_svg_logo_color($list_id = false) {

				$svg_logo = '<g transform="translate(39.000000, 75.000000)"><path fill="#dc7e28" d="M64 4.7v15.2L61.9 22V7.1z"/><linearGradient id="a" gradientUnits="userSpaceOnUse" x1="47.916" y1="6.722" x2="61.684" y2="21.891"><stop offset=".006" stop-color="#f8c51b"/><stop offset="1" stop-color="#e78f24"/></linearGradient><path fill="url(#a)" d="M48.1 7.1H62V22H48.1z"/><path fill="#fdd600" d="M50.2 4.7l-2.1 2.4h13.8L64 4.7z"/><path fill="#90c7ea" d="M33 5.4l13.7-1.9-1.1 3.2-13.5 1.9z"/><linearGradient id="b" gradientUnits="userSpaceOnUse" x1="32.569" y1="8.573" x2="46.768" y2="19.975"><stop offset=".006" stop-color="#1e8aca"/><stop offset="1" stop-color="#01659e"/></linearGradient><path fill="url(#b)" d="M45.6 6.7l1.8 13.4L34 22 32.1 8.6z"/><path fill="#0b98d5" d="M45.6 6.7l1.8 13.4 1.1-2.9-1.8-13.6v-.1"/><linearGradient id="c" gradientUnits="userSpaceOnUse" x1="24.826" y1="9.191" x2="24.826" y2="21.634"><stop offset="0" stop-color="#269f49"/><stop offset="1" stop-color="#008a54"/></linearGradient><path fill="url(#c)" d="M21.5 4.2l12.1 5.5-5.5 12.4-12-5.5z"/><path fill="#95cea0" d="M20.3 0l1.2 4.2 12.1 5.5-1-4.1z"/><path fill="#69bf6d" d="M20.3 0l-4.9 11.5.7 5.1 5.4-12.4z"/><linearGradient id="d" gradientUnits="userSpaceOnUse" x1="6.776" y1="7.974" x2="8.916" y2="20.892"><stop offset="0" stop-color="#d81f32"/><stop offset="1" stop-color="#bb2637"/></linearGradient><path fill="url(#d)" d="M13.5 6.7l2 13.1-13.4 2.1L0 8.8z"/><path fill="#761216" d="M14.4 2l2 13.2-.9 4.6-2-13.1z"/><path d="M59.8 12.4c-.3-.7-.7-1.3-1.2-1.9-.5-.5-1.1-.9-1.7-1.2-.7-.2-1.4-.3-2.1-.3-.7 0-1.4.1-2.1.4-.6.3-1.2.7-1.7 1.2s-.9 1.2-1.2 1.8c-.3.7-.4 1.4-.4 2.3 0 .8.1 1.6.4 2.2.3.7.7 1.3 1.2 1.9.5.5 1.1.9 1.7 1.2.6.3 1.3.4 2.1.4.7 0 1.4-.1 2.1-.4.6-.3 1.2-.7 1.7-1.2.5-.6.9-1.2 1.2-1.9.3-.7.4-1.4.4-2.3 0-.7-.1-1.5-.4-2.2zM57 17.3c-.5.6-1.2 1-2.2 1-.9 0-1.6-.3-2.2-1-.5-.6-.8-1.5-.8-2.6 0-1.2.3-2 .8-2.7.5-.6 1.3-1 2.2-1 .9 0 1.6.3 2.2 1 .5.6.8 1.5.8 2.7 0 1.1-.3 2-.8 2.6zm-48.1-3zm3.2 3.4c-.1-.1-.2-.2-.4-.3-.2-.1-.4-.1-.6-.1-.2 0-.5 0-.8.1l-3.5.6c0-.3.2-.6.4-1.1.3-.5.7-1.1 1.2-1.8.2-.2.3-.4.4-.5.1-.1.2-.2.3-.4.8-1.1 1.3-1.9 1.5-2.6.1-.4.2-.8.2-1.2v-.3c0-.3 0-.6-.1-.8 0-.2-.1-.5-.2-.6-.1-.2-.2-.3-.3-.4-.3.1-.5 0-.7 0-.2 0-.5 0-.8.1l-4.1.7c-.5.1-.8.2-1.1.4-.2.2-.3.4-.3.7v.2c.1.3.2.6.5.7.2.1.4.1.6.1.2 0 .3 0 .5-.1l3.4-.6v.2c0 .2 0 .4-.1.6-.1.3-.4.7-.8 1.2.1.3-.1.5-.3.8-.9 1.1-1.6 2.1-2 2.9-.3.5-.5 1.1-.6 1.6-.1.3-.1.6-.1.9v.6c0 .3.1.5.2.7.1.2.2.3.3.4.1.1.3.1.6.1s.8-.1 1.4-.1l3.7-.6c.7-.1 1.1-.3 1.4-.5.2-.2.3-.4.3-.7v-.2c0-.4 0-.5-.1-.7zm16.7-8.8c-.5-.5-1.1-.9-1.7-1.2-.7-.3-1.4-.4-2-.4-.7 0-1.4.2-2.1.5-.7.3-1.3.7-1.8 1.3-.5.6-1 1.2-1.3 2-.3.8-.5 1.5-.5 2.3v.3c0 .7.1 1.3.3 2 .2.7.6 1.3 1.1 1.8.5.5 1.1.9 1.8 1.2.7.3 1.3.4 2 .4s1.4-.2 2.1-.4c.7-.3 1.3-.8 1.9-1.3.5-.6 1-1.2 1.3-2 .3-.8.5-1.5.5-2.3v-.2c0-.7-.1-1.4-.4-2-.3-.9-.7-1.5-1.2-2zm-1.3 5.4c-.5 1.1-1.1 1.8-1.8 2.2-.4.2-.8.3-1.2.3-.4 0-.7-.1-1.1-.2-.8-.4-1.4-.9-1.6-1.8-.1-.3-.1-.6-.1-.9 0-.6.2-1.3.5-2 .5-1.1 1.1-1.8 1.8-2.3.4-.2.8-.3 1.2-.3.4 0 .8.1 1.2.3.8.4 1.4.9 1.6 1.8v.9c0 .6-.2 1.2-.5 2zm15.8-5c-.1-.4-.2-.7-.4-.9-.2-.2-.4-.2-.7-.2H42c-.3.1-.6.2-.8.5-.1.1-.2.3-.2.6v.3l.5 3.5-3.9.6-.5-3.5c-.1-.4-.2-.7-.4-.9-.2-.2-.4-.3-.6-.3h-.2c-.4.1-.6.2-.8.5-.1.2-.2.4-.2.7v.3l1.3 9c.1.4.2.7.4.9.2.2.4.2.7.2h.2c.3 0 .6-.2.7-.5.1-.2.2-.4.2-.7v-.3l-.5-3.6 3.9-.6.5 3.6c.1.4.2.7.4.9.2.2.4.2.7.2h.2c.3-.1.6-.2.8-.5.1-.2.2-.4.2-.7v-.3l-1.3-8.8z" fill="#fff"/><path fill="#ec473f" d="M0 8.8l.3-1.3.6-3.4L14.4 2l-.9 4.7z"/></g>';

				return $svg_logo;
			}
		}
	});